// maCdaOSX.c

/*
 * MoreAmp cd reading for MacOS X
 * Copyright (C) 2004-2010 moreforge
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include "maConfig.h" // included by all moreamp files

#ifdef __MAOSX__

#ifndef MANOCDA

#ifdef __cplusplus
extern "C" {
#endif

#include <Types.h>
#include <Files.h>
#include <Gestalt.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <Devices.h>

#include "maPlayer.h"
#include "maUtils.h"
#include "maCda.h"

#define MALOGf(x)

cdadrive g_cdadrive[CDAMAXNUMDRIVES + 1];
char *g_cdabuf;
int g_cdabufblocks = FIFOSIZE / CDABLOCKSIZE;
int g_cdaactive; // 1-based index
int g_cdanumdrives;

int CdaBCDtoHex(char SrcByte);
char CdaHextoBCD(int SrcInt);
int Cdamsf_to_frames(int minutes, int seconds, int frames);
int Cdamsf_to_lba(int minutes, int seconds, int frames);

cdadisk g_cd[CDAMAXNUMDRIVES + 1];

int CdaBCDtoHex(char SrcByte)
{
	int x, y;

	x = SrcByte % 16;
	y = SrcByte / 16;
	y = y * 10;
	return x + y;
}


char CdaHextoBCD(int SrcInt)
{
	int x, y;
	char z;

	z = 0;
	x = SrcInt % 256;
	y = x;
	if(y >= 10)
	{
		z = z + 16;
		do
		{
			y = y - 10;
            if(y >= 10)
				z = z + 16;
		} while(y >= 10);
	}
	
	return y + z;

} // CdaHextoBCD


int Cdamsf_to_frames(int minutes, int seconds, int frames)
{
  return minutes * 4500 + seconds * 75 + frames;
}


int Cdamsf_to_lba(int minutes, int seconds, int frames)
{
  if(Cdamsf_to_frames(minutes, seconds, frames) > 150)
	return Cdamsf_to_frames(minutes, seconds, frames) - 150;
   
  return 0;
}


int CdaCleanup(int drive)
{
	char cstr[256];

	if(drive < 1 || drive > CDAMAXNUMDRIVES)
	{
		sprintf(cstr, "CdaCleanup bad drive index=%d", drive);
		MALOGf(cstr);
		return 1;
	}
	
	/* this screws up Plextor drive on USB
	// killIO
	ParamBlockRec PB = {0};
	OSErr err;
	PB.fileParam.ioCompletion = nil;
	PB.cntrlParam.ioCRefNum = g_cdadrive[drive].driverRefNum;
	PB.cntrlParam.csCode = 1; // KillIO
	PBCONTROL(&PB, false);
	*/

	// set block size back to default value
	if(CdaSetBlockSize(drive, 512))
	{
		sprintf(cstr, "CdaCleanup CdaSetBlockSize 512 failed");
		MALOGf(cstr);
	}
	
	/*
	// goodbye
	memset(&PB, 0, sizeof(PB));
	PB.fileParam.ioCompletion = nil;
	PB.cntrlParam.ioCRefNum = g_cdadrive[drive].driverRefNum;
	PB.cntrlParam.csCode = -1; // GoodBye
	err = PBCONTROL(&PB, false);
	if(err != noErr)
	{
		sprintf(cstr, "CdaCleanup PBControl err=%d", err);
		MALOGf(cstr);
		return err;
	}
	*/

	/*
	// flush volume
	memset(&PB, 0, sizeof(PB));
	PB.fileParam.ioNamePtr = NULL; // specify volume by vrefnum
	PB.fileParam.ioVRefNum = g_cdadrive[drive].fsvolrefnum;
	err = PBFlushVolSync((ParamBlockRec *)&PB); // ignore errors from flush
	if(err != noErr)
	{
		sprintf(cstr, "CdaCleanup PBFlushVolSync err=%d", err);
		MALOGf(cstr);
	}
	*/
	
	return noErr;
	
} // CdaCleanup


int CdaEject(int drive)
{
	OSErr err;
	char cstr[256];

	if(drive < 1 || drive > CDAMAXNUMDRIVES)
	{
		sprintf(cstr, "CdaEject bad drive index=%d", drive);
		MALOGf(cstr);
		return 1;
	}
	
	err = CdaSetBlockSize(drive, 512);
	if(err != noErr)
	{
		sprintf(cstr, "CdaEject CdaSetBlockSize err=%d", err);
		MALOGf(cstr);
	}

#if defined(MA_CPU_I86)
	pid_t dissenter;
	err = FSEjectVolumeSync(g_cdadrive[drive].fsvolrefnum, 0, &dissenter);
	if(err != noErr)
	{
 		sprintf(cstr, "CdaEject vol=%d FSEjectVolumeSync err=%d", g_cdadrive[drive].fsvolrefnum, err);
		MALOGf(cstr);
	}
#else
	ParamBlockRec PB;
	memset (&PB, 0, sizeof(ParamBlockRec));
	PB.fileParam.ioNamePtr = NULL; // specify volume by vrefnum
	PB.fileParam.ioVRefNum = g_cdadrive[drive].fsvolrefnum; // DrvNum for eject
	PBFlushVolSync((ParamBlockRec *)&PB); // ignore errors from flush
 #if !TARGET_API_MAC_CARBON // eject is done by PBUnmountVol in Carbon
	PB.ioParam.ioNamePtr = NULL;
	PB.ioParam.ioVRefNum = g_cdadrive[drive].fsvolrefnum; // DrvNum for eject
	err = PBEject((ParamBlockRec *) &PB);
	if(err != noErr)
	{
		sprintf(cstr, "CdaEject vol=%d PBEject err=%d", g_cdadrive[drive].fsvolrefnum, err);
		MALOGf(cstr);
	}
 #else // !TARGET_API_MAC_CARBON
	PB.ioParam.ioNamePtr = NULL;
	PB.ioParam.ioVRefNum = g_cdadrive[drive].fsvolrefnum; // DrvNum for unmount
	err = PBUnmountVol((ParamBlockRec *) &PB);
	if(err != noErr)
	{
		sprintf(cstr, "CdaEject vol=%d PBUnmountVol err=%d", g_cdadrive[drive].fsvolrefnum, err);
		MALOGf(cstr);
	}
 #endif
#endif

	return err;
	
} // CdaEject


#if !defined(kAudioCDFilesystemID)
 #define kAudioCDFilesystemID  (UInt16)(('J' << 8) | 'H') // $4A48 'JH'; this avoids compiler warning
#endif
int GetCdInfo(int drive);

int CdaGetCDDrives() // updates g_cd and returns number of cd drives containing audio media
{
	int numdrives = 0;
	int i;
	int isaudio;
	int y = kAudioCDFilesystemID;
	
	OSErr err;
	ItemCount volumeIndex;
	long systemVersion;
	HFSUniStr255 volumeName;
	FSVolumeInfo volumeInfo;
	FSVolumeRefNum fsvolrefnum = 0;
	FSVolumeRefNum *pfsvolrefnum = &fsvolrefnum;
	char cstr[256];
	char name[256];

	if(Gestalt(gestaltSystemVersion, &systemVersion) != noErr)
		systemVersion = 0;

	err = noErr;
	for(volumeIndex = 1; err == noErr || err != nsvErr; volumeIndex++)
	{
		memset((void *) &volumeInfo, 0, sizeof(volumeInfo));
		err = FSGetVolumeInfo(	kFSInvalidVolumeRefNum,
								volumeIndex,
								pfsvolrefnum,
								kFSVolInfoFSInfo | kFSVolInfoDriveInfo,
								&volumeInfo,
								&volumeName,
								NULL); 
		if(err == noErr)
		{
			for(i = 0; i < volumeName.length; i++)
						name[i] = volumeName.unicode[i];
			name[volumeName.length] = 0;
			/*
			if((systemVersion >= 0x00001000 && systemVersion < 0x00001010 &&
							volumeInfo.signature	== kAudioCDFilesystemID) ||
							volumeInfo.filesystemID == kAudioCDFilesystemID) // is an audio CD
			*/
			isaudio = 0;
			if(systemVersion >= 0x00001000 && systemVersion < 0x00001010 && volumeInfo.signature == y)
				isaudio = 1;
			else if(volumeInfo.filesystemID == y)
				isaudio = 1;
			
			if(isaudio)
			{
				numdrives++;
				if(numdrives <= CDAMAXNUMDRIVES)
				{
					g_cd[numdrives].exists = 1;
					g_cd[numdrives].isready = 1;
					g_cd[numdrives].driverRefNum = volumeInfo.driverRefNum;
					g_cd[numdrives].fsvolrefnum = fsvolrefnum;
					
					for(i = 0; i < volumeName.length; i++)
						g_cd[numdrives].name[i] = volumeName.unicode[i];
					
					g_cd[numdrives].name[volumeName.length] = 0;
					
					GetCdInfo(numdrives);
				}
			}
		}
		else if(err != nsvErr)
		{
			sprintf(cstr, "CdaGetCDDrives FSGetVolumeInfo vol=%d err=%d", (int)volumeIndex, (int)err);
			MALOGf(cstr);
			return 0;
		}
			
	} // for volumeIndex
		
	return numdrives; // ok
		
} // CdaGetCDDrives


int CdaInitCDRom()
{
	int numready = 0;
	int whichready = 0;
	int i;
	char cstr[256];

	for(i = 1; i <= CDAMAXNUMDRIVES; i++)
	{
		if(g_cd[i].cdtrack)
		{
			MAFREE(g_cd[i].cdtrack);
			g_cd[i].cdtrack = NULL;
		}
		
		g_cd[i].exists		= g_cdadrive[i].exists = 0;
		g_cd[i].isactive	= g_cdadrive[i].isactive = 0;
		g_cd[i].isready		= g_cdadrive[i].isready = 0;
		g_cd[i].name[0]		= g_cdadrive[i].name[0] = 0;		
	}
	
	g_cdaactive = 0;

	if((g_cdanumdrives = CdaGetCDDrives()) <= 0)
	{
		MALOGf("CdaInitCDRom no cdrom drive found");
		return 1;
	}

	if(g_cdanumdrives > CDAMAXNUMDRIVES)
	{
		sprintf(cstr, "CdaInitCDRom found %d cdrom drives, limited to first %d", 
													g_cdanumdrives, CDAMAXNUMDRIVES);
		MALOGf(cstr);
		g_cdanumdrives = CDAMAXNUMDRIVES;
	}

	for(i = 1; i <= g_cdanumdrives; i++) // legacy
	{
		g_cdadrive[i].driverRefNum  = g_cd[i].driverRefNum;	
		g_cdadrive[i].fsvolrefnum   = g_cd[i].fsvolrefnum;
		g_cdadrive[i].exists		= g_cd[i].exists;
		g_cdadrive[i].isready		= g_cd[i].isready;
		g_cdadrive[i].track			= g_cd[i].track;
		g_cdadrive[i].tracklast		= g_cd[i].tracklast;		
		strcpy(g_cdadrive[i].name, g_cd[i].name);
	}

	for(i = 1; i <= g_cdanumdrives; i++)
	{
		if(g_cdadrive[i].exists)
			if(g_cdadrive[i].isready)
			{
				numready++;
				whichready = i;
			}
	}
		
	if(numready < 1)
		return 1;

	whichready = 1;
	
	g_cdadrive[whichready].isactive = 1;
	g_cdaactive = whichready ;

	return 0;

} // CdaInitCDRom


/*
	File:		AudioCDSample.c
	
	Description:	This sample demonstrates how to discover audio CDs and access the TOC information
                        presented by the CD-DA filesystem.
				
	Copyright:		 Copyright 2001 Apple Computer, Inc. All rights reserved.
	
	Disclaimer:		IMPORTANT:  This Apple software is supplied to you by Apple Computer, Inc.
                        ("Apple") in consideration of your agreement to the following terms, and your
                        use, installation, modification or redistribution of this Apple software
                        constitutes acceptance of these terms.  If you do not agree with these terms,
                        please do not use, install, modify or redistribute this Apple software.

                        In consideration of your agreement to abide by the following terms, and subject
                        to these terms, Apple grants you a personal, non-exclusive license, under Apples
                        copyrights in this original Apple software (the "Apple Software"), to use,
                        reproduce, modify and redistribute the Apple Software, with or without
                        modifications, in source and/or binary forms; provided that if you redistribute
                        the Apple Software in its entirety and without modifications, you must retain
                        this notice and the following text and disclaimers in all such redistributions of
                        the Apple Software.  Neither the name, trademarks, service marks or logos of
                        Apple Computer, Inc. may be used to endorse or promote products derived from the
                        Apple Software without specific prior written permission from Apple.  Except as
                        expressly stated in this notice, no other rights or licenses, express or implied,
                        are granted by Apple herein, including but not limited to any patent rights that
                        may be infringed by your derivative works or by other works in which the Apple
                        Software may be incorporated.

                        The Apple Software is provided by Apple on an "AS IS" basis.  APPLE MAKES NO
                        WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
                        WARRANTIES OF NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
                        PURPOSE, REGARDING THE APPLE SOFTWARE OR ITS USE AND OPERATION ALONE OR IN
                        COMBINATION WITH YOUR PRODUCTS.

                        IN NO EVENT SHALL APPLE BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL OR
                        CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
                        GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
                        ARISING IN ANY WAY OUT OF THE USE, REPRODUCTION, MODIFICATION AND/OR DISTRIBUTION
                        OF THE APPLE SOFTWARE, HOWEVER CAUSED AND WHETHER UNDER THEORY OF CONTRACT, TORT
                        (INCLUDING NEGLIGENCE), STRICT LIABILITY OR OTHERWISE, EVEN IF APPLE HAS BEEN
                        ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
				
	Change History (most recent first):
		
            <2>		09/28/01	Updated for Mac OS X 10.1. Work around bug in 10.0.x where
                                        FSGetVolumeInfo swapped the filesystem ID and signature.
                                        
            <1>	 	04/05/01	New sample.
		
*/

#include <CoreFoundation/CoreFoundation.h>
#include <Carbon/Carbon.h>
#include <string.h>

// Constants
#if !defined(kAudioCDFilesystemID)
 #define kAudioCDFilesystemID  (UInt16)(('J' << 8) | 'H') // $4A48 'JH'; this avoids compiler warning
#endif

// XML PList keys
#define kRawTOCDataString			"Format 0x02 TOC Data"
#define kSessionsString				"Sessions"
#define kSessionTypeString			"Session Type"
#define kTrackArrayString			"Track Array"
#define kFirstTrackInSessionString  "First Track"
#define kLastTrackInSessionString   "Last Track"
#define kLeadoutBlockString			"Leadout Block"
#define	kDataKeyString 				"Data"
#define kPointKeyString				"Point"
#define kSessionNumberKeyString		"Session Number"
#define kStartBlockKeyString		"Start Block"

// Enums for trackType (digital data or audio)
enum
{
	kDigitalDataBit	= 2,
	kDigitalDataMask	= (1 << kDigitalDataBit)
};

enum
{
	kSecondsPerMinute	= 60
};

//
//	SubQTOCInfo - 	Structure which describes the SubQTOCInfo defined in
//		 MMC-2 NCITS T10/1228D SCSI MultiMedia Commands Version 2
//		  rev 9.F April 1, 1999, p. 215				
//		 ATAPI SFF-8020 rev 1.2 Feb 24, 1994, p. 149
//

struct SubQTOCInfo
{	
	UInt8	sessionNumber;
	UInt8	address:4;
	UInt8	control:4;
	UInt8	tno;
	UInt8	point;
	UInt8	ATIP[3];
	UInt8	zero;
	union {
		
		struct {
            UInt8	minutes;
            UInt8	seconds;
            UInt8	frames;
		} startPosition;
		
		struct {
            UInt8	firstTrackNum;
            UInt8	discType;
            UInt8	reserved;
		} A0PMSF;
		
		struct {
            UInt8	lastTrackNum;
            UInt8	reserved[2];
		} A1PMSF;
		
		struct {
            UInt8	minutes;
            UInt8	seconds;
            UInt8	frames;
		} leadOutStartPosition;
		
		struct {
            UInt8	skipIntervalPointers;
            UInt8	skipTrackPointers;
            UInt8	reserved;
		} B1;

		struct {
            UInt8	skipNumber[3];
		} B2toB4;

		struct {
            UInt8	minutes;
            UInt8	seconds;
            UInt8	frames;
		} hybridDiscLeadInArea;

	} PMSF;
	
};
typedef struct SubQTOCInfo SubQTOCInfo;
typedef struct SubQTOCInfo * SubQTOCInfoPtr;


//
// QTOCDataFormat10 - Structure which describes the QTOCDataFormat10 defined in
//		 MMC-2 NCITS T10/1228D SCSI MultiMedia Commands Version 2
//		  rev 9.F April 1, 1999, p. 215				
//		 ATAPI SFF-8020 rev 1.2 Feb 24, 1994, p. 149
//


struct QTOCDataFormat10
{	
	UInt16		TOCDataLength;
	UInt8		firstSessionNumber;
	UInt8		lastSessionNumber;
	SubQTOCInfo		trackDescriptors[1];
};
typedef struct QTOCDataFormat10 QTOCDataFormat10;
typedef struct QTOCDataFormat10 * QTOCDataFormat10Ptr;


int GetTOCDataX(const FSRef *theFSRef, int drive);
UInt32 CalculateNumberOfDescriptors(const QTOCDataFormat10Ptr TOCDataPtr);
Boolean IsAudioTrack(const SubQTOCInfoPtr trackDescriptorPtr);

int GetTOCDataX(const FSRef *theFSRef, int drive)
{
	HFSUniStr255 dataForkName;
	OSErr theErr;
	SInt16 forkRefNum;
	SInt64 forkSize;
	Ptr forkData;
	ByteCount actualRead;
	CFDataRef dataRef = 0;
	CFPropertyListRef propertyListRef = 0;
	int ret = 0;
	char cstr[256];
	
	theErr = FSGetDataForkName(&dataForkName);
	if(theErr != noErr)
	{
		maLog("GetTOCData error getting data fork name");
		return 1;
	}
	
	theErr = FSOpenFork(theFSRef, dataForkName.length, dataForkName.unicode, fsRdPerm, &forkRefNum);
	if(theErr != noErr)
	{
		maLog("GetTOCData error opening data fork");
		return 1;
	}
	
	theErr = FSGetForkSize(forkRefNum, &forkSize);
	if(theErr != noErr)
	{
		maLog("GetTOCData error getting fork size");
		return 1;
	}
	
	// Allocate some memory for the XML data
	forkData = NewPtr(forkSize);
	if(forkData == NULL)
	{
		maLog("GetTOCData error allocating memory for XML data");
		return 1;
	}
	
	theErr = FSReadFork(forkRefNum, fsFromStart, 0 /* offset location */, forkSize, forkData, &actualRead);
	if(theErr != noErr)
	{
		maLog("GetTOCData error opening data fork");
		return 1;
	}
	
	dataRef = CFDataCreate(kCFAllocatorDefault, (const UInt8 *)forkData, forkSize);
	if(dataRef != 0)
	{
		CFStringRef	errorString;
		propertyListRef = CFPropertyListCreateFromXMLData(kCFAllocatorDefault,
                                                            dataRef,
                                                            kCFPropertyListImmutable,
                                                            &errorString);
		if(errorString != 0)
			CFRelease (errorString);
		
		// Now we got the Property List in memory. Parse it.
		
		// First, make sure the root item is a CFDictionary. If not, release and bail.
		if(CFGetTypeID(propertyListRef) == CFDictionaryGetTypeID ())
		{
            CFDataRef	theRawTOCDataRef	= 0;
            CFArrayRef	theSessionArrayRef	= 0;
            CFIndex		numSessions			= 0;
            UInt32		index				= 0;
            
            // This is how we get the Raw TOC Data
            theRawTOCDataRef = (CFDataRef)CFDictionaryGetValue((CFDictionaryRef)propertyListRef,
																			CFSTR(kRawTOCDataString));
            
            // Get the session array info.
            theSessionArrayRef = (CFArrayRef)CFDictionaryGetValue((CFDictionaryRef)propertyListRef,
																				CFSTR(kSessionsString));
            
            // Find out how many sessions there are.
            numSessions = CFArrayGetCount(theSessionArrayRef);
            
			if(numSessions > 1)
				maLog("GetTOCData multiple sessions on this CD, using first session");
			
            //u printf("Number of sessions = %ld\n", (UInt32) numSessions);
            //u printf("");
            
            for(index = 0; index < 1 /*u numSessions*/; index++)
            {                    
				CFDictionaryRef	theSessionDict = 0;
				//CFNumberRef	firstTrackNumber = 0;
				CFNumberRef	lastTrackNumber = 0;
				CFNumberRef	leadoutBlock = 0;
				//CFNumberRef	sessionNumber = 0;
				//CFNumberRef	sessionType = 0;
				CFArrayRef trackArray = 0;
				CFIndex numTracks = 0;
				UInt32 trackIndex = 0;
				UInt32 value = 0;
				UInt32 leadoutvalue = 0;
				
				theSessionDict = (CFDictionaryRef)CFArrayGetValueAtIndex(theSessionArrayRef, index);
				
				//firstTrackNumber = (CFNumberRef)CFDictionaryGetValue(theSessionDict, CFSTR(kFirstTrackInSessionString));
				
				lastTrackNumber = (CFNumberRef)CFDictionaryGetValue(theSessionDict, CFSTR(kLastTrackInSessionString));

				leadoutBlock = (CFNumberRef)CFDictionaryGetValue(theSessionDict,  CFSTR(kLeadoutBlockString));
				
				/*u
				sessionType = (CFNumberRef)CFDictionaryGetValue(theSessionDict, CFSTR(kSessionTypeString));
				
				sessionNumber = (CFNumberRef)CFDictionaryGetValue(theSessionDict, CFSTR(kSessionNumberKeyString));
				
				if(CFNumberGetValue(sessionNumber, kCFNumberSInt32Type, &value))
				{
                    //u printf("-----------------------------------");
                    //u printf("Session Number %ld\n", value);
                    //u printf("-----------------------------------");
				}

				if(CFNumberGetValue(firstTrackNumber, kCFNumberSInt32Type, &value))
				{
                    //u printf("First Track in session = %ld\n", value);
					if(value != 1)
					{
						sprintf(cstr,"firstTrackNumber is not 1 !!!, is %d", value);
						maLog(cstr);
					}
				}
				*/

				if(CFNumberGetValue(lastTrackNumber, kCFNumberSInt32Type, &value))
				{
                    //u printf("Last Track in session = %ld\n", value);
					g_cd[drive].tracklast = value;
				}
				
				if(CFNumberGetValue(leadoutBlock, kCFNumberSInt32Type, &leadoutvalue))
				{
                    //u printf("Leadout block for session = %ld\n", value);
				}

				/*
				if(CFNumberGetValue(sessionType, kCFNumberSInt32Type, &value))
				{
                    if(value == 0)
                    {
                        //u printf("Session Type is CD-DA or CD-ROM with first track in Mode 1");
                    }
                    else if(value == 16)
                    {
                        //u printf("Session Type is CD-I");
                    }
                    else if(value == 32)
                    {
                        //u printf("Session Type is CD-ROM XA with first track in Mode 2");
                    }
				}
				*/
				
				trackArray = (CFArrayRef)CFDictionaryGetValue(theSessionDict, CFSTR(kTrackArrayString));
				
				numTracks = CFArrayGetCount(trackArray);

				if(!(g_cd[drive].cdtrack = (cdatrack *)MAMALLOC(sizeof(cdatrack) * (numTracks + 1))))
				{
					sprintf(cstr, "GetTOCDataX cannot MAMALLOC g_cd.ptrack for %d tracks", (int)numTracks);
					MALOGf(cstr);
					ret = 1;
					goto GetTOCDataExit;
				}
				
				memset((void *)g_cd[drive].cdtrack, 0, sizeof(cdatrack) * (numTracks + 1));
				
				for(trackIndex = 0; (int)trackIndex < numTracks; trackIndex++)
				{
                    CFDictionaryRef	theTrackDict = 0;
                    // CFNumberRef trackNumber	= 0;
					// CFNumberRef sessionNumber = 0;
                    CFNumberRef startBlock = 0;
                    //CFBooleanRef isDataTrack = kCFBooleanFalse;
                    UInt32 value = 0;
 
                    theTrackDict = (CFDictionaryRef)CFArrayGetValueAtIndex(trackArray, trackIndex);
                    
                    //trackNumber = (CFNumberRef)CFDictionaryGetValue(theTrackDict, CFSTR(kPointKeyString));
                    
                    //sessionNumber = (CFNumberRef)CFDictionaryGetValue(theTrackDict, CFSTR(kSessionNumberKeyString));

                    startBlock = (CFNumberRef)CFDictionaryGetValue(theTrackDict, CFSTR(kStartBlockKeyString));
                    
                    //isDataTrack = (CFBooleanRef)CFDictionaryGetValue(theTrackDict, CFSTR(kDataKeyString));
                    /*                                        
                    if(CFNumberGetValue(trackNumber, kCFNumberSInt32Type, &value))
                    {
                        //u printf("Track Number = %ld\n", value);
                    }
                    
				    if(CFNumberGetValue(sessionNumber, kCFNumberSInt32Type, &value))
                    {
                        //u printf("It resides in session %ld\n", value);
                    }
                    */
                    if(CFNumberGetValue(startBlock, kCFNumberSInt32Type, &value))
                    {
                        //u printf("Track starts at block = %ld\n", value);
						g_cd[drive].cdtrack[trackIndex + 1].lbafirst = value;
						
						if(trackIndex > 0)											// not first track
						{
							g_cd[drive].cdtrack[trackIndex].lbalast = value - 1;	// for precedigng track
							if((int)(trackIndex + 1) >= numTracks)					// for last track
								g_cd[drive].cdtrack[trackIndex + 1].lbalast = leadoutvalue - 1;
						}
                    }

					/*
                    if(isDataTrack == kCFBooleanTrue)
                    {
                        //u printf("This is a DATA track");
                    }
                    else
                    {
                        //u printf("This is an AUDIO track");
                    }
					*/
					
				} // for trackIndex

            } // for session
		}

		CFRelease(propertyListRef);		
	}

GetTOCDataExit:
	
	CFRelease(dataRef);
	DisposePtr(forkData);
	FSCloseFork(forkRefNum);
	
	return ret;
	
} // GetTOCData


//
// CalculateNumberOfDescriptors - Calculate the number of SubQTOCInfo entries
//							     in the given TOC
//

UInt32 CalculateNumberOfDescriptors(const QTOCDataFormat10Ptr TOCDataPtr)
{	
	UInt32	numberOfDescriptors = 0;
	UInt32	length = 0;
	
	// Get the length of the TOC
	length = TOCDataPtr->TOCDataLength;

	// Remove the first and last session numbers so all we are left with are track descriptors
	length -= (sizeof (TOCDataPtr->firstSessionNumber) + sizeof (TOCDataPtr->lastSessionNumber));
	
	// Divide the length by the size of a single track descriptor to get total number
	numberOfDescriptors = length / (sizeof (SubQTOCInfo));
	
	return numberOfDescriptors;
}


//
//	IsAudioTrack -	Checks the arguments passed in to find out if specified
//					track is audio or not
//

Boolean IsAudioTrack(const SubQTOCInfoPtr trackDescriptorPtr)
{	
	// Check to make sure the point is between 1 and 99 (inclusive)
	if(trackDescriptorPtr->point < 100 && trackDescriptorPtr->point > 0)
	{            
		// Do we have digital data?
		if((trackDescriptorPtr->control & kDigitalDataMask) == 0)
		{
            // Found an audio track
            return true;
		}
	}	
	return false;
}


int GetCdInfo(int drive)
{
	OSErr result = noErr;
	FSRefParam fsRefPB;
	FSRef tocPlistFSRef;
	//HParamBlockRec pb;
				
	// get stuff from .TOC.plist                                                   
	fsRefPB.ioCompletion = NULL;
	fsRefPB.ioNamePtr = "\p.TOC.plist";
	fsRefPB.ioVRefNum = g_cd[drive].fsvolrefnum;
	fsRefPB.ioDirID = 0;
	fsRefPB.newRef = &tocPlistFSRef;
	
	result = PBMakeFSRefSync(&fsRefPB);
				
	if(result != noErr)
        maLog("GetCdInfo PBMakeFSRefSync error");
	else
		GetTOCDataX(&tocPlistFSRef, drive);
						
	return 0;
	
} // GetCdInfo


int CdaGetTrackStartEnd(int drive) // legacy
{	
	g_cdadrive[drive].lbafirst = g_cd[drive].cdtrack[g_cdadrive[drive].track].lbafirst;
	g_cdadrive[drive].lbalast  = g_cd[drive].cdtrack[g_cdadrive[drive].track].lbalast;

	return 0;
	
} // CdaGetTrackStartEnd


int CdaReadTrack(int drive, long *pbytesread)
{
	OSErr err;
	//FSRef tocPlistFSRef;
	//SInt64 forkSize;
	ByteCount actualRead;
	char cstr[256];
	
	*pbytesread = 0;
	
	if(drive < 1 || drive > CDAMAXNUMDRIVES)
	{
		sprintf(cstr, "CdaReadTrack bad drive index=%d", drive);
		MALOGf(cstr);
		return 1;
	}
	
	if(!g_cdabuf)
	{
		sprintf(cstr, "CdaReadTrack drive=%d NULL buf (lba=%d)", drive, g_cdadrive[drive].lba);
		MALOGf(cstr);
		return 1;
	}
	
	err = FSReadFork(   g_cd[drive].forkRefNum, fsFromStart,
						(g_cdadrive[drive].lba - g_cdadrive[drive].lbafirst) * CDABLOCKSIZE, // offset
						g_cdabufblocks * CDABLOCKSIZE, g_cdabuf, &actualRead);	
						
	if(err != noErr)
	{
	 	sprintf(cstr, "CdaReadTrack drive=%d lba=%d lbasperread=%d err=%d",
	 						drive, g_cdadrive[drive].lba, g_cdabufblocks, err);
		MALOGf(cstr);
	}
	else
		*pbytesread = actualRead;
		
	if(err == -39) // eofErr
		err = noErr;
		
	return err;
	
} // CdaReadTrack


int CdaSetBlockSize(int drive, int blocksize)
{
	/*
	ParamBlockRec PB = {0};
	OSErr err;
	char cstr[256];

	if(drive < 1 || drive > CDAMAXNUMDRIVES)
	{
		sprintf(cstr, "CdaSetBlockSize bad drive index=%d", drive);
		MALOGf(cstr);
		return 1;
	}
	
	PB.fileParam.ioCompletion = nil;
	PB.cntrlParam.ioCRefNum = g_cdadrive[drive].driverRefNum;
	PB.cntrlParam.csCode = 79;
	PB.cntrlParam.csParam[0] = blocksize;
	err = PBCONTROL(&PB, false);
	if(err != noErr)
	{
		sprintf(cstr, "CdaSetBlockSize %d err=%d", blocksize, err);
		MALOGf(cstr);
		return err;
	}
	
	return noErr;
	*/
	return 0;
		
} // CdaSetBlockSize


int CdaSetSpindleSpeed(int drive, int speed)
{
/*u
	ParamBlockRec PB = {0};
	OSErr err;
	char cstr[256];

	if(drive < 1 || drive > CDAMAXNUMDRIVES)
	{
		sprintf(cstr, "CdaSetSpindleSpeed bad drive index=%d", drive);
		MALOGf(cstr);
		return 1;
	}
	
	PB.fileParam.ioCompletion = nil;
	PB.cntrlParam.ioCRefNum = g_cdadrive[drive].driverRefNum;
	PB.cntrlParam.csCode = 114;
	PB.cntrlParam.csParam[0] = speed ? 0xff : 0; // 0 normal or 0xff for max
	err = PBCONTROL(&PB, false);
	if(err != noErr)
	{
		sprintf(cstr, "CdaSetSpindleSpeed PBControl err=%d", err);
		MALOGf(cstr);
		return err;
	}
	
	return noErr;
*/
	return 0;
} // CdaSetSpindleSpeed


int CdaPBOpenFile(int drive)
{
	FSRefParam fsRefPB;
	HFSUniStr255 dataForkName;
	FSRef tocPlistFSRef;
	char ioname[255];
	OSErr err;

	// finder filename
	//sprintf(ioname, "\p%d Audio Track.aiff", g_cdadrive[drive].track);
	sprintf(&ioname[1], "%d Audio Track.aiff", g_cdadrive[drive].track);
	ioname[0] = strlen(&ioname[1]);
	
	fsRefPB.ioCompletion = NULL;
	fsRefPB.ioNamePtr = (const unsigned char*)ioname; // "\p1 Audio Track.aiff";
	fsRefPB.ioVRefNum = g_cd[drive].fsvolrefnum; // fsvolrefnum set in CdaInit
	fsRefPB.ioDirID = 0;
	fsRefPB.newRef = &tocPlistFSRef;
	
	err = PBMakeFSRefSync(&fsRefPB);
	if(err != noErr)
        maLog("CdaPBOpenFile PBMakeFSRefSync error");

	err = FSGetDataForkName(&dataForkName);
	if(err != noErr)
	{
		maLog("CdaPBOpenFile FSGetDataForkName error");
		return 1;
	}

	err = FSOpenFork(&tocPlistFSRef, dataForkName.length, dataForkName.unicode, fsRdPerm, &g_cd[drive].forkRefNum);
	if(err != noErr)
	{
		maLog("CdaPBOpenFile FSOpenFork error");
		return 1;
	}

	return 0;
	
} // CdaPBOpenFile


int CdaPBCloseFile(int drive)
{
	FSCloseFork(g_cd[drive].forkRefNum);
	
	return 0;
	
} // CdaPBCloseFile


#ifdef __cplusplus
}
#endif

#endif // MANOCDA

#endif // __MAOSX__

//eof

