//  maConfig.h

/*
 * MoreAmp
 * Copyright (C) 2004-2010 moreforge
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

// this file is included first by all moreamp ma*.c and ma*.cpp files

// Platform		Dev environment		defined in environment  defined here	config file

// Mac OS X		Apple XCode			__XCODE__				__MAOSX__		maConfigMacOSX.h
// Mac OS X		darwin bash/gcc		__XCODE__				__MAOSX__		maConfigMacOSX.h		
// Mac OS X		darwin bash/gcc		__MACUNIX__				__MAUNIX__		maConfigUnix.h		
// Win32		MS VC6				_WIN32 _MSC_VER			__MAWIN__		maConfigWin32.h	
// unix/linux   bash/gcc			__UNIX__				__MAUNIX__		maConfigUnix.h

#ifndef __MACONFIG_H__
#define __MACONFIG_H__

// only one of these :
#ifdef __XCODE__ // set in configure.in and by XCode
 #define __MAOSX__
 #ifdef __UNIX__ // gcc 4 now sets this?
  #undef __UNIX__
 #endif
#endif

#ifdef __UNIX__ // set with -D__UNIX__ in configure.in and by wxGTK and wxX11
 #define __MAUNIX__
#endif

#if defined(_WIN32) || defined(_MSC_VER) // set by MSVC++
 #define __MAWIN__
#endif

// sanity check
#if !defined(__MAOSX__) && !defined(__MAUNIX__) && !defined(__MAWIN__)
 #error missing definition __MAOSX__ __MAUNIX__ or __MAWIN__
#endif

#if defined(__MAOSX__) && ( defined(__MAUNIX__) || defined(__MAWIN__))
 #error multiple definition with __MAOSX__
#endif

#if defined(__MAUNIX__) && (defined(__MAOSX__) || defined(__MAWIN__))
 #error multiple definition with __MAUNIX__
#endif

#if defined(__MACUNIX__) && !defined(__MAUNIX__)
 #error __MAUNIX__ not defined with __MACUNIX__
#endif

#ifdef __MAOSX__
 #include "../maConfigMacOSX.h"
#else
 #ifdef __MAWIN__
  #include "../maConfigWin32.h"
 #else // __MAUNIX__
  #include "../maConfigUnix.h"
 #endif
#endif

// program name
#ifdef MOREAMP_NAME
 #undef MOREAMP_NAME
#endif
#if defined(__MAUNIX__)
 #define MOREAMP_NAME "moreamp"
#else
 #define MOREAMP_NAME "MoreAmp"
#endif

// moreamp dir name
#ifdef MOREAMP_DIRNAME
 #undef MOREAMP_DIRNAME
#endif
#if defined(__MAUNIX__) && !defined(__MACUNIX__)
 #define MOREAMP_DIRNAME "moreamp"
#else
 #define MOREAMP_DIRNAME "MoreAmp"
#endif

// version also needs updating in several places :
// in the MacOSX XCode projects and in mac/Info.plist. 
#define MOREAMP_VERSION_STRING "0.1.25" // VERSIONSTRINGSIZE - 1 chars max

#define MOREAMP_HOME_URL "http://sourceforge.net/projects/moreamp/"

// audio check
#if defined(__MAOSX__) || defined(__MACUNIX__)
 #if !defined(AUDIO_MAC_CORE) && !defined(AUDIO_PORTAUDIO)
  #error missing definition AUDIO_MAC_CORE or AUDIO_PORTAUDIO
 #endif
 #if defined(AUDIO_PORTAUDIO) && defined(AUDIO_MAC_CORE)
  #error AUDIO_PORTAUDIO and AUDIO_MAC_CORE are both defined
 #endif
#endif

#if defined(__MAUNIX__) && !defined(__MACUNIX__)
 #if !defined(AUDIO_PORTAUDIO)
  #error AUDIO_PORTAUDIO not defined
 #endif
 #if defined(AUDIO_MAC_CORE)
  #error AUDIO_MAC_CORE is defined
 #endif
#endif

#if defined(__MAWIN__) 
 #if defined(AUDIO_MAC_CORE)
  #error AUDIO_MAC_CORE is defined
 #endif
 #if defined(AUDIO_WIN_WAVE) && defined(AUDIO_PORTAUDIO)
  #error both AUDIO_WIN_WAVE and AUDIO_PORTAUDIO are defined
 #endif
 #if !defined(AUDIO_WIN_WAVE)
  #define AUDIO_PORTAUDIO // default
 #endif
#endif

// decoders can be excluded :

#if !defined(USE_AUDIOIN) // || !defined(AUDIO_PORTAUDIO)
 #define NODECODEAUDIOIN
#endif

#ifndef USE_AIF
 #define NODECODEAIF
#endif

#ifndef USE_WAV
 #define NODECODEWAV
#endif

#ifndef USE_WMA
 #define NODECODEWMA
#endif

#ifndef USE_LIBMAD
 #define NODECODEMP3 // also mp1 and mp2
#endif

#ifndef USE_LIBVORBIS
 #define NODECODEOGG
#endif

#ifndef USE_LIBFAAD
 #define NODECODEAAC
#endif

#if !defined(USE_LIBFAAD) || !defined(USE_LIBMP4V2)
 #define NODECODEMP4
#endif

#ifndef USE_LIBFLAC
 #define NODECODEFLAC
#endif

#if !defined(USE_CDREADER) || defined(__MACUNIX__) // no cdparanoia on macunix
 #define MANOCDA
#endif

#ifndef USE_LIBID3TAG
 #define NOID3TAG
#endif

// encoders must be specifically included :
// binaries of some encoders should not be distributed without sources :

// these four encoders are always legal
#ifdef USE_LIBVORBIS		// this is defined, or not, in file maConfigXXXX.h generated by configure
 #define ENCODEOGG		
#endif

#ifdef USE_LIBFLAC			// this is defined, or not, in file maConfigXXXX.h generated by configure
 #define ENCODEFLAC
#endif

#ifdef USE_WAV_ENC			// this is defined, or not, in file maConfigXXXX.h generated by configure
 #define ENCODEWAV
#endif

#ifdef USE_AIF_ENC			// this is defined, or not, in file maConfigXXXX.h generated by configure
 #define ENCODEAIF
#endif

// #define BINARYDISTRIB //ss
#ifdef BINARYDISTRIB
#else
 #ifdef USE_LIBLAME			// this is defined, or not, in file maConfigXXXX.h generated by configure
  #define ENCODEMP3
 #endif

 #ifdef USE_LIBFAAC			// this is defined, or not, in file maConfigXXXX.h generated by configure
  #define ENCODEAAC	
 #endif

 #ifdef USE_LIBMP4V2		// this is defined, or not, in file maConfigXXXX.h generated by configure
  #ifdef USE_LIBMP4V2_ENC	// this is defined, or not, in file maConfigXXXX.h generated by configure
   #define ENCODEMP4	
  #endif
 #endif

 #ifdef USE_LIBWMA_ENC		// probably never
  #define ENCODEWMA
 #endif
#endif // !BINARYDISTRIB

// OSX moreampdir location
#if defined(__MAOSX__)
 #define MAOSXHOME	// this puts the moreampdir in the users home dir, like on unix
#endif				// else the moreampdir is the dir with the MoreAmp binary, like on win

// manage multiple instances, or not
#define MAINSTANCES
#if defined(MAINSTANCES)
 #if defined(__MAOSX__)
  #if !defined(MAOSXHOME)
   #define MAOSXHOME // moreampdir must be in the user's home dir
  #endif
 #endif
 #define MAUSESHM	// use shared memory
#endif

// kludges
#ifdef __MAWIN__
 #define LROUND(x) x
#else
 #define LROUND(x) lround(floor(x))
#endif

#ifdef __MAUNIX__ // redhat linux
 #ifndef __cplusplus
  long int lrintf(float x);
  long int lround(double x);
  long int lroundf(float x);
 #endif
#endif

// avoid unused formal parameter warnings
#ifdef __MAWIN__
 #define MAUNUSEDWIN
#else
 #define MAUNUSED
#endif

// OSX multitasking
#if defined(__MAOSX__)
 //#define MAUSEMPSERVICES	// use mac Carbon Multiprocessing services
#endif						// else use bsd as on moreamp unix

// override some maConfigWin32.h definitions
#if defined(__MAWIN__)
 #if defined(MAWIN_NOGUI) // in VC6++ project
  #if defined(USE_GUI)
   #undef USE_GUI
  #endif
 #endif
#endif

 // override some maConfigMacOSX.h definitions (as set by 'configure')
#if defined(__MAOSX__)
 #if defined(MAOSXN_GUI) // in MoreAmpN.pbproj
  #if !defined(USE_GUI)
   #define USE_GUI 1
  #endif
  #if !defined(USE_GUI_NATIVE)
   #define USE_GUI_NATIVE
  #endif
  #if defined(USE_GUI_NATIVE2)
   #undef USE_GUI_NATIVE2
  #endif
  #if defined(USE_GUI_WX)
   #undef USE_GUI_WX
  #endif
 #endif
#endif

// gui
#if !defined(USE_GUI)
 #define MAGUINONE
#elif defined(USE_GUI_NATIVE)
 #define MAGUINATIVE
#elif defined(USE_GUI_NATIVE2)
 #define MAGUINATIVE2
#elif defined(USE_GUI_WX) // wxWidgets
 #include "wx/version.h"
 #ifndef wxMAJOR_VERSION
  #error wxMAJOR_VERSION undefined
 #endif
 #ifndef wxMINOR_VERSION
  #error wxMINOR_VERSION undefined
 #endif
 #if(wxMAJOR_VERSION != 2)
   #error wxMAJOR_VERSION is not 2
 #endif
 #if(wxMINOR_VERSION == 4)
  #define MAGUIWX 4
  //#warning wxMINOR_VERSION is 4
 #elif(wxMINOR_VERSION == 6)
  #define MAGUIWX 6
  //#warning wxMINOR_VERSION is 6
 #elif(wxMINOR_VERSION == 8)
  #define MAGUIWX 8
  //#warning wxMINOR_VERSION is 8
 #else
  #error wxMINOR_VERSION is not 4 or 6 or 8
 #endif
#else
 #error USE_GUI defined without USE_GUI_NATIVE or USE_GUI_NATIVE2 or USE_GUI_WX
#endif

#if defined(MAGUINATIVE)
 #if defined(__MAOSX__)
  #define MAGUIOSX
 #elif defined(__MAUNIX__)
  #define MAGUIGTK
 #else
  #error USE_GUI_NATIVE is only implemented on Mac OS X and on Unix/Linux/GTK+1.2
 #endif
#elif defined(MAGUINATIVE2)
 #if defined(__MAUNIX__)
  #define MAGUIGTK
  #define MAGUIGTK2
 #else
  #error USE_GUI_NATIVE2 is only implemented on Unix/Linux/GTK+2
 #endif
#endif

// custom buttons under wx
#if defined(USE_GUI_WX) && !defined(USE_WXBUTTONS)
 #define MAUSEMABUTTON 1
#endif

// custom sliders under wx
#if defined(USE_GUI_WX) && !defined(USE_WXSLIDERS)
 #define MAUSEMASLIDER 1
#endif

// custom checkboxes under wx
#if defined(USE_GUI_WX) && !defined(USE_WXCHECKBOXES)
 #define MAUSEMACHECKBOX 1
#endif

 // allow use of keyboard shortcuts
#ifdef USE_KEYBOARD
 #define MAKEYBOARD
#endif

// mousewheel volume control
#if defined(USE_MOUSEWHEEL)
 #define MAMOUSEWHEEL
#endif

#if defined(USE_SMALLFONT) || (defined(__MAOSX__) && defined(__WXMAC__))
 #define MASMALLERFONT
#endif

#if defined(__MAOSX__)
 #if defined(MA_CPU_PPC) || defined(MA_CPU_PPC64)
  #ifndef WORDS_BIGENDIAN
   #error WORDS_BIGENDIAN not defined
  #endif
 #endif
 #if defined(MA_CPU_I86)
  #ifdef WORDS_BIGENDIAN
   #error WORDS_BIGENDIAN defined
  #endif
 #endif
#endif

#ifdef __MAWIN__
 #ifdef WORDS_BIGENDIAN
  #error WORDS_BIGENDIAN defined
 #endif
#endif

// disable multiple instances while testing
#if(0) // defined(MAGUIGTK)
 #if defined(MAINSTANCES)
  #undef MAINSTANCES
  #if defined(MAUSESHM)
   #undef MAUSESHM
  #endif
 #endif
#endif

// allow cd reader on mac unix native (gtk) builds for partial testing
#if defined(MANOCDA)
 #if defined(USE_CDREADER) && defined(__MACUNIX__) && (defined(USE_GUI_NATIVE) || defined(USE_GUI_NATIVE2))
  #undef MANOCDA
 #endif
#endif

// mem test
//#define MAMEMTRACE
#if defined(MAMEMTRACE)
 #ifdef __cplusplus
 extern "C"
 {
 #endif
 extern int g_nmamallocs;
 extern int g_nmacallocs;
 extern int g_ndupallocs;
 extern int g_nmafrees;
 extern int g_ndupfrees;
 void *mamalloc(size_t a);
 void *macalloc(size_t a, size_t b);
 void mafree(void *a);
 void madupfree(void *a);
 #define MAMALLOC mamalloc
 #define MACALLOC macalloc
 #define MAFREE mafree
 #define MADUPFREE madupfree
 #ifdef __cplusplus
 }
 #endif
#else
 #define MAMALLOC malloc
 #define MACALLOC calloc
 #define MAFREE free
 #define MADUPFREE free
#endif

// these are defined by configure and are redefined in some lib includes
// we don't use them so we undef them to avoid compiler warnings
#ifdef PACKAGE_BUGREPORT
 #undef PACKAGE_BUGREPORT
#endif
#ifdef PACKAGE_NAME
 #undef PACKAGE_NAME
#endif
#ifdef PACKAGE_STRING
 #undef PACKAGE_STRING
#endif
#ifdef PACKAGE_TARNAME
 #undef PACKAGE_TARNAME
#endif
#ifdef PACKAGE_VERSION
 #undef PACKAGE_VERSION
#endif
#ifdef PACKAGE_BUGREPORT
 #undef PACKAGE_BUGREPORT
#endif
#ifdef PACKAGE_NAME
 #undef PACKAGE_NAME
#endif
#ifdef PACKAGE_STRING
 #undef PACKAGE_STRING
#endif
#ifdef PACKAGE_TARNAME
 #undef PACKAGE_TARNAME
#endif
#ifdef PACKAGE_VERSION
 #undef PACKAGE_VERSION
#endif

#endif // __MACONFIG_H__
