// maControls.h

/*
 * MoreAmp
 * Copyright (C) 2004-2010 moreforge
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#ifndef __MACONTROLS_H__
#define __MACONTROLS_H__

#if defined(MAUSEMABUTTON) || defined(MAUSEMASLIDER) || defined(MAUSEMACHECKBOX)

// this code contains extracts from audacity :

/**********************************************************************

  Audacity: A Digital Audio Editor

  AButton.cpp

  Dominic Mazzoni

  This is a custom button class for Audacity.  The main feature it
  supports that a wxButton does not is mouseovers.  It uses an image
  for all of its states: up, over, down, and disabled, allowing any
  sort of customization you want.  Currently it does not support
  transparency effects, so the images must be rectangular and
  opaque.

**********************************************************************/

#include "wx/wxprec.h"

#ifndef WX_PRECOMP
 #include "wx/wx.h"
#endif

#include "wx/image.h"
#include "wx/dcclient.h"
#include "wx/dcmemory.h"
#include "wx/control.h"
#include <math.h>
#include "wx/defs.h"
#include "wx/msgdlg.h"
#include "wx/panel.h"
#include "wx/tooltip.h"

#ifdef __MAWIN__
 #include "wx/bmpbuttn.h"
#endif

#include "wx/colour.h"
#include "wx/brush.h"
#include "wx/pen.h"

// define this for the platforms which don't support wxBitmapButton
// (such as Motif), else a wxBitmapButton will be used
#ifdef __WXMOTIF__
 //#define USE_MDI_PARENT_FRAME 1
 #ifdef USE_MDI_PARENT_FRAME
  #include "wx/mdi.h"
 #endif
 #define USE_STATIC_BITMAP
#endif

// This looks at the first pixel in the image, and shifts
// the entire image by the vector difference between that 
// pixel and the dstColour.  For better control, use
// ChangeImageColour(wxImage, wxColour*, wxColour*) below

wxImage *ChangeImageColour(wxImage * srcImage, wxColour & dstColour); 

// This function takes a source image, which it assumes to
// be grayscale, and smoothly changes the overall color
// to the specified color, and returns the result as a
// new image.  This works well for grayscale 3D images.
// Use these routines to make tbuttons adapt to the color
// scheme of the user.
wxImage *ChangeImageColour(wxImage * srcImage,
                           wxColour & srcColour,
                           wxColour & dstColour);
     
// Takes a background image, foreground image, and mask
// returns a new image where the foreground has been
// overlaid onto the background, at location (xoff, yoff).
wxImage *OverlayImage(wxImage * background, wxImage * foreground,
                      wxImage * mask, int xoff, int yoff);

// Creates an image with a solid background color
wxImage *CreateBackground(int width, int height, wxColour colour);

class AColor
{
 public:
   static void Init();

   static void Bevel(wxDC & dc, bool up, wxRect & r);

   static void Light(wxDC * dc, bool selected);
   static void Medium(wxDC * dc, bool selected);
   static void Dark(wxDC * dc, bool selected);

   static void CursorColor(wxDC * dc);
   static void IndicatorColor(wxDC * dc, bool recording);

   static void Mute(wxDC * dc, bool on, bool selected, bool soloing);
   static void Solo(wxDC * dc, bool on, bool selected);

   static void MIDIChannel(wxDC * dc, int channel /* 1 - 16 */ );
   static void LightMIDIChannel(wxDC * dc, int channel /* 1 - 16 */ );
   static void DarkMIDIChannel(wxDC * dc, int channel /* 1 - 16 */ );

   static wxBrush lightBrush[2];
   static wxBrush mediumBrush[2];
   static wxBrush darkBrush[2];
   static wxPen lightPen[2];
   static wxPen mediumPen[2];
   static wxPen darkPen[2];
   
   static wxPen cursorPen;
   static wxPen indicatorPen[2];

   static wxBrush muteBrush[2];
   static wxBrush soloBrush;

   static wxPen envelopePen;
   static wxBrush envelopeBrush;

 private:
   static bool inited;

};

#if defined(MAUSEMABUTTON)
class maButton : public wxControl/*wxWindow*/
{
 public:

	maButton(wxWindow * parent, wxWindowID id,
			  const wxPoint & pos,
			  const wxSize & size,
			  wxImage *up,
			  wxImage *over,
			  wxImage *down,
			  wxImage *dis,
			  bool processdownevents);

	maButton(wxWindow * parent, wxWindowID id,
			  const wxPoint & pos,
			  const wxSize & size,
			  char **upXPM, char **overXPM, char **downXPM, char **disXPM,
			  bool processdownevents			  
			  );

	virtual ~ maButton();

	virtual bool Enable(bool enable);
	
	virtual bool Disable(); // not equivalent to 'Enable(FALSE)'
  
	void SetEnabled(bool state)
	{
		state ? Enable(TRUE) : Disable();
	}

	virtual void PushDown();
	virtual void PopUp();
	virtual void Command(wxCommandEvent & event);

	virtual void OnPaint(wxPaintEvent & event);
	virtual void OnMouseEvent(wxMouseEvent & event);

	virtual bool WasShiftDown();	// returns true if shift was held down
									// the last time the button was clicked

	void PaintUp();		// Paints ButtonUp if button is enabled
						// (ButtonDown is also used for disabled buttons)

	void PaintDown();	// Paints ButtonDown

	virtual bool AcceptsFocus() const { return FALSE; }
    
 private:

	enum AButtonState {
		AButtonUp,
		AButtonOver,
		AButtonDown,
		AButtonDis
	};

	int mWidth;
	int mHeight;

	bool mWasShiftDown;

	bool mButtonIsDown;
	bool mIsClicking;
	bool mEnabled;
	bool mProcessDownEvents;	// This bool, if true, makes the button able to process 
								// events when it is in the down state, and moving to
								// the opposite state when it is clicked. It is used for the Pause
								// button, and possibly others. If false, it (should)
								// behave just like a standard button.

	bool mPaintDown;			// Paints ButtonDown

	AButtonState mButtonState;

	wxBitmap *mBitmap[4];

 public:

	DECLARE_EVENT_TABLE()
};
#endif // !MAUSEMABUTTON


#ifdef __MAUNIX__
 #if defined(__WXGTK__) && defined(USE_WXGTK2) // wx built with gtk+-2.0
  #define CHECKBOXWIDTH 17
  #define CHECKBOXHEIGHt 15 
 #else
  #define CHECKBOXWIDTH 12
  #define CHECKBOXHEIGHt 11
 #endif
#else
 #define CHECKBOXWIDTH 11
 #define CHECKBOXHEIGHt 11
#endif
#define MACHECKBOXCLASS wxBitmapButton

class maCheckBox : public MACHECKBOXCLASS
{
public:
	maCheckBox(wxWindow *parent, wxWindowID id, const wxString& label,
			const wxPoint& pos = wxDefaultPosition,
			const wxSize& size = wxDefaultSize, long style = 0,
			const wxValidator& validator = wxDefaultValidator,
			const wxString& name = wxCheckBoxNameStr);
	
	virtual ~maCheckBox();

	wxBitmap Create(wxWindow *parent, wxWindowID id, const wxString& label,
			const wxPoint& pos = wxDefaultPosition,
			const wxSize& size = wxDefaultSize, long style = 0,
			const wxValidator& validator = wxDefaultValidator,
			const wxString& name = wxCheckBoxNameStr);
	
	void SetLabel(wxString label);
	void SetValue(bool);
	bool GetValue();
	bool IsChecked();
	void OnSize(wxSizeEvent& event);
	void OnMouseEvent(wxMouseEvent &event);
	void OnButton(wxCommandEvent& event);
	void Command(wxCommandEvent& event);
	void MacHandleControlClick(WXWidget control, wxInt16 controlpart);
	void LabelHide();
	void LabelShow();
	/*
	void wxCheckBox::Command (wxCommandEvent & event);
	void wxCheckBox::MacHandleControlClick( WXWidget WXUNUSED(control), wxInt16 WXUNUSED(controlpart));
	*/

	virtual bool AcceptsFocus() const { return FALSE; }

private:
	wxBitmap CreateBitmapForButton(int on = 0);
	wxBitmap m_bmBlue;
	wxBitmap m_bmGreen;
	wxBitmap m_bmRed;

	bool m_onalt; // use alternate on color
	bool m_value;
	wxStaticText *m_plabel;
	
	wxWindow *m_parent;
	wxWindowID m_id;
	wxString m_label;
	wxPoint m_pos;
	wxSize m_size;
	long m_style;
	//wxValidator m_ytvalidator;
	wxString m_name;
	
#ifdef USE_STATIC_BITMAP
	wxButton /*wxStaticBitmap*/ *m_button; // needs work...
#else
	wxBitmapButton *m_button;
#endif

	DECLARE_EVENT_TABLE()
};


class LWSlider
{
	friend class maSlider;

 public:

	LWSlider(wxWindow * parent,
				wxString name,
				const wxPoint &pos,
				const wxSize &size,
				int style,
				bool heavyweight=false);

	virtual ~LWSlider();

	void SetId(wxWindowID id);

	float Get();
	void Set(float value);

	int GetValue();
	void SetValue(int value);

	void Move(const wxPoint &newpos);

	void OnPaint(wxDC &dc, bool selected);
	void OnMouseEvent(wxMouseEvent &event);
	void Refresh();

	void RecreateTipWin();

	virtual bool AcceptsFocus() const { return FALSE; }

 private:

	void FormatPopWin();
	void SetPopWinPosition();
	void CreatePopWin();

	wxWindow *mParent;

	bool mHW;			// is heavyweight (in a window)

	int mStyle;

	int mLeft;			// passed pos.x value of slider
	int mTop;			// passed pos.y value of slider

	int mWidth;			// passed size.x value of slider
	int mHeight;		// passed size.y value of slider

	int mAxisX;			// x coord of axis of vertical   slider rel to mLeft
	int mAxisY;			// y coord of axis of horizontal slider rel to mTop

	int mThumbWidth;	// width of thumb bitmap
	int mHalfThumbWidth;

	int mThumbHeight;	// height of thumb bitmap
	int mHalfThumbHeight;

	int mRightX;		// x coord of limit of thumb center rel to mLeft
	int mBottomY;		// y coord of limit of thumb center rel to mTop

	int mRange;			// range of slider (mWidth - mThumbWidth - 1, or Height)

	int mValue;			// slider center position 0...range

	int mClickValue;	// temp new mValue when slider is clicked
	
	int mClickX;		// x coord of mouse rel to mLeft
	int mClickY;		// y coord of mouse rel to mTop
	
	wxWindowID mID;

	wxWindow *mPopWin;

	bool mIsDragging;

	wxBitmap *mBitmap;
	wxBitmap *mSelBitmap;
	wxBitmap *mThumbBitmap;
	wxBitmap *mSelThumbBitmap;

	wxString mName;
};

class maSlider : public wxWindow
{
 public:

  maSlider(wxWindow *parent, wxWindowID id,
           int value, int minValue, int maxValue,
           const wxPoint& pos = wxDefaultPosition,
           const wxSize& size = wxDefaultSize,
           long style = wxSL_HORIZONTAL,
           const wxValidator& validator = wxDefaultValidator,
           const wxString& name = wxSliderNameStr);

	virtual ~maSlider();
	
	float Get();
	void Set(float value);

	int GetValue();
	void SetValue(int value);

	void OnPaint(wxPaintEvent & event);
	void OnMouseEvent(wxMouseEvent & event);
	void OnLeftDClick(wxMouseEvent &event);

	void RecreateTipWin();

	int m_range;
	int m_xSize;
	int m_ySize;
	int m_Style;
	wxString m_name;
	
	wxWindow *m_parent;
	wxSize m_parentSize;
	wxWindowID m_id;

	virtual bool AcceptsFocus() const { return FALSE; }

 private:
	LWSlider *mLWSlider;

 public:
	DECLARE_EVENT_TABLE()
};

// bitmaps
extern const char *BMSliderThumb[];
extern const char *BMSliderThumbA[];
extern const char *BMUpButton[];
extern const char *BMDownButton[];
extern const char *BMHiliteButton[];

// Player controls
extern const char *BMFFwdMask[];
extern const char *BMFFwdDisabled[];
extern const char *BMFFwd[];

extern const char *BMPlayMask[];
extern const char *BMPlayDisabled[];
extern const char *BMPlay[];

extern const char *BMQuitMask[];
extern const char *BMQuitDisabled[];
extern const char *BMQuit[];

extern const char *BMRewindMask[];
extern const char *BMRewindDisabled[];
extern const char *BMRewind[];

extern const char *BMStopMask[];
extern const char *BMStopDisabled[];
extern const char *BMStop[];

extern const char *BMPause[];
extern const char *BMPauseDisabled[];
extern const char *BMPauseMask[];

// Eq controls
extern const char *BMMinusMask[];
extern const char *BMMinusDisabled[];
extern const char *BMMinus[];

extern const char *BMPlusMask[];
extern const char *BMPlusDisabled[];
extern const char *BMPlus[];

extern const char *BMMid[];
//extern const char *BMMidDisabled[];
extern const char *BMMidMask[];

extern const char *BMCopyMask[];
extern const char *BMCopyDisabled[];
extern const char *BMCopy[];

extern const char *BMPasteMask[];
extern const char *BMPasteDisabled[];
extern const char *BMPaste[];

extern const char *BMResetMask[];
extern const char *BMResetDisabled[];
extern const char *BMReset[];

extern const char *BMShiftLMask[];
//extern const char *BMShiftLDisabled[];
extern const char *BMShiftL[];

extern const char *BMShiftRMask[];
//extern const char *BMShiftRDisabled[];
extern const char *BMShiftR[];

extern const char *BMInvertMask[];
//extern const char *BMInvertDisabled[];
extern const char *BMInvert[];

// List controls
extern const char *BMInfoMask[];
extern const char *BMInfoDisabled[];
extern const char *BMInfo[];

extern const char *BMLoadedMask[];
extern const char *BMLoadedDisabled[];
extern const char *BMLoaded[];

extern const char *BMSonginfo[];
extern const char *BMSonginfoDisabled[];
extern const char *BMSonginfoMask[];

extern const char *BMSonglistMask[];
extern const char *BMSonglistDisabled[];
extern const char *BMSonglist[];

#endif // !(MAUSEMABUTTON || MAUSEMASLIDER || MAUSEMACHECKBOX)

#endif // __MACONTROLS_H__

// eof
