// maMemFile.c

/*
 * MoreAmp
 * Copyright (C) 2004-2010 moreforge
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

// maMemFile is not mmap()

#include "maConfig.h" // included by all moreamp files

#ifdef __MAWIN__
 #ifdef MAUNUSEDWIN
  #pragma warning( disable : 4100 ) // unreferenced formal parameter
 #endif
 #pragma warning( disable : 4514 ) // 'mad_f_mul_inline' : unreferenced inline function has been removed
 #pragma warning( disable : 4706 ) // assignment within conditional expression
 #pragma warning( disable : 4996 ) // The POSIX name for this item is deprecated.
#endif

#ifdef __cplusplus
extern "C"
{
#endif

#include <stdio.h>
#include <errno.h>
#include <string.h>

#ifdef __MAUNIX__
 #include <unistd.h>
#endif

#ifdef __MAWIN__
 #include <io.h>
 #if defined _MSC_VER
  #if _MSC_VER <= 1600
   #define fseeko fseek
   #define ftello ftell
  #endif
 #endif
#endif

#include <sys/stat.h>

#if defined(__MAOSX__)
 #include <unistd.h>
#endif

#include "maPlayer.h"
#include "maUtils.h"
#include "maMemFile.h"

#define MAXMEMFP 8 // seven available

typedef struct memFILE
{
	int errorind;
	int isopen;
	char *songbeg;
	char *songpos;
	char *songend;
} memFILE;

void memShowErr(const char * msg);

int g_noamem = 1;
int g_nobmem = 1;
int g_noimem = 1;
memFILE g_memfp[MAXMEMFP]; // element 0 unused because 0 == INVALID_HANDLE_VALUE

#ifdef __MAWIN__
 #ifndef MAX_PATH
  #define MAX_PATH 260
 #endif
#endif

// mam* functions use g_noamem (called from libFLAC)
FILE *mamfopen(const char *filename, const char *mode)
{
	if(g_noamem)
		return fopen((char *)filename, mode);
		
	return memfopen(filename, mode);
}


int mamfseek(FILE *stream, long offset, int whence)
{
	if(g_noamem)
		return fseek(stream, offset, whence);
		
	return memfseek(stream, offset, whence);
}


int mamfseeko(FILE *stream, off_t offset, int whence)
{
	long loffset;
	
	if(g_noamem)
		return fseeko(stream, offset, whence);
	else
	{
		loffset = offset;
		return memfseek(stream, loffset, whence);
	}
}


int mamlseek(int stream, off_t offset, int whence)
{
	if(g_noamem)
		return lseek(stream, offset, whence);
		
	return memlseek(stream, offset, whence);
}


size_t mamfread(void *ptr, size_t size, size_t nmemb, FILE *stream)
{
	if(g_noamem)
		return fread(ptr, size, nmemb, stream);
		
	return memfread(ptr, size, nmemb, stream);
}


int mamfclose(FILE *stream)
{
	if(g_noamem)
		return fclose(stream);
		
	return memfclose(stream);
}


long int mamftell(FILE *stream)
{
	if(g_noamem)
		return ftell(stream);
		
	return memftell(stream);
}


off_t mamftello(FILE *stream)
{
	long int ret;
	
	if(g_noamem)
		return ftello(stream);
	else
	{
		ret = memftell(stream);
		return ret;
	}
}


void mamrewind(FILE *stream)
{
	if(g_noamem)
		rewind(stream);
	else
		memrewind(stream);
}


int mamread(int fd, char *ptr, int size)
{
	if(g_noamem)
		return read(fd, ptr, size);
		
	return memread(fd, ptr, size);
}


// int	stat __P((const char *, struct stat *));
int mamstat(const char *filename, struct stat *buf)
{
	if(g_noamem)
		return stat(filename, buf);
	else
	{
		int i;
		for(i = 0; i < g_loadednum; i++)
		{
			if(!strcmp(filename, g_hdsongpath[i]))
				break;
		}
		
		if(g_loadednum <= 0 || i >= g_loadednum)
			return -1;
			
		return memfstat(i + 1, buf);	
	}
}


// int	fstat __P((int, struct stat *));
int mamfstat(int j, struct stat *buf)
{
	if(g_noamem)
		return fstat(j, buf);

	return memfstat(j, buf);
}


// int fileno(FILE *stream);
int mamfileno(FILE *stream)
{
	if(g_noamem)
		return fileno(stream);

	return memfileno(stream);
}


// int	 feof(FILE *);
int mamfeof(FILE *stream)
{
	if(g_noamem)
		return feof(stream);

	return memfeof(stream);
}


// int	 ferror(FILE *);
int mamferror(FILE *stream)
{
	if(g_noamem)
		return ferror(stream);

	return memferror(stream);
}


void mamclearerr(FILE *stream)
{
	if(g_noamem)
		clearerr(stream);
	else
		memclearerr(stream);
}


int mamfgetpos(FILE *stream, fpos_t *pos)
{
	if(g_noamem)
		return fgetpos(stream, pos);

	return memfgetpos(stream, pos);
}


int mamfsetpos(FILE *stream, const fpos_t *pos)
{
	if(g_noamem)
		return fsetpos(stream, pos);

	return memfsetpos(stream, pos);
}


// mbm* functions use g_nobmem (called from libFLAC)
FILE *mbmfopen(const char *filename, const char *mode)
{
	if(g_nobmem)
		return fopen((char *)filename, mode);

	return memfopen(filename, mode);
}


int mbmfseek(FILE *stream, long offset, int whence)
{
	if(g_nobmem)
		return fseek(stream, offset, whence);

	return memfseek(stream, offset, whence);
}


int mbmfseeko(FILE *stream, off_t offset, int whence)
{
	long loffset;
	
	if(g_nobmem)
		return fseeko(stream, offset, whence);
	else
	{
		loffset = offset;
		return memfseek(stream, loffset, whence);
	}
}


int mbmlseek(int stream, off_t offset, int whence)
{
	if(g_nobmem)
		return lseek(stream, offset, whence);

	return memlseek(stream, offset, whence);
}


size_t mbmfread(void *ptr, size_t size, size_t nmemb, FILE *stream)
{
	if(g_nobmem)
		return fread(ptr, size, nmemb, stream);

	return memfread(ptr, size, nmemb, stream);
}


int mbmfclose(FILE *stream)
{
	if(g_nobmem)
		return fclose(stream);

	return memfclose(stream);
}


long int mbmftell(FILE *stream)
{
	if(g_nobmem)
		return ftell(stream);
		
	return memftell(stream);
}


off_t mbmftello(FILE *stream)
{
	long int ret;
	
	if(g_nobmem)
		return ftello(stream);
	else
	{
		ret = memftell(stream);
		return ret;
	}
}


void mbmrewind(FILE *stream)
{
	if(g_nobmem)
		rewind(stream);
	else
		memrewind(stream);
}


int mbmread(int fd, char *ptr, int size)
{
	if(g_nobmem)
		return read(fd, ptr, size);
		
	return memread(fd, ptr, size);
}


// int	stat __P((const char *, struct stat *));
int mbmstat(const char *filename, struct stat *buf)
{
	if(g_nobmem)
		return stat(filename, buf);
	else
	{
		int i;
		for(i = 0; i < g_loadednum; i++)
		{
			if(!strcmp(filename, g_hdsongpath[i]))
				break;
		}
		
		if(g_loadednum <= 0 || i >= g_loadednum)
			return -1;
			
		return memfstat(i + 1, buf);	
	}
}


// int	fstat __P((int, struct stat *));
int mbmfstat(int j, struct stat *buf)
{
	if(g_nobmem)
		return fstat(j, buf);
		
	return memfstat(j, buf);
}


// int	 feof(FILE *);
int mbmfeof(FILE *stream)
{
	if(g_nobmem)
		return feof(stream);
		
	return memfeof(stream);
}


// int	 ferror(FILE *);
int mbmferror(FILE *stream)
{
	if(g_nobmem)
		return ferror(stream);
		
	return memferror(stream);
}


void mbmclearerr(FILE *stream)
{
	if(g_nobmem)
		clearerr(stream);
	else
		memclearerr(stream);
}


int mbmfgetpos(FILE *stream, fpos_t *pos)
{
	if(g_nobmem)
		return fgetpos(stream, pos);
		
	return memfgetpos(stream, pos);
}


int mbmfsetpos(FILE *stream, const fpos_t *pos)
{
	if(g_nobmem)
		return fsetpos(stream, pos);
		
	return memfsetpos(stream, pos);
}


// mim* functions use g_noimem (called from libid3tag)
FILE *mimfopen(const char *filename, const char *mode)
{
	if(g_noimem)
		return fopen((char *)filename, mode);
		
	return memfopen(filename, mode);
}


int mimfseek(FILE *stream, long offset, int whence)
{
	if(g_noimem)
		return fseek(stream, offset, whence);
		
	return memfseek(stream, offset, whence);
}


int mimlseek(int stream, off_t offset, int whence)
{
	if(g_noimem)
		return lseek(stream, offset, whence);
		
	return memlseek(stream, offset, whence);
}


size_t mimfread(void *ptr, size_t size, size_t nmemb, FILE *stream)
{
	if(g_noimem)
		return fread(ptr, size, nmemb, stream);
		
	return memfread(ptr, size, nmemb, stream);
}


int mimfclose(FILE *stream)
{
	if(g_noimem)
		return fclose(stream);
		
	return memfclose(stream);
}


long int mimftell(FILE *stream)
{
	if(g_noimem)
		return ftell(stream);
		
	return memftell(stream);
}


void mimrewind(FILE *stream)
{
	if(g_noimem)
		rewind(stream);
	else
		memrewind(stream);
}


int mimread(int fd, char *ptr, int size)
{
	if(g_noimem)
		return read(fd, ptr, size);
		
	return memread(fd, ptr, size);
}


// int	stat __P((const char *, struct stat *));
int mimstat(const char *filename, struct stat *buf)
{
	if(g_noimem)
		return stat(filename, buf);
	else
	{
		int i;
		for(i = 0; i < g_loadednum; i++)
		{
			if(!strcmp(filename, g_hdsongpath[i]))
				break;
		}
		
		if(g_loadednum <= 0 || i >= g_loadednum)
			return -1;
			
		return memfstat(i + 1, buf);	
	}
}


// int	fstat __P((int, struct stat *));
int mimfstat(int j, struct stat *buf)
{
	if(g_noimem)
		return fstat(j, buf);
		
	return memfstat(j, buf);
}


// int	 feof(FILE *);
int mimfeof(FILE *stream)
{
	if(g_noimem)
		return feof(stream);
		
	return memfeof(stream);
}

int mimfgetpos(FILE *stream, fpos_t *pos)
{
	if(g_noimem)
		return fgetpos(stream, pos);
		
	return memfgetpos(stream, pos);
}


int mimfsetpos(FILE *stream, const fpos_t *pos)
{
	if(g_noimem)
		return fsetpos(stream, pos);
		
	return memfsetpos(stream, pos);
}


int mimferror(FILE *stream)
{
	if(g_noimem)
		return ferror(stream);
		
	return memferror(stream);
}


void mimclearerr(FILE *stream)
{
	if(g_noimem)
		clearerr(stream);
	else
		memclearerr(stream);
}


// mem* functions operate on files in memory
void memShowErr(const char * msg)
{
	maLog(msg);
	
} // memShowErr


FILE *memfopen(const char *filename, const char *mode) // okret == FILE *, else NULL
{
#ifdef MAUNUSED
 (void)mode;
#endif

	return maCreateFile((char *)filename, 0, 0, 0, 0, 0, 0);
}


int memopen(char *filename, int mode) // okret == fd, else -1
{
#ifdef MAUNUSED
 (void)mode;
#endif
	
	int fd;
	
	 if((fd = (int)maCreateFile((char *)filename, 0, 0, 0, 0, 0, 0)))
		return fd;
	
	return -1;
}


int memfclose(FILE *stream) // okret == 0, else -1
{
	int j = (int)stream;
	if(j < 1 || j >= MAXMEMFP)
		return -1;
	g_memfp[j].isopen = 0;
	return 0;
}


// fseek() returns zero if it is successful and returns a nonzero value if it fails
int memfseek(FILE *stream, long offset, int whence) // okret == 0
{
	int j = (int)stream;
	if(j < 1 || j >= MAXMEMFP)
		return -1;
		
	if(maSetFilePointer((FILE *)j, offset, 0, whence, 0) >= 0)
		return 0;

	return -1;
}

// lseek() returns the new file position indicator value.
int memlseek(int stream, off_t offset, int whence)
{
	return maSetFilePointer((FILE *)stream, (int)offset, 0, whence, 0);
}


// ftell(), when successful, returns the current file position indicator value.
// If it fails, ftell() returns -1L and sets the global variable errno to a nonzero value
long int memftell(FILE *stream) // ret == file pos
{
	long int pos;
	int j = (int)stream;
	if(j < 1 || j >= MAXMEMFP)
		return -1L;
		
	if((pos = (g_memfp[j].songpos - g_memfp[j].songbeg)) >= 0 &&
		pos <= (g_memfp[j].songend - g_memfp[j].songbeg))
		return pos;
		
	return -1L;
}


// fread() returns the number of items read successfully
size_t memfread(void *ptr, size_t size, size_t nmemb, FILE *stream)
{
	unsigned long bytesread;
	int j = (int)stream;
	size_t nm;
	char *pos;
	
	if(j < 1 || j >= MAXMEMFP || g_memfp[j].songpos < g_memfp[j].songbeg)
	{
		return 0;
	}
		
	//g_memfp[j].errorind = 0;
	
	if((bytesread = size * nmemb) < 1)
	{
		g_memfp[j].errorind = 1;
		return 0;
	}
	
	if((pos = g_memfp[j].songpos + bytesread) <= g_memfp[j].songend)
	{
		memcpy(ptr, g_memfp[j].songpos, bytesread);
		g_memfp[j].songpos = pos;			
		return nmemb;
	}

	// past eof :
	 
	bytesread = (unsigned long)(g_memfp[j].songend - g_memfp[j].songpos);
		
	nm = bytesread / size;
	bytesread = size * nm;
	memcpy(ptr, g_memfp[j].songpos, bytesread);
	g_memfp[j].songpos = g_memfp[j].songend;				
	return nm;
	
} // memfread


// read() returns the number of bytes actually read from the file.
// In case of an error a value of negative one is returned and errno is set
int memread(int j, char *ptr, int size)
{
	int bytesread;

	if(j < 1 || j >= MAXMEMFP)
		return -1;
	
	if((bytesread = size) < 0)
		return -1;
	
	if(g_memfp[j].songpos >= g_memfp[j].songend) // eof
		return 0;
		
	if(bytesread > g_memfp[j].songend - g_memfp[j].songpos) // eof
		bytesread = g_memfp[j].songend - g_memfp[j].songpos;
	
	if(g_memfp[j].songpos < g_memfp[j].songbeg)
		return -1;

	memcpy(ptr, g_memfp[j].songpos, (size_t)bytesread);
	g_memfp[j].songpos += bytesread;
				
	return bytesread;
}


void memrewind(FILE *stream)
{
	int j = (int)stream;
	if(j < 1 || j >= MAXMEMFP)
		return;
	g_memfp[j].songpos = g_memfp[j].songbeg;
}


// If it is successful, fstat() returns zero.
// If it encounters an error, fstat() returns -1 and sets errno
int memfstat(int j, struct stat *buf) // okret = 0, else -1 and errno
{
	if(j < 1 || j >= MAXMEMFP)
		return -1;
		
	buf->st_mode = S_IFREG;
	buf->st_size = g_memfp[j].songend - g_memfp[j].songbeg;

	return 0;
}


// int feof(FILE *stream) tests the end-of-file indicator for the stream
//     pointed to by stream, returning non-zero if it is set.
int memfeof(FILE *stream)
{
	int j = (int)stream;

	if(j < 1 || j >= MAXMEMFP || g_memfp[j].songpos >= g_memfp[j].songend)
		return 1;
	 
	return 0;
}


// int ferror(FILE *stream) tests the error indicator for the stream pointed to
//     by stream, returning non-zero if it is set.  The error indicator can only
//     be reset by the clearerr() function.
int memferror(FILE *stream)
{
	int j = (int)stream;
	
	if(j < 1 || j >= MAXMEMFP)
		return 1;
		
	return g_memfp[j].errorind;
}


// fileno() examines the argument stream and returns its integer descriptor.
int memfileno(FILE *stream)
{
	int j = (int)stream;
	
	if(j < 1 || j >= MAXMEMFP)
		return 1;
		
	return j;
}


void memclearerr(FILE *stream)
{
	int j = (int)stream;
	
	if(j < 1 || j >= MAXMEMFP)
		return;
		
	g_memfp[j].errorind = 0;
}


/* some compilers have any of these in unistd.h:
typedef struct
{
 __off_t __pos;
 __mbstate_t __state;
} _G_fpos_t;
typedef _G_fpos_t fpos_t;
typedef __off_t off_t;
typedef __off64_t off_t;
typedef long long fpos_t; // 64 bit file sizes & positions
typedef unsigned long fpos_t;
*/

// fgetpos() returns zero when successful and returns a nonzero value when it fails
int memfgetpos(FILE *stream, fpos_t *pos)
{
	long lpos;
	int j = (int)stream;

	if(j < 1 || j >= MAXMEMFP || !pos)
		return -1; // error
	 
	if((lpos = memftell((FILE *)j)) < 0)
	 	return -1;
	 
	*((off_t *)pos) = (off_t)lpos;
	 
	return 0;
}


// fsetpos() returns zero if it is successful and returns a nonzero value if it fails
int memfsetpos(FILE *stream, const fpos_t *pos)
{
	int j = (int)stream;
	
	if(j < 1 || j >= MAXMEMFP || !pos)
		return -1; // error
		
	if(memfseek((FILE *)j, (long)(*((off_t *)pos)), SEEK_SET))
    	return -1;

	return 0;
}


// pc functions :

extern char *g_songmemptr[];
extern char g_hdsongpath[][MAX_PATH];
extern int g_loadednum;
FILE *maCreateFile(char *path, int what1, int what2, int what3,
									char * beg, char * end, int nomem)
{
#ifdef MAUNUSED
 (void)what1; (void)what2; (void)what3; (void)beg; (void)end;
#endif

	FILE *filep;
	char cstr[256];

	if(!nomem)
	{
		int i, j;
		
		for(i = 0; i < g_loadednum; i++)
		{
			if(!strcmp(path, g_hdsongpath[i]))
				break;
		}
		
		if(g_loadednum <= 0 || i >= g_loadednum)
		{
			sprintf(cstr, "maCreateFile not loaded : '%s'", path);
			memShowErr(cstr);
			return 0;
		}
		
		if(g_songmemptr == NULL)
		{
			memShowErr("maCreateFile null g_songmemptr");
			return 0;
		}
		
		for(j = 1; j < MAXMEMFP; j++)
			if(!g_memfp[j].isopen)
				break;
		
		if(j >= MAXMEMFP)
		{
			memShowErr("maCreateFile failed, no more FILE handles available");
			return 0;
		}
		
		g_memfp[j].errorind = 0;
		g_memfp[j].isopen = 1;
		g_memfp[j].songbeg = g_memfp[j].songpos = g_songmemptr[i];
		g_memfp[j].songend = g_songmemptr[i + 1];

		return (FILE *)j;
		
	} // USEMEM
		
	if(!path || !path[0])
	{
		memShowErr("maCreateFile null path (CREATE_ALWAYS)");
		return 0;
	}
		
	if((filep = mafopen(path, "rb")) == NULL) // read only
	{
		sprintf(cstr, "maCreateFile mafopen '%s'", path);
		memShowErr(cstr);
		return 0;
	}
		
#ifdef ASYNCREAD
	g_ar_offset = 0;
#endif

	return filep;		
	
} // maCreateFile


// called from maPlayMp3 only
int maReadFile(FILE *file, unsigned char *buffer, int bytes, int *pbytes, int what, int nomem)
{
#ifdef MAUNUSED
 (void)what;
#endif

	int bytesread;
	char cstr[256];

	if(nomem)
	{	
		bytesread = (int)fread(buffer, 1, (size_t)bytes, file);			
	}
	else
	{
		int j = (int)file;
		if(j < 1 || j >= MAXMEMFP)
		{
			sprintf(cstr, "maReadFile bad memfile pointer=%d", j);
			memShowErr(cstr);
			*pbytes = 0;
			return 0; // error
		}

		if((bytesread = bytes) > (g_memfp[j].songend - g_memfp[j].songpos))
					  bytesread = g_memfp[j].songend - g_memfp[j].songpos;
			
		memcpy(buffer, g_memfp[j].songpos, (unsigned long)bytesread);
		g_memfp[j].songpos += bytesread;
	}
	
	*pbytes = bytesread;
	
	return 1; // true == success
		
} // maReadFile


int maSetFilePointer(FILE *file, int position, int what, int method, int nomem)
{
#ifdef MAUNUSED
 (void)what;
#endif
				
	char cstr[256];
	
	if(nomem)
	{
		// SEEK_SET 0, SEEK_CUR 1, SEEK_END 2	
		if(fseek(file, position, method))
		{
			sprintf(cstr,  "maSetFilePointer fseek file=x%x method=%d position=%d",
														(int)file, method, position);
			memShowErr(cstr);
			return -1;
		}
	
#ifdef ASYNCREAD
		g_ar_offset = position;
#endif // ASYNCREAD
	
		return ftell(file); // ok		
	}
	else
	{
		int j = (int)file;
		if(j < 1 || j >= MAXMEMFP)
		{
			sprintf(cstr, "maSetFilePointer (g_usesongmem) bad memfile pointer=%d", (int)file);
			memShowErr(cstr);
			return -1;
		}

		switch (method)
		{
			case SEEK_SET :
				g_memfp[j].songpos = g_memfp[j].songbeg + position;
			break;	

			case SEEK_CUR :
				g_memfp[j].songpos += position;
			break;	

			case SEEK_END : // position s.b. <= 0
				g_memfp[j].songpos = g_memfp[j].songend + position;
			break;
			
			default:
				memShowErr("maSetFilePointer fseek bad method");
				return -1;
				
		} // switch

		if(g_memfp[j].songpos  > g_memfp[j].songend)
			g_memfp[j].songpos  = g_memfp[j].songend - 1;
		if(g_memfp[j].songpos  < g_memfp[j].songbeg)
			g_memfp[j].songpos  = g_memfp[j].songbeg;

		return g_memfp[j].songpos - g_memfp[j].songbeg; // memftell

	} // if(!nomem)
	
} // maSetFilePointer


int maGetFileSize(FILE *file, int what, int nomem)
{
#ifdef MAUNUSED
 (void)what;
#endif

	int size;
	
	if(nomem)
	{	
		int current;
	
		if((current = ftell(file)) == -1)
		{
			memShowErr("maGetFileSize ftell failed");
			return 0;
		}

		fseek(file, 0, SEEK_END);
		size = ftell(file);
		fseek(file, current, SEEK_SET);
	
		return size;		
	}
	else
	{	
		int current;
	
		if((current = memftell(file)) == -1)
		{
			memShowErr("maGetFileSize memftell failed");
			return 0;
		}

		memfseek(file, 0, SEEK_END);
		size = memftell(file);
		memfseek(file, current, SEEK_SET);
	
		return size;
		
	} // if(!nomem)
	
} // maGetFileSize


#ifdef __cplusplus
}
#endif

// eof
