// maMemFile.h

/*
 * MoreAmp
 * Copyright (C) 2004-2010 moreforge
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#ifndef _MAMEMFILE_H_
#define _MAMEMFILE_H_

#ifdef __MAWIN__
 #pragma warning( disable : 4115 ) // named type definition in parentheses
#endif

#ifdef __cplusplus
extern "C"
{
#endif

#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>

#ifdef __MAWIN__
 #include <sys/types.h>
#else
 #include <unistd.h>
#endif

extern int g_noamem;
extern int g_nobmem;
extern int g_noimem;

// mam* functions use g_noamem 
FILE *mamfopen(const char *filename, const char *mode);
int mamfseek(FILE *stream, long offset, int whence);
int mamfseeko(FILE *stream, off_t offset, int whence);
int mamlseek(int stream, off_t offset, int whence);
size_t mamfread(void *ptr, size_t size, size_t nmemb, FILE *stream);
int mamfclose(FILE *stream);
long int mamftell(FILE *stream);
off_t mamftello(FILE *stream);
void mamrewind(FILE *stream);
int mamread(int fd, char *ptr, int size);
int mamstat(const char *filename, struct stat *buf);
int mamfstat(int j, struct stat *buf);
int mamfileno(FILE *stream);
int mamfeof(FILE *stream);
int mamfgetpos(FILE *stream, fpos_t *pos);
int mamfsetpos(FILE *stream, const fpos_t *pos);
int mamferror(FILE *stream);
void mamclearerr(FILE *stream);

// mbm* functions use g_nobmem 
FILE *mbmfopen(const char *filename, const char *mode);
int mbmfseek(FILE *stream, long offset, int whence);
int mbmfseeko(FILE *stream, off_t offset, int whence);
int mbmlseek(int stream, off_t offset, int whence);
size_t mbmfread(void *ptr, size_t size, size_t nmemb, FILE *stream);
int mbmfclose(FILE *stream);
long int mbmftell(FILE *stream);
off_t mbmftello(FILE *stream);
void mbmrewind(FILE *stream);
int mbmread(int fd, char *ptr, int size);
int mbmstat(const char *filename, struct stat *buf);
int mbmfstat(int j, struct stat *buf);
int mbmfeof(FILE *stream);
int mbmfgetpos(FILE *stream, fpos_t *pos);
int mbmfsetpos(FILE *stream, const fpos_t *pos);
int mbmferror(FILE *stream);
void mbmclearerr(FILE *stream);

// mim* functions use g_noimem
FILE *mimfopen(const char *filename, const char *mode);
int mimfseek(FILE *stream, long offset, int whence);
int mimlseek(int stream, off_t offset, int whence);
size_t mimfread(void *ptr, size_t size, size_t nmemb, FILE *stream);
int mimfclose(FILE *stream);
long int mimftell(FILE *stream);
void mimrewind(FILE *stream);
int mimread(int fd, char *ptr, int size);
int mimstat(const char *filename, struct stat *buf);
int mimfstat(int j, struct stat *buf);
int mimfeof(FILE *stream);
int mimfgetpos(FILE *stream, fpos_t *pos);
int mimfsetpos(FILE *stream, const fpos_t *pos);
int mimferror(FILE *stream);
void mimclearerr(FILE *stream);

// mem* functions operate on files in memory
FILE *memfopen(const char *filename, const char *mode);
int memfseek(FILE *stream, long offset, int whence);
int memlseek(int stream, off_t offset, int whence);
size_t memfread(void *ptr, size_t size, size_t nmemb, FILE *stream);
int memfclose(FILE *stream);
long int memftell(FILE *stream);
void memrewind(FILE *stream);
int memopen(char *filename, int mode);
int memread(int fd, char *ptr, int size);
int memfstat(int j, struct stat *buf);
int memfeof(FILE *stream);
int memferror(FILE *stream);
int memfileno(FILE *stream);
void memclearerr(FILE *stream);
int memfgetpos(FILE *stream, fpos_t *pos);
int memfsetpos(FILE *stream, const fpos_t *pos);

int maSetFilePointer(FILE * file, int position, int what, int method, int nomem);
int maGetFileSize(FILE * file, int what, int nomem);
int maReadFile(FILE *file, unsigned char *buffer, int bytes, int *pbytes, int what, int nomem);
int maWriteFile(FILE *file, unsigned char *buffer, int bytes, int *pbytes, int what);

FILE *maCreateFile(char *path, int what1, int what2, int what3,
															char *beg, char *end, int nomem);

#ifdef __cplusplus
}
#endif

#endif // _MAMEMFILE_H_

// eof
