// maPlayAac.c

/*
 * MoreAmp
 * Copyright (C) 2004-2010 moreforge
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include "maConfig.h" // included by all moreamp files

#ifndef NODECODEAAC

#ifdef __MAWIN__
 #ifdef MAUNUSEDWIN
  #pragma warning( disable : 4100 ) // unreferenced formal parameter
 #endif
 #pragma warning( disable : 4127 ) // conditional expression is constant
 #pragma warning( disable : 4201 ) // nonstandard extension used : nameless struct/union
 #pragma warning( disable : 4244 ) // conversion from 'double ' to 'int ', possible loss of data
 #pragma warning( disable : 4701 ) // local variable 'mp4ASC' may be used without having been initialized
 #pragma warning( disable : 4706 ) // assignment within conditional expression
 #pragma warning( disable : 4761 ) // integral size mismatch in argument; conversion supplied
 #define __RPCASYNC_H__ // rpcasync.h(45) : warning C4115: '_RPC_ASYNC_STATE' : named type definition in parentheses
#endif

#ifdef __cplusplus
extern "C"
{
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#ifdef __MAWIN__
 #include <stdlib.h>
 #include <limits.h>
 #include <time.h>
#endif

#ifdef __MAUNIX__
 #include "time.h"
#endif

#if defined(__MAOSX__)
 #include <MacTypes.h>
#endif

#include "maDefs.h"
#include "maPlayer.h"
#include "maUtils.h"
#include "maPlayAac.h"
#include "maEqualizer.h"
#include "maMemFile.h"

#ifdef __MAWIN__
 #include "maAudioWin.h"
#endif

#ifdef __MAOSX__
 #include "maAudioMac.h"
#endif

#ifdef __MAUNIX__
 #include "maAudioUnix.h"
#endif

#include "faad.h"

typedef struct
{
	long bytes_into_buffer;
	long bytes_consumed;
	long file_offset;
	long fileread;
	unsigned char *buffer;
	int at_eof;
	FILE *infilefp;
} aac_buffer;

typedef struct
{
	faacDecHandle hcodec;
	aac_buffer b;
	faacDecFrameInfo frameInfo;
	faacDecConfigurationPtr config;
	MASTATE *pstate;
} AacContext;

int aaccurrentpossec();
void maPlayAac();
int AacOpenFile(AacContext *paac);
void AacCloseFile(AacContext *paac);

void aacadvance_buffer(aac_buffer *b, int bytes);
int aac_fill_buffer(AacContext *paac);
int aacadts_parse(AacContext *paac, int *bitrate, float *length);

AacContext g_aaccontext;
AacContext *g_paacpcontext;

long g_aacseek = -1;
void aacseek(int pos)
{
	if(g_paacpcontext == NULL || g_aacseek >= 0 || pos < 0)
		return;

	g_aacseek = pos;	

} // aacseek


#define MAX_CHANNELS 6 /* make this higher to support files with more channels */

void maPlayAac()
{
	void *sample_buffer;
	char cstr[256];

	if(!g_state.inpath)
	{
		maLog("maPlayAac error null filename");
		return;
	}
	
	if(g_paacpcontext)
		maLog("maPlayAac unclosed context");
	
	// format of this decoder output
#ifdef WORDS_BIGENDIAN
	g_state.inwordsbigendian = 1;
#else
	g_state.inwordsbigendian = 0;
#endif
	g_state.ininterleaved = 1;

	// open
	g_paacpcontext = &g_aaccontext;
	memset(g_paacpcontext, 0, sizeof(AacContext));
	g_paacpcontext->pstate = &g_state;
	
	if(AacOpenFile(g_paacpcontext))
	{
		sprintf(cstr, "maPlayAac AacOpenFile failed on '%s'", g_paacpcontext->pstate->inpath);
		maLog(cstr);
		g_paacpcontext = NULL;
		goto maPlayAacexit;
	}
		
	if(maOutputOpen() < 0)
	{
		maLog("maPlayAac maOutputOpen failed");
		goto maPlayAacexit;
	}

	if(!g_posendon)
		g_posend = g_state.duration;
	
	// aac decode loop
	while(1)
	{	
		ThreadYield();
		
		if(g_state.stop)
			break;
				
		if(g_aacseek >= 0) // no seek in aac raw, could seek if adts
			g_aacseek = -1;

		sample_buffer = faacDecDecode(g_paacpcontext->hcodec, &g_paacpcontext->frameInfo,
							g_paacpcontext->b.buffer, g_paacpcontext->b.bytes_into_buffer);

		/* update buffer indices */
		aacadvance_buffer(&g_paacpcontext->b, g_paacpcontext->frameInfo.bytesconsumed);

		if(g_paacpcontext->frameInfo.error > 0)
		{
			sprintf(cstr, "faacDecDecode failed : %s '%s'",
				faacDecGetErrorMessage(g_paacpcontext->frameInfo.error), g_paacpcontext->pstate->inpath);
			maLog(cstr);
			break;
		}

		if((g_paacpcontext->frameInfo.error == 0) && (g_paacpcontext->frameInfo.samples > 0))
		{
			g_state.inchannels = g_paacpcontext->frameInfo.channels;

			// Write to fifo
			if(maOutputWrite((char *)sample_buffer, g_paacpcontext->frameInfo.samples << 1, &g_state) != 0)
				break;

		}

		/* fill buffer */
		aac_fill_buffer(g_paacpcontext);

		if(g_paacpcontext->b.bytes_into_buffer == 0)
			sample_buffer = NULL; /* to make sure it stops now */

		g_state.position = LROUND(((double)g_paacpcontext->b.file_offset * g_state.duration) /
																	g_paacpcontext->b.fileread); 
		// NULL at end of stream
		if(sample_buffer == NULL)
			break;

	} // while 1
	
	// finish

	FlushFifo();

maPlayAacexit:
	if(g_paacpcontext)
	{
		AacCloseFile(g_paacpcontext);
		g_paacpcontext = NULL;
	}
	
	g_state.position = 0;
	g_aacseek = -1;
	
} // maPlayAac


int aac_fill_buffer(AacContext *paac)
{
	int bread;

	if(paac->b.bytes_consumed > 0)
	{
		if(paac->b.bytes_into_buffer)
		{
			memmove((void*)paac->b.buffer, (void*)(paac->b.buffer + paac->b.bytes_consumed),
											paac->b.bytes_into_buffer * sizeof(unsigned char));
		}

		if(!paac->b.at_eof)
		{
			if(paac->pstate->nomem)
				bread = fread((void*)(paac->b.buffer + paac->b.bytes_into_buffer), 1,
													paac->b.bytes_consumed, paac->b.infilefp);
			else
				bread = memfread((void*)(paac->b.buffer + paac->b.bytes_into_buffer), 1,
													paac->b.bytes_consumed, paac->b.infilefp);

			if(bread != paac->b.bytes_consumed)
				paac->b.at_eof = 1;

			paac->b.bytes_into_buffer += bread;
		}

		paac->b.bytes_consumed = 0;

		if(paac->b.bytes_into_buffer > 3)
		{
			if(memcmp(paac->b.buffer, "TAG", 3) == 0)
				paac->b.bytes_into_buffer = 0;
		}
		if(paac->b.bytes_into_buffer > 11)
		{
			if(memcmp(paac->b.buffer, "LYRICSBEGIN", 11) == 0)
				paac->b.bytes_into_buffer = 0;
		}
		if(paac->b.bytes_into_buffer > 8)
		{
			if(memcmp(paac->b.buffer, "APETAGEX", 8) == 0)
				paac->b.bytes_into_buffer = 0;
		}
	}

	return 1;
}

void aacadvance_buffer(aac_buffer *b, int bytes)
{
	b->file_offset += bytes;
	b->bytes_consumed = bytes;
	b->bytes_into_buffer -= bytes;
}

static int adts_sample_rates[] =
		{96000,88200,64000,48000,44100,32000,24000,22050,16000,12000,11025,8000,7350,0,0,0};

int aacadts_parse(AacContext *paac, int *bitrate, float *length)
{
	int frames, frame_length;
	int t_framelength = 0;
	int samplerate = 44100;
	float frames_per_sec, bytes_per_frame;
	aac_buffer *b = &paac->b;
	
	/* Read all frames to ensure correct time and bitrate */
	for(frames = 0; /* */; frames++)
	{
		aac_fill_buffer(paac);

		if(b->bytes_into_buffer > 7)
		{
			/* check syncword */
			if(!((b->buffer[0] == 0xFF)&&((b->buffer[1] & 0xF6) == 0xF0)))
				break;

			if(frames == 0)
				samplerate = adts_sample_rates[(b->buffer[2]&0x3c)>>2];

			frame_length = ((((unsigned int)b->buffer[3] & 0x3)) << 11)
				| (((unsigned int)b->buffer[4]) << 3) | (b->buffer[5] >> 5);

			t_framelength += frame_length;

			if(frame_length > b->bytes_into_buffer)
				break;

			aacadvance_buffer(b, frame_length);
		}
		else
		{
			break;
		}
	}

	frames_per_sec = (float)samplerate / 1024.0f;
	if(frames != 0)
		bytes_per_frame = (float)t_framelength / (float)(frames*1000);
	else
		bytes_per_frame = 0;
	*bitrate = (int)(8. * bytes_per_frame * frames_per_sec + 0.5);
	if(frames_per_sec != 0)
		*length = (float)frames / frames_per_sec;
	else
		*length = 1;

	return 0;
}


int AacOpenFile(AacContext *paac)
{
	int lnomem;
	int object_type = LC;
	int def_srate = 44100;
	int downMatrix = 0;
	int outputFormat = FAAD_FMT_16BIT;
	int old_format = 1;
	char cstr[256];

	int tagsize;
	unsigned long samplerate;
	unsigned char channels;
	int bread;
	int header_type = 0;
	int bitrate = 0;
	float length = 0;

	if(paac == NULL)
	{
		maLog((char *)"AacOpenFile null context pointer");
		return 1;
	}

	lnomem = ((g_sets.buffertype == BUFUSEMEM) && !paac->pstate->nomem) ? 0 : 1;
	
	if(lnomem)
	{
		if((paac->b.infilefp = mafopen(paac->pstate->inpath, "rb")) == NULL)
		{
			sprintf(cstr, "AacOpenFile mafopen failed on '%s'", paac->pstate->inpath);
			maLog(cstr);
			return 1;
		}
	}
	else
	{
		if((paac->b.infilefp = memfopen(paac->pstate->inpath, "rb")) == NULL)
		{
			sprintf(cstr, "AacOpenFile memfopen failed on '%s'", paac->pstate->inpath);
			maLog(cstr);
			return 1;
		}
	}

	//unsigned long cap = faacDecGetCapabilities();

	if(FileNameHasExtension(paac->pstate->inpath, EXTM4A))
		old_format = 0;
		
	if(lnomem)
	{
		fseek(paac->b.infilefp, 0, SEEK_END);
		paac->b.fileread = ftell(paac->b.infilefp);
		fseek(paac->b.infilefp, 0, SEEK_SET);
	}
	else
	{
		memfseek(paac->b.infilefp, 0, SEEK_END);
		paac->b.fileread = memftell(paac->b.infilefp);
		memfseek(paac->b.infilefp, 0, SEEK_SET);
	}

	if(!(paac->b.buffer = (unsigned char*)MAMALLOC(FAAD_MIN_STREAMSIZE * MAX_CHANNELS)))
	{
		sprintf(cstr, "AacOpenFile Memory allocation error");
		maLog(cstr);
		return 0;
	}
		
	memset(paac->b.buffer, 0, FAAD_MIN_STREAMSIZE * MAX_CHANNELS);

	if(lnomem)
		bread = fread(paac->b.buffer, 1, FAAD_MIN_STREAMSIZE * MAX_CHANNELS, paac->b.infilefp);
	else
		bread = memfread(paac->b.buffer, 1, FAAD_MIN_STREAMSIZE * MAX_CHANNELS, paac->b.infilefp);
		
	paac->b.bytes_into_buffer = bread;
	paac->b.bytes_consumed = 0;
	paac->b.file_offset = 0;

	if(bread != FAAD_MIN_STREAMSIZE * MAX_CHANNELS)
		paac->b.at_eof = 1;

	tagsize = 0;
	if(!memcmp(paac->b.buffer, "ID3", 3))
	{
		/* high bit is not used */
		tagsize = (paac->b.buffer[6] << 21) | (paac->b.buffer[7] << 14) |
						(paac->b.buffer[8] <<  7) | (paac->b.buffer[9] <<  0);

		tagsize += 10;
		aacadvance_buffer(&paac->b, tagsize);
		aac_fill_buffer(paac);
	}

	paac->hcodec = faacDecOpen();

	/* Set the default object type and samplerate */
	/* This is useful for RAW AAC files */
	paac->config = faacDecGetCurrentConfiguration(paac->hcodec);
	if(def_srate)
		paac->config->defSampleRate = def_srate;
	paac->config->defObjectType = object_type;
	paac->config->outputFormat = outputFormat;
	paac->config->downMatrix = downMatrix;
	paac->config->useOldADTSFormat = old_format;
	paac->config->dontUpSampleImplicitSBR = 1;
	faacDecSetConfiguration(paac->hcodec, paac->config);

	/* get AAC infos for printing */
	header_type = 0;
	if((paac->b.buffer[0] == 0xFF) && ((paac->b.buffer[1] & 0xF6) == 0xF0))
	{
		aacadts_parse(paac, &bitrate, &length);
		
		if(lnomem)
		{
			fseek(paac->b.infilefp, tagsize, SEEK_SET);
			bread = fread(paac->b.buffer, 1, FAAD_MIN_STREAMSIZE * MAX_CHANNELS, paac->b.infilefp);
		}
		else
		{
			memfseek(paac->b.infilefp, tagsize, SEEK_SET);
			bread = memfread(paac->b.buffer, 1, FAAD_MIN_STREAMSIZE * MAX_CHANNELS, paac->b.infilefp);
		}
			
		if(bread != FAAD_MIN_STREAMSIZE * MAX_CHANNELS)
			paac->b.at_eof = 1;
		else
			paac->b.at_eof = 0;
				
		paac->b.bytes_into_buffer = bread;
		paac->b.bytes_consumed = 0;
		paac->b.file_offset = tagsize;

		header_type = 1;
	}
	else if(memcmp(paac->b.buffer, "ADIF", 4) == 0)
	{
		int skip_size = (paac->b.buffer[4] & 0x80) ? 9 : 0;
		bitrate = ((unsigned int)(paac->b.buffer[4 + skip_size] & 0x0F)<<19) |
			((unsigned int)paac->b.buffer[5 + skip_size]<<11) |
			((unsigned int)paac->b.buffer[6 + skip_size]<<3) |
			((unsigned int)paac->b.buffer[7 + skip_size] & 0xE0);

		length = (double)paac->b.fileread;
		if(length != 0)
		{
			length = ((double)length * 8) / bitrate + 0.5;
		}

		bitrate = LROUND(((double)bitrate/1000.0 + 0.5));

		header_type = 2;
	}
	else // raw, no adts headers
	{
		//length = 60; // 1.00 minute arbitrary for display
		length = (double)paac->b.fileread / 12991; // approx value for files ecoded at q=100
	}
	
	paac->pstate->duration = LROUND(length * 1000);
	
	aac_fill_buffer(paac);
	
	if((bread = faacDecInit(paac->hcodec, paac->b.buffer,
					paac->b.bytes_into_buffer, &samplerate, &channels)) < 0)
	{
		/* If some error initializing occured, skip the file */
		sprintf(cstr, "AacOpenFile error initializing decoder library.");
		maLog(cstr);
		if(paac->b.buffer)
			MAFREE(paac->b.buffer);
		faacDecClose(paac->hcodec);
		goto aacopenerrexit;
	}
	
	aacadvance_buffer(&paac->b, bread);
	aac_fill_buffer(paac);
	paac->pstate->insamplerate = samplerate;

	paac->pstate->position = 0;

	return 0;

aacopenerrexit:
		if(lnomem)
			fclose(paac->b.infilefp);
		else
			memfclose(paac->b.infilefp);
					
		paac->b.infilefp = NULL;
		
		return 1;

} // AacOpenFile


void AacCloseFile(AacContext *paac)
{
	if(!paac)
	{
		maLog((char *)"AacCloseFile null context pointer");
		return ;
	}

	// close paac->hcodec
	if(paac->hcodec)
	{
		faacDecClose(paac->hcodec);
		paac->hcodec = NULL;
	}
		
	// close file
	if(paac->b.infilefp)
	{
		if(paac->pstate->nomem)
			fclose(paac->b.infilefp);
		else
			memfclose(paac->b.infilefp);
					
		paac->b.infilefp = NULL;
	}
	
	if(paac->b.buffer)
	{
		MAFREE(paac->b.buffer);
		paac->b.buffer = NULL;
	}

} // AacCloseFile


signed long ThreadPlayAac(void *parameter)
{
#ifdef MAUNUSED
 (void)parameter;
#endif

	g_waitaudiodone = 1;

	maPlayAac();
		
	g_state.stop |= STOP_PLAYDECODE;
	
	// wait for output owned by this thread to finish
	while(!g_stopB && g_waitaudiodone)
		maSleep(100);

	return 0;

} // ThreadPlayAac


//void DumpTagsAac(MASTATE *pstate, FILE *dumpfp, int nomem);

int DumpInfoAac(MASTATE *pstate, FILE *dumpfp)
{
	char tstr[32];
	char cstrx[512];
	
	AacContext aaccontext;
	int object_type = LC;
	int def_srate = 44100;
	int downMatrix = 0;
	int outputFormat = FAAD_FMT_16BIT;
	int old_format = 1;
	char cstr[256];

	int tagsize;
	unsigned long samplerate;
	unsigned char channels;
	int bread;
	int header_type = 0;
	int bitrate = 0;
	float length = 0;

	// open
	memset(&aaccontext, 0, sizeof(AacContext));
	aaccontext.pstate = pstate;
	
	if(pstate->nomem)
	{
		if((aaccontext.b.infilefp = mafopen(aaccontext.pstate->inpath, "rb")) == NULL)
		{
			sprintf(cstr, "DumpInfoAac mafopen failed on '%s'", aaccontext.pstate->inpath);
			maLog(cstr);
			return 1;
		}
	}
	else
	{
		if((aaccontext.b.infilefp = memfopen(aaccontext.pstate->inpath, "rb")) == NULL)
		{
			sprintf(cstr, "DumpInfoAac memfopen failed on '%s'", aaccontext.pstate->inpath);
			maLog(cstr);
			return 1;
		}
	}

	if(FileNameHasExtension(aaccontext.pstate->inpath, EXTM4A))
		old_format = 0;
		
	if(pstate->nomem)
	{
		fseek(aaccontext.b.infilefp, 0, SEEK_END);
		aaccontext.b.fileread = ftell(aaccontext.b.infilefp);
		fseek(aaccontext.b.infilefp, 0, SEEK_SET);
	}
	else
	{
		memfseek(aaccontext.b.infilefp, 0, SEEK_END);
		aaccontext.b.fileread = memftell(aaccontext.b.infilefp);
		memfseek(aaccontext.b.infilefp, 0, SEEK_SET);
	}

	if(!(aaccontext.b.buffer = (unsigned char*)MAMALLOC(FAAD_MIN_STREAMSIZE * MAX_CHANNELS)))
	{
		sprintf(cstr, "DumpInfoAac Memory allocation error");
		maLog(cstr);
		return 0;
	}
		
	memset(aaccontext.b.buffer, 0, FAAD_MIN_STREAMSIZE * MAX_CHANNELS);

	if(pstate->nomem)
		bread = fread(aaccontext.b.buffer, 1, FAAD_MIN_STREAMSIZE * MAX_CHANNELS, aaccontext.b.infilefp);
	else
		bread = memfread(aaccontext.b.buffer, 1, FAAD_MIN_STREAMSIZE * MAX_CHANNELS, aaccontext.b.infilefp);
		
	aaccontext.b.bytes_into_buffer = bread;
	aaccontext.b.bytes_consumed = 0;
	aaccontext.b.file_offset = 0;

	if(bread != FAAD_MIN_STREAMSIZE * MAX_CHANNELS)
		aaccontext.b.at_eof = 1;

	tagsize = 0;
	if(!memcmp(aaccontext.b.buffer, "ID3", 3))
	{
		/* high bit is not used */
		tagsize = (aaccontext.b.buffer[6] << 21) | (aaccontext.b.buffer[7] << 14) |
						(aaccontext.b.buffer[8] <<  7) | (aaccontext.b.buffer[9] <<  0);

		tagsize += 10;
		aacadvance_buffer(&aaccontext.b, tagsize);
		aac_fill_buffer(&aaccontext);
	}

	aaccontext.hcodec = faacDecOpen();

	aaccontext.config = faacDecGetCurrentConfiguration(aaccontext.hcodec);
	if(def_srate)
		aaccontext.config->defSampleRate = def_srate;
	aaccontext.config->defObjectType = object_type;
	aaccontext.config->outputFormat = outputFormat;
	aaccontext.config->downMatrix = downMatrix;
	aaccontext.config->useOldADTSFormat = old_format;
	aaccontext.config->dontUpSampleImplicitSBR = 1;
	faacDecSetConfiguration(aaccontext.hcodec, aaccontext.config);

	/* get AAC infos for printing */
	header_type = 0;
	if((aaccontext.b.buffer[0] == 0xFF) && ((aaccontext.b.buffer[1] & 0xF6) == 0xF0))
	{
		aacadts_parse(&aaccontext, &bitrate, &length);
		
		if(pstate->nomem)
		{
			fseek(aaccontext.b.infilefp, tagsize, SEEK_SET);
			bread = fread(aaccontext.b.buffer, 1, FAAD_MIN_STREAMSIZE * MAX_CHANNELS, aaccontext.b.infilefp);
		}
		else
		{
			memfseek(aaccontext.b.infilefp, tagsize, SEEK_SET);
			bread = memfread(aaccontext.b.buffer, 1, FAAD_MIN_STREAMSIZE * MAX_CHANNELS, aaccontext.b.infilefp);
		}
			
		if(bread != FAAD_MIN_STREAMSIZE * MAX_CHANNELS)
			aaccontext.b.at_eof = 1;
		else
			aaccontext.b.at_eof = 0;
				
		aaccontext.b.bytes_into_buffer = bread;
		aaccontext.b.bytes_consumed = 0;
		aaccontext.b.file_offset = tagsize;

		header_type = 1;
	}
	else if(memcmp(aaccontext.b.buffer, "ADIF", 4) == 0)
	{
		int skip_size = (aaccontext.b.buffer[4] & 0x80) ? 9 : 0;
		bitrate = ((unsigned int)(aaccontext.b.buffer[4 + skip_size] & 0x0F)<<19) |
			((unsigned int)aaccontext.b.buffer[5 + skip_size]<<11) |
			((unsigned int)aaccontext.b.buffer[6 + skip_size]<<3) |
			((unsigned int)aaccontext.b.buffer[7 + skip_size] & 0xE0);

		length = (double)aaccontext.b.fileread;
		if(length != 0)
		{
			length = ((double)length * 8) / bitrate + 0.5;
		}

		bitrate = LROUND(((double)bitrate/1000.0 + 0.5));

		header_type = 2;
	}
	else // raw, no adts headers
	{
		//length = 60; // 1.00 minute arbitrary for display
		length = (double)aaccontext.b.fileread / 12991; // approx value for files ecoded at q=100
	}
	
	pstate->duration = LROUND(length);
	
	aac_fill_buffer(&aaccontext);
	
	if((bread = faacDecInit(aaccontext.hcodec, aaccontext.b.buffer,
					aaccontext.b.bytes_into_buffer, &samplerate, &channels)) < 0)
	{
		sprintf(cstr, "DumpInfoAac error initializing decoder library.");
		maLog(cstr);
	}
	
	if(aaccontext.b.buffer)
		MAFREE(aaccontext.b.buffer);
		
	faacDecClose(aaccontext.hcodec);

	if(pstate->nomem)
		fclose(aaccontext.b.infilefp);
	else
		memfclose(aaccontext.b.infilefp);
					
	sprintf(cstrx, "Duration : %s", SecondsToStrHMS(tstr, pstate->duration));
	DumpString(dumpfp, cstrx);
	
	sprintf(cstrx, "Channels : %d", channels);
	DumpString(dumpfp, cstrx);
	
	sprintf(cstrx, "Bitrate : %d", bitrate * 1000);
	DumpString(dumpfp, cstrx);
	
	sprintf(cstrx, "Samplerate : %lu", samplerate);
	DumpString(dumpfp, cstrx);
	
	//DumpTagsAac(songfpath, *dumpfp, pstate->nomem);

	return 0;
	
} // DumpInfoAac


#ifdef __cplusplus
}
#endif

#endif

// eof
