// maPlayMp3.c

/*
 * MoreAmp
 * Copyright (C) 2004-2010 moreforge
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

// this code contains extracts from madplay/mad winamp :
/*
 * madplay - MPEG audio decoder and player
 * Copyright (C) 2000-2004 Robert Leslie
 */

#include "maConfig.h" // included by all moreamp files

#ifdef __MAWIN__
 #ifdef MAUNUSEDWIN
  #pragma warning( disable : 4100 ) // unreferenced formal parameter
 #endif
 #pragma warning( disable : 4706 ) // assignment within conditional expression
 #pragma warning( disable : 4127 ) // conditional expression is constant
 #define __RPCASYNC_H__ // rpcasync.h(45) : warning C4115: '_RPC_ASYNC_STATE' : named type definition in parentheses
#endif

#ifndef NODECODEMP3

#ifdef __cplusplus
extern "C"
{
#endif

#include <stdio.h>
#include <fcntl.h>
#include <stdarg.h>
#include <string.h>
#include <locale.h>
#include <math.h>

#if defined(__MAOSX__)
 #include <errno.h>
 #include <Files.h>
#endif

#include "maEqualizer.h"

#if defined(__MAWIN__)
 #include "../libs/libmad/msvc++/mad.h"
#elif (defined(__MACUNIX__) || defined(__MAOSX__)) && !defined(USE_LIBMAD_SYSTEM)
 #if defined(MA_CPU_I86)
  #include "../libs/libmad/osxi/mad.h"
 #else
  #include "../libs/libmad/osx/mad.h"
 #endif
#else
 #include "mad.h"
#endif
#ifndef NOID3TAG
 #include "id3tag.h"
#endif
#include "maPlayMp3.h"
#include "maMemFile.h"
#include "maPlayer.h"
#include "maUtils.h"

#ifdef __MAWIN__
 #include "maAudioWin.h"
#endif

#ifdef __MAOSX__
 #include "maAudioMac.h"
#endif

#ifdef __MAUNIX__
 #include "maAudioUnix.h"
#endif

// global
unsigned char *g_mp3inpbuf = NULL;
unsigned char *g_mp3decbuf = NULL;

// local to this module
#define RESOLUTION 16
#define BYTES 2
#define PCM_CHUNK 576
#define NEW_DITHER

typedef struct
{
  long flags;				// valid fields (see below)
  unsigned long frames;		// total number of frames
  unsigned long bytes;		// total number of bytes
  unsigned char toc[100];	// 100-point seek table
  long scale;				// ??
} XING;

enum
{
  XING_FRAMES = 0x00000001L,
  XING_BYTES  = 0x00000002L,
  XING_TOC    = 0x00000004L,
  XING_SCALE  = 0x00000008L
};

int madscanheader(FILE *madfilep, struct mad_header *header, XING *xing, int nomem, char *inpath);
int xing_parse(XING *, struct mad_bitptr, unsigned int);
void madattenfilter(struct mad_frame *frame, mad_fixed_t scalefactor);
void madeqfilter(struct mad_frame *frame, mad_fixed_t eqfactor);

#ifdef NEW_DITHER
 struct dither
 {
	mad_fixed_t error[3];
	mad_fixed_t random;
 };
 unsigned long prng(unsigned long state);
 signed long madlineardither(mad_fixed_t sample, struct dither *dither,
								unsigned long *clipped, mad_fixed_t *clipping);
#else
 signed long madlineardither(mad_fixed_t sample, mad_fixed_t *error,
								unsigned long *clipped, mad_fixed_t *clipping);
#endif
int maderror(struct mad_stream *stream, struct mad_frame *frame, int *last_error);
mad_fixed_t madeqfactor(double db);

#if !defined(__MAOSX__)
 #ifndef OSErr
  #define OSErr signed long
 #endif
 #ifndef noErr
  #define noErr 0
 #endif
#endif

typedef struct
{
	mad_fixed_t madattenuation;	// mad attenuation factor
	int madequalizer;			// mad equalizer on or off
	mad_fixed_t madeqfactor;	// mad equalizer settings (used as preamp)
} MADSTATE;

MADSTATE g_madstate;

typedef struct
{
	XING mp3xing;
	struct mad_header mp3header;
} LSTATE;

LSTATE l_lstate;


void madattenfilter(struct mad_frame *frame, mad_fixed_t scalefactor)
{
	unsigned int nch, ch, ns, s, sb;

	nch = MAD_NCHANNELS(&frame->header);
	ns	= MAD_NSBSAMPLES(&frame->header);

	for(ch = 0; ch < nch; ++ch)
	{
		for(s = 0; s < ns; ++s)
		{
			for(sb = 0; sb < 32; ++sb)
			{
				frame->sbsample[ch][s][sb] =
				mad_f_mul(frame->sbsample[ch][s][sb], scalefactor);
			}
		}
	}
} // madattenfilter


void madeqfilter(struct mad_frame *frame, mad_fixed_t eqfactor)
{
	unsigned int nch, ch, ns, s, sb;

	nch = MAD_NCHANNELS(&frame->header);
	ns	= MAD_NSBSAMPLES(&frame->header);

	for(ch = 0; ch < nch; ++ch)
	{
		for(s = 0; s < ns; ++s)
		{
			for(sb = 0; sb < 32; ++sb)
			{
				frame->sbsample[ch][s][sb] =
							mad_f_mul(frame->sbsample[ch][s][sb], eqfactor/*[sb]*/);
			}
		}
	}

} // madeqfilter


#ifdef NEW_DITHER
// 32-bit pseudo-random number generator
unsigned long prng(unsigned long ulstate)
{
	return (ulstate * 0x0019660dL + 0x3c6ef35fL) & 0xffffffffL;
}


signed long madlineardither(mad_fixed_t sample, struct dither *dither,
							unsigned long *clipped, mad_fixed_t *clipping)
{
	unsigned int scalebits;
	mad_fixed_t output, mask, random;

	enum {
			MADMIN = -MAD_F_ONE,
			MADMAX = MAD_F_ONE - 1
	};

	// noise shape
	sample += dither->error[0] - dither->error[1] + dither->error[2];

	dither->error[2] = dither->error[1];
	dither->error[1] = dither->error[0] / 2;

	// bias
	output = sample + (1L << (MAD_F_FRACBITS + 1 - RESOLUTION - 1));

	scalebits = MAD_F_FRACBITS + 1 - RESOLUTION;
	mask = (1L << scalebits) - 1;

	// dither
	random	= prng(dither->random);
	output += (random & mask) - (dither->random & mask);

	dither->random = random;

	// clip
	if(output > MADMAX)
	{
		if(g_madstate.madequalizer & CLIPPING)
		{
			++*clipped;
			if(output - MADMAX > *clipping)
				*clipping = output - MADMAX;

			output = MADMAX;
		}
		
		if(sample > MADMAX)
			sample = MADMAX;
	}
	else if(output < MADMIN)
	{
		if(g_madstate.madequalizer & CLIPPING)
		{
			++*clipped;
			if(MADMIN - output > *clipping)
				*clipping = MADMIN - output;

			output = MADMIN;
		}
		
		if(sample < MADMIN)
			sample = MADMIN;
	}

	// quantize
	output &= ~mask;

	// error feedback
	dither->error[0] = sample - output;

	// scale
	return output >> scalebits;
}
#else
signed long madlineardither(mad_fixed_t sample, mad_fixed_t *error,
							unsigned long *clipped, mad_fixed_t *clipping)
{
	mad_fixed_t quantized, check;

	// dither
	sample += *error;

	// clip
	quantized = sample;
	check = (sample >> MAD_F_FRACBITS) + 1;
	if(check & ~1)
	{
		if(state->equalizer & CLIPPING)
		{
			if(sample >= MAD_F_ONE)
			{
				quantized = MAD_F_ONE - 1;
				++*clipped;
				if(sample - quantized > *clipping && mad_f_abs(*error) <
								(MAD_F_ONE >> (MAD_F_FRACBITS + 1 - RESOLUTION)))
					*clipping = sample - quantized;
			}
			else if(sample < -MAD_F_ONE)
			{
				quantized = -MAD_F_ONE;
				++*clipped;
				if(quantized - sample > *clipping && mad_f_abs(*error) <
								(MAD_F_ONE >> (MAD_F_FRACBITS + 1 - RESOLUTION)))
					*clipping = quantized - sample;
			}
		}
	  	else
  		{
  			if(sample >= MAD_F_ONE)
  				quantized = MAD_F_ONE - 1;
  			else if(sample < -MAD_F_ONE)
				quantized = -MAD_F_ONE;
	  	}
	}

	// quantize
	quantized &= ~((1L << (MAD_F_FRACBITS + 1 - RESOLUTION)) - 1);

	// error
	*error = sample - quantized;

	// scale
	return quantized >> (MAD_F_FRACBITS + 1 - RESOLUTION);
}
#endif // !NEW_DITHER


#if defined(__MAOSX__)
 #define _MADPACKPCM_
#endif
#ifdef __MAUNIX__
 #define _MADPACKPCM_
#endif
#ifdef _MADPACKPCM_
/*
signed long no_dither(mad_fixed_t sample, struct dither *dither,
						unsigned long *clipped, mad_fixed_t *clipping);
signed long no_dither(mad_fixed_t sample, struct dither *dither,
						unsigned long *clipped, mad_fixed_t *clipping)
{
#ifdef MAUNUSED
 (void)dither;
#endif

	unsigned int scalebits;
	mad_fixed_t output, mask;

	enum {
			MADMIN = -MAD_F_ONE,
			MADMAX = MAD_F_ONE - 1
	};

	output = sample + (1L << (MAD_F_FRACBITS + 1 - RESOLUTION - 1));

	scalebits = MAD_F_FRACBITS + 1 - RESOLUTION;
	mask = (1L << scalebits) - 1;

	// clip
	if(output > MADMAX)
	{
		if(g_madstate.madequalizer & CLIPPING)
		{
			++*clipped;
			if(output - MADMAX > *clipping)
				*clipping = output - MADMAX;

			output = MADMAX;
		}
		
		if(sample > MADMAX)
			sample = MADMAX;
	}
	else if(output < MADMIN)
	{
		if(g_madstate.madequalizer & CLIPPING)
		{
			++*clipped;
			if(MADMIN - output > *clipping)
				*clipping = MADMIN - output;

			output = MADMIN;
		}
		
		if(sample < MADMIN)
			sample = MADMIN;
	}
  	else
  	{
  		if(output > MADMAX)
  			output = MADMAX;
  		else if(output < MADMIN)
  			output = MADMIN;

  		if(sample > MADMAX)
  			sample = MADMAX;
  		else if(sample < MADMIN)
  			sample = MADMIN;
  	}

	// quantize
	output &= ~mask;

	// scale
	return output >> scalebits;
	
} // no_dither
*/


unsigned int madpackpcm(unsigned char *data, unsigned int nsamples,
						mad_fixed_t const *left, mad_fixed_t const *right,
						unsigned long *clipped, mad_fixed_t *clipping);
unsigned int madpackpcm(unsigned char *data, unsigned int nsamples,
						mad_fixed_t const *left, mad_fixed_t const *right,
						unsigned long *clipped, mad_fixed_t *clipping)
{
 #ifdef NEW_DITHER
	static struct dither left_dither, right_dither;
 #else
	static mad_fixed_t left_dither, right_dither;
 #endif
	unsigned char const *start;
	register signed long sample0, sample1;

	start = data;

	//if(g_sets.mp3dither)
	//{
		if(right)
		{   // stereo
			while(nsamples--)
			{
				sample0 = madlineardither(*left++, &left_dither, clipped, clipping);
				sample1 = madlineardither(*right++, &right_dither, clipped, clipping);			
 #ifndef WORDS_BIGENDIAN
						data[        1] = sample0 >> 8; // k16BitLittleEndianFormat
						data[BYTES + 1] = sample1 >> 8;
						data[        0] = sample0;
						data[BYTES + 0] = sample1;
 #else
						data[        0] = sample0 >> 8; // k16BitBigEndianFormat
						data[BYTES + 0] = sample1 >> 8;
						data[        1] = sample0;
						data[BYTES + 1] = sample1;
 #endif
				data += BYTES * 2;
			}
		}
		else
		{   // mono
			while(nsamples--)
			{
				sample0 = madlineardither(*left++, &left_dither, clipped, clipping);
 #ifndef WORDS_BIGENDIAN
						data[1] = sample0 >> 8;
						data[0] = sample0;
 #else
						data[0] = sample0 >> 8;
						data[1] = sample0;
 #endif
				data += BYTES;
			}
		}
	/*
	}
	else // AUDIO_MODE_ROUND
	{
		if(right)
		{   // stereo
			while(nsamples--)
			{
				sample0 = no_dither(*left++, &left_dither, clipped, clipping);
				sample1 = no_dither(*right++, &right_dither, clipped, clipping);			
 #ifndef WORDS_BIGENDIAN						
						data[        1] = sample0 >> 8; // k16BitLittleEndianFormat
						data[BYTES + 1] = sample1 >> 8;
						data[        0] = sample0;
						data[BYTES + 0] = sample1;
 #else
						data[        0] = sample0 >> 8; // k16BitBigEndianFormat
						data[BYTES + 0] = sample1 >> 8;
						data[        1] = sample0;
						data[BYTES + 1] = sample1;
 #endif
				data += BYTES * 2;
			}
		}
		else
		{   // mono
			while(nsamples--)
			{
				sample0 = no_dither(*left++, &left_dither, clipped, clipping);
 #ifndef WORDS_BIGENDIAN
						data[1] = sample0 >> 8;
						data[0] = sample0;
 #else
						data[0] = sample0 >> 8;
						data[1] = sample0;
 #endif
				data += BYTES;
			}
		}
	}
	*/
		
	return data - start;
  
} // madpackpcm
#endif // _MADPACKPCM_


#ifdef __MAWIN__
unsigned int madpackpcm(unsigned char *data, unsigned int nsamples,
						mad_fixed_t const *left, mad_fixed_t const *right,
						unsigned long *clipped, mad_fixed_t *clipping)
{
 #ifdef NEW_DITHER
	static struct dither left_dither, right_dither;
 #else
	static mad_fixed_t left_dither, right_dither;
 #endif
	unsigned char const *start;
	register signed long sample0, sample1;

	start = data;

	if(right)
	{	// stereo
    	while(nsamples--)
		{
			sample0 = madlineardither(*left++, &left_dither, clipped, clipping);
			sample1 = madlineardither(*right++, &right_dither, clipped, clipping);

			data[        1] = (unsigned char)(sample0 >> 8);
			data[BYTES + 1] = (unsigned char)(sample1 >> 8);
			data[        0] = (unsigned char)sample0;
			data[BYTES + 0] = (unsigned char)sample1;

			data += BYTES * 2;
		}
	}
	else
	{	// mono
		while(nsamples--)
		{
			sample0 = madlineardither(*left++, &left_dither, clipped, clipping);

			data[1] = (unsigned char)(sample0 >> 8);
			data[0] = (unsigned char)sample0;

			data += BYTES;
		}
	}

	return data - start;
}
#endif


int maderror(struct mad_stream *stream, struct mad_frame *frame, int *last_error)
{
	int do_continue = 1;
#ifndef NOID3TAG
	signed long tagsize;
#endif

	switch (stream->error)
	{
		case MAD_ERROR_BADCRC:
			if(last_error)
			{
				if(*last_error)
				{
					if(frame)
						mad_frame_mute(frame);
				}
				else
					*last_error = 1;
			}

			do_continue = 0;
		break;

		case MAD_ERROR_LOSTSYNC:
#ifndef NOID3TAG
			tagsize = id3_tag_query(stream->this_frame, stream->bufend - stream->this_frame);
			if(tagsize > 0)
				mad_stream_skip(stream, tagsize);
			else
#endif
		break;

		default:
		break;
	}

	return do_continue;
	
} // maderror


int madscanheader(FILE *madfilep, struct mad_header *header, XING *xing, int nomem, char *inpath)
{
	struct mad_stream stream;
	struct mad_frame frame;
	unsigned char buffer[8192]; // not necessarily MPEG_BUFSZ
	unsigned int buflen = 0;
	int count = 0, result = 0;
	int bytes;
	char cstr[64 + MAX_PATH];

	mad_stream_init(&stream);
	mad_frame_init(&frame);

	if(xing)
		xing->flags = 0;

	while(1)
	{
		if(buflen < sizeof(buffer))
		{
			if(maReadFile(madfilep, buffer + buflen, sizeof(buffer) - buflen,
																&bytes, 0, nomem) == 0)
			{
				result = -1;
				if(inpath)
				{
					sprintf(cstr, "madscanheader maReadFile 1 failed on '%s'", inpath);
					maLog(cstr);
				}
				break;
			}

			if(bytes <= 0)
			{
				// if(inpath)
				// {
				//	sprintf(cstr, "madscanheader maReadFile 2 failed on '%s'", inpath);
				//	maLog(cstr); //, IDS_WARN_NOHEADER);
				// }
				break;
			}

			buflen += bytes;
		}

		mad_stream_buffer(&stream, buffer, buflen);

		while(1)
		{
			if(mad_frame_decode(&frame, &stream) == -1)
			{
				if(!MAD_RECOVERABLE(stream.error))
					break;

				if(maderror(&stream, 0, 0))
					continue;
			}

			if(count++ || (xing && xing_parse(xing, stream.anc_ptr, stream.anc_bitlen) == -1))
				break;
		}

		if(count || stream.error != MAD_ERROR_BUFLEN)
			break;

		memmove(buffer, stream.next_frame,
		buflen = &buffer[buflen] - stream.next_frame);
	}

	if(count)
	{
		if(header)
			*header = frame.header;
	}
	else
	{
		result = -1;
		if(inpath)
		{
			sprintf(cstr, "madscanheader maReadFile 3 failed on '%s'", inpath);
			maLog(cstr); //, IDS_WARN_NOHEADER);
		}	
	}
	
	mad_frame_finish(&frame);
	mad_stream_finish(&stream);

	return result;
	
} // madscanheader


void maPlayMp3()
{
	mad_fixed_t clipping;
	unsigned long clipped;
	FILE *madfilep;
	struct mad_header header;
	XING xing;
	int eof;
	int bytes;
	int seek_skip;
	unsigned int input_length;
	unsigned int output_length;
	unsigned int input_size;
	struct mad_stream stream;
	struct mad_frame frame;
	struct mad_synth synth;
	mad_timer_t timer;
	mad_timer_t rduration;
	unsigned char *input_buffer;
	unsigned char *output_buffer;
	int last_error;
	mad_fixed_t const *ch1;
	mad_fixed_t const *ch2;
	int attsensitivity;
	
	char cstr[256];

	if(!g_state.inpath || !g_state.inpath[0])
	{
		maLog("maPlayMp3 error null filename");
		return;
	}

	// format of this decoder output
#ifdef WORDS_BIGENDIAN
	g_state.inwordsbigendian = 1;
#else
	g_state.inwordsbigendian = 0;
#endif
	g_state.ininterleaved = 1;
	
	attsensitivity = MAD_F(0x0c000000);	// auto attenuation sensitivity
	g_madstate.madattenuation = MAD_F_ONE;
	MadDoParamChange();

	if((madfilep = maCreateFile(g_state.inpath, 0, 0, 0, 0, 0, g_state.nomem)) == NULL)
	{
		sprintf(cstr, "maPlayMp3 maCreateFile failed on '%s'", g_state.inpath);
		maLog(cstr);
		return;
	}

	if(madscanheader(madfilep, &header, &xing, g_state.nomem, g_state.inpath) == -1)
	{
		sprintf(cstr, "maPlayMp3 madscanheader failed on '%s'", g_state.inpath);
		maLog(cstr);
		goto playmp3exit;
	}

	maSetFilePointer(madfilep, 0, 0, SEEK_SET, g_state.nomem);

	g_state.size = maGetFileSize(madfilep, 0, g_state.nomem);

	g_state.bitrate = 0;

	if(xing.flags & XING_FRAMES)
	{
		mad_timer_t timer;
		timer = header.duration;
		mad_timer_multiply(&timer, xing.frames);
		g_state.duration = mad_timer_count(timer, MAD_UNITS_MILLISECONDS);
		if((xing.flags & XING_BYTES) && g_state.duration > 0)
				g_state.bitrate = xing.bytes * 8 / g_state.duration;
	}
	else
	{
		if(header.bitrate <= 0)
		{
			sprintf(cstr, "maPlayMp3 zero bitrate on '%s'", g_state.inpath);
			maLog(cstr);
			goto playmp3exit;
		}
		g_state.duration = (long)((g_state.size * 8 / (header.bitrate / 1000)));
	}
	
	g_state.position = 0;
	
	if(g_posbeg > 0)
		g_state.seek = g_posbeg; //mp3seek(g_posbeg);
	else
		g_state.seek = -1;
		
	g_state.stop = 0;
	
	g_state.insamplerate = header.samplerate;
	
	if(maOutputOpen() < 0)
	{
		maLog("maPlayMp3 maOutputOpen failed");
		goto playmp3exit;
	}

	// mp3 decoder input buffer
	input_size = MP3INPBUFSIZE;
	
	if((input_buffer = g_mp3inpbuf) == NULL)
	{
		maLog("maPlayMp3 MAMALLOC g_mp3inpbuf failed");
		goto playmp3exit;
	}
		
	// mp3 decoder output buffer
	if((output_buffer = g_mp3decbuf) == NULL)
	{
		maLog("maPlayMp3 MAMALLOC g_mp3decbuf failed");
		goto playmp3exit;
	}

	mad_stream_init(&stream);
	mad_frame_init(&frame);
	mad_synth_init(&synth);

	timer = rduration = mad_timer_zero;

	clipped = 0; clipping = 0;

	// MadDecodeStream:
	eof = 0;
	input_length = 0;
	seek_skip = 0;
	last_error = 0;

	if(!g_posendon)
		g_posend = g_state.duration;

	// mp3 decode loop
	while(!eof)
	{
		ThreadYield();
		
		if(input_length < input_size / 4)
		{
			if(maReadFile(madfilep, input_buffer + input_length,
							input_size - input_length, &bytes, 0, g_state.nomem) == 0)
			{
				sprintf(cstr, "maPlayMp3 maReadFile failed on '%s'", g_state.inpath);
				maLog(cstr);
				break;
			}
			
			if(bytes == 0)
			{
				eof = 1;
				while(bytes < MAD_BUFFER_GUARD)
					input_buffer[input_length + bytes++] = 0;
			}

			input_length += bytes;
		}
		
		mad_stream_buffer(&stream, input_buffer, input_length);

		if(seek_skip)
		{
			int skip;

			skip = 2;
			do
			{			
				if(mad_frame_decode(&frame, &stream) == 0)
				{
					mad_timer_add(&timer, frame.header.duration);
					if(--skip == 0)
						mad_synth_frame(&synth, &frame);
				}
				else if(!MAD_RECOVERABLE(stream.error))
					break;
					
			} while(skip);

			//maOutputFlush(seek_skip);

			seek_skip = 0;
		}

		while(!g_state.stop)
		{
			if(g_state.seek != -1 && g_state.duration > 0)
			{
				int new_position;

				if(g_state.seek < 0)
					new_position = (int)((double) g_state.duration * -g_state.seek / 1000);
				else
					new_position = g_state.seek;

				g_state.position = new_position;
				g_state.seek = -1;

				if(maSetFilePointer(madfilep, (int)((double)new_position * g_state.size / g_state.duration),
																		0, SEEK_SET, g_state.nomem) != -1)
				{
					mad_timer_set(&timer, 0, new_position, 1000);
				
					mad_frame_mute(&frame);
					mad_synth_mute(&synth);

					input_length = stream.next_frame - &input_buffer[0];
					stream.error = MAD_ERROR_BUFLEN;
					stream.sync = 0;

					if(new_position > 0)
						seek_skip = new_position;
					//else
					//	maOutputFlush(0);
					break;
				}
				
			} // if seek

			if(mad_frame_decode(&frame, &stream) == -1)
			{
				if(!MAD_RECOVERABLE(stream.error))
					break;

				if(maderror(&stream, &frame, &last_error))
					continue;
			}
			else
				last_error = 0;
		
			if(g_madstate.madattenuation != MAD_F_ONE)
				madattenfilter(&frame, g_madstate.madattenuation);
				
			if(g_madstate.madequalizer & PREAMP)
				madeqfilter(&frame, g_madstate.madeqfactor);
				
			mad_synth_frame(&synth, &frame);
		
			g_state.inchannels = synth.pcm.channels;
			ch1 = synth.pcm.samples[0];
			if(g_state.inchannels == 1)
				ch2 = 0;
			else
				ch2 = synth.pcm.samples[1];
			
			output_length = madpackpcm(output_buffer, synth.pcm.length, ch1, ch2,
								&clipped, &clipping);

			if(clipping > 0)
			{
				g_madstate.madattenuation =
					mad_f_tofixed(mad_f_todouble(g_madstate.madattenuation) /
					mad_f_todouble(MAD_F_ONE + mad_f_mul(clipping, attsensitivity)));
				g_state.mpegattenuation = mad_f_todouble(g_madstate.madattenuation);
				clipping = 0;
			}
			else if(g_state.mpegattenuation == MP3ATTENUATIONFLAG)	// MP3ATTENUATIONFLAG set by maplayer
			{														// as a flag to reset attenuation
				g_madstate.madattenuation = MAD_F_ONE;
				g_state.mpegattenuation = 1.0;
			}
			
			if(g_state.inchannels == 1)
				mad_timer_set(&rduration, 0, output_length >> 1, synth.pcm.samplerate);
			else
				mad_timer_set(&rduration, 0, output_length >> 2, synth.pcm.samplerate);
			mad_timer_add(&timer, rduration);
			g_state.position = mad_timer_count(timer, MAD_UNITS_MILLISECONDS);

			// Write to fifo
			if(maOutputWrite((char *)output_buffer, output_length, &g_state) != 0)
			{
				g_state.stop |= STOP_FLAG;
				break;
			}
			
		} // while(!g_state.stop)

		if(g_state.stop || stream.error != MAD_ERROR_BUFLEN)
			break;
	
		memmove(input_buffer, stream.next_frame,
			input_length = &input_buffer[input_length] - stream.next_frame);
			
		if(eof && g_reptsong)
		{
			eof = 0;
			mp3seek(g_posbeg);
		}
			
	} // while(!eof)

	mad_synth_finish(&synth);
	mad_frame_finish(&frame);
	mad_stream_finish(&stream);

	FlushFifo();

playmp3exit:

	// close input file
	if(madfilep)
	{
		if(g_state.nomem)
			fclose(madfilep);
		else
			memfclose(madfilep);

		madfilep = 0;
	}
	
} // maPlayMp3


void mp3seek(int position)
{
	int seek;

	if(g_state.duration >= 0)
		g_state.seek = position;
	else
	{
		seek = (int)((double)position * 1000 / g_state.duration);
		
		if(seek >= 0)
			seek = position;
		else if(seek == -1)
			seek = -2;
		g_state.seek = seek;
	}
}


mad_fixed_t madeqfactor(double db)
{
	if(db > 18)
		db = 18;
	return mad_f_tofixed(pow(10, db / 20));
}


void mp3eqset(int flags, double preampdb) // flags are PREAMP | CLIPPING
{
	double base;
	
	if(flags & PREAMP)
		base = preampdb;
	else
		base = 0;
		
	if(!(flags & CLIPPING))
		g_madstate.madattenuation = MAD_F_ONE;
		
	if(base != 0)
		g_madstate.madeqfactor = madeqfactor(base);
	else
		g_madstate.madeqfactor = MAD_F_ONE;
		
	g_madstate.madequalizer = flags;
	
} // mp3eqset


signed long ThreadPlayMp3(void *parameter)
{
#ifdef MAUNUSED
 (void)parameter;
#endif

	g_waitaudiodone = 1;

	maPlayMp3();
		
	g_state.stop |= STOP_PLAYDECODE;
	
	// wait for output owned by this thread to finish
	while(!g_stopB && g_waitaudiodone)
		maSleep(100);

	return 0;

} // ThreadPlayMp3


// xing VBR
#define XING_MAGIC (('X' << 24) | ('i' << 16) | ('n' << 8) | 'g')


int xing_parse(XING *xing, struct mad_bitptr ptr, unsigned int bitlen)
{
	if(bitlen < 64 || mad_bit_read(&ptr, 32) != XING_MAGIC)
    goto fail;

	xing->flags = mad_bit_read(&ptr, 32);
	bitlen -= 64;

	if(xing->flags & XING_FRAMES) {
    if(bitlen < 32)
      goto fail;

    xing->frames = mad_bit_read(&ptr, 32);
    bitlen -= 32;
	}

	if(xing->flags & XING_BYTES) {
    if(bitlen < 32)
      goto fail;

    xing->bytes = mad_bit_read(&ptr, 32);
    bitlen -= 32;
	}

	if(xing->flags & XING_TOC) {
    int i;

    if(bitlen < 800)
      goto fail;

    for(i = 0; i < 100; ++i)
      xing->toc[i] = (unsigned char)mad_bit_read(&ptr, 8);

    bitlen -= 800;
	}

	if(xing->flags & XING_SCALE) {
    if(bitlen < 32)
      goto fail;

    xing->scale = mad_bit_read(&ptr, 32);
    bitlen -= 32;
	}

	return 0;

 fail:
	xing->flags = 0;
	return -1;
}


int Mp3GetHeaderInfo(FILE *filep, int nomem, MASTATE *pstate, LSTATE *plstate);
int Mp3GetHeaderInfo(FILE *filep, int nomem, MASTATE *pstate, LSTATE *plstate)
{
	mad_timer_t timer;
	char cstr[MAX_PATH];
	
	memset(&plstate->mp3header, 0, sizeof(struct mad_header));
	maSetFilePointer(filep, 0, 0, SEEK_SET, nomem);
			
	if(madscanheader(filep, &plstate->mp3header, &plstate->mp3xing, nomem, (char *)&pstate->inpath) == -1)
	{
		if(&pstate->inpath)
		{
			sprintf(cstr, "DumpSongInfo madscanheader failed on '%s'", (char *)&pstate->inpath);
			maLog(cstr);
		}
		pstate->duration = -1;
		return 1;
	}

	if(plstate->mp3xing.flags & XING_FRAMES)
	{		
		timer = plstate->mp3header.duration;
		mad_timer_multiply(&timer, plstate->mp3xing.frames);
		pstate->duration = mad_timer_count(timer, MAD_UNITS_MILLISECONDS);
		if((plstate->mp3xing.flags & XING_BYTES) && pstate->duration > 0)
			pstate->bitrate = plstate->mp3xing.bytes * 8 / pstate->duration;			
	}
	else
		pstate->duration = pstate->size * 8 / (plstate->mp3header.bitrate / MSPERSEC);

	pstate->insamplerate = plstate->mp3header.samplerate;

	return 0;
	
} // Mp3GetHeaderInfo


int DumpInfoMp3(MASTATE *pstate, FILE *dumpfp)
{
	char tstr[32];
	char cstrx[512];
	LSTATE lstate;
	
	Mp3GetHeaderInfo(pstate->fp, pstate->nomem, pstate, &lstate);
	
	sprintf(cstrx, "Duration : %s", SecondsToStrHMS(tstr, (pstate->duration / MSPERSEC)));
	DumpString(dumpfp, cstrx);

	if(lstate.mp3xing.flags & XING_FRAMES) // XING_FRAMES = 0x00000001L
	{
		DumpString(dumpfp, (char *)"Xing:");

		sprintf(cstrx, "Frames : %lu", lstate.mp3xing.frames);
		DumpString(dumpfp, cstrx);
	
		sprintf(cstrx, "Bytes : %lu", lstate.mp3xing.bytes);
		DumpString(dumpfp, cstrx);
	}	

	sprintf(cstrx, "Layer : %d", lstate.mp3header.layer);
	DumpString(dumpfp, cstrx);

	switch(lstate.mp3header.mode)
	{
		case 0:
			sprintf(cstrx, (char *)"Mode : single channel");
		break;
		case 1:
			sprintf(cstrx, (char *)"Mode : dual channel");
		break;
		case 2:
			sprintf(cstrx, (char *)"Mode : joint (MS/intensity) stereo");
		break;
		case 3:
			sprintf(cstrx, (char *)"Mode : normal LR stereo");
		break;
		default:
			sprintf(cstrx, (char *)"Mode : unknown %d", lstate.mp3header.mode);
		break;
	}
	DumpString(dumpfp, cstrx);
	
	sprintf(cstrx, "Bitrate : %ld", lstate.mp3header.bitrate);
	DumpString(dumpfp, cstrx);

	sprintf(cstrx, "Samplerate : %d", pstate->insamplerate);
	DumpString(dumpfp, cstrx);
			
#ifndef NOID3TAG
	DumpTagsID3(pstate, dumpfp);
#endif
			
	return 0;
	
} // DumpInfoMp3


#ifndef NOID3TAG
// from mad 0.15.0b release
struct id3_file {
	FILE *iofile;
	enum id3_file_mode mode;
	int flags;
	int options;
	struct id3_tag *primary;
	unsigned int ntags;
	struct filetag *tags;
};

void show_id3(struct id3_tag const *tag, FILE *dumpfp);
void DumpTagsID3(MASTATE *pstate, FILE *dumpfp)
{
	struct id3_tag *tag;
	struct id3_file *file;
	
	g_noimem = pstate->nomem;
	
	if(pstate->nomem)
	{
#if defined(__MAOSX__)
		char bsdname[MAX_PATH];
		strcpy(bsdname, pstate->inpath);
		StrOSXtoStrBSD(bsdname);
    	file = id3_file_open(bsdname, ID3_FILE_MODE_READONLY);		// this calls fopen
#else
		file = id3_file_open(pstate->inpath, ID3_FILE_MODE_READONLY); // this calls fopen
#endif
	}
	else
    	file = id3_file_open(pstate->inpath, ID3_FILE_MODE_READONLY); // this calls mimfopen
	
    if(file == 0)
    {
		DumpString(dumpfp, (char *)"DumpTagsID3 id3_file_open failed");
		return;
    }

	tag = file->primary;
	if(!tag)
	{
		DumpString(dumpfp, (char *)"no id3 tag");
		id3_file_close(file);
		return;
	}
	
	DumpString(dumpfp, (char *)"id3 tag:");	
	show_id3(tag, dumpfp);
	id3_tag_delete(tag);
	id3_file_close(file);

} // DumpTagsID3


void show_id3(struct id3_tag const *tag, FILE *dumpfp)
{
	unsigned int i;
	struct id3_frame const *frame;
	id3_ucs4_t const *ucs4;
	id3_latin1_t *latin1;
	char const spaces[] = "          ";
	char cstrx[512];

	static struct {
		char const *id;
		char const *name;
	} const info[] =
	{
		{ID3_FRAME_TITLE, "Title"},
		{"TIT3",			0},			// Subtitle
		{"TCOP",			0},			// Copyright
		{"TPRO",			0},			// Produced
		{"TCOM",			"Composer"},
		{ID3_FRAME_ARTIST,	"Artist"},
		{"TPE2",			"Orchestra"},
		{"TPE3",			"Conductor"},
		{"TEXT",			"Lyricist"},
		{ID3_FRAME_ALBUM,	"Album"},
		{ID3_FRAME_TRACK,	"Track"},
		{ID3_FRAME_YEAR,	"Year"},
		{"TPUB",			"Publisher"},
		{ID3_FRAME_GENRE,	"Genre"},
		{"TRSN",			"Station"},
		{"TENC",			"Encoder"}
	};

	// text information
	for(i = 0; i < sizeof(info) / sizeof(info[0]); ++i)
	{
		union id3_field const *field;
		unsigned int nstrings, namelen, j;
		char const *name;

		frame = id3_tag_findframe(tag, info[i].id, 0);
		if(frame == 0)
			continue;

		field = id3_frame_field(frame, 1);
		nstrings = id3_field_getnstrings(field);

		name = info[i].name;
		namelen = name ? strlen(name) : 0;
		//u assert(namelen < sizeof(spaces));

		for(j = 0; j < nstrings; ++j)
		{
			ucs4 = id3_field_getstrings(field, j);
			//u assert(ucs4);

			if(strcmp(info[i].id, ID3_FRAME_GENRE) == 0)
				ucs4 = id3_genre_name(ucs4);

			latin1 = id3_ucs4_latin1duplicate(ucs4);
			if(latin1 == 0)
				goto fail;

#if defined(MAMEMTRACE)
 g_ndupallocs++; // latin1
#endif

			if(j == 0 && name)
			{
				sprintf(cstrx, "%s%s: %s", &spaces[namelen], name, latin1);
				DumpString(dumpfp, cstrx);
			}
			else
			{
				if(strcmp(info[i].id, "TCOP") == 0 ||
					strcmp(info[i].id, "TPRO") == 0)
				{
					sprintf(cstrx, "%s	%s %s\n", spaces, (info[i].id[1] == 'C') ?
										"Copyright (C)" : "Produced (P)", latin1);
					DumpString(dumpfp, cstrx);
				}
				else
				{
					sprintf(cstrx, "%s	%s", spaces, latin1);
					DumpString(dumpfp, cstrx);
				}
			}

			MADUPFREE(latin1);
		}
	}

	// comments
	i = 0;
	while((frame = id3_tag_findframe(tag, ID3_FRAME_COMMENT, i++)))
	{
		id3_latin1_t *ptr, *newline;
		int first = 1;

		ucs4 = id3_field_getstring(id3_frame_field(frame, 2));
		//u assert(ucs4);

		if(*ucs4)
			continue;

		ucs4 = id3_field_getfullstring(id3_frame_field(frame, 3));
		//u assert(ucs4);

		latin1 = id3_ucs4_latin1duplicate(ucs4);
		if(latin1 == 0)
			goto fail;

#if defined(MAMEMTRACE)
 g_ndupallocs++; // latin1
#endif

		ptr = latin1;
		while(*ptr)
		{
			newline = (unsigned char *)strchr((char *)ptr, '\n');
			if(newline)
				*newline = 0;

			if(strlen((const char*)ptr) > 66)
			{
				id3_latin1_t *linebreak;

				linebreak = ptr + 66;

				while(linebreak > ptr && *linebreak != ' ')
						--linebreak;

				if(*linebreak == ' ')
				{
					if(newline)
						*newline = '\n';

					newline = linebreak;
					*newline = 0;
				}
			}

			if(first)
			{
				char const *name;
				unsigned int namelen;

				name = "Comment";
				namelen = strlen(name);
				//u assert(namelen < sizeof(spaces));
				{
					sprintf(cstrx, "%s%s: %s", &spaces[namelen], name, ptr);
					DumpString(dumpfp, cstrx);
				}
				first = 0;
			}
			else 
			{
				sprintf(cstrx, "%s	%s", spaces, ptr);
				DumpString(dumpfp, cstrx);
			}
			ptr += strlen((const char *)ptr) + (newline ? 1 : 0);
		}

		MADUPFREE(latin1);
		break;
	}

	if(0)
	{
fail:
		maLog("show_id3 not enough memory to display tag");
	}
	
} // show_id3
#endif

#ifdef __cplusplus
}
#endif

#endif // NODECODEMP3

// eof
