// maPlayMp4.c

/*
 * MoreAmp
 * Copyright (C) 2004-2010 moreforge
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include "maConfig.h" // included by all moreamp files

#ifndef NODECODEMP4

#ifdef __MAWIN__
 #ifdef MAUNUSEDWIN
  #pragma warning( disable : 4100 ) // unreferenced formal parameter
 #endif
 #pragma warning( disable : 4127 ) // conditional expression is constant
 #pragma warning( disable : 4142 ) // benign redefinition of type
 #pragma warning( disable : 4201 ) // nonstandard extension used : nameless struct/union
 #pragma warning( disable : 4244 ) // conversion from 'double ' to 'int ', possible loss of data
 #pragma warning( disable : 4701 ) // local variable 'mp4ASC' may be used without having been initialized
 #pragma warning( disable : 4706 ) // assignment within conditional expression
 #pragma warning( disable : 4761 ) // integral size mismatch in argument; conversion supplied
 #define __RPCASYNC_H__ // rpcasync.h(45) : warning C4115: '...' : named type definition in parentheses
#endif

#ifdef __cplusplus
extern "C"
{
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#ifdef __MAWIN__
 #include <stdlib.h>
 #include <limits.h>
 #include <time.h>
 #include <wtypes.h>
 typedef long int32_t;
 typedef unsigned long u_int32_t;
 typedef unsigned __int64 u_int64_t;
 typedef __int64 int64_t;
 typedef unsigned short u_int16_t;
 typedef unsigned char u_int8_t;
 typedef char int8_t;
#endif

#if defined(__MAOSX__)
 #include <MacTypes.h>
#endif

#include "faad.h"
#include "mp4ff.h"

#define __SYSTEMS_H__
#include "mp4.h"
#undef __SYSTEMS_H__

#ifdef __MAUNIX__
 #include "time.h"
#endif

#include "maDefs.h"
#include "maPlayer.h"
#include "maUtils.h"
#include "maPlayMp4.h"
#include "maEqualizer.h"
#include "maMemFile.h"

#ifdef __MAWIN__
 #include "maAudioWin.h"
#endif

#ifdef __MAOSX__
 #include "maAudioMac.h"
#endif

#ifdef __MAUNIX__
 #include "maAudioUnix.h"
#endif

typedef struct
{
	long bytes_into_buffer;
	long bytes_consumed;
	long file_offset;
	long fileread;
	unsigned char *buffer;
	int at_eof;
	FILE *infilefp;
} mp4_buffer;

typedef struct
{
	//char *infilepath;
	faacDecHandle hcodec;
	mp4_buffer b;
	faacDecFrameInfo frameInfo;
	faacDecConfigurationPtr config;
	MP4SampleId sampleId;
	MP4SampleId numSamples;
	unsigned char *buffer;
	int buffer_size;
	mp4ff_t *infile;
	int track;
	int noGapless;
	unsigned int initial;
	mp4ff_callback_t acb;
	mp4ff_callback_t *cb;
	unsigned long samplerate;
	unsigned char channels;
	unsigned int useAacLength;
	unsigned int framesize;
	unsigned long timescale;
	MASTATE *pstate;
} Mp4Context;

void maPlayMp4();
int Mp4OpenFile(Mp4Context *pmp4);
void Mp4CloseFile(Mp4Context *pmp4);

uint32_t mp4read_callback(void *user_data, void *buffer, uint32_t length);
uint32_t mp4seek_callback(void *user_data, uint64_t position);
int GetMp4Track(mp4ff_t *infile);

long g_mp4seek = -1;
void mp4seek(int pos)
{
	if(/* g_pmp4pcontext == NULL || */ g_mp4seek >= 0 || pos < 0)
		return;

	g_mp4seek = pos;	

} // mp4seek

#define MAX_CHANNELS 6 /* make this higher to support files with more channels */


void maPlayMp4()
{
	void *sample_buffer;
	int pcmblocksize;
	unsigned int i;
	int j;
	int rc;
	long dur;
	unsigned int sample_count;
	//unsigned int delay = 0;
	MP4SampleId sampleId;
	Mp4Context mp4context;
	Mp4Context *pmp4;
	short *sample_buffer16;
	char *data = NULL;
	size_t s = 0;
	size_t datasize = 0;
	char cstr[256];
	
	if(!g_state.inpath)
	{
		maLog("maPlayMp4 error null filename");
		return;
	}
	
	// format of this decoder output
#ifdef WORDS_BIGENDIAN
	g_state.inwordsbigendian = 1;
#else
	g_state.inwordsbigendian = 0;
#endif
	g_state.ininterleaved = 1;
	
	// open
	pmp4 = &mp4context;
	memset(pmp4, 0, sizeof(Mp4Context));
	pmp4->pstate = &g_state;
	
	if(Mp4OpenFile(pmp4))
	{
		sprintf(cstr, "maPlayMp4 Mp4OpenFile failed on '%s'", pmp4->pstate->inpath);
		maLog(cstr);
		pmp4 = NULL;
		goto maPlayMp4exit;
	}
		
	for(j = 0; j < mp4ff_total_tracks(pmp4->infile); j++)
		if((g_state.insamplerate = mp4ff_get_sample_rate(pmp4->infile, j)))
			break; // use first non-zero samplerate
	
	if(maOutputOpen() < 0)
	{
		maLog("maPlayMp4 maOutputOpen failed");
		goto maPlayMp4exit;
	}

	if(!g_posendon)
		g_posend = g_state.duration;

	// mp4 decode loop
	while(1)
	{	
		ThreadYield();
		
		if(g_state.stop)
			break;
				
		if(g_mp4seek >= 0)
		{
			if((sampleId = LROUND(((double)g_mp4seek * pmp4->numSamples) / g_state.duration)) >= pmp4->numSamples)
			{
				if(g_reptsong)
				{
					g_mp4seek = g_posbeg;
					if((sampleId = LROUND(((double)g_mp4seek * pmp4->numSamples) / g_state.duration)) >= pmp4->numSamples)
						break;
				}
				else
					break;
			}
			if((pmp4->sampleId = sampleId) <= 0)
				pmp4->sampleId = 1;
			g_mp4seek = -1;
		}

		pmp4->buffer = NULL;
		pmp4->buffer_size = 0;

		dur = mp4ff_get_sample_duration(pmp4->infile, pmp4->track, pmp4->sampleId);
		rc = mp4ff_read_sample(pmp4->infile, pmp4->track, pmp4->sampleId, &pmp4->buffer,
															(unsigned int*)&pmp4->buffer_size);
		if(rc == 0)
		{
			sprintf(cstr, "maPlayMp4 mp4ff_read_sample failed '%s'", pmp4->pstate->inpath);
			maLog(cstr);
			break;
		}

		if(pmp4->buffer == NULL)
		{
			sprintf(cstr, "maPlayMp4 mp4ff_read_sample null buffer '%s'", pmp4->pstate->inpath);
			maLog(cstr);
			break;
		}
			
		sample_buffer = faacDecDecode(pmp4->hcodec, &pmp4->frameInfo,
															pmp4->buffer, pmp4->buffer_size);

		if(pmp4->buffer)
			MAFREE(pmp4->buffer);

		if(!pmp4->noGapless)
		{
			if(pmp4->sampleId == 0) dur = 0;

			if(pmp4->useAacLength || (pmp4->timescale != pmp4->samplerate))
				sample_count = pmp4->frameInfo.samples;
			else
			{
				sample_count = (unsigned int)(dur * pmp4->frameInfo.channels);

				if(!pmp4->useAacLength && !pmp4->initial && (pmp4->sampleId < pmp4->numSamples/2) &&
												(sample_count != pmp4->frameInfo.samples))
				{
					sprintf(cstr, "maPlayMp4 incorrect frame duration, using values from AAC data, '%s'",
																pmp4->pstate->inpath);
					maLog(cstr);
					pmp4->useAacLength = 1;
					sample_count = pmp4->frameInfo.samples;
				}
			}
		}
		else
			sample_count = pmp4->frameInfo.samples;

		if(pmp4->initial && sample_count > 0)
			pmp4->initial = 0;

		if((pmp4->frameInfo.error == 0) && (sample_count > 0))
        {
			if((s = pmp4->frameInfo.samples << 1) > datasize)
			{
				if(data)
					MAFREE(data);
				datasize = s;
				data = (char *)MAMALLOC(datasize);
				if(data == NULL)
				{
					sprintf(cstr, "maPlayMp4 MAMALLOC failed on %d bytes for '%s'",
												(int)datasize, pmp4->pstate->inpath);
					maLog(cstr);
					break;
				}
			}
			sample_buffer16 = (short*)sample_buffer;

#ifdef __MAWIN__
			for(i = 0; i < pmp4->frameInfo.samples; i++) // ENDIAN
			{
				data[i * 2    ] = (char) (sample_buffer16[i]       & 0xFF);
				data[i * 2 + 1] = (char)((sample_buffer16[i] >> 8) & 0xFF);
			}
#else			
			for(i = 0; i < pmp4->frameInfo.samples; i++)
			{
				data[i * 2 + 1] = (char) (sample_buffer16[i]       & 0xFF);
				data[i * 2    ] = (char)((sample_buffer16[i] >> 8) & 0xFF);
			}
#endif

			pcmblocksize = sample_count << 1;
			if(pcmblocksize > 0)
			{
				g_state.inchannels = pmp4->frameInfo.channels;

				// Write to fifo
				if(maOutputWrite((char *)sample_buffer, pcmblocksize, &g_state) != 0)
				{
					sprintf(cstr,
					"maPlayMp4 maOutputWrite failed : len=%d g_state.stop=x%x g_fifoready=%d",
												pcmblocksize, g_state.stop, g_fifoready);
					maLog(cstr);
					break;
				}

			} //if(pcmblocksize > 0)

        } // (frameInfo.error == 0) && (sample_count > 0)

        if(pmp4->frameInfo.error > 0)
        {
            sprintf(cstr, "maPlayMp4 Warning: %s, '%s'", faacDecGetErrorMessage(pmp4->frameInfo.error),
																	pmp4->pstate->inpath);
			maLog(cstr);
        }

		g_state.position = LROUND(((double)pmp4->sampleId * g_state.duration) / pmp4->numSamples);

		if((++pmp4->sampleId >= pmp4->numSamples))
		{
			if(g_reptsong)
				g_mp4seek = g_posbeg;
			else
				break;
		}

	} // while 1

	// finish

	FlushFifo();

maPlayMp4exit:
	if(pmp4)
	{
		Mp4CloseFile(pmp4);
		pmp4 = NULL;
	}
	
	if(data)
		MAFREE(data);

	g_state.position = 0;
	g_mp4seek = -1;
	
} // maPlayMp4


uint32_t mp4read_callback(void *user_data, void *buffer, uint32_t length)
{
	if(((Mp4Context *)user_data)->pstate->nomem)
		return fread(buffer, 1, length, (FILE *)((Mp4Context *)user_data)->b.infilefp);

	return memfread(buffer, 1, length, (FILE *)((Mp4Context *)user_data)->b.infilefp);
}


uint32_t mp4seek_callback(void *user_data, uint64_t position)
{
	if(((Mp4Context *)user_data)->pstate->nomem)
		return fseek((FILE *)((Mp4Context *)user_data)->b.infilefp, position, SEEK_SET);
		
	return memfseek((FILE *)((Mp4Context *)user_data)->b.infilefp, position, SEEK_SET);
}


int GetMp4Track(mp4ff_t *infile)
{
	/* find AAC track */
	int i, rc;
	int numTracks = mp4ff_total_tracks(infile);

	for(i = 0; i < numTracks; i++)
	{
		unsigned char *buff = NULL;
		int buff_size = 0;
		mp4AudioSpecificConfig mp4ASC;

		mp4ff_get_decoder_config((const mp4ff_t *)infile, (const int)i,
						(unsigned char**)&buff, (unsigned int*)&buff_size);

		if(buff)
		{
			rc = AudioSpecificConfig(buff, buff_size, &mp4ASC);
			MAFREE(buff);

			if(rc < 0)
				continue;
			return i;
		}
	}

	/* can't decode this */
	return -1;
}


int Mp4OpenFile(Mp4Context *pmp4)
{
	int lnomem;
	int downMatrix = 0;
	int outputFormat = FAAD_FMT_16BIT;
	size_t dummy;
	unsigned char header[8];
	double f;
	mp4AudioSpecificConfig mp4ASC;
	char cstr[256];
	
	// error if null
	if(pmp4 == NULL)
	{
		maLog((char *)"Mp4OpenFile null context pointer");
		return 1;
	}

	pmp4->initial = 1;

	lnomem = pmp4->pstate->nomem;
	
	if(lnomem)
	{
		if((pmp4->b.infilefp = mafopen(pmp4->pstate->inpath, "rb")) == NULL)
		{
			sprintf(cstr, "Mp4OpenFile mafopen failed on '%s'", pmp4->pstate->inpath);
			maLog(cstr);
			return 1;
		}
	}
	else
	{
		if((pmp4->b.infilefp = memfopen(pmp4->pstate->inpath, "rb")) == NULL)
		{
			sprintf(cstr, "Mp4OpenFile memfopen failed on '%s'", pmp4->pstate->inpath);
			maLog(cstr);
			return 1;
		}
	}

	// check for mp4 file
	if(lnomem)
	{
		dummy = fread(header, 1, 8, pmp4->b.infilefp);
		rewind(pmp4->b.infilefp);
	}
	else
	{
		memfread(header, 1, 8, pmp4->b.infilefp);
		memrewind(pmp4->b.infilefp);
	}

	if(header[4] != 'f' || header[5] != 't' || header[6] != 'y' || header[7] != 'p') // decodeMP4file
	{
		sprintf(cstr, "Mp4OpenFile not an MP4 file '%s'", pmp4->pstate->inpath);
		maLog(cstr);
		return 1;
	}

	/* initialise the callback structure */
	pmp4->cb = &pmp4->acb; //(mp4ff_callback_t *)MAMALLOC(sizeof(mp4ff_callback_t));

	pmp4->cb->read = mp4read_callback;
	pmp4->cb->seek = mp4seek_callback;
	pmp4->cb->user_data = pmp4; //pmp4->b.infilefp;
	
	pmp4->hcodec = faacDecOpen();

	/* Set configuration */
	pmp4->config = faacDecGetCurrentConfiguration(pmp4->hcodec);
	pmp4->config->outputFormat = outputFormat;
	pmp4->config->downMatrix = downMatrix;
	pmp4->config->dontUpSampleImplicitSBR = 1;
	faacDecSetConfiguration(pmp4->hcodec, pmp4->config);

	pmp4->infile = mp4ff_open_read(pmp4->cb);
	if(!pmp4->infile)
	{
		/* unable to open file */
		sprintf(cstr, "Mp4OpenFile error opening file '%s'", pmp4->pstate->inpath);
		maLog(cstr);
		return 1;
	}

	if((pmp4->track = GetMp4Track(pmp4->infile)) < 0)
	{
		sprintf(cstr, "Mp4OpenFile unable to find correct AAC sound track in MP4 file '%s'",
																pmp4->pstate->inpath);
		maLog(cstr);
		faacDecClose(pmp4->hcodec);
		mp4ff_close(pmp4->infile);
		MAFREE(pmp4->cb);
		goto mp4openerrexit;
	}

	pmp4->buffer = NULL;
	pmp4->buffer_size = 0;
	mp4ff_get_decoder_config((const mp4ff_t *)pmp4->infile, (const int)pmp4->track,
					(unsigned char**)&pmp4->buffer, (unsigned int*)&pmp4->buffer_size);

	if((signed char)faacDecInit2(pmp4->hcodec, pmp4->buffer, pmp4->buffer_size,
											&pmp4->samplerate, &pmp4->channels) < 0)
	{
		/* If some error initializing occured, skip the file */
		sprintf(cstr, "Mp4OpenFile error initializing aac decoder library.");
		maLog(cstr);
		faacDecClose(pmp4->hcodec);
		mp4ff_close(pmp4->infile);
		MAFREE(pmp4->cb);
		goto mp4openerrexit;
	}

	pmp4->timescale = mp4ff_time_scale(pmp4->infile, pmp4->track);
	pmp4->framesize = 1024;
	pmp4->useAacLength = 0;

	if(pmp4->buffer)
	{
		if((signed char)AudioSpecificConfig(pmp4->buffer, pmp4->buffer_size, &mp4ASC) >= 0)
		{
			if(mp4ASC.frameLengthFlag == 1)
				pmp4->framesize = 960;
			if(mp4ASC.sbr_present_flag == 1)
				pmp4->framesize *= 2;
		}
			
		MAFREE(pmp4->buffer);
	}

	pmp4->numSamples = mp4ff_num_samples(pmp4->infile, pmp4->track);

	f = 1024.0;
		
	if(mp4ASC.sbr_present_flag == 1)
		f = f * 2.0;

	pmp4->pstate->duration = LROUND(1000 * ((double)pmp4->numSamples * (f - 1.0) / mp4ASC.samplingFrequency));

	pmp4->pstate->position = 0;
	pmp4->sampleId = 0;
	
	return 0;

mp4openerrexit:

		if(lnomem)
			fclose(pmp4->b.infilefp);
		else
			memfclose(pmp4->b.infilefp);
					
		pmp4->b.infilefp = NULL;
		
		return 1;

} // Mp4OpenFile


void Mp4CloseFile(Mp4Context *pmp4)
{
	if(!pmp4)
	{
		maLog((char *)"Mp4CloseFile null context pointer");
		return;
	}

	if(pmp4->hcodec)
	{
		faacDecClose(pmp4->hcodec);
		pmp4->hcodec = NULL;
	}
		
	if(pmp4->b.infilefp)
	{
		if(pmp4->pstate->nomem)
			fclose(pmp4->b.infilefp);
		else
			memfclose(pmp4->b.infilefp);
					
		pmp4->b.infilefp = NULL;
	}
	
	if(pmp4->b.buffer)
	{
		MAFREE(pmp4->b.buffer);
		pmp4->b.buffer = NULL;
	}
	
} // Mp4CloseFile


signed long ThreadPlayMp4(void *parameter)
{
#ifdef MAUNUSED
 (void)parameter;
#endif

	g_waitaudiodone = 1;

	maPlayMp4();
		
	g_state.stop |= STOP_PLAYDECODE;
	
	// wait for output owned by this thread to finish
	while(!g_stopB && g_waitaudiodone)
		maSleep(100);

	return 0;

} // ThreadPlayMp4


int DumpInfoMp4(MASTATE *pstate, FILE *dumpfp)
{
	Mp4Context mp4context;
	Mp4Context *pmp4;
	int ret = 0;
	int i;
	int numtracks;
	int item;
	char *pvalue;
	char *pitem;
	char tstr[64];
	char cstrx[512];
	
	if(!pstate || !pstate->inpath)
	{
		maLog("DumpInfoMp4 error null filename");
		return 1;
	}
	
	// open
	pmp4 = &mp4context;
	memset(pmp4, 0, sizeof(Mp4Context));
	pmp4->pstate = pstate;

	if(Mp4OpenFile(pmp4))
	{
		sprintf(cstrx, "DumpInfoMp4 Mp4OpenFile failed on '%s'", pmp4->pstate->inpath);
		maLog(cstrx);
		pmp4 = NULL;
		ret = 1;
		goto DumpInfoMp4exit;
	}

	// dump
	if((numtracks = mp4ff_total_tracks(pmp4->infile)) < 0)
	{
		sprintf(cstrx, "DumpInfoMp4 mp4ff_total_tracks failed on '%s'", pmp4->pstate->inpath);
		maLog(cstrx);
	}
	else
	{
		//sprintf(cstrx, "numtracks : %d", numtracks);
		//DumpString(dumpfp, cstrx);
		
		for(i = 0; i < numtracks; i++)
		{
			sprintf(cstrx, "Duration : %s", SecondsToStrHMS(tstr, pstate->duration / MSPERSEC));
			DumpString(dumpfp, cstrx);

			//sprintf(cstrx, "numsamples : %d", pmp4->numSamples);
			//DumpString(dumpfp, cstrx);
			
			if((item = mp4ff_get_channel_count(pmp4->infile, i)) < 0)
			{
				sprintf(cstrx, "DumpInfoMp4 mp4ff_get_channel_count failed on track %d of file '%s'", i, pmp4->pstate->inpath);
				maLog(cstrx);
			}
			else
			{
				sprintf(cstrx, "channels : %d", item);
				DumpString(dumpfp, cstrx);
			}
		
			if((item = mp4ff_get_sample_rate(pmp4->infile, i)) < 0)
			{
				sprintf(cstrx, "DumpInfoMp4 mp4ff_get_sample_rate failed on track %d of file '%s'", i, pmp4->pstate->inpath);
				maLog(cstrx);
			}
			else
			{
				sprintf(cstrx, "samplerate : %d", item);
				DumpString(dumpfp, cstrx);
			}
		
			if((item = mp4ff_get_avg_bitrate(pmp4->infile, i)) < 0)
			{
				sprintf(cstrx, "DumpInfoMp4 mp4ff_get_avg_bitrate failed on track %d of file '%s'", i, pmp4->pstate->inpath);
				maLog(cstrx);
			}
			else
			{
				sprintf(cstrx, "avgbitrate : %d", item);
				DumpString(dumpfp, cstrx);
			}
		
			if((item = mp4ff_get_max_bitrate(pmp4->infile, i)) < 0)
			{
				sprintf(cstrx, "DumpInfoMp4 mp4ff_get_max_bitrate failed on track %d of file '%s'", i, pmp4->pstate->inpath);
				maLog(cstrx);
			}
			else
			{
				sprintf(cstrx, "maxbitrate : %d", item);
				DumpString(dumpfp, cstrx);
			}
		
			if((item = mp4ff_get_track_type(pmp4->infile, i)) < 0)
			{
				sprintf(cstrx, "DumpInfoMp4 mp4ff_get_track_type failed on track %d of file '%s'", i, pmp4->pstate->inpath);
				maLog(cstrx);
			}
			else
			{
				sprintf(cstrx, "tracktype : %d", item);
				DumpString(dumpfp, cstrx);
			}
			
			if((item = mp4ff_get_audio_type(pmp4->infile, i)) < 0)
			{
				sprintf(cstrx, "DumpInfoMp4 mp4ff_get_audio_type failed on track %d of file '%s'", i, pmp4->pstate->inpath);
				maLog(cstrx);
			}
			else
			{
				sprintf(cstrx, "audiotype : %d", item);
				DumpString(dumpfp, cstrx);
			}
			
		} // for i
	}
	
	if((item = mp4ff_meta_get_num_items(pmp4->infile)) > 0)
	{
		for(i = 0; i < item; i++)
		{
			if(mp4ff_meta_get_by_index(pmp4->infile, (unsigned int)i, &pitem, &pvalue))
			{
				if(pitem && pvalue)
				{
					if(strlen(pvalue) > 495)
						pvalue[495] = 0;
					if(strlen(pitem) > 15)
						pvalue[15] = 0;
					sprintf(cstrx, "%s : %s", pitem, pvalue);
					DumpString(dumpfp, cstrx);
					MAFREE(pitem);
					MAFREE(pvalue);
				}
				else
				{
					if(pitem)
						MAFREE(pitem);
					if(pvalue)
						MAFREE(pvalue);
					sprintf(cstrx, "DumpInfoMp4 mp4ff_meta_get_num_items failed on '%s'", pmp4->pstate->inpath);
					maLog(cstrx);
				}
			}
		}
	}
	
	// close
	Mp4CloseFile(&mp4context);
	
	return 0;
	
DumpInfoMp4exit:
	if(pmp4)
		Mp4CloseFile(pmp4);
		
	return ret;
					
} // DumpInfoMp4

#ifdef __cplusplus
}
#endif

#endif

// eof
