// maSongList.c

/*
 * MoreAmp
 * Copyright (C) 2004-2010 moreforge
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include "maConfig.h" // included by all moreamp files

#ifdef __MAWIN__
 #ifdef MAUNUSEDWIN
  #pragma warning( disable : 4100 ) // unreferenced formal parameter
 #endif
 #pragma warning( disable : 4127 ) // conditional expression is constant
 #pragma warning( disable : 4706 ) // assignment within conditional expression
 #define __RPCASYNC_H__ // rpcasync.h(45) : warning C4115: '_RPC_ASYNC_STATE' : named type definition in parentheses
#endif

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>

#ifdef __MAOSX__
 #include <Files.h>
 #include <sys/stat.h>
 #include <dirent.h>
#endif

#ifdef __MAUNIX__
 #include <unistd.h>
 #include <sys/stat.h>
 #include <sys/param.h>
 #include <dirent.h>
#endif

#if defined(__MAOSX__)
 #include <AEDataModel.h>
 #include <AppleEvents.h>
 #include <CarbonEvents.h>
#endif

#ifdef __MAWIN__
 #include <windows.h>
#endif

// Application includes
#include "maDefs.h"
#include "maGUI.h"
#include "maPlayer.h"
#include "maUtils.h"
#include "maSongList.h"

#ifndef MANOCDA
 #include "maCda.h"
 #include "maPlayCda.h"
#endif

#if(0) // defined(__MAOSX__) || defined(__MAUNIX__) // untested on other platforms
 #define SONG_MAX_PATH 512
 #define SONG_MAX_PATH_REAL (SONG_MAX_PATH - 128)
#else
 #define SONG_MAX_PATH MAX_PATH
 #define SONG_MAX_PATH_REAL MAX_PATH
#endif
#define CSTR_MAX_PATH (128 + SONG_MAX_PATH)

#ifdef __cplusplus
 extern "C"
 {
#endif

int maSongListAddItems(FILE *fp, int nitems, char **pitems, int *pntracks, int *pnfiles, int *pnlists, int *pnslfiles, char *sglist);
int maSongListResolveItem(FILE *fp, char *p, int *pntracks, int *pnfiles, int *pnlists, int *pnslfiles, char *sglist);
int maSongListResolveFolder(FILE *fp, char *dirname, int *pntracks, int *pnfiles, int *pnlists, int *pnslfiles, char *sglist);
int maSongListResolveFile(FILE *fp, char *outfname, int *pntracks, int *pnfiles, int *pnlists, int *pnslfiles, char *sglist);
int maSongListAddFromSongList(FILE *fp, char *infname, int *pnfiles, int *pnslfiles);
int maSongListAddFromSongListm3u(FILE *fp, char *infname, int *pnfiles, int *pnslfiles);
// These functions are used to resolve songfile paths from items (songfiles, links,
// dirs, cd, songlists), passed as command line arguments or dragged-and-dropped,
// into a new songlist.
//
// A cd generates a list of tracks.
//
// If passed items resolve to a single file and it is a songlist then it becomes the
// 'new' songlist, otherwise passed songlists' contents are added to the new songlist.
//
// args to these functions :
// 
// fp        : FILE* to the new songlist being written, e.g. g_songlisttemp
// nitems	 : the number of items passed or dropped
// pitems	 : an array of pointers to the items passed or dropped
// pntracks  : int* for the number of cd tracks resolved
// pnfiles   : int* for the number of songfiles resolved (not songfiles from songlists)
// pnlists   : int* for the number of songlists resolved
// pnslfiles : int* for the number of songfiles from songlists
// sglist	 : the songlist that is a single resolved file
//


#if !defined(__MAOSX__)
 int DoArgsOpenDoc();
#endif

char g_newsonglist[SONG_MAX_PATH];
int l_onlym3uindir = 1; // process only m3u files in a dir

int maSongListAddItems(FILE *fp, int nitems, char **pitems, int *pntracks, int *pnfiles, int *pnlists, int *pnslfiles, char *sglist)
{
	int n = nitems;
	char **p = pitems;
	char cstr[CSTR_MAX_PATH];
	
	while(n-- > 0 && p && *p)
	{
		if(maSongListResolveItem(fp, *p, pntracks, pnfiles, pnlists, pnslfiles, sglist))
		{
			sprintf(cstr, "maSongListAddItems maSongListResolveItem failed on '%s'", *p);
			maLog(cstr);
		}
		p++;
	}
	
	return 0;
	
} // maSongListAddItems


#if defined(__MAOSX__)
char l_rootvolumepath[SONG_MAX_PATH] = {0};
int maSongListResolveItem(FILE *fp, char *p, int *pntracks, int *pnfiles, int *pnlists, int *pnslfiles, char *sglist)
{
	OSStatus err;
	FSRef theRef;
	Boolean isDirectory;
	Boolean targetIsFolder;
	Boolean wasAliased;
	char bsdfname[SONG_MAX_PATH];
	char outfname[SONG_MAX_PATH];
	char cstr[CSTR_MAX_PATH + MAXLOGLINELEN];
	
	if(!p)
	{
		sprintf(cstr, "maSongListResolveItem null parameter p");
		maLog(cstr);
		return 1;
	}

	if(strlen(p) >= SONG_MAX_PATH_REAL)
	{
		sprintf(cstr, "maSongListResolveItem file path > max %d : %d ", SONG_MAX_PATH_REAL, (int)strlen(p));
		maLog(cstr);
		if(strlen(p) < MAXLOGLINELEN)
			maLog(p);
		else
		{
			strncpy(cstr, p, MAXLOGLINELEN - 1);
			cstr[MAXLOGLINELEN - 1] = 0;
			maLog(cstr);
			strncpy(cstr, p + MAXLOGLINELEN, CSTR_MAX_PATH - 1);
			cstr[CSTR_MAX_PATH - 1] = 0;
			maLog(cstr);
		}
		return 1;
	}

#if !defined(MANOCDA)
	int drive;
	int i;
	if((drive = IsAudioCd(p)))
	{
		for(i = 1; i <= g_cdadrive[drive].tracklast; i++)
		{
			if(fprintf(fp, "%s%cTrack%02d%s\n", p, FILESEP, i, EXTCDA) <= 0)
			{
				sprintf(cstr, "maSongListResolveItem fprintf errno=%d", errno);
				maLog(cstr);
				return 1;
			}
			(*pntracks)++;
		}
		return 0;
	}
#endif

	strcpy(bsdfname, p);
	StrOSXtoStrBSD(bsdfname);
	
	// all our paths need to start from "/Volumes/VolumeName/"
	if(!l_rootvolumepath[0])
	{
		if(maGetRootVolumeName(cstr))
		{
			strcpy(l_rootvolumepath, "/Volumes/Macintosh HD");
			sprintf(cstr, "maSongListResolveItem maGetRootVolumeName failed, using '%s'", l_rootvolumepath);
			maLog(cstr);
		}
		else
			sprintf(l_rootvolumepath, "/Volumes/%s", cstr);
	}
	
	if(strncmp(bsdfname, "/Volumes/", 9))
	{
		sprintf(cstr, "%s%s", l_rootvolumepath, bsdfname);
		strcpy(bsdfname, cstr);
	}

	err = FSPathMakeRef((const UInt8 *)bsdfname, &theRef, &isDirectory);
	if(err)
	{
		if(err == fnfErr) // -43
			sprintf(cstr, "maSongListResolveItem File not found '%s'", bsdfname);
		else
			sprintf(cstr, "maSongListResolveItem FSPathMakeRef err=%d '%s'", (int)err, bsdfname);
		maLog(cstr);
		return 1;
	}
	
	if((err = FSResolveAliasFile(&theRef, true, &targetIsFolder, &wasAliased)))
	{
		sprintf(cstr, "maSongListResolveItem ResolveAliasFile err=%d '%s'", (int)err, p);
		maLog(cstr);
		return 1;
	}
	
	if(wasAliased) // get the target name
	{
		err = FSRefMakePath(&theRef, (UInt8 *)bsdfname, sizeof(bsdfname));
		if(err)
		{
			sprintf(cstr, "maSongListResolveItem FSRefMakePath err=%d '%s'", (int)err, bsdfname);
			maLog(cstr);
			return 1;
		}
		StrBSDtoStrOSX(bsdfname);
		if(strncmp(bsdfname, "/Volumes/", 9))
			sprintf(outfname, "%s%s", l_rootvolumepath, bsdfname);
		else
			strcpy(outfname, bsdfname);
		
#if !defined(MANOCDA)
		if((drive = IsAudioCd(outfname)))
		{
			for(i = 1; i <= g_cdadrive[drive].tracklast; i++)
			{
				if(fprintf(fp, "%s%cTrack%02d%s\n", outfname, FILESEP, i, EXTCDA) <= 0)
				{
					sprintf(cstr, "maSongListResolveItem fprintf errno=%d", errno);
					maLog(cstr);
					return 1;
				}
				(*pntracks)++;
			}
			return 0;
		}
#endif
	}
	else
		strcpy(outfname, bsdfname);

	if(targetIsFolder)
	{
		if(maSongListResolveFolder(fp, outfname, pntracks, pnfiles, pnlists, pnslfiles, sglist))
		{
			sprintf(cstr, "maSongListResolveItem maSongListResolveFolder failed on '%s'", outfname);
			maLog(cstr);
			return 1;
		}
		return 0;
	}
	
	// not a mac audio cd, not an alias and not a folder, must be a file
	return maSongListResolveFile(fp, outfname, pntracks, pnfiles, pnlists, pnslfiles, sglist);

 } // maSongListResolveItem // __MAOSX__
 
 
int maSongListResolveFolder(FILE *fp, char *dirname, int *pntracks, int *pnfiles, int *pnlists, int *pnslfiles, char *sglist)
{
	char *p;
	int n;
	char ldirname[256];
	strcpy(ldirname, dirname);
	StrOSXtoStrBSD(ldirname);

	DIR *dir;
	struct dirent *dt;
	char file[SONG_MAX_PATH];
	char cstr[CSTR_MAX_PATH];
	
	n = *pnfiles;

	if(g_sets.usem3u && l_onlym3uindir) // scan for and use only m3u
	{
		if((dir = opendir(ldirname)) == NULL)
		{
			sprintf(cstr, "maSongListResolveFolder opendir failed on '%s'", ldirname);
			maLog(cstr);
			return 0;
		}
		p = NULL;
		while((dt = readdir(dir)))
		{
			if(FileNameHasExtension(dt->d_name, EXTM3U))
			{
				sprintf(file, "%s%c%s", ldirname, FILESEP, dt->d_name);
				p = file;
				maSongListResolveItem(fp, p, pntracks, pnfiles, pnlists, pnslfiles, sglist);
				// break;
			}
		}
		if(closedir(dir) == -1)
		{
			sprintf(cstr, "maSongListResolveFolder closedir errno=%d on '%s'", errno, ldirname);
			maLog(cstr);
		}
		if(p)
			return 0;
	}
	
	if((dir = opendir(ldirname)) == NULL)
	{
		sprintf(cstr, "maSongListResolveFolder opendir failed on '%s'", ldirname);
		maLog(cstr);
		return 0;
	}

	while((dt = readdir(dir)))
	{
 		if(strcmp(dt->d_name, ".") && strcmp(dt->d_name, "..") &&
		   strcmp(dt->d_name, ".DS_Store") && strcmp(dt->d_name, "Icon\015") && strcmp(dt->d_name, "Icon"))
		{
			if(g_sets.usem3u || !FileNameHasExtension(dt->d_name, EXTM3U))
			{
				sprintf(file, "%s%c%s", ldirname, FILESEP, dt->d_name);
				p = file;
				maSongListResolveItem(fp, p, pntracks, pnfiles, pnlists, pnslfiles, sglist);
			}
		}
	}
	n = *pnfiles - n;
	
	if(closedir(dir) == -1)
	{
		sprintf(cstr, "maSongListResolveFolder closedir errno=%d on '%s'", errno, ldirname);
		maLog(cstr);
	}

#if !defined(MANOCDA)
	size_t i;
	if(n <= 0) // 'empty' dir is possibly a cd where the os did not enumerate the tracks
	{
		int drive;
		if((drive = IsAudioCd(dirname)))
		{
			for(i = 1; i <= (size_t)g_cdadrive[drive].tracklast; i++)
			{
				if(fprintf(fp, "%s%cTrack%02d%s\n", dirname, FILESEP, (int)i, EXTCDA) <= 0)
				{
					char cstr[CSTR_MAX_PATH];
					sprintf(cstr, "maSongListResolveFolder fprintf errno=%d'", errno);
					maLog(cstr);
					return 1;
				}
				(*pntracks)++;
			}
		}
	}
#endif
			
	return 0;

} // maSongListResolveFolder __MAOSX__
#endif


#if defined(__MAUNIX__)
int maSongListResolveItem(FILE *fp, char *p, int *pntracks, int *pnfiles, int *pnlists, int *pnslfiles, char *sglist)
{
	struct stat finfo;
	char linkpath[PATH_MAX]; // SONG_MAX_PATH
	char cstr[CSTR_MAX_PATH];

#if !defined(MANOCDA)
	int drive;
	int i;
	if((drive = IsAudioCd(p)))
	{
		for(i = 1; i <= g_cdadrive[drive].tracklast; i++)
		{
			if(fprintf(fp, "%s/Track%02d%s\n", g_cdadrive[drive].name, i, EXTCDA) <= 0)
			{
				sprintf(cstr, "maSongListResolveItem fprintf errno=%d", errno);
				maLog(cstr);
				return 1;
			}
			(*pntracks)++;
		}
		return 0;
	}
#endif

	if(lstat(p, &finfo))
	{
		if(errno == ENOENT) // 2
			sprintf(cstr, "maSongListResolveItem File not found '%s'", p);
		else
			sprintf(cstr, "maSongListResolveItem lstat failed on '%s' errno=%d", p, errno);
		maLog(cstr);
		return 1;
	}

	switch(finfo.st_mode & S_IFMT)
	{
		case S_IFREG : // regular file
		return maSongListResolveFile(fp, p, pntracks, pnfiles, pnlists, pnslfiles, sglist);
			
		case S_IFDIR : // directory
			if(maSongListResolveFolder(fp, p, pntracks, pnfiles, pnlists, pnslfiles, sglist))
			{
				sprintf(cstr, "maSongListResolveItem maSongListResolveFolder failed on '%s'", p);
				maLog(cstr);
				return 1;
			}
		return 0;
		
		case S_IFLNK : // symbolic link
			if(realpath((const char *)p, linkpath) == NULL)
			{
				sprintf(cstr, "maSongListResolveItem realpath failed on '%s' errno=%d", p, errno);
				maLog(cstr);
				return 1;
			}
		return maSongListResolveItem(fp, linkpath, pntracks, pnfiles, pnlists, pnslfiles, sglist);

		default:
		break;
	}
		
	return -1; // other

 } // maSongListResolveItem // __MAUNIX__
 
 
int maSongListResolveFolder(FILE *fp, char *dirname, int *pntracks, int *pnfiles, int *pnlists, int *pnslfiles, char *sglist)
{
	char *p;
	int n;
	
#if defined(MACUNIX)
	char ldirname[256];
	strcpy(ldirname, dirname);
	StrOSXtoStrBSD(ldirname);
#else
	char *ldirname = dirname;
#endif
	
	DIR *dir;
	struct dirent *dt;
	char file[SONG_MAX_PATH];
	char cstr[CSTR_MAX_PATH];

	n = *pnfiles;

	if(g_sets.usem3u && l_onlym3uindir) // scan for and use only m3u
	{
		if((dir = opendir(ldirname)) == NULL)
		{
			sprintf(cstr, "maSongListResolveFolder opendir failed on '%s'", ldirname);
			maLog(cstr);
			return 0;
		}

		p = NULL;
		while((dt = readdir(dir)))
		{
			if(FileNameHasExtension(dt->d_name, EXTM3U))
			{
				sprintf(file, "%s%c%s", ldirname, FILESEP, dt->d_name);
				p = file;
				maSongListResolveItem(fp, p, pntracks, pnfiles, pnlists, pnslfiles, sglist);
				// break;
			}
		}
		if(closedir(dir) == -1)
		{
			sprintf(cstr, "maSongListResolveFolder closedir errno=%d on '%s'", errno, ldirname);
			maLog(cstr);
		}
		if(p)
			return 0;
	}

	if((dir = opendir(ldirname)) == NULL)
	{
		sprintf(cstr, "maSongListResolveFolder opendir failed on '%s'", ldirname);
		maLog(cstr);
		return 0;
	}
	
	while((dt = readdir(dir)))
	{
 		if(strcmp(dt->d_name, ".") && strcmp(dt->d_name, ".."))
		{
			if(g_sets.usem3u || !FileNameHasExtension(dt->d_name, EXTM3U))
			{
				sprintf(file, "%s%c%s", ldirname, FILESEP, dt->d_name);
				p = file;
				maSongListResolveItem(fp, p, pntracks, pnfiles, pnlists, pnslfiles, sglist);
			}
		}
	}
	n = *pnfiles - n;
	
	if(closedir(dir) == -1)
	{
		sprintf(cstr, "maSongListResolveFolder closedir errno=%d on '%s'", errno, ldirname);
		maLog(cstr);
	}
			
	return 0;

} // maSongListResolveFolder // __MAUNIX__
#endif


#ifdef __MAWIN__
 #define EXTWINLNK ".lnk"
int maSongListResolveItem(FILE *fp, char *p, int *pntracks, int *pnfiles, int *pnlists, int *pnslfiles, char *sglist) // __MAWIN__
{
	char outfname[SONG_MAX_PATH];
	char cstr[CSTR_MAX_PATH];

	int sz = maGetLongPathName(p, outfname, SONG_MAX_PATH);
	if(!sz)
		strcpy(outfname, p);

	if(FileNameHasExtension(outfname, EXTWINLNK)) // e.g ...tune.mp3.lnk
	{
		if(maWinResolveShortcut(outfname, outfname))	
		{
			sprintf(cstr, "maSongListResolveItem maWinResolveShortcut failed on '%s'", outfname);
			maLog(cstr);
			return 1;
		}
	}

	if(maDirExists(outfname))
	{
		if(maSongListResolveFolder(fp, outfname, pntracks, pnfiles, pnlists, pnslfiles, sglist))
		{
			sprintf(cstr, "maSongListResolveItem maSongListResolveFolder failed on '%s'", outfname);
			maLog(cstr);
			return 1;
		}
		return 0;
	}
	
	// maybe an audio cd e.g. "D:/", in case maDirExists did not catch a cd
	if(strlen(outfname) == 3 && IsAudioCd(outfname))
	{
		if(maSongListResolveFolder(fp, outfname, pntracks, pnfiles, pnlists, pnslfiles, sglist))
		{
			sprintf(cstr, "maSongListResolveItem maSongListResolveFolder failed on '%s'", outfname);
			maLog(cstr);
			return 1;
		}
		return 0;
	}

	// not a link and not a folder and not an audio cd, must be a file
	return maSongListResolveFile(fp, outfname, pntracks, pnfiles, pnlists, pnslfiles, sglist);

} // maSongListResolveItem // __MAWIN__


int maSongListResolveFolder(FILE *fp, char *dirname, int *pntracks, int *pnfiles, int *pnlists, int *pnslfiles, char *sglist)
{
	HANDLE hFind;
	WIN32_FIND_DATA FindData;
	char dirwild[SONG_MAX_PATH];
	char file[SONG_MAX_PATH];
	
#if !defined(MANOCDA)
	int i;
	int drive;
	if(strlen(dirname) == 3 && dirname[1] == ':' && dirname[2] == '\\') // e.g. 'D:\"
	{
		char fname[256];
		char cstr[CSTR_MAX_PATH];
		sprintf(fname, "%sTrack01%s", dirname, EXTCDA);
		if((drive = IsAudioCd(fname)))
		{
			for(i = 1; i <= g_cdadrive[drive].tracklast; i++)
			{
				if(fprintf(fp, "%sTrack%02d%s\n", dirname, i, EXTCDA) <= 0)
				{
					sprintf(cstr, "maSongListResolveFolder fprintf errno=%d'", errno);
					maLog(cstr);
					return 1;
				}
				(*pntracks)++;
			}
			return 0;
		}
	}
#endif

	sprintf(dirwild, "%s%c*", dirname, FILESEP);

	if(g_sets.usem3u && l_onlym3uindir) // scan for and use only m3u
	{
		hFind = FindFirstFile(dirwild, &FindData);
		if(hFind == INVALID_HANDLE_VALUE)
			return 1;
		file[0] = 0;
		while(1)
		{
			if(FileNameHasExtension(FindData.cFileName, EXTM3U))
			{
				sprintf(file, "%s%c%s", dirname, FILESEP, FindData.cFileName);
				maSongListResolveItem(fp, file, pntracks, pnfiles, pnlists, pnslfiles, sglist);
				//FindClose(hFind);
				// return 0;
			}
			if(!FindNextFile(hFind, &FindData))
				break;
		}
		FindClose(hFind);
		if(file[0])
			return 0;
	}
	
	hFind = FindFirstFile(dirwild, &FindData);
	if(hFind == INVALID_HANDLE_VALUE)
		return 1;
	while(1)
	{
		if(strcmp(FindData.cFileName, ".") && strcmp(FindData.cFileName, ".."))
		{
			if(g_sets.usem3u || !FileNameHasExtension(FindData.cFileName, EXTM3U))
			{
				sprintf(file, "%s%c%s", dirname, FILESEP, FindData.cFileName);
				maSongListResolveItem(fp, file, pntracks, pnfiles, pnlists, pnslfiles, sglist);
			}
		}
		if(!FindNextFile(hFind, &FindData))
			break;
	}
	FindClose(hFind);

	return 0;

} // maSongListResolveFolder __MAWIN__
#endif
		

int maSongListResolveFile(FILE *fp, char *outfname, int *pntracks, int *pnfiles, int *pnlists, int *pnslfiles, char *sglist)
{
	char cstr[CSTR_MAX_PATH];

	if(FileNameHasExtension(outfname, EXTMAF)) // a moreamp songlist
	{
		if(strcmp(outfname, g_songlistcda) && strcmp(outfname, g_songlistdrop) &&
			strcmp(outfname, g_songlisttemp))
		{
			if((*pnlists)++ == 0)
				strcpy(sglist, outfname); // one single songlist resolved to this
				
			if(maSongListAddFromSongList(fp, outfname, pnfiles, pnslfiles))
			{
				sprintf(cstr, "maSongListResolveItem maSongListAddFromSongList failed on '%s'", outfname);
				maLog(cstr);
				return 1;
			}
		}
	}
	else if(FileNameHasExtension(outfname, EXTM3U)) // a standard m3u playlist (not an Extended M3u playlist)
	{
		if(maSongListAddFromSongListm3u(fp, outfname, pnfiles, pnslfiles))
		{
			sprintf(cstr, "maSongListResolveItem maSongListAddFromSongListm3u failed on '%s'", outfname);
			maLog(cstr);
			return 1;
		}
	}
	else if(FileIsPlayable(outfname)) // a song file or audio cd track
	{
#if !defined(MANOCDA)
 #if defined(__MAWIN__) // a cd is a dir with tracks as files
		if(IsAudioCd(outfname))
			(*pntracks)++;
		else
 #endif
#endif
		(*pnfiles)++;

		if(fprintf(fp, "%s\n", outfname) <= 0)
		{
			sprintf(cstr, "maSongListResolveItem fprintf errno=%d", errno);
			maLog(cstr);
			return 1;
		}
		
		/* // test
		if(g_w[WINDOWAPP])
		{
			sprintf(cstr, "Scanning...%d", *pnfiles + *pnslfiles);
			WindowSetTitle(WINDOWAPP, cstr); //SetAppWindowTitle(cstr);
		}
		*/
	}

	return 0;

} // maSongListResolveFile


int maSongListAddFromSongList(FILE *fp, char *infname, int *pnfiles, int *pnslfiles)
{
	FILE *infp;
	char item[SONG_MAX_PATH];
	char cstr[CSTR_MAX_PATH];
	
	if((infp = mafopen(infname, "rb")) == NULL)
	{
		sprintf(cstr, "maSongListAddFromSongList mafopen failed on '%s'", infname);
		maLog(cstr);
		return 1;
	}
	
	while(mafgets(item, SONG_MAX_PATH, infp) != NULL)
	{
		// skip lines with '#' or files that are not playable or don't exist
		if(item[0] != '#')
		{
			if(FileIsPlayable(item)) // && SongFileExists(item))
			{
				(*pnslfiles)++;
						
				if(fprintf(fp, "%s\n", item) <= 0)
				{
					sprintf(cstr, "maSongListAddFromSongList fprintf errno=%d", errno);
					maLog(cstr);
					fclose(infp);
					return 1;
				}
				
				if(g_w[WINDOWAPP])
				{
					sprintf(cstr, "Scanning...%d", *pnfiles + *pnslfiles);
					WindowSetTitle(WINDOWAPP, cstr); //SetAppWindowTitle(cstr);
				}
			}
			/*
			else
			{
				sprintf(cstr, "maSongListAddFromSongList file not playable or not found: '%s'", item);
				maLog(cstr);
			}
			*/
		} // if !#
		
	} // while
			
	fclose(infp);
	
	return 0;
	
} // maSongListAddFromSongList


int maSongListAddFromSongListm3u(FILE *fp, char *infname, int *pnfiles, int *pnslfiles)
{
	FILE *infp;
	char *p, *p2;
	char item[SONG_MAX_PATH];
	char m3udir[SONG_MAX_PATH];
	char m3uitem[SONG_MAX_PATH];
	char cstr[CSTR_MAX_PATH];
	int ntracks = 0;
	int nlists = 0;
	
	if((infp = mafopen(infname, "rb")) == NULL)
	{
		sprintf(cstr, "maSongListAddFromSongListm3u mafopen failed on '%s'", infname);
		maLog(cstr);
		return 1;
	}
	
	// the dir holding the m3u file
	strcpy(m3udir, infname);
	p = m3udir;
	p2 = NULL;
	while((p = strchr(p, FILESEP)) != NULL)
		p2 = p++;
	if(p2 == NULL)
	{
		sprintf(cstr, "maSongListAddFromSongListm3u failed to get m3u dir from '%s'", infname);
		maLog(cstr);
		fclose(infp);
		return 1;
	}
	*p2 = 0;
	
	while(mafgets(item, SONG_MAX_PATH, infp) != NULL)
	{
		// skip lines with '#' or files that are not playable or don't exist
		if(item[0] != '#')
		{
			if(maIsFullPath(item))
				p = item;
			else
			{
				sprintf(m3uitem, "%s%c%s", m3udir, FILESEP, item);
				p = m3uitem;
			}
			
			if(maSongListResolveItem(fp, p, &ntracks, pnfiles, &nlists, pnslfiles, (char *)""))
			{
				sprintf(cstr, "maSongListAddFromSongListm3u maSongListResolveItem failed on '%s'", p);
				maLog(cstr);
			}
			if(g_w[WINDOWAPP])
			{
				sprintf(cstr, "Scanning...%d", *pnfiles + *pnslfiles);
				WindowSetTitle(WINDOWAPP, cstr); //SetAppWindowTitle(cstr);
			}
		}
		
	} // while
			
	fclose(infp);
	
	return 0;
	
} // maSongListAddFromSongListm3u


int maSongListMake(int nitems, char **pitems, char *newsonglist)
{
	int iscda = 0;
	int isfile = 0;
	int islist = 0;
	int isrefile = 0;
	int concat = 0;
	int err;
	FILE *fpsonglisttemp;
	char selectedpath[SONG_MAX_PATH];
	char lnewsonglist[SONG_MAX_PATH];
	char cstr[CSTR_MAX_PATH];

	if(nitems == 1 && FileNameHasExtension(*pitems, EXTMAF)) // one single songlist
	{
		strcpy(newsonglist, *pitems); // no item to resolve or new list to build
		return 0;
	}
	
	// open a file for writing a temp songlist
	if((fpsonglisttemp = mafopen(g_songlisttemp, (char *)"w")) == NULL)
	{
		sprintf(cstr, "maSongListMake mafopen failed on '%s'", g_songlisttemp);
		maLog(cstr);
		return 1;
	}

#if defined(__MAOSX__) || defined(__MAUNIX__)
	machmod(g_songlisttemp, ACCESSPERMS); // 0777
#endif

	if((err = maSongListAddItems(fpsonglisttemp, nitems, pitems, &iscda, &isfile, &islist, &isrefile, lnewsonglist)))
	{
		sprintf(cstr, "maSongListMake maSongListAddItems failed on '%s'", pitems[0]);
		maLog(cstr);
	}

	fclose(fpsonglisttemp);
	fpsonglisttemp = NULL;

	if(err)
		return 1;

	if(!iscda && !isfile && !islist && !isrefile)
	{
#if !defined(NODECODEAUDIOIN)
		if(nitems == 1 && !FileNameHasExtension(*pitems, EXTMAA))
#endif
		{
			if(g_sets.addmode)
				maLog("maSongListMake no new items for the songlist");
			else
				maLog("maSongListMake no items for a new songlist");
			return 1;
		}
	}
	
	if(!iscda && !isfile && islist == 1 && isrefile) // one single songlist resolved to lnewsonglist
	{
		if(newsonglist)
			strcpy(newsonglist, lnewsonglist);
		return 0;
	}
	
	if(newsonglist && newsonglist[0]) // prompt for new songlist name
	{
		char extmaf[64];
		sprintf(extmaf, "%s file (%s)|%s", g_prognametitle, g_extstr[kMAMAF], g_extstr[kMAMAF]);
		if(FilenameFromPathname(newsonglist, newsonglist))
		{
			sprintf(cstr, "maSongListMake failed to get filename from '%s'", newsonglist);
			maLog(cstr);
			newsonglist[0] = 0;
		}					

		if(maFileDialog(selectedpath, WINDOWAPP,
				(char *)"Enter a new songlist name or select a list to overwrite", g_songlistsdir, newsonglist,
				extmaf, maSAVE | maOVERWRITE_PROMPT))
			return 1;

		// unless user selected SONGLISTTMP, check to overwrite
		if(mastrcasecmp(g_songlisttemp, selectedpath)) 
		{
			// append extension if missing
			if(!FileNameHasExtension(selectedpath, g_extstr[kMAMAF]))
			{
				strcat(selectedpath, g_extstr[kMAMAF]);
	
				// overwrite ? (maOVERWRITE_PROMPT missed the name that user typed in without an extension)
				if(maFileExists(selectedpath))
				{
					sprintf(cstr, "This songlist already exists and will be overwritten\n\n'%s'", selectedpath);
					if(maMessageBox(cstr, (char *)"CONTINUE ?", maYES_NO | maICON_QUESTION) != maYES)
						return 1;
				}
			}
		}
		strcpy(lnewsonglist, selectedpath);
	}
	
#ifndef MANOCDA
	else if(iscda && !isfile)
		strcpy(lnewsonglist, g_songlistcda);
#endif
	else
	{
		strcpy(lnewsonglist, g_songlistdrop);
		if(g_sets.addmode)
			concat = 1; // do not clear current list before adding dropped files
	}
		
	// we are stopped when doing this but lnewsonglist may still be open
	if(!mastrcasecmp(lnewsonglist, g_songlist))
	{		
		if(g_fpsonglist)
		{
			fclose(g_fpsonglist);
			g_fpsonglist = NULL;
		}	
	}
		
	// unless user selected SONGLISTTMP, rename it
	if(mastrcasecmp(g_songlisttemp, selectedpath)) 
	{
		// delete lnewsonglist if it already exists
		if(maFileExists(lnewsonglist))
		{
			if(!concat && maunlink(lnewsonglist))
			{
				sprintf(cstr, "maSongListMake maunlink errno=%d '%s'", errno, lnewsonglist);
				maLog(cstr);
				return 1;
			}		
		}
		else
			concat = 0; // no file to concat to
		
		// copy file songlisttemp to new songlist
		if((err = maFileCopy(g_songlisttemp, lnewsonglist, concat)) != 0)
		{
			sprintf(cstr, "maSongListMake error %d copying '%s' to '%s'", err, g_songlisttemp, lnewsonglist);
			maLog(cstr);
			return 1;
		}

		// delete temp song file
		if(maunlink(g_songlisttemp))
		{
			sprintf(cstr, "maSongListMake maunlink errno=%d '%s'", errno, g_songlisttemp);
			maLog(cstr);
		}
	}
	
	if(newsonglist)
		strcpy(newsonglist, lnewsonglist);
		
	return 0;
	
} // maSongListMake	



int SaveSongListAs()
{
	int err;
	char selectedpath[SONG_MAX_PATH];
	char newsettingsfile[SONG_MAX_PATH];
	char cstr[CSTR_MAX_PATH];

	// prompt for file name to save to
	char extmaf[64];
	sprintf(extmaf, "%s file (%s)|%s", g_prognametitle, g_extstr[kMAMAF], g_extstr[kMAMAF]);
	if(maFileDialog(selectedpath, WINDOWAPP,
			(char *)"Enter a new songlist name or select a list to overwrite", g_songlistsdir, (char *)"",
				extmaf, maSAVE | maOVERWRITE_PROMPT))
		return 1;

	// append extension if missing
	if(!FileNameHasExtension(selectedpath, g_extstr[kMAMAF]))
	{
		strcat(selectedpath, g_extstr[kMAMAF]);
	
		// overwrite ? (maOVERWRITE_PROMPT missed the name that user typed in without an extension)
		if(maFileExists(selectedpath))
		{
			sprintf(cstr, "This songlist already exists and will be overwritten\n\n'%s'", selectedpath);
			if(maMessageBox(cstr, (char *)"CONTINUE ?", maYES_NO | maICON_QUESTION) != maYES)
				return 1;
		}
	}
	
	// if the songlist selected is the current songlist, just save settings
	if(!mastrcasecmp(selectedpath, g_songlist)) 
	{
		memcpy(&g_setsbu, &g_sets, sizeof(MASETS));
		SaveSettings();
		return 0;
	}
	
	// close the old file
	if(g_fpsonglist)
	{
		fclose(g_fpsonglist);
		g_fpsonglist = NULL;
	}
	
	// delete file name selected if it already exists
	if(maFileExists(selectedpath))
	{
		if(maunlink(selectedpath))
		{
			sprintf(cstr, "SaveSongListAs failed to delete '%s'", selectedpath);
			maLog(cstr);
			return 1;
		}		
	}
		
	// copy current g_songlist to file name selected from prompt
	if((err = maFileCopy(g_songlist, selectedpath, 0)) != 0)
	{
		sprintf(cstr, "%d SaveSongListAs copying '%s' to '%s'", err, g_songlist, selectedpath);
		maLog(cstr);
		return 1;
	}

	// copy saved settings file to new settings file
	if(FilenameFromPathname(selectedpath, cstr))
	{	
		sprintf(cstr, "SaveSongListAs failed to get filename from '%s'", selectedpath);
		maLog(cstr);
		return 1;
	}
	FileNameRemoveExtension(cstr);
	sprintf(newsettingsfile, "%s%c%s%s", g_settingsdir, FILESEP, cstr, EXTMAS);
	if(maFileExists(newsettingsfile))
		maunlink(newsettingsfile);
	if(maFileExists(g_settingspath))
	{
		if((err = maFileCopy(g_settingspath, newsettingsfile, 0)) != 0)
		{
			sprintf(cstr, "SaveSongListAs error %d copying '%s' to '%s'", err, g_settingspath, newsettingsfile);
			maLog(cstr);
			return 1;
		}
	}
	
	// save current active settings to new settings file
	strcpy(cstr, g_settingspath);
	strcpy(g_settingspath, newsettingsfile);
	memcpy(&g_setsbu, &g_sets, sizeof(MASETS));
	SaveSettings();
	strcpy(g_settingspath, cstr);
	
	// new songlist becomes current songlist
	if(ChangeSongList(selectedpath))
		return 1;

	UpdateRecent(g_songlist, 0);

	SetAppWindowTitle((char *)"Stopped");
	
	return 0;

} // SaveSongListAs


#if defined(__MAOSX__)
 // make a songlist from files, dirs or links passed in an AEvent to the program
 void DoAEOpenDoc(void *inAppleEvent)
 {
	int nitems;
	int arraysize;
	int i;
	int err = 0;
	char **p;
	char *q;
	char **pitems = NULL;
	char *items = NULL; // null separated items
	char fname[SONG_MAX_PATH + SONG_MAX_PATH];
	SInt32 numDocs;
	AEDescList docList;
	AEKeyword theKey;
	DescType theType;
	FSRef theRef;
	Size theSize;

	// get CDRom drives
 #ifndef MANOCDA
	CdaInitCDRom();
 #endif

	// build an array of numDocs null spearated items
		
	if(AEGetParamDesc((const AppleEvent *)inAppleEvent, keyDirectObject, typeWildCard, &docList))
	{
		maLog("DoAEOpenDoc AEGetParamDesc failed");
		goto errorret;
	}
	
	if(AECountItems(&docList, &numDocs))
	{
		maLog("DoAEOpenDoc AECountItems failed");
		goto errorret;
	}
				
	// get arraysize needed for items
	arraysize = numDocs;
	for(i = 1; i <= numDocs; i++)
	{
		if(AEGetNthPtr(&docList, i, typeFSRef, &theKey, &theType,
							(Ptr)&theRef, sizeof(FSRef), &theSize) != noErr)
		{
			maLog("DoAEOpenDoc AEGetNthPtr failed");
			goto errorret;
		}
		
		if(FSRefMakePath(&theRef, (UInt8 *)fname, sizeof(fname)) != noErr)
		{
			maLog("DoAEOpenDoc FSRefMakePath failed");
			goto errorret;
		}

		StrBSDtoStrOSX(fname);			
		arraysize += strlen(fname);
	} // for i
		
	nitems = numDocs;
	
	// allocate array members
	if((pitems = (char **)MAMALLOC(numDocs * sizeof(char *))) == NULL)
	{
		maLog("DoAEOpenDoc MAMALLOC failed");
		goto errorret;
	}
	
	if((items = (char *)MAMALLOC(arraysize)) == NULL)
	{
		maLog("DoAEOpenDoc MAMALLOC failed");
		goto errorret;
	}
	memset(items, 0, arraysize);
		
	// fill in array
	p = pitems;
	q = items;
	for(i = 1; i <= numDocs; i++, p++)
	{
		if(AEGetNthPtr(&docList, i, typeFSRef, &theKey, &theType,
							(Ptr)&theRef, sizeof(FSRef), &theSize))
		{
			maLog("DoAEOpenDoc AEGetNthPtr failed");
			goto errorret;
		}
		
		if(FSRefMakePath(&theRef, (UInt8 *)fname, sizeof(fname)) != noErr)
		{
			maLog("DoAEOpenDoc AEGetNthPtr failed");
			goto errorret;
		}

		StrBSDtoStrOSX(fname);			
		*p = q;
		strcpy(q, fname);
		q += strlen(fname) + 1;
	} // for i
		
	AEDisposeDesc(&docList);
		
	// make a new songlist, either madrop or macda
	fname[0] = 0;
	err = maSongListMake(nitems, pitems, fname); // DoAEOpenDoc

errorret:
	if(items)
		MAFREE(items);
	if(pitems)
		MAFREE(pitems);

	if(!err)
		PlaySonglist(fname);
	else if(!g_w[WINDOWAPP])
	{
		LoadDefaultSettings(&g_sets);
		AppWindowMakeA();
	}
	else
		ResetAppTitle();
	 
 } // DoAEOpenDoc // __MAOSX__

#else
// make a songlist from files, dirs or links passed as args to the program
 int DoArgsOpenDoc()
 {
	int err = 0;
	char newsonglist[SONG_MAX_PATH];

	// make a new songlist, either madrop or macda
	newsonglist[0] = 0;
	err = maSongListMake(g_argc - 1, &g_argv[1], newsonglist); // DoArgsOpenDoc
	
	if(!err)
		PlaySonglist(newsonglist);
	else if(!g_w[WINDOWAPP])
	{
		LoadDefaultSettings(&g_sets);
		AppWindowMakeA();
	}			
	else
		ResetAppTitle();
	
	return 0;
	
 } // DoArgsOpenDoc
#endif


int DoPassedArgs()
{
#if defined(__MAOSX__)
 EventRecord event;
	return EventAvail(highLevelEventMask, &event) && event.message == kCoreEventClass &&
				*(OSType *)&event.where == kAEOpenDocuments; // DoAEOpenDoc() gets called
 /* EventAvail above is deprecated, use: // in 10.6, 10.5?
 
	//const EventTypeSpec inList[] = {{/ * kCoreEventClass * / kEventClassApplication, / * kAEOpenDocuments * / kEventAppleEvent}};
	//const EventTypeSpec inList[] = {{kEventClassAppleEvent, kEventAppleEvent}};

	//if(AcquireFirstMatchingEventInQueue(GetCurrentEventQueue(), 1, inList, kEventQueueOptionsNone))
	EventRecord event;
	if(EventAvail(highLevelEventMask, &event) && event.message == kCoreEventClass &&
		   *(OSType *)&event.where == kAEOpenDocuments) // DoAEOpenDoc() gets called
	{
		//ShowErr("AcquireFirstMatchingEventInQueue YES");
		printf("AcquireFirstMatchingEventInQueue YES\n"); fflush(stdout);
		return 1;
	}
	//ShowErr("AcquireFirstMatchingEventInQueue NO");
	printf("AcquireFirstMatchingEventInQueue NO\n"); fflush(stdout);
	return 0;
 */ 
 /*
 AcquireFirstMatchingEventInQueue
 Obtains the first event that matches the specified list of event classes and kinds.
 
 EventRef AcquireFirstMatchingEventInQueue (
 EventQueueRef inQueue,
 ItemCount inNumTypes,
 const EventTypeSpec *inList,
 OptionBits inOptions
 );
 Parameters
 inQueue
 The queue to check.
 
 inNumTypes
 The number of event kinds for which to search. You may pass 0 if you also pass NULL for inList.
 
 inList
 The list of event classes and kinds to search for in the queue. You may pass NULL if inNumTypes is 0. Doing so effectively matches any event in the queue and causes this function to return the first event in the queue.
 
 inOptions
 Must be kEventQueueOptionsNone.
 
 Return Value
 An event reference, or NULL if no events match. The reference count for the event has been incremented (that is, it has been retained), so you must release the event reference.
 
 Discussion
 This function does not remove the event from the queue. To remove the event, call RemoveEventFromQueue.
 
 This function does not call the run loop, so no timers fire as a result of calling this function. This function does not cause any window flushing to occur, but it does get new events from the window server.
 
 This function should have better performance characteristics than the older EventAvail API.
 
 Availability
 Available in Mac OS X v10.3 and later.
 Declared In
 CarbonEventsCore.h
 
 ---
 GetCurrentEventQueue
 Obtains the current event queue.
 
 EventQueueRef GetCurrentEventQueue (
 void
 );
 Return Value
 An event queue reference.
 
 Discussion
 This function obtains the event queue for the current thread. If the current thread is a cooperative thread, the main event queue is returned.
 
 Availability
 Available in Mac OS X v10.0 and later.
 ---
 struct EventTypeSpec {
 UInt32 eventClass;
 UInt32 eventKind;
 };
 typedef struct EventTypeSpec EventTypeSpec;
 
 typedef UInt32 EventClass;
 enum {
 kEventClassMouse = 'mous',
 kEventClassKeyboard = 'keyb',
 kEventClassTextInput = 'text',
 kEventClassApplication = 'appl',
 kEventClassAppleEvent = 'eppc',
 kEventClassMenu = 'menu',
 kEventClassWindow = 'wind',
 kEventClassControl = 'cntl',
 kEventClassCommand = 'cmds',
 kEventClassTablet = 'tblt',
 kEventClassVolume = 'vol ',
 kEventClassAppearance = 'appm',
 kEventClassService = 'serv',
 kEventClassToolbar = 'tbar',
 kEventClassToolbarItem = 'tbit',
 kEventClassToolbarItemView = 'tbiv',
 kEventClassAccessibility = 'acce'.
 kEventClassSystem = 'macs',
 kEventClassInk = 'ink ',
 kEventClassTSMDocumentAccess = 'tdac'
 };
 ---
RemoveEventFromQueue
 Removes an event from the event queue.
 
 OSStatus RemoveEventFromQueue (
 EventQueueRef inQueue,
 EventRef inEvent
 );
 Parameters
 inQueue
 The queue to remove the event from.
 
 inEvent
 The event to remove.
 
 Return Value
 A result code. See “Carbon Event Manager Result Codes.”
 
 Discussion
 Removes the given event from the queue on which it was posted and decrements its retain count.
 
 Availability
 Available in Mac OS X v10.0 and later.
 */
	
#else
	if(g_argc <= 1) // no passed args
		return 0;
	DoArgsOpenDoc();
	return 1;
#endif
}

#ifdef __cplusplus
 }
#endif

// eof
