// maWriteAif.c

/*
 * MoreAmp
 * Copyright (C) 2004-2010 moreforge
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

// extracts from madplay

#include "maConfig.h" // included by all moreamp files

#ifdef __MAWIN__
 #pragma warning( disable : 4206 ) // nonstandard extension used : translation unit is empty
 #pragma warning( disable : 4514 ) // mad_f_mul_inline : unreferenced inline function has been removed
 #pragma warning( disable : 4706 ) // assignment within conditional expression
#endif

#ifdef ENCODEAIF

#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <stdlib.h>
#include <ctype.h>
#include <time.h>
#include <errno.h>

#ifdef __MAOSX__
 #include <sys/stat.h>
#endif

#include "maDefs.h"
#include "maPlayer.h"
#include "maUtils.h"
#include "maWriteAif.h"
 
#ifdef __cplusplus
extern "C"
{
#endif

static unsigned long nsamples;
static unsigned long form_len, ssnd_len;
static long comm_chunk, ssnd_chunk;

#define UNKNOWN_LENGTH "\xff\xff\xff\xff"

#if defined(__MAOSX__)
 #define WAVE_FORMAT_PCM 0x0001
#endif

#ifdef __MAUNIX__
 #define WAVE_FORMAT_PCM 0x0001
#endif

void int32(unsigned char *ptr, unsigned long num);
void int16(unsigned char *ptr, unsigned int num);
void float80(unsigned char *ptr, signed long num);
int patch_int32(long address, unsigned long num);

int OpenWriteAif()
{
	unsigned char chunk[8 + 18];
	char cstr[256];

	// group ID and file type ID header
	if(fwrite("FORM" UNKNOWN_LENGTH "AIFF", 4 + 4 + 4, 1, g_outputfilefp) != 1)
		return 1;
		
	comm_chunk = ftell(g_outputfilefp);

	memcpy(&chunk[0], (char *)"COMM", 4);	// chunkID
	int32(&chunk[4], 18);					// chunkSize
	int16(&chunk[8], MAOUTCHANNELS);		// numChannels
	int32(&chunk[10], ~0UL);				// numSampleFrames
	int16(&chunk[14], MAOUTBITS);			// bitdepth // sampleSize
	float80(&chunk[16], g_state.insamplerate);

	if(fwrite(chunk, 8 + 18, 1, g_outputfilefp) != 1)
	{
		sprintf(cstr, "OpenWriteAif fwrite '%s'", g_outputfile);
		maLog(cstr);
		return 1;
	}

	form_len += 8 + 18;

	// Sound Data chunk
	ssnd_chunk = ftell(g_outputfilefp);

	memcpy(&chunk[0], (char *)"SSND", 4);	// chunkID
	int32(&chunk[4], (unsigned long)~0L);	// chunkSize

	int32(&chunk[8], 0);					// offset
	int32(&chunk[12], 0);					// blockSize

	if(fwrite(chunk, 8 + 8, 1, g_outputfilefp) != 1)
	{
		sprintf(cstr, "OpenWriteAif fwrite '%s'", g_outputfile);
		maLog(cstr);
		return 1;
	}

	form_len += 8 + 8;
	ssnd_len = 8;

	nsamples = 0;

	return 0;

} // OpenWriteAif


int WriteAif(void* pcmblock, unsigned long frames, unsigned long len)
{
	unsigned char *data = (unsigned char *)pcmblock;
	char cstr[256];
	
#ifndef WORDS_BIGENDIAN
	// aif is big-endian, pcmblock is little-endian
	unsigned char d;
	unsigned long i;	
	for(i = 0; i < len; i += 2)
	{
		d = data[i];
		data[i] = data[i + 1];
		data[i + 1] = d;	
	}
#endif

	if(fwrite(data, len, 1, g_outputfilefp) != 1)
	{
		sprintf(cstr, (char *)"WriteAif fwrite error");
		maLog(cstr);
		return 1;
	}
 
	nsamples += frames;

	ssnd_len += len;
	form_len += len;
	
	return 0;
	
} // WriteAif


int CloseWriteAif()
{
	int result = 0;
	char cstr[256];

	if(ssnd_len & 1)
	{
		if(fputc(0, g_outputfilefp) == EOF && result == 0)
		{
			sprintf(cstr, "CloseWriteAif fputc '%s'", g_outputfile);
			maLog(cstr);
			return 1;
		}
		
		++form_len;
	}

	if(result == 0 &&
			((comm_chunk != -1 && patch_int32(comm_chunk + 10, nsamples) == -1) ||
			 (ssnd_chunk != -1 && patch_int32(ssnd_chunk +  4, ssnd_len) == -1)))
		result = -1;

	if(result == 0)
		patch_int32(4, form_len);

	return 0;

} // CloseWriteAif


void int32(unsigned char *ptr, unsigned long num)
{
	ptr[0] = (unsigned char)(num >> 24);
	ptr[1] = (unsigned char)(num >> 16);
	ptr[2] = (unsigned char)(num >>  8);
	ptr[3] = (unsigned char)(num >>  0);
}

void int16(unsigned char *ptr, unsigned int num)
{
	ptr[0] = (unsigned char)(num >> 8);
	ptr[1] = (unsigned char)(num >> 0);
}

void float80(unsigned char *ptr, signed long num)
{
	if(num == 0)
		memset(ptr, 0, 10);
	else {
		int exp;
		unsigned int s, e;
		unsigned long abs, sr, f;

		enum {
			BIAS = 0x3fff
		};

		s = num < 0;
		abs = s ? -num : num;

		for(exp = -1, sr = abs; sr; sr >>= 1)
			++exp;

		e = exp + BIAS;

		// write normalized value; high bit of f (i) is always set
		f = abs << (32 - exp - 1);

		ptr[0] = (unsigned char)((s <<  7) | (e >> 8));
		ptr[1] = (unsigned char)((e >>  0));
		ptr[2] = (unsigned char)((f >> 24));
		ptr[3] = (unsigned char)((f >> 16));
		ptr[4] = (unsigned char)((f >>	8));
		ptr[5] = (unsigned char)((f >>	0));
		ptr[6] = 0;
		ptr[7] = 0;
		ptr[8] = 0;
		ptr[9] = 0;
	}
	
} // float80

int patch_int32(long address, unsigned long num)
{
	unsigned char dword[4];
	char cstr[256];

	if(fseek(g_outputfilefp, address, SEEK_SET) == -1)
	{
		sprintf(cstr, "patch_int32 fseek '%s'", g_outputfile);
		maLog(cstr);
		return 1;
	}

	int32(dword, num);

	if(fwrite(dword, sizeof(dword), 1, g_outputfilefp) != 1)
	{
		sprintf(cstr, "patch_int32 fwrite '%s'", g_outputfile);
		maLog(cstr);
		return 1;
	}

	if(fseek(g_outputfilefp, 0, SEEK_END) == -1)
	{
		sprintf(cstr, "patch_int32 fseek '%s'", g_outputfile);
		maLog(cstr);
		return 1;
	}

	return 0;

} // patch_int32

#ifdef __cplusplus
}
#endif

#endif

// eof
