// maWriteMp3.c

/*
 * MoreAmp
 * Copyright (C) 2004-2010 moreforge
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include "maConfig.h" // included by all moreamp files

#ifdef __MAWIN__
 #ifdef MAUNUSEDWIN
  #pragma warning( disable : 4100 ) // unreferenced formal parameter
 #endif
 #pragma warning( disable : 4206 ) // nonstandard extension used : translation unit is empty
 #pragma warning( disable : 4514 ) // mad_f_mul_inline : unreferenced inline function has been removed
 #pragma warning( disable : 4706 ) // assignment within conditional expression
#endif

#ifdef ENCODEMP3

#ifdef __cplusplus
extern "C"
{
#endif

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#ifdef __MAOSX__
 #include <sys/stat.h>
#endif

#include "maDefs.h"
#include "maWriteMp3.h"
#include "maPlayer.h"
#include "maUtils.h"

#if defined(__MAOSX__) || defined(__MAWIN__)
 #include "lame.h"
#else
 #if defined(USE_LIBLAME_SYSTEM)
  #include "lame/lame.h" // e.g. /usr/local/include/lame/lame.h
 #else
  #include "lame.h" // libs/lame/include/lame.h
 #endif
#endif

lame_global_flags *g_lame = NULL;
void lamedebugf(const char *format, va_list ap);


int OpenWriteMp3()
{
	MPEG_mode mode;
	int brate;
	int quality;
	int samplerate;
	int ret;
	char cstr[256];
	
	if(g_lame)
	{
		lame_close(g_lame);
		g_lame = NULL;
	}
	
	if((g_lame = lame_init()) == NULL)
	{
		sprintf(cstr, "OpenWriteMp3 lame_init failed '%s'", g_outputfile);
		maLog(cstr);
		return 1;
	}

	lame_set_errorf(g_lame, lamedebugf);
	lame_set_debugf(g_lame, lamedebugf);
	lame_set_msgf  (g_lame, lamedebugf);

	switch(g_sets.fileout[g_k[kOUTMP3]].mode)
	{
		case 1 : mode = JOINT_STEREO; break;
		case 2 : mode = DUAL_CHANNEL; break;
		case 3 : mode = MONO; break;
		default : mode = STEREO; break;
	}
	
	switch(g_sets.fileout[g_k[kOUTMP3]].bitrate)
	{
		case 0 : brate = 32; break;
		case 1 : brate = 40; break;
		case 2 : brate = 48; break;
		case 3 : brate = 56; break;
		case 4 : brate = 64; break;
		case 5 : brate = 80; break;
		case 6 : brate = 96; break;
		case 7 : brate = 112; break;
		case 9 : brate = 160; break;
		case 10 : brate = 192; break;
		case 11 : brate = 224; break;
		case 12 : brate = 256; break;
		case 13 : brate = 320; break;
		default : brate = 128; break;
	}
	
	switch(g_sets.fileout[g_k[kOUTMP3]].quality)
	{
		case 0 : quality = 0; break;
		case 1 : quality = 1; break;
		case 3 : quality = 3; break;
		case 4 : quality = 4; break;
		case 5 : quality = 5; break;
		case 6 : quality = 6; break;
		case 7 : quality = 7; break;
		case 8 : quality = 8; break;
		case 9 : quality = 9; break;
		case 10 : quality = 10; break;
		default : quality = 2; break;
	}
	
	switch(g_sets.fileout[g_k[kOUTMP3]].samplerate)
	{
		case 0 : samplerate = 0; break; // automatic
		//case 0 : samplerate = 8000; break; // crashes on OSX
		case 1 : samplerate = 11025; break;
		case 2 : samplerate = 12000; break;
		case 3 : samplerate = 16000; break;
		case 4 : samplerate = 22050; break;
		case 5 : samplerate = 24000; break;
		case 6 : samplerate = 32000; break;
		case 8 : samplerate = 48000; break;
		case 9 : samplerate = g_state.insamplerate; break;
		default : samplerate = 44100; break;
	}
	
	lame_set_quality(g_lame, quality);	
	lame_set_mode(g_lame, mode);
	lame_set_brate(g_lame, brate);
	lame_set_in_samplerate(g_lame, g_state.insamplerate);
	lame_set_out_samplerate(g_lame, samplerate);
	
	id3tag_init(g_lame);
	id3tag_v2_only(g_lame);
	id3tag_add_v2(g_lame);
	id3tag_set_title(g_lame, "any title" );
	id3tag_set_artist(g_lame, "any artist" );
	id3tag_set_album(g_lame, "any album" );
	id3tag_set_year(g_lame, "2010" );
	sprintf(cstr, "%s %s / lame %s", MOREAMP_NAME, MOREAMP_VERSION_STRING, get_lame_version());
	id3tag_set_comment(g_lame, cstr);
	id3tag_set_track(g_lame, "99" );
	id3tag_set_genre(g_lame, "1" );		

	lame_set_bWriteVbrTag(g_lame, 0); // don't add Xing VBR tag frame
	
	if((ret = lame_init_params(g_lame)) < 0)
	{
		sprintf(cstr, "OpenWriteMp3 lame_init_params error=%d '%s'", ret, g_outputfile);
		maLog(cstr);
		return 1;
	}
	
	return 0;

} // OpenWriteMp3


int WriteMp3(void* pcmblock, unsigned long frames, unsigned long len)
{
#ifdef MAUNUSED
 (void)len;
#endif
	
	int bytes;
	int writ;	
	char cstr[256];

	if(g_lame == NULL)
	{
		sprintf(cstr, "WriteMp3 lame not initialized for '%s'", g_loadedname[g_loadedcurrent]);
		maLog(cstr);
		return 1;
	}

	 bytes = lame_encode_buffer_interleaved(g_lame, (short int*)pcmblock, frames, (unsigned char*)g_sndbuf[0], SOUNDBUFSIZE);
	
	if(bytes < 0)
	{
		sprintf(cstr, "WriteMp3 lame_encode_buffer_interleaved error=%d '%s'", bytes, g_loadedname[g_loadedcurrent]);
		maLog(cstr);
		return 1;
	}
	
	if(bytes > 0)
		if((writ = fwrite((unsigned char*)g_sndbuf[0], 1, bytes, g_outputfilefp)) != bytes)
		{
			sprintf(cstr, (char *)"WriteMp3 fwrite error, wrote %d instead of %d", writ, bytes);
			maLog(cstr);
			return 1;
		}

	return 0;
	
} // WriteMp3


int CloseWriteMp3()
{
	int bytes, writ;
	char cstr[256];
	
	if(!g_lame)
	{
		maLog("CloseWriteMp3 null g_lame");
		return 1;
	}
	
	if((bytes = lame_encode_flush(g_lame, (unsigned char*)g_sndbuf[0], SOUNDBUFSIZE)) > 0)
		if((writ = fwrite(g_sndbuf[0], 1, bytes, g_outputfilefp)) != bytes)
		{
			sprintf(cstr, (char *)"CloseWriteMp3 fwrite error, wrote %d instead of %d", writ, bytes);
			maLog(cstr);
		}
			
	//lame_mp3_tags_fid(g_lame, g_outputfilefp); // add Xing VBR tag frame
	lame_mp3_tags_fid(g_lame, NULL); // don't add Xing VBR tag frame

	lame_close(g_lame);
	g_lame = NULL;
	
	return 0;
	
} // CloseWriteMp3


void lamedebugf(const char *format, va_list ap)
{
	char cstr[1024];
	( void ) vsprintf(cstr, format, ap);
	maLog(cstr);
}


#ifdef __MAWIN__
 #ifndef HAVE_MPGLIB
  // these two are here to satisfy the linker because libmp3lame 3.96.1
  // looks for them even when HAVE_MPGLIB is undefined
  int CDECL lame_decode_init()
  {
	return 0;
  }

  int lame_decode1_unclipped(
        unsigned char*  mp3buf,
        int             len,
        short           pcm_l[],
        short           pcm_r[] )
  {
	return 0;
  }

  // these three are here to satisfy the linker because libmp3lame 3.98.4
  // looks for them even when HAVE_MPGLIB is undefined
  hip_t hip_decode_init(void)
  {
    return NULL;
  }

  int hip_decode_exit(hip_t hip)
  {
    return 0;
  }

  typedef float sample_t;
  int hip_decode1_unclipped(hip_t hip, unsigned char *buffer, size_t len, sample_t pcm_l[], sample_t pcm_r[])
  {
    return 0;
  }
 #endif
#endif

#ifdef __cplusplus
}
#endif

#endif

// eof
