// maWriteWav.c

/*
 * MoreAmp
 * Copyright (C) 2004-2010 moreforge
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

// extracts from madplay

#include "maConfig.h" // included by all moreamp files

#ifdef __MAWIN__
 #ifdef MAUNUSEDWIN
  #pragma warning( disable : 4100 ) // unreferenced formal parameter
 #endif
 #pragma warning( disable : 4206 ) // nonstandard extension used : translation unit is empty
 #pragma warning( disable : 4514 ) // mad_f_mul_inline : unreferenced inline function has been removed
 #pragma warning( disable : 4706 ) // assignment within conditional expression
#endif

#ifdef ENCODEWAV

#ifdef __cplusplus
extern "C"
{
#endif

#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <stdlib.h>
#include <ctype.h>
#include <time.h>
#include <errno.h>

#ifdef __MAOSX__
 #include <sys/stat.h>
#endif

#include "maDefs.h"
#include "maWriteWav.h"
#include "maPlayer.h"
#include "maUtils.h"

#ifdef __MAWIN__
 #define WAVE_FORMAT_PCM 1 // from maWriteAif.h
#endif

#ifdef __MAUNIX__
 #define WAVE_FORMAT_PCM 1 // from maWriteAif.h
#endif

static unsigned long riff_len, data_len;
static long data_chunk;

#define UNKNOWN_LENGTH "\xff\xff\xff\xff"

#if defined(__MAOSX__)
 #define WAVE_FORMAT_PCM 0x0001
#endif

int patch_int32w(long address, unsigned long num);
void int16w(unsigned char *ptr, unsigned int num);
void int32w(unsigned char *ptr, unsigned long num);

int OpenWriteWav()
{
	unsigned int block_al, bitdepth;
	unsigned long bytes_ps;
	unsigned char chunk[8 + 18];
	char cstr[256];
		
	// RIFF header and (WAVE) data type identifier
	if(fwrite("RIFF" UNKNOWN_LENGTH "WAVE", 4 + 4 + 4, 1, g_outputfilefp) != 1)
	{
		sprintf(cstr, "OpenWriteWav fwrite '%s'", g_outputfile);
		maLog(cstr);
		return 1;
	}

	riff_len = 4;
	data_chunk = -1;

	// Format chunk
	bitdepth = MAOUTBITS;
	block_al = MAOUTCHANNELS * ((bitdepth + 7) / 8);
	bytes_ps = g_state.insamplerate * block_al;

	memcpy(&chunk[0], (char *)"fmt ", 4);			// chunkID
	int32w(&chunk[4], 16);							// chunkSize

	int16w(&chunk[8], WAVE_FORMAT_PCM);				// wFormatTag
	int16w(&chunk[10], MAOUTCHANNELS);				// wChannels
	int32w(&chunk[12], g_state.insamplerate);   // dwSamplesPerSec
	int32w(&chunk[16], bytes_ps);					// dwAvgBytesPerSec
	int16w(&chunk[20], block_al);					// wBlockAlign

	// PCM-format-specific
	int16w(&chunk[22], bitdepth);					// wBitsPerSample

	if(fwrite(chunk, 8 + 16, 1, g_outputfilefp) != 1)
	{
		sprintf(cstr, "OpenWriteWav fwrite '%s'", g_outputfile);
		maLog(cstr);
		return 1;
	}

	riff_len += 8 + 16;

	// Data chunk
	data_chunk = ftell(g_outputfilefp);

	if(fwrite("data" UNKNOWN_LENGTH, 8 + 0, 1, g_outputfilefp) != 1)
	{
		sprintf(cstr, "OpenWriteWav fwrite '%s'", g_outputfile);
		maLog(cstr);
		return 1;
	}

	riff_len += 8 + 0;
	data_len = 0;

	return 0;

} // OpenWriteWav


int WriteWav(void* pcmblock, unsigned long frames, unsigned long len)
{
#ifdef MAUNUSED
 (void)frames;
#endif
	
	unsigned char *data = (unsigned char *)pcmblock;
	char cstr[256];
	
#ifdef WORDS_BIGENDIAN
	// wav is little-endian, pcmblock is big-endian
	unsigned char d;
	unsigned int i;
	for(i = 0; i < len; i += 2)
	{
		d = data[i];
		data[i] = data[i + 1];
		data[i + 1] = d;	
	}
#endif

	if(fwrite(data, len, 1, g_outputfilefp) != 1)
	{
		sprintf(cstr, (char *)"WriteWav fwrite error");
		maLog(cstr);
		return 1;
	}

	data_len += len;
	riff_len += len;
	
	return 0;

} // WriteWav


int CloseWriteWav()
{
	int result = 0;
	char cstr[256];

	if(data_len & 1)
	{
		if(fputc(0, g_outputfilefp) == EOF && result == 0)
		{
			sprintf(cstr, "CloseWriteWav fputc '%s'", g_outputfile);
			maLog(cstr);
			return 1;
		}
		
		++riff_len;
	}

	if(result == 0 &&
			data_chunk != -1 && patch_int32w(data_chunk + 4, data_len) == -1)
		result = 1;

	if(result == 0)
		patch_int32w(4, riff_len);

	return 0;

} // CloseWriteWav


void int16w(unsigned char *ptr, unsigned int num)
{
	ptr[0] = (unsigned char)(num >> 0);
	ptr[1] = (unsigned char)(num >> 8);
}

void int32w(unsigned char *ptr, unsigned long num)
{
	ptr[0] = (unsigned char)(num >>	0);
	ptr[1] = (unsigned char)(num >>	8);
	ptr[2] = (unsigned char)(num >> 16);
	ptr[3] = (unsigned char)(num >> 24);
}

int patch_int32w(long address, unsigned long num)
{
	char cstr[256];
	
	unsigned char dword[4];

	if(fseek(g_outputfilefp, address, SEEK_SET) == -1)
	{
		sprintf(cstr, "patch_int32w fseek '%s'", g_outputfile);
		maLog(cstr);
		return 1;
	}

	int32w(dword, num);

	if(fwrite(dword, sizeof(dword), 1, g_outputfilefp) != 1)
	{
		sprintf(cstr, "patch_int32w fwrite '%s'", g_outputfile);
		maLog(cstr);
		return 1;
	}

	if(fseek(g_outputfilefp, 0, SEEK_END) == -1)
	{
		sprintf(cstr, "patch_int32w fseek '%s'", g_outputfile);
		maLog(cstr);
		return 1;
	}

	return 0;

} // patch_int32w

#ifdef __cplusplus
}
#endif

#endif

// eof
