// maPlayWma.c

/*
 * MoreAmp
 * Copyright (C) 2004-2010 moreforge
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

// this code contains extracts from ffmpeg :
// * Copyright (c) 2000-2003 Fabrice Bellard

#include "maConfig.h" // included by all moreamp files

#ifdef __MAWIN__
 #ifdef MAUNUSEDWIN
  #pragma warning( disable : 4100 ) // unreferenced formal parameter
 #endif
 #pragma warning( disable : 4127 ) // conditional expression is constant
 #pragma warning( disable : 4146 ) // unary minus operator applied to unsigned type, result still unsigned
 #pragma warning( disable : 4201 ) // nonstandard extension used : nameless struct/union
 #pragma warning( disable : 4244 ) // conversion from 'float ' to 'int ', possible loss of data
 #pragma warning( disable : 4305 ) // truncation from 'const double ' to 'const float '
 #pragma warning( disable : 4514 ) // unreferenced inline function has been removed
 #pragma warning( disable : 4706 ) // assignment within conditional expression
 #pragma warning( disable : 4996 ) // The POSIX name for this item is deprecated.
 #define __RPCASYNC_H__ // rpcasync.h(45) : warning C4115: '_RPC_ASYNC_STATE' : named type definition in parentheses
#endif

#ifndef NODECODEWMA

#if defined(__cplusplus)
extern "C"
{
#endif

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>
#include <math.h>
#include <stdarg.h>

#ifdef __MAOSX__
 #include <MacTypes.h>
 #include <inttypes.h>
 #include <unistd.h>
 #include <sys/ioctl.h>
 #include <sys/time.h>
 #include <termios.h>
 #include <sys/resource.h>
 #include <signal.h>
 #define HAVE_LRINTF 1
#endif

#ifdef __MAUNIX__
 #include <unistd.h>
 #include <sys/ioctl.h>
 #include <sys/time.h>
 #include <termios.h>
 #include <sys/resource.h>
 #include <signal.h>
 #include <inttypes.h>
 #if !defined(__MACUNIX__)
  #include <strings.h>
 #endif
 #define HAVE_LRINTF 1
#endif

#ifdef __MAWIN__
 #include <limits.h>
 #include <sys/types.h>
 #include <sys/timeb.h>
 #include <io.h>
 #define strcasecmp stricmp
 #define open(fname,oflag,pmode) _open(fname,oflag,pmode)
 #define CONFIG_WIN32
#endif
 
#define restrict __restrict__

#include "maMemFile.h"
#include "maPlayer.h"
#include "maEqualizer.h"
#include "maUtils.h"
#include "maPlayWma.h"

#ifdef __MAWIN__
 #include "maAudioWin.h"
#endif

#ifdef __MAOSX__
 #include "maAudioMac.h"
#endif

#ifdef __MAUNIX__
 #include "maAudioUnix.h"
#endif

void maPlayWma();
int wmamain(MASTATE *pstate);

// common.h:
/**
 * @file common.h
 * common internal api header.
 */

#ifndef COMMON_H
#define COMMON_H

#define ALT_BITSTREAM_READER

#ifndef M_PI
 #define M_PI 3.14159265358979323846
#endif

#if defined(__GNUC__) && (__GNUC__ > 3 || __GNUC__ == 3 && __GNUC_MINOR__ > 0)
 #define always_inline __attribute__((always_inline)) inline
#else
 #define always_inline inline
#endif

#ifdef CONFIG_WIN32

/* windows */

typedef unsigned short uint16_t;
typedef signed short int16_t;
typedef unsigned char uint8_t;
typedef unsigned int uint32_t;
typedef unsigned __int64 uint64_t;
typedef signed char int8_t;
typedef signed int int32_t;
typedef signed __int64 int64_t;

#define int64_t_C(c) (c ## i64)
#define uint64_t_C(c) (c ## i64)
#define inline __inline
#define snprintf _snprintf
#define vsnprintf _vsnprintf

/* CONFIG_WIN32 end */
#else

/* unix, osx */

 #ifndef int64_t_C
  #define int64_t_C(c) (c ## LL)
  #define uint64_t_C(c) (c ## ULL)
 #endif

#endif /* !CONFIG_WIN32 */

#ifdef ALT_BITSTREAM_READER
 #define NEG_SSR32(a,s) ((( int32_t)(a))>>(32-(s)))
 #define NEG_USR32(a,s) (((uint32_t)(a))>>(32-(s)))
#endif

/* bit input */

typedef struct GetBitContext
{
    const uint8_t *buffer, *buffer_end;
#ifdef ALT_BITSTREAM_READER
    int index;
#endif
    int size_in_bits;
} GetBitContext;

static inline int get_bits_count(GetBitContext *s);

#define VLC_TYPE int16_t

typedef struct VLC
{
    int bits;
    VLC_TYPE (*table)[2]; ///< code, bits
    int table_size, table_allocated;
} VLC;

/* used to avoid missaligned exceptions on some archs (alpha, ...) */
#ifdef __GNUC__
 static inline uint32_t unaligned32(const void *v)
 {
    struct Unaligned
	{
		uint32_t i;
    } __attribute__((packed));

    return ((const struct Unaligned *) v)->i;
 }
#else
 static inline uint32_t unaligned32(const void *v)
 {
    return *(const uint32_t *) v;
 }
#endif

#ifdef ALT_BITSTREAM_READER

#ifdef WORDS_BIGENDIAN
 #define be2me_32(x) (x)
#else
 // code from bits/byteswap.h (C) 1997, 1998 Free Software Foundation, Inc.
 #define bswap_32(x) \
     ((((x) & 0xff000000) >> 24) | (((x) & 0x00ff0000) >>  8) | \
      (((x) & 0x0000ff00) <<  8) | (((x) & 0x000000ff) << 24))
 #define be2me_32(x) bswap_32(x)
#endif

static inline int unaligned32_be(const void *v)
{
	return be2me_32( unaligned32(v)); //original
}

#   define MIN_CACHE_BITS 25

#   define OPEN_READER(name, gb)\
        int name##_index= (gb)->index;\
        int name##_cache= 0;\

#   define CLOSE_READER(name, gb)\
        (gb)->index= name##_index;\

#   define UPDATE_CACHE(name, gb)\
        name##_cache= unaligned32_be( ((uint8_t *)(gb)->buffer)+(name##_index>>3) ) << (name##_index&0x07);\

#   define SKIP_CACHE(name, gb, num)\
        name##_cache <<= (num);\

// FIXME name?
#   define SKIP_COUNTER(name, gb, num)\
        name##_index += (num);\

#   define SKIP_BITS(name, gb, num)\
        {\
            SKIP_CACHE(name, gb, num)\
            SKIP_COUNTER(name, gb, num)\
        }\

#   define LAST_SKIP_BITS(name, gb, num) SKIP_COUNTER(name, gb, num)
#   define LAST_SKIP_CACHE(name, gb, num) ;

#   define SHOW_UBITS(name, gb, num)\
        NEG_USR32(name##_cache, num)

#   define SHOW_SBITS(name, gb, num)\
        NEG_SSR32(name##_cache, num)

#   define GET_CACHE(name, gb)\
        ((uint32_t)name##_cache)

static inline int get_bits_count(GetBitContext *s){
    return s->index;
}
#endif

/**
 * reads 0-17 bits.
 * Note, the alt bitstream reader can read upto 25 bits, but the libmpeg2 reader cant
 */
static inline unsigned int get_bits(GetBitContext *s, int n){
    register int tmp;
    OPEN_READER(re, s)
    UPDATE_CACHE(re, s)
    tmp= SHOW_UBITS(re, s, n);
    LAST_SKIP_BITS(re, s, n)
    CLOSE_READER(re, s)
    return tmp;
}

static inline void skip_bits(GetBitContext *s, int n){
 //Note gcc seems to optimize this to s->index+=n for the ALT_READER :))
    OPEN_READER(re, s)
    UPDATE_CACHE(re, s)
    LAST_SKIP_BITS(re, s, n)
    CLOSE_READER(re, s)
}

void init_get_bits(GetBitContext *s,
                   const uint8_t *buffer, int buffer_size);

void align_get_bits(GetBitContext *s);
int init_vlc(VLC *vlc, int nb_bits, int nb_codes,
             const void *bits, int bits_wrap, int bits_size,
             const void *codes, int codes_wrap, int codes_size);
void free_vlc(VLC *vlc);

/**
 *
 * if the vlc code is invalid and max_depth=1 than no bits will be removed
 * if the vlc code is invalid and max_depth>1 than the number of bits removed
 * is undefined
 */
#define GET_VLC(code, name, gb, table, bits, max_depth)\
{\
    int n, index, nb_bits;\
\
    index= SHOW_UBITS(name, gb, bits);\
    code = table[index][0];\
    n    = table[index][1];\
\
    if(max_depth > 1 && n < 0){\
        LAST_SKIP_BITS(name, gb, bits)\
        UPDATE_CACHE(name, gb)\
\
        nb_bits = -n;\
\
        index= SHOW_UBITS(name, gb, nb_bits) + code;\
        code = table[index][0];\
        n    = table[index][1];\
        if(max_depth > 2 && n < 0){\
            LAST_SKIP_BITS(name, gb, nb_bits)\
            UPDATE_CACHE(name, gb)\
\
            nb_bits = -n;\
\
            index= SHOW_UBITS(name, gb, nb_bits) + code;\
            code = table[index][0];\
            n    = table[index][1];\
        }\
    }\
    SKIP_BITS(name, gb, n)\
}

// deprecated, dont use get_vlc for new code, use get_vlc2 instead or use GET_VLC directly
static inline int get_vlc(GetBitContext *s, VLC *vlc)
{
    int code;
    VLC_TYPE (*table)[2]= vlc->table;
    
    OPEN_READER(re, s)
    UPDATE_CACHE(re, s)

    GET_VLC(code, re, s, table, vlc->bits, 3)    

    CLOSE_READER(re, s)
    return code;
}

/**
 * parses a vlc code, faster then get_vlc()
 * @param bits is the number of bits which will be read at once, must be 
 *             identical to nb_bits in init_vlc()
 * @param max_depth is the number of times bits bits must be readed to completly
 *                  read the longest vlc code 
 *                  = (max_vlc_length + bits - 1) / bits
 */
/*x
static always_inline int get_vlc2(GetBitContext *s, VLC_TYPE (*table)[2],
                                  int bits, int max_depth)
{
    int code;
    
    OPEN_READER(re, s)
    UPDATE_CACHE(re, s)

    GET_VLC(code, re, s, table, bits, max_depth)

    CLOSE_READER(re, s)
    return code;
}
*/

/* misc math functions */
extern const uint8_t ff_log2_tab[256];

static inline int av_log2(unsigned int v)
{
    int n;

    n = 0;
    if(v & 0xffff0000)
	{
        v >>= 16;
        n += 16;
    }
    if(v & 0xff00)
	{
        v >>= 8;
        n += 8;
    }
    n += ff_log2_tab[v];

    return n;
}

/* avoid usage of various functions */
// #define MAMALLOC please_use_av_malloc
// #define free please_use_av_free
// #define realloc please_use_av_realloc

#endif /* COMMON_H */

// moved here from maPlayWma.c
#define NB_LSP_COEFS 10 
typedef struct CoefVLCTable
{
	int n; /* total number of codes */
	const uint32_t *huffcodes; /* VLC bit values */
	const uint8_t *huffbits;   /* VLC bit size */
	const uint16_t *levels; /* table to build run/level tables */
} CoefVLCTable;

// bswap.h:
/**
 * @file bswap.h
 * byte swap.
 */

#ifndef __BSWAP_H__
#define __BSWAP_H__

#define bswap_16(x) (((x) & 0x00ff) << 8 | ((x) & 0xff00) >> 8)
			
// code from bits/byteswap.h (C) 1997, 1998 Free Software Foundation, Inc.
#define bswap_32(x) \
     ((((x) & 0xff000000) >> 24) | (((x) & 0x00ff0000) >>  8) | \
      (((x) & 0x0000ff00) <<  8) | (((x) & 0x000000ff) << 24))

static inline uint64_t ByteSwap64(uint64_t x)
{
    union
	{ 
        uint64_t ll;
        uint32_t l[2]; 
    } w, r;
    w.ll = x;
    r.l[0] = bswap_32 (w.l[1]);
    r.l[1] = bswap_32 (w.l[0]);
    return r.ll;
}
#define bswap_64(x) ByteSwap64(x)

// be2me ... BigEndian to MachineEndian
// le2me ... LittleEndian to MachineEndian

#ifdef WORDS_BIGENDIAN
 #define be2me_32(x) (x)
#else
 #define be2me_32(x) bswap_32(x)
#endif

#endif /* __BSWAP_H__ */

// avio.h:
#ifndef AVIO_H
#define AVIO_H

/* output byte stream handling */

typedef int64_t offset_t;

/* unbuffered I/O */

struct URLContext
{
    int flags;        
    int max_packet_size;  /* if non zero, the stream is packetized with this max packet size */
    void *priv_data;
    char filename[1]; /* specified filename */
};

typedef struct URLContext URLContext;

#define URL_RDONLY 0
#define URL_WRONLY 1

int url_open(URLContext **h, const char *filename, int flags, int nomem);
int url_read(URLContext *h, unsigned char *buf, int size, int nomem);
offset_t url_seek(URLContext *h, offset_t pos, int whence, int nomem);
int url_close(URLContext *h, int nomem);
offset_t url_filesize(URLContext *h, int nomem);
int url_get_max_packet_size(URLContext *h);

typedef struct
{
    unsigned char *buffer;
    int buffer_size;
    unsigned char *buf_ptr, *buf_end;
    void *opaque;
    int (*read_packet)(void *opaque, uint8_t *buf, int buf_size, int nomem);
    int (*seek)(void *opaque, offset_t offset, int whence, int nomem);
    offset_t pos; /* position in the file of the current buffer */
    int eof_reached; /* true if eof reached */
    int max_packet_size;
} ByteIOContext;

int init_put_byte(ByteIOContext *s,
                  unsigned char *buffer,
                  int buffer_size,
                  void *opaque,
                  int (*read_packet)(void *opaque, uint8_t *buf, int buf_size, int nomem),
                  int (*seek)(void *opaque, offset_t offset, int whence, int nomem));

offset_t url_fseek(ByteIOContext *s, offset_t offset, int whence, int nomem);
void url_fskip(ByteIOContext *s, offset_t offset, int nomem);
offset_t url_ftell(ByteIOContext *s, int nomem);
int url_feof(ByteIOContext *s);

int get_buffer(ByteIOContext *s, unsigned char *buf, int size, int nomem);
int get_byte(ByteIOContext *s, int nomem);
unsigned int get_le32(ByteIOContext *s, int nomem);
uint64_t get_le64(ByteIOContext *s, int nomem);
unsigned int get_le16(ByteIOContext *s, int nomem);

int url_fdopen(ByteIOContext *s, URLContext *h);
int url_setbufsize(ByteIOContext *s, int buf_size);
int url_fopen(ByteIOContext *s, const char *filename, int flags, int nomem);
int url_fclose(ByteIOContext *s, int nomem);
URLContext *url_fileno(ByteIOContext *s);

#endif // AVIO_H


// avcodec.h:
#ifndef AVCODEC_H
#define AVCODEC_H

/**
 * @file avcodec.h
 * external api header.
 */
enum CodecID
{
    CODEC_ID_NONE, 
    CODEC_ID_WMAV1,
    CODEC_ID_WMAV2
};

enum CodecType
{
    CODEC_TYPE_UNKNOWN = -1,
    CODEC_TYPE_AUDIO
};

/* in bytes */
#define AVCODEC_MAX_AUDIO_FRAME_SIZE 131072

/**
 * Required number of additionally allocated bytes at the end of the input bitstream for decoding.
 * this is mainly needed because some optimized bitstream readers read 
 * 32 or 64 bit at once and could read over the end<br>
 * Note, if the first 23 bits of the additional bytes are not 0 then damaged
 * MPEG bitstreams could cause overread and segfault
 */
#define FF_INPUT_BUFFER_PADDING_SIZE 8

/**
 * main external api structure.
 */
typedef struct AVCodecContext
{
    /**
     * the average bitrate.
     * - encoding: set by user. unused for constant quantizer encoding
     * - decoding: set by lavc. 0 or some bitrate if this info is available in the stream 
     */
    int bit_rate;

    /**
     * some codecs need / can use extra-data like huffman tables.
     * mjpeg: huffman tables
     * rv10: additional flags
     * mpeg4: global headers (they can be in the bitstream or here)
     * - encoding: set/allocated/freed by lavc.
     * - decoding: set/allocated/freed by user.
     */
    void *extradata;
    int extradata_size;
    
    /* video only */
    /**
     * frames per sec multiplied by frame_rate_base.
     * for variable fps this is the precission, so if the timestamps 
     * can be specified in msec precssion then this is 1000*frame_rate_base
     * - encoding: MUST be set by user
     * - decoding: set by lavc. 0 or the frame_rate if available
     */
    int frame_rate;
    
    /* audio only */
    int sample_rate; ///< samples per sec 
    int channels;
    int sample_fmt;  ///< sample format, currenly unused 

    /* the following data should not be initialized */
    int frame_size;     ///< in samples, initialized when calling 'init' 
    int frame_number;   ///< audio or video frame number 
    
    struct AVCodec *codec;
    
    void *priv_data;

    char codec_name[32];
    enum CodecType codec_type; /* see CODEC_TYPE_xxx */
    enum CodecID codec_id; /* see CODEC_ID_xxx */
    
    /**
     * fourcc (LSB first, so "ABCD" -> ('D'<<24) + ('C'<<16) + ('B'<<8) + 'A').
     * this is used to workaround some encoder bugs
     * - encoding: set by user, if not then the default based on codec_id will be used
     * - decoding: set by user, will be converted to upper case by lavc during init
     */
    unsigned int codec_tag;
    
    int block_align; ///< used by some WAV based audio codecs
    
    /**
     * bits per sample/pixel from the demuxer (needed for huffyuv).
     * - encoding: set by lavc
     * - decoding: set by user
     */
     int bits_per_sample;
    
    /**
     * fourcc from the AVI stream header (LSB first, so "ABCD" -> ('D'<<24) + ('C'<<16) + ('B'<<8) + 'A').
     * this is used to workaround some encoder bugs
     * - encoding: unused
     * - decoding: set by user, will be converted to upper case by lavc during init
     */
    unsigned int stream_codec_tag;

} AVCodecContext;


/**
 * AVCodec.
 */
typedef struct AVCodec
{
    const char *name;
    enum CodecType type;
    int id;
    int priv_data_size;
    int (*init)(AVCodecContext *);
    int (*encode)(AVCodecContext *, uint8_t *buf, int buf_size, void *data);
    int (*close)(AVCodecContext *);
    int (*decode)(AVCodecContext *, void *outdata, int *outdata_size,
                  uint8_t *buf, int buf_size);
    int capabilities;
    struct AVCodec *next;
    void (*flush)(AVCodecContext *);
} AVCodec;

extern AVCodec wmav1_decoder;
extern AVCodec wmav2_decoder;

/* external high level API */

extern AVCodec *first_avcodec;
void register_avcodec(AVCodec *format);
AVCodec *avcodec_find_decoder(enum CodecID id);

/**
 * opens / inits the AVCodecContext.
 * not thread save!
 */
int avcodec_open(AVCodecContext *avctx, AVCodec *codec);

int avcodec_decode_audio(AVCodecContext *avctx, int16_t *samples, 
                         int *frame_size_ptr,
                         uint8_t *buf, int buf_size);

int avcodec_close(AVCodecContext *avctx);

/* memory */
void *av_malloc(unsigned int size);
void *av_mallocz(unsigned int size);
void *av_realloc(void *ptr, unsigned int size);
void av_free(void *ptr);
void __av_freep(void **ptr);
#define av_freep(p) __av_freep((void **)(p))

#endif /* AVCODEC_H */

// avi.h:
#ifndef FFMPEG_AVI_H
#define FFMPEG_AVI_H

typedef struct CodecTag
{
    int id;
    unsigned int tag;
    unsigned int invalid_asf : 1;
} CodecTag;

int wav_codec_get_id(unsigned int tag, int bps);
void get_wav_header(ByteIOContext *pb, AVCodecContext *codec, int size, int nomem); 

extern const CodecTag codec_wav_tags[];

#endif /* FFMPEG_AVI_H */


// avformat.h:
#ifndef AVFORMAT_H
#define AVFORMAT_H

/* packet functions */

#ifndef MAXINT64
#define MAXINT64 int64_t_C(0x7fffffffffffffff)
#endif

#ifndef MININT64
#define MININT64 int64_t_C(0x8000000000000000)
#endif

#define AV_NOPTS_VALUE MININT64
#define AV_TIME_BASE 1000000

typedef struct AVPacket
{
    int64_t pts; /* presentation time stamp in stream units (set av_set_pts_info) */
    uint8_t *data;
    int   size;
    int   stream_index;
    int   flags;
    int   duration;
    void  (*destruct)(struct AVPacket *);
    void  *priv;
} AVPacket; 
#define PKT_FLAG_KEY   0x0001

static inline void av_init_packet(AVPacket *pkt)
{
    pkt->pts   = AV_NOPTS_VALUE;
    pkt->flags = 0;
    pkt->stream_index = 0;
}

int av_new_packet(AVPacket *pkt, int size);

/**
 * Free a packet
 *
 * @param pkt packet to free
 */
static inline void av_free_packet(AVPacket *pkt)
{
    if(pkt && pkt->destruct)
		pkt->destruct(pkt);
}

/*************************************************/
/* fractional numbers for exact pts handling */

/* the exact value of the fractional number is: 'val + num / den'. num
   is assumed to be such as 0 <= num < den */
typedef struct AVFrac
{
    int64_t val, num, den; 
} AVFrac;

void av_frac_init(AVFrac *f, int64_t val, int64_t num, int64_t den);
void av_frac_add(AVFrac *f, int64_t incr);

/*************************************************/
/* input/output formats */

struct AVFormatContext;

/* this structure contains the data a format has to probe a file */
typedef struct AVProbeData
{
    const char *filename;
    unsigned char *buf;
    int buf_size;
} AVProbeData;

#define AVPROBE_SCORE_MAX 100

typedef struct AVFormatParameters
{
    int frame_rate;
    int sample_rate;
    int channels;
} AVFormatParameters;

#define AVFMT_NOFILE        0x0001 /* no file should be opened */
#define AVFMT_NOHEADER      0x0004 /* signal that no header is present
                                      (streams are added dynamically) */

typedef struct AVInputFormat
{
    const char *name;
    const char *long_name;
    /* size of private data so that it can be allocated in the wrapper */
    int priv_data_size;
    /* tell if a given file has a chance of being parsing by this format */
    int (*read_probe)(AVProbeData *);
    /* read the format header and initialize the AVFormatContext
       structure. Return 0 if OK. 'ap' if non NULL contains
       additionnal paramters. Only used in raw format right
       now. 'av_new_stream' should be called to create new streams.  */
    int (*read_header)(struct AVFormatContext *,
                       AVFormatParameters *ap, int nomem);
    /* read one packet and put it in 'pkt'. pts and flags are also
       set. 'av_new_stream' can be called only if the flag
       AVFMT_NOHEADER is used. */
    int (*read_packet)(struct AVFormatContext *, AVPacket *pkt, int nomem);
    /* close the stream. The AVFormatContext and AVStreams are not
       freed by this function */
    int (*read_close)(struct AVFormatContext *);
    /* seek at or before a given pts (given in microsecond). The pts
       origin is defined by the stream */
    int (*read_seek)(struct AVFormatContext *, int64_t pts);
    /* can use flags: AVFMT_NOFILE, AVFMT_NEEDNUMBER, AVFMT_NOHEADER */
    int flags;
    /* if extensions are defined, then no probe is done. You should
       usually not use extension format guessing because it is not
       reliable enough */
    const char *extensions;
    /* general purpose read only value that the format can use */
    int value;
    /* private fields */
    struct AVInputFormat *next;
} AVInputFormat;

typedef struct AVStream
{
    int index;    /* stream index in AVFormatContext */
    int id;       /* format specific stream id */
    AVCodecContext codec; /* codec context */
    void *priv_data;
    /* internal data used in av_find_stream_info() */
    int codec_info_state;     
    /* PTS generation when outputing stream */
    AVFrac pts;
    /* decoding: position of the first frame of the component, in AV_TIME_BASE fractional seconds. */
    int64_t start_time; 
    /* decoding: duration of the stream, in AV_TIME_BASE fractional seconds. */
    int64_t duration;
} AVStream;

/* format I/O context */
typedef struct AVFormatContext
{
    /* can only be iformat or oformat, not both at the same time */
    struct AVInputFormat *iformat;
    void *priv_data;
    ByteIOContext pb;
    AVStream *stream;
    char filename[1024]; /* input or output filename */
    /* stream info */
    char title[512];
    char author[512];
    char copyright[512];
    char comment[512];
    char album[512];
    int year;  /* ID3 year, 0 if none */
    int track; /* track number, 0 if none */
    char genre[32]; /* ID3 genre */

    int flags; /* format specific flags */
    /* private data for pts handling (do not modify directly) */
    int pts_wrap_bits; /* number of bits in pts (used for wrapping control) */
    int pts_num, pts_den; /* value to convert to seconds */
    /* This buffer is only needed when packets were already buffered but
       not decoded, for example to get the codec parameters in mpeg
       streams */
    struct AVPacketList *packet_buffer;

    /* decoding: position of the first frame of the component, in
       AV_TIME_BASE fractional seconds. NEVER set this value directly:
       it is deduced from the AVStream values.  */
    int64_t start_time; 
    /* decoding: duration of the stream, in AV_TIME_BASE fractional
       seconds. NEVER set this value directly: it is deduced from the
       AVStream values.  */
    int64_t duration;
    /* decoding: total file size. 0 if unknown */
    int64_t file_size;
    /* decoding: total stream bitrate in bit/s, 0 if not
       available. Never set it directly if the file_size and the
       duration are known as ffmpeg can compute it automatically. */
    int bit_rate;
} AVFormatContext;

typedef struct AVPacketList
{
    AVPacket pkt;
    struct AVPacketList *next;
} AVPacketList;

extern AVInputFormat *first_iformat;

/* asf.c */
int asf_init(AVInputFormat *p);

/* utils.c */
void av_register_input_format(AVInputFormat *format);
static int av_has_timings(AVFormatContext *ic);
static void av_update_stream_timings(AVFormatContext *ic);
static void fill_all_stream_timings(AVFormatContext *ic);
static void av_estimate_timings_from_bit_rate(AVFormatContext *ic);
static void av_estimate_timings(AVFormatContext *ic, int nomem);

/* media file input */
AVInputFormat *av_probe_input_format(AVProbeData *pd, int is_opened);
int opt_input_file(const char *filename, int nomem,
		AVFormatContext **pinput_file, AVInputFormat **pfile_iformat,
		int *paudio_channels, int *paudio_sample_rate);
int av_open_input_file(AVFormatContext **ic_ptr, const char *filename, 
                       AVInputFormat *fmt,
                       int buf_size,
                       AVFormatParameters *ap, int nomem);

#define AVERROR_UNKNOWN     (-1)  /* unknown error */
#define AVERROR_IO          (-2)  /* i/o error */
#define AVERROR_NUMEXPECTED (-3)  /* number syntax expected in filename */
#define AVERROR_INVALIDDATA (-4)  /* invalid data found */
#define AVERROR_NOMEM       (-5)  /* not enough memory */
#define AVERROR_NOFMT       (-6)  /* unknown format */

int av_find_stream_info(AVFormatContext *ic, int nomem);
int av_read_packet(AVFormatContext *s, AVPacket *pkt, int nomem);
void av_close_input_file(AVFormatContext *s, int nomem);
AVStream *av_new_stream(AVFormatContext *s, int id);
void av_set_pts_info(AVFormatContext *s, int pts_wrap_bits, int pts_num, int pts_den);
void pstrcpy(char *buf, int buf_size, const char *str);

#endif /* AVFORMAT_H */

// dsputil.h:
/**
 * @file dsputil.h
 * DSP utils.
 * note, many functions in here may use MMX which trashes the FPU state, it is
 * absolutely necessary to call emms_c() between dsp & float/double code
 */

#ifndef DSPUTIL_H
#define DSPUTIL_H


/* FFT computation */

/* NOTE: soon integer code will be added, so you must use the FFTSample type */
typedef float FFTSample;

typedef struct FFTComplex
{
    FFTSample re, im;
} FFTComplex;

typedef struct FFTContext
{
    int nbits;
    int inverse;
    uint16_t *revtab;
    FFTComplex *exptab;
    FFTComplex *exptab1; /* only used by SSE code */
    void (*fft_calc)(struct FFTContext *s, FFTComplex *z);
} FFTContext;

int wmafft_init(FFTContext *s, int nbits, int inverse);
void fft_calc_c(FFTContext *s, FFTComplex *z);

static inline void fft_calc(FFTContext *s, FFTComplex *z)
{
    s->fft_calc(s, z);
}
void fft_end(FFTContext *s);

/* MDCT computation */

typedef struct MDCTContext
{
    int n;  /* size of MDCT (i.e. number of input data * 2) */
    int nbits; /* n = 2^nbits */
    /* pre/post rotation tables */
    FFTSample *tcos;
    FFTSample *tsin;
    FFTContext fft;
} MDCTContext;

int ff_mdct_init(MDCTContext *s, int nbits, int inverse);
void ff_imdct_calc(MDCTContext *s, FFTSample *output,
                const FFTSample *input, FFTSample *tmp);
void ff_mdct_calc(MDCTContext *s, FFTSample *out,
               const FFTSample *input, FFTSample *tmp);
void ff_mdct_end(MDCTContext *s);

#ifndef HAVE_LRINTF
 /* XXX: add ISOC specific test to avoid specific BSD testing. */
 /* better than nothing implementation. */
 /* btw, rintf() is existing on fbsd too -- alex */
 static inline long int lrintf(float x)
 {
  #ifdef CONFIG_WIN32
    /* XXX: incorrect, but make it compile */
    return (int)(x);
  #else
    return (int)(rint(x));
  #endif
 }
#endif

#endif // DSPUTIL_H

// wmadata.h
/**
 * @file wmadata.h
 * Various WMA tables.
 */

static const uint16_t wma_critical_freqs[25] = {
    100,   200,  300, 400,   510,  630,  770,    920,
    1080, 1270, 1480, 1720, 2000, 2320, 2700,   3150,
    3700, 4400, 5300, 6400, 7700, 9500, 12000, 15500,
    24500,
};

/* first value is number of bands */
static const uint8_t exponent_band_22050[3][25] = {
    { 10, 4, 8, 4, 8, 8, 12, 20, 24, 24, 16, },
    { 14, 4, 8, 8, 4, 12, 12, 16, 24, 16, 20, 24, 32, 40, 36, },
    { 23, 4, 4, 4, 8, 4, 4, 8, 8, 8, 8, 8, 12, 12, 16, 16, 24, 24, 32, 44, 48, 60, 84, 72, },
};

static const uint8_t exponent_band_32000[3][25] = {
    { 11, 4, 4, 8, 4, 4, 12, 16, 24, 20, 28, 4, },
    { 15, 4, 8, 4, 4, 8, 8, 16, 20, 12, 20, 20, 28, 40, 56, 8, },
    { 16, 8, 4, 8, 8, 12, 16, 20, 24, 40, 32, 32, 44, 56, 80, 112, 16, },
};

static const uint8_t exponent_band_44100[3][25] = {
    { 12,  4,   4,   4,   4,   4,   8,   8,   8,  12,  16,  20,  36, },
    { 15,  4,   8,   4,   8,   8,   4,   8,   8,  12,  12,  12,  24,  28,  40,  76, },
    { 17,  4,   8,   8,   4,  12,  12,   8,   8,  24,  16,  20,  24,  32,  40,  60,  80, 152, },
};

static const uint16_t hgain_huffcodes[37] = {
 0x00003, 0x002e7, 0x00001, 0x005cd, 0x0005d, 0x005c9, 0x0005e, 0x00003,
 0x00016, 0x0000b, 0x00001, 0x00006, 0x00001, 0x00006, 0x00004, 0x00005,
 0x00004, 0x00007, 0x00003, 0x00007, 0x00004, 0x0000a, 0x0000a, 0x00002,
 0x00003, 0x00000, 0x00005, 0x00002, 0x0005f, 0x00004, 0x00003, 0x00002,
 0x005c8, 0x000b8, 0x005ca, 0x005cb, 0x005cc,
};

static const uint8_t hgain_huffbits[37] = {
 10, 12, 10, 13,  9, 13,  9,  8,
  7,  5,  5,  4,  4,  3,  3,  3,
  4,  3,  4,  4,  5,  5,  6,  8,
  7, 10,  8, 10,  9,  8,  9,  9,
 13, 10, 13, 13, 13,
};

static const float lsp_codebook[NB_LSP_COEFS][16] = {
 { 1.98732877, 1.97944528, 1.97179088, 1.96260549, 1.95038374, 1.93336114, 1.90719232, 1.86191415, },
 { 1.97260000, 1.96083160, 1.94982586, 1.93806164, 1.92516608, 1.91010199, 1.89232331, 1.87149812,
   1.84564818, 1.81358067, 1.77620070, 1.73265264, 1.67907855, 1.60959081, 1.50829650, 1.33120330, },
 { 1.90109110, 1.86482426, 1.83419671, 1.80168452, 1.76650116, 1.72816320, 1.68502700, 1.63738256,
   1.58501580, 1.51795181, 1.43679906, 1.33950585, 1.24176208, 1.12260729, 0.96749668, 0.74048265, },
 { 1.76943864, 1.67822463, 1.59946365, 1.53560582, 1.47470796, 1.41210167, 1.34509536, 1.27339507,
   1.19303814, 1.09765169, 0.98818722, 0.87239446, 0.74369172, 0.59768184, 0.43168630, 0.17977021, },
 { 1.43428349, 1.32038354, 1.21074086, 1.10577988, 1.00561746, 0.90335924, 0.80437489, 0.70709671,
   0.60427395, 0.49814048, 0.38509539, 0.27106800, 0.14407416, 0.00219910, -0.16725141, -0.36936085, },
 { 0.99895687, 0.84188166, 0.70753739, 0.57906595, 0.47055563, 0.36966965, 0.26826648, 0.17163380,
   0.07208392, -0.03062936, -1.40037388, -0.25128968, -0.37213937, -0.51075646, -0.64887512, -0.80308031, },
 { 0.26515280, 0.06313551, -0.08872080, -0.21103548, -0.31069678, -0.39680323, -0.47223474, -0.54167135,
   -0.61444740, -0.68943343, -0.76580211, -0.85170082, -0.95289061, -1.06514703, -1.20510707, -1.37617746, },
 { -0.53940301, -0.73770929, -0.88424876, -1.01117930, -1.13389091, -1.26830073, -1.42041987, -1.62033919,
   -1.10158808, -1.16512566, -1.23337128, -1.30414401, -1.37663312, -1.46853845, -1.57625798, -1.66893638, },
 { -0.38601997, -0.56009350, -0.66978483, -0.76028471, -0.83846064, -0.90868087, -0.97408881, -1.03694962, },
 { -1.56144989, -1.65944032, -1.72689685, -1.77857740, -1.82203011, -1.86220079, -1.90283983, -1.94820479, },
};

static const uint32_t scale_huffcodes[121] = {
 0x3ffe8, 0x3ffe6, 0x3ffe7, 0x3ffe5, 0x7fff5, 0x7fff1, 0x7ffed, 0x7fff6,
 0x7ffee, 0x7ffef, 0x7fff0, 0x7fffc, 0x7fffd, 0x7ffff, 0x7fffe, 0x7fff7,
 0x7fff8, 0x7fffb, 0x7fff9, 0x3ffe4, 0x7fffa, 0x3ffe3, 0x1ffef, 0x1fff0,
 0x0fff5, 0x1ffee, 0x0fff2, 0x0fff3, 0x0fff4, 0x0fff1, 0x07ff6, 0x07ff7,
 0x03ff9, 0x03ff5, 0x03ff7, 0x03ff3, 0x03ff6, 0x03ff2, 0x01ff7, 0x01ff5,
 0x00ff9, 0x00ff7, 0x00ff6, 0x007f9, 0x00ff4, 0x007f8, 0x003f9, 0x003f7,
 0x003f5, 0x001f8, 0x001f7, 0x000fa, 0x000f8, 0x000f6, 0x00079, 0x0003a,
 0x00038, 0x0001a, 0x0000b, 0x00004, 0x00000, 0x0000a, 0x0000c, 0x0001b,
 0x00039, 0x0003b, 0x00078, 0x0007a, 0x000f7, 0x000f9, 0x001f6, 0x001f9,
 0x003f4, 0x003f6, 0x003f8, 0x007f5, 0x007f4, 0x007f6, 0x007f7, 0x00ff5,
 0x00ff8, 0x01ff4, 0x01ff6, 0x01ff8, 0x03ff8, 0x03ff4, 0x0fff0, 0x07ff4,
 0x0fff6, 0x07ff5, 0x3ffe2, 0x7ffd9, 0x7ffda, 0x7ffdb, 0x7ffdc, 0x7ffdd,
 0x7ffde, 0x7ffd8, 0x7ffd2, 0x7ffd3, 0x7ffd4, 0x7ffd5, 0x7ffd6, 0x7fff2,
 0x7ffdf, 0x7ffe7, 0x7ffe8, 0x7ffe9, 0x7ffea, 0x7ffeb, 0x7ffe6, 0x7ffe0,
 0x7ffe1, 0x7ffe2, 0x7ffe3, 0x7ffe4, 0x7ffe5, 0x7ffd7, 0x7ffec, 0x7fff4,
 0x7fff3,
};

static const uint8_t scale_huffbits[121] = {
 18, 18, 18, 18, 19, 19, 19, 19,
 19, 19, 19, 19, 19, 19, 19, 19,
 19, 19, 19, 18, 19, 18, 17, 17,
 16, 17, 16, 16, 16, 16, 15, 15,
 14, 14, 14, 14, 14, 14, 13, 13,
 12, 12, 12, 11, 12, 11, 10, 10,
 10,  9,  9,  8,  8,  8,  7,  6,
  6,  5,  4,  3,  1,  4,  4,  5,
  6,  6,  7,  7,  8,  8,  9,  9,
 10, 10, 10, 11, 11, 11, 11, 12,
 12, 13, 13, 13, 14, 14, 16, 15,
 16, 15, 18, 19, 19, 19, 19, 19,
 19, 19, 19, 19, 19, 19, 19, 19,
 19, 19, 19, 19, 19, 19, 19, 19,
 19, 19, 19, 19, 19, 19, 19, 19,
 19,
};

static const uint32_t coef0_huffcodes[666] = {
 0x00258, 0x0003d, 0x00000, 0x00005, 0x00008, 0x00008, 0x0000c, 0x0001b,
 0x0001f, 0x00015, 0x00024, 0x00032, 0x0003a, 0x00026, 0x0002c, 0x0002f,
 0x0004a, 0x0004d, 0x00061, 0x00070, 0x00073, 0x00048, 0x00052, 0x0005a,
 0x0005d, 0x0006e, 0x00099, 0x0009e, 0x000c1, 0x000ce, 0x000e4, 0x000f0,
 0x00093, 0x0009e, 0x000a2, 0x000a1, 0x000b8, 0x000d2, 0x000d3, 0x0012e,
 0x00130, 0x000de, 0x0012d, 0x0019b, 0x001e4, 0x00139, 0x0013a, 0x0013f,
 0x0014f, 0x0016d, 0x001a2, 0x0027c, 0x0027e, 0x00332, 0x0033c, 0x0033f,
 0x0038b, 0x00396, 0x003c5, 0x00270, 0x0027c, 0x0025a, 0x00395, 0x00248,
 0x004bd, 0x004fb, 0x00662, 0x00661, 0x0071b, 0x004e6, 0x004ff, 0x00666,
 0x0071c, 0x0071a, 0x0071f, 0x00794, 0x00536, 0x004e2, 0x0078e, 0x004ee,
 0x00518, 0x00535, 0x004fb, 0x0078d, 0x00530, 0x00680, 0x0068f, 0x005cb,
 0x00965, 0x006a6, 0x00967, 0x0097f, 0x00682, 0x006ae, 0x00cd0, 0x00e28,
 0x00f13, 0x00f1f, 0x009f5, 0x00cd3, 0x00f11, 0x00926, 0x00964, 0x00f32,
 0x00f12, 0x00f30, 0x00966, 0x00d0b, 0x00a68, 0x00b91, 0x009c7, 0x00b73,
 0x012fa, 0x0131d, 0x013f9, 0x01ca0, 0x0199c, 0x01c7a, 0x0198c, 0x01248,
 0x01c74, 0x01c64, 0x0139e, 0x012fd, 0x00a77, 0x012fc, 0x01c7b, 0x012ca,
 0x014cc, 0x014d2, 0x014e3, 0x014dc, 0x012dc, 0x03344, 0x02598, 0x0263c,
 0x0333b, 0x025e6, 0x01a1c, 0x01e3c, 0x014e2, 0x033d4, 0x01a11, 0x03349,
 0x03cce, 0x014e1, 0x01a34, 0x0273e, 0x02627, 0x0273f, 0x038ee, 0x03971,
 0x03c67, 0x03c61, 0x0333d, 0x038c2, 0x0263f, 0x038cd, 0x02638, 0x02e41,
 0x0351f, 0x03348, 0x03c66, 0x03562, 0x02989, 0x027d5, 0x0333c, 0x02e4f,
 0x0343b, 0x02ddf, 0x04bc8, 0x029c0, 0x02e57, 0x04c72, 0x025b7, 0x03547,
 0x03540, 0x029d3, 0x04c45, 0x025bb, 0x06600, 0x04c73, 0x04bce, 0x0357b,
 0x029a6, 0x029d2, 0x0263e, 0x0298a, 0x07183, 0x06602, 0x07958, 0x04b66,
 0x0537d, 0x05375, 0x04fe9, 0x04b67, 0x0799f, 0x04bc9, 0x051fe, 0x06a3b,
 0x05bb6, 0x04fa8, 0x0728f, 0x05376, 0x0492c, 0x0537e, 0x0795a, 0x06a3c,
 0x0e515, 0x07887, 0x0683a, 0x051f9, 0x051fd, 0x0cc6a, 0x06a8a, 0x0cc6d,
 0x05bb3, 0x0683b, 0x051fc, 0x05378, 0x0728e, 0x07886, 0x05bb7, 0x0f2a4,
 0x0795b, 0x0683c, 0x09fc1, 0x0683d, 0x0b752, 0x09678, 0x0a3e8, 0x06ac7,
 0x051f0, 0x0b759, 0x06af3, 0x04b6b, 0x0f2a0, 0x0f2ad, 0x096c3, 0x0e518,
 0x0b75c, 0x0d458, 0x0cc6b, 0x0537c, 0x067aa, 0x04fea, 0x0343a, 0x0cc71,
 0x0967f, 0x09fc4, 0x096c2, 0x0e516, 0x0f2a1, 0x0d45c, 0x0d45d, 0x0d45e,
 0x12fb9, 0x0967e, 0x1982f, 0x09883, 0x096c4, 0x0b753, 0x12fb8, 0x0f2a8,
 0x1ca21, 0x096c5, 0x0e51a, 0x1ca27, 0x12f3c, 0x0d471, 0x0f2aa, 0x0b75b,
 0x12fbb, 0x0f2a9, 0x0f2ac, 0x0d45a, 0x0b74f, 0x096c8, 0x16e91, 0x096ca,
 0x12fbf, 0x0d0a7, 0x13103, 0x0d516, 0x16e99, 0x12cbd, 0x0a3ea, 0x19829,
 0x0b755, 0x29ba7, 0x1ca28, 0x29ba5, 0x16e93, 0x1982c, 0x19828, 0x25994,
 0x0a3eb, 0x1ca29, 0x16e90, 0x1ca25, 0x1982d, 0x1ca26, 0x16e9b, 0x0b756,
 0x0967c, 0x25997, 0x0b75f, 0x198d3, 0x0b757, 0x19a2a, 0x0d45b, 0x0e517,
 0x1ca24, 0x1ca23, 0x1ca22, 0x0b758, 0x16e97, 0x0cd14, 0x13100, 0x00007,
 0x0003b, 0x0006b, 0x00097, 0x00138, 0x00125, 0x00173, 0x00258, 0x00335,
 0x0028e, 0x004c6, 0x00715, 0x00729, 0x004ef, 0x00519, 0x004ed, 0x00532,
 0x0068c, 0x00686, 0x00978, 0x00e5d, 0x00e31, 0x009f4, 0x00b92, 0x012f8,
 0x00d06, 0x00a67, 0x00d44, 0x00a76, 0x00d59, 0x012cd, 0x01c78, 0x01c75,
 0x0199f, 0x0198f, 0x01c67, 0x014c6, 0x01c79, 0x01c76, 0x00b94, 0x00d1b,
 0x01e32, 0x01e31, 0x01ab0, 0x01a05, 0x01aa1, 0x0333a, 0x025e5, 0x02626,
 0x03541, 0x03544, 0x03421, 0x03546, 0x02e55, 0x02e56, 0x0492d, 0x02dde,
 0x0299b, 0x02ddc, 0x0357a, 0x0249c, 0x0668b, 0x1c77f, 0x1ca20, 0x0d45f,
 0x09886, 0x16e9a, 0x0f2a7, 0x0b751, 0x0a3ee, 0x0cf59, 0x0cf57, 0x0b754,
 0x0d0a6, 0x16e98, 0x0b760, 0x06ac6, 0x0a3f0, 0x12fbe, 0x13104, 0x0f2a5,
 0x0a3ef, 0x0d472, 0x12cba, 0x1982e, 0x16e9c, 0x1c77e, 0x198d0, 0x13105,
 0x16e92, 0x0b75d, 0x0d459, 0x0001a, 0x000c0, 0x0016c, 0x003cd, 0x00350,
 0x0067b, 0x0051e, 0x006a9, 0x009f4, 0x00b72, 0x00d09, 0x01249, 0x01e3d,
 0x01ca1, 0x01a1f, 0x01721, 0x01a8a, 0x016e8, 0x03347, 0x01a35, 0x0249d,
 0x0299a, 0x02596, 0x02e4e, 0x0298b, 0x07182, 0x04c46, 0x025ba, 0x02e40,
 0x027d6, 0x04fe8, 0x06607, 0x05310, 0x09884, 0x072e1, 0x06a3d, 0x04b6a,
 0x04c7a, 0x06603, 0x04c7b, 0x03428, 0x06605, 0x09664, 0x09fc0, 0x071de,
 0x06601, 0x05bb2, 0x09885, 0x0a3e2, 0x1c61f, 0x12cbb, 0x0b750, 0x0cf58,
 0x0967d, 0x25995, 0x668ad, 0x0b75a, 0x09fc2, 0x0537f, 0x0b75e, 0x13fae,
 0x12fbc, 0x00031, 0x001c4, 0x004c5, 0x005b8, 0x00cf4, 0x0096f, 0x00d46,
 0x01e57, 0x01a04, 0x02625, 0x03346, 0x028f9, 0x04c47, 0x072e0, 0x04b69,
 0x03420, 0x07957, 0x06639, 0x0799e, 0x07959, 0x07881, 0x04b68, 0x09fc3,
 0x09fd6, 0x0cc70, 0x0a3f1, 0x12cbe, 0x0e30e, 0x0e51b, 0x06af2, 0x12cbc,
 0x1c77d, 0x0f2ab, 0x12fbd, 0x1aa2f, 0x0a3ec, 0x0d473, 0x05377, 0x0a3e9,
 0x1982b, 0x0e300, 0x12f3f, 0x0cf5f, 0x096c0, 0x38c3c, 0x16e94, 0x16e95,
 0x12f3d, 0x29ba4, 0x29ba6, 0x1c77c, 0x6a8ba, 0x3545c, 0x33457, 0x668ac,
 0x6a8bb, 0x16e9d, 0x0e519, 0x25996, 0x12f3e, 0x00036, 0x0033e, 0x006ad,
 0x00d03, 0x012c8, 0x0124a, 0x03c42, 0x03ccd, 0x06606, 0x07880, 0x06852,
 0x06a3a, 0x05bb4, 0x0f2a2, 0x09fc7, 0x12cb9, 0x0cc6c, 0x0a6e8, 0x096c1,
 0x0004a, 0x00355, 0x012f9, 0x014e8, 0x01abe, 0x025b6, 0x0492e, 0x09fc6,
 0x051ff, 0x0cc6f, 0x096cb, 0x0d071, 0x198d1, 0x12cb8, 0x38c3d, 0x13faf,
 0x096c9, 0x0009d, 0x00539, 0x012ce, 0x0341f, 0x029c1, 0x04b33, 0x0a3e3,
 0x0d070, 0x16e96, 0x0b763, 0x000a0, 0x009ce, 0x038cc, 0x0343d, 0x051fa,
 0x09888, 0x12fba, 0x000df, 0x00a75, 0x029a7, 0x09fc5, 0x0e301, 0x0967b,
 0x001e7, 0x012c9, 0x051fb, 0x09889, 0x0f2a6, 0x0016f, 0x01cb9, 0x0cf5a,
 0x12cbf, 0x09679, 0x00272, 0x01a15, 0x0967a, 0x003cb, 0x025f6, 0x0b762,
 0x0028d, 0x03c60, 0x0cf5e, 0x00352, 0x03ccc, 0x0072f, 0x07186, 0x004ec,
 0x05379, 0x0068e, 0x09887, 0x006a7, 0x06af1, 0x00e29, 0x0cf5b, 0x00f31,
 0x0d470, 0x009c6, 0x013fb, 0x13102, 0x019a5, 0x13101, 0x01983, 0x01c65,
 0x0124f, 0x014c7, 0x01726, 0x01abf, 0x03304, 0x02624, 0x03c41, 0x027d7,
 0x02ddd, 0x02e54, 0x0343c, 0x06604, 0x07181, 0x0663a, 0x04fa9, 0x0663b,
 0x05311, 0x0537a, 0x06839, 0x05bb5, 0x0492f, 0x06af0, 0x096c7, 0x0cc6e,
 0x0537b, 0x0cf5c, 0x0cf56, 0x198d2, 0x0cf5d, 0x0a3ed, 0x0f2a3, 0x1982a,
 0x0b761, 0x096c6,
};

static const uint8_t coef0_huffbits[666] = {
 11,  6,  2,  3,  4,  5,  5,  5,
  5,  6,  6,  6,  6,  7,  7,  7,
  7,  7,  7,  7,  7,  8,  8,  8,
  8,  8,  8,  8,  8,  8,  8,  8,
  9,  9,  9,  9,  9,  9,  9,  9,
  9,  9,  9,  9,  9, 10, 10, 10,
 10, 10, 10, 10, 10, 10, 10, 10,
 10, 10, 10, 11, 11, 11, 10, 11,
 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 12, 12, 11, 12,
 12, 12, 12, 11, 12, 12, 12, 12,
 12, 12, 12, 12, 12, 12, 12, 12,
 12, 12, 12, 12, 12, 13, 13, 12,
 12, 12, 13, 13, 13, 13, 13, 13,
 13, 13, 13, 13, 13, 13, 13, 14,
 13, 13, 13, 13, 13, 13, 13, 14,
 14, 14, 14, 14, 14, 14, 14, 14,
 14, 14, 14, 13, 14, 14, 14, 14,
 14, 14, 14, 14, 14, 14, 14, 14,
 14, 14, 14, 14, 14, 14, 14, 15,
 15, 14, 14, 15, 15, 15, 14, 15,
 15, 15, 15, 15, 15, 15, 15, 15,
 15, 15, 15, 15, 15, 15, 15, 15,
 15, 15, 14, 15, 15, 15, 15, 16,
 16, 16, 15, 16, 15, 15, 16, 16,
 16, 16, 15, 16, 16, 16, 15, 16,
 16, 15, 16, 16, 16, 16, 16, 16,
 16, 16, 16, 16, 15, 15, 16, 16,
 15, 16, 16, 16, 17, 17, 17, 16,
 16, 17, 16, 16, 16, 16, 17, 16,
 17, 17, 16, 16, 15, 15, 15, 16,
 17, 16, 17, 16, 16, 17, 17, 17,
 17, 17, 17, 16, 17, 17, 17, 16,
 17, 17, 16, 17, 17, 17, 16, 17,
 17, 16, 16, 17, 17, 17, 18, 17,
 17, 17, 17, 17, 18, 18, 17, 17,
 17, 19, 17, 19, 18, 17, 17, 18,
 17, 17, 18, 17, 17, 17, 18, 17,
 17, 18, 17, 17, 17, 17, 17, 16,
 17, 17, 17, 17, 18, 16, 17,  4,
  6,  8,  9,  9, 10, 10, 10, 10,
 11, 11, 11, 11, 12, 12, 12, 12,
 12, 12, 12, 12, 12, 13, 13, 13,
 13, 13, 13, 13, 13, 13, 13, 13,
 13, 13, 13, 14, 13, 13, 13, 13,
 13, 13, 14, 14, 14, 14, 14, 14,
 15, 15, 15, 15, 15, 15, 16, 15,
 15, 15, 15, 15, 15, 17, 17, 17,
 16, 18, 16, 17, 17, 16, 16, 17,
 17, 18, 17, 16, 17, 17, 17, 16,
 17, 17, 18, 17, 18, 17, 17, 17,
 18, 17, 17,  5,  8, 10, 10, 11,
 11, 12, 12, 12, 13, 13, 14, 13,
 13, 14, 14, 14, 14, 14, 14, 15,
 15, 15, 15, 15, 15, 15, 15, 15,
 15, 15, 15, 16, 16, 15, 16, 16,
 15, 15, 15, 15, 15, 16, 16, 15,
 15, 16, 16, 17, 17, 18, 17, 16,
 17, 18, 19, 17, 16, 16, 17, 17,
 17,  6,  9, 11, 12, 12, 13, 13,
 13, 14, 14, 14, 15, 15, 15, 16,
 15, 15, 15, 15, 15, 15, 16, 16,
 16, 16, 17, 18, 16, 16, 16, 18,
 17, 16, 17, 18, 17, 17, 16, 17,
 17, 16, 17, 16, 17, 18, 18, 18,
 17, 19, 19, 17, 20, 19, 18, 19,
 20, 18, 16, 18, 17,  7, 10, 12,
 13, 13, 14, 14, 14, 15, 15, 16,
 16, 16, 16, 16, 18, 16, 17, 17,
  8, 11, 13, 14, 14, 15, 16, 16,
 16, 16, 17, 17, 17, 18, 18, 17,
 17,  8, 12, 14, 15, 15, 15, 17,
 17, 18, 17,  9, 12, 14, 15, 16,
 16, 17,  9, 13, 15, 16, 16, 17,
  9, 13, 16, 16, 16, 10, 13, 16,
 18, 17, 10, 14, 17, 10, 14, 17,
 11, 14, 16, 11, 14, 11, 15, 12,
 16, 12, 16, 12, 16, 12, 16, 12,
 17, 13, 13, 17, 13, 17, 13, 13,
 14, 14, 14, 14, 14, 14, 14, 15,
 15, 15, 15, 15, 15, 15, 16, 15,
 16, 16, 16, 16, 16, 16, 17, 16,
 16, 16, 16, 17, 16, 17, 16, 17,
 17, 17,
};

static const uint32_t coef1_huffcodes[555] = {
 0x00115, 0x00002, 0x00001, 0x00000, 0x0000d, 0x00007, 0x00013, 0x0001d,
 0x00008, 0x0000c, 0x00023, 0x0002b, 0x0003f, 0x00017, 0x0001b, 0x00043,
 0x00049, 0x00050, 0x00055, 0x00054, 0x00067, 0x00064, 0x0007b, 0x0002d,
 0x00028, 0x0002a, 0x00085, 0x00089, 0x0002b, 0x00035, 0x00090, 0x00091,
 0x00094, 0x00088, 0x000c1, 0x000c6, 0x000f2, 0x000e3, 0x000c5, 0x000e2,
 0x00036, 0x000f0, 0x000a7, 0x000cd, 0x000fb, 0x00059, 0x00116, 0x00103,
 0x00108, 0x0012b, 0x0012d, 0x00188, 0x0012e, 0x0014c, 0x001c3, 0x00187,
 0x001e7, 0x0006f, 0x00094, 0x00069, 0x001e6, 0x001ca, 0x00147, 0x00195,
 0x000a7, 0x00213, 0x00209, 0x00303, 0x00295, 0x00289, 0x0028c, 0x0028d,
 0x00312, 0x00330, 0x0029b, 0x00308, 0x00328, 0x0029a, 0x0025e, 0x003c5,
 0x00384, 0x0039f, 0x00397, 0x00296, 0x0032e, 0x00332, 0x003c6, 0x003e6,
 0x0012d, 0x000d1, 0x00402, 0x000dd, 0x00161, 0x0012b, 0x00127, 0x0045d,
 0x00601, 0x004ab, 0x0045f, 0x00410, 0x004bf, 0x00528, 0x0045c, 0x00424,
 0x00400, 0x00511, 0x00618, 0x0073d, 0x0063a, 0x00614, 0x0073c, 0x007c0,
 0x007cf, 0x00802, 0x00966, 0x00964, 0x00951, 0x008a0, 0x00346, 0x00803,
 0x00a52, 0x0024a, 0x007c1, 0x0063f, 0x00126, 0x00406, 0x00789, 0x008a2,
 0x00960, 0x00967, 0x00c05, 0x00c70, 0x00c79, 0x00a5d, 0x00c26, 0x00c4d,
 0x00372, 0x008a5, 0x00c08, 0x002c5, 0x00f11, 0x00cc4, 0x00f8e, 0x00e16,
 0x00496, 0x00e77, 0x00f9c, 0x00c25, 0x00f1e, 0x00c27, 0x00f1f, 0x00e17,
 0x00ccd, 0x00355, 0x00c09, 0x00c78, 0x00f90, 0x00521, 0x00357, 0x00356,
 0x0068e, 0x00f9d, 0x00c04, 0x00e58, 0x00a20, 0x00a2c, 0x00c4c, 0x0052f,
 0x00f8d, 0x01178, 0x01053, 0x01097, 0x0180f, 0x0180d, 0x012fb, 0x012aa,
 0x0202a, 0x00a40, 0x018ed, 0x01ceb, 0x01455, 0x018e3, 0x012a1, 0x00354,
 0x00353, 0x00f1c, 0x00c7b, 0x00c37, 0x0101d, 0x012cb, 0x01142, 0x0197d,
 0x01095, 0x01e3b, 0x0186b, 0x00588, 0x01c2a, 0x014b8, 0x01e3a, 0x018ec,
 0x01f46, 0x012fa, 0x00a53, 0x01ce8, 0x00a55, 0x01c29, 0x0117b, 0x01052,
 0x012a0, 0x00589, 0x00950, 0x01c2b, 0x00a50, 0x0208b, 0x0180e, 0x02027,
 0x02556, 0x01e20, 0x006e7, 0x01c28, 0x0197a, 0x00684, 0x020a2, 0x01f22,
 0x03018, 0x039cf, 0x03e25, 0x02557, 0x0294c, 0x028a6, 0x00d11, 0x028a9,
 0x02979, 0x00d46, 0x00a56, 0x039ce, 0x030cc, 0x0329a, 0x0149d, 0x0510f,
 0x0451c, 0x02028, 0x03299, 0x01ced, 0x014b9, 0x00f85, 0x00c7a, 0x01800,
 0x00341, 0x012ca, 0x039c8, 0x0329d, 0x00d0d, 0x03e20, 0x05144, 0x00d45,
 0x030d0, 0x0186d, 0x030d5, 0x00d0f, 0x00d40, 0x04114, 0x020a1, 0x0297f,
 0x03e24, 0x032f1, 0x04047, 0x030d4, 0x028a8, 0x00d0e, 0x0451d, 0x04044,
 0x0297e, 0x04042, 0x030d2, 0x030cf, 0x03e21, 0x03e26, 0x028a5, 0x0451a,
 0x00d48, 0x01a16, 0x00d44, 0x04518, 0x0149b, 0x039ca, 0x01498, 0x0403d,
 0x0451b, 0x0149c, 0x032f3, 0x030cb, 0x08073, 0x03e22, 0x0529a, 0x020aa,
 0x039cc, 0x0738a, 0x06530, 0x07389, 0x06193, 0x08071, 0x04043, 0x030ce,
 0x05147, 0x07388, 0x05145, 0x08072, 0x04521, 0x00d47, 0x0297c, 0x030cd,
 0x030ca, 0x0000b, 0x0000c, 0x00083, 0x000e4, 0x00048, 0x00102, 0x001cc,
 0x001f5, 0x00097, 0x0020b, 0x00124, 0x00453, 0x00627, 0x00639, 0x00605,
 0x00517, 0x001b8, 0x00663, 0x00667, 0x007c3, 0x00823, 0x00961, 0x00963,
 0x00e5a, 0x00e59, 0x00a2b, 0x00cbf, 0x00292, 0x00a2d, 0x007d0, 0x00953,
 0x00cc5, 0x00f84, 0x004ab, 0x014a7, 0x0068a, 0x0117a, 0x0052e, 0x01442,
 0x0052c, 0x00c77, 0x00f8f, 0x004aa, 0x01094, 0x01801, 0x012c4, 0x0297b,
 0x00952, 0x01f19, 0x006a5, 0x01149, 0x012c5, 0x01803, 0x022f2, 0x0329b,
 0x04520, 0x0149e, 0x00d13, 0x01f16, 0x01ce9, 0x0101c, 0x006e6, 0x039c9,
 0x06191, 0x07c8e, 0x06192, 0x0ca63, 0x039cd, 0x06190, 0x06884, 0x06885,
 0x07382, 0x00d49, 0x00d41, 0x0450c, 0x0149a, 0x030d1, 0x08077, 0x03e23,
 0x01a15, 0x0e701, 0x0e702, 0x08079, 0x0822a, 0x0a218, 0x07887, 0x0403f,
 0x0520b, 0x0529b, 0x0e700, 0x04519, 0x00007, 0x000e0, 0x000d0, 0x0039b,
 0x003e5, 0x00163, 0x0063e, 0x007c9, 0x00806, 0x00954, 0x01044, 0x01f44,
 0x0197c, 0x01f45, 0x00a51, 0x01f47, 0x00951, 0x0052d, 0x02291, 0x0092f,
 0x00a54, 0x00d12, 0x0297d, 0x00d0c, 0x01499, 0x0329e, 0x032f0, 0x02025,
 0x039c6, 0x00a57, 0x03e46, 0x00d42, 0x0738b, 0x05146, 0x04046, 0x08078,
 0x0510e, 0x07886, 0x02904, 0x04156, 0x04157, 0x06032, 0x030d3, 0x08bce,
 0x04040, 0x0403e, 0x0a414, 0x10457, 0x08075, 0x06887, 0x07c8f, 0x039c7,
 0x07387, 0x08070, 0x08bcf, 0x1482a, 0x10456, 0x1482b, 0x01a17, 0x06886,
 0x0450d, 0x00013, 0x0006b, 0x00615, 0x0080b, 0x0082b, 0x00952, 0x00e5b,
 0x018e2, 0x0186c, 0x01f18, 0x0329f, 0x00d43, 0x03e29, 0x05140, 0x05141,
 0x0ca62, 0x06033, 0x03c42, 0x03e28, 0x0450f, 0x0a21a, 0x07384, 0x0a219,
 0x0e703, 0x0a21b, 0x01a14, 0x07383, 0x045e6, 0x0007a, 0x0012c, 0x00ccc,
 0x0068f, 0x01802, 0x00a52, 0x00953, 0x04045, 0x01a20, 0x0451f, 0x000a4,
 0x00735, 0x01cec, 0x02029, 0x020a3, 0x0451e, 0x00069, 0x00c24, 0x02024,
 0x032f2, 0x05142, 0x00196, 0x00523, 0x000a6, 0x0197b, 0x0030b, 0x0092e,
 0x003e9, 0x03e27, 0x00160, 0x05143, 0x00652, 0x04041, 0x00734, 0x028a7,
 0x0080f, 0x01483, 0x0097c, 0x00340, 0x0068b, 0x00522, 0x01054, 0x01096,
 0x01f17, 0x0202b, 0x01cea, 0x020a0, 0x02978, 0x02026, 0x0297a, 0x039cb,
 0x03e2b, 0x0149f, 0x0329c, 0x07385, 0x08074, 0x0450e, 0x03e2a, 0x05149,
 0x08076, 0x07386, 0x05148,
};

static const uint8_t coef1_huffbits[555] = {
  9,  5,  2,  4,  4,  5,  5,  5,
  6,  6,  6,  6,  6,  7,  7,  7,
  7,  7,  7,  7,  7,  7,  7,  8,
  8,  8,  8,  8,  8,  8,  8,  8,
  8,  8,  8,  8,  8,  8,  8,  8,
  9,  8,  8,  8,  8,  9,  9,  9,
  9,  9,  9,  9,  9,  9,  9,  9,
  9, 10, 10, 10,  9,  9,  9,  9,
 10, 10, 10, 10, 10, 10, 10, 10,
 10, 10, 10, 10, 10, 10, 10, 10,
 10, 10, 10, 10, 10, 10, 10, 10,
 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11,
 11, 12, 12, 12, 12, 12, 12, 12,
 12, 12, 11, 11, 11, 11, 11, 12,
 12, 12, 12, 12, 12, 12, 12, 12,
 13, 12, 12, 12, 12, 12, 12, 12,
 13, 12, 12, 12, 12, 12, 12, 12,
 12, 13, 12, 12, 12, 13, 13, 13,
 13, 12, 12, 12, 12, 12, 12, 13,
 12, 13, 13, 13, 13, 13, 13, 13,
 14, 14, 13, 13, 13, 13, 13, 13,
 13, 12, 12, 12, 13, 13, 13, 13,
 13, 13, 13, 13, 13, 13, 13, 13,
 13, 13, 14, 13, 14, 13, 13, 13,
 13, 13, 14, 13, 14, 14, 13, 14,
 14, 13, 14, 13, 13, 14, 14, 13,
 14, 14, 14, 14, 14, 14, 14, 14,
 14, 15, 14, 14, 14, 14, 15, 15,
 15, 14, 14, 13, 13, 12, 12, 13,
 13, 13, 14, 14, 15, 14, 15, 15,
 14, 13, 14, 15, 15, 15, 14, 14,
 14, 14, 15, 14, 14, 15, 15, 15,
 14, 15, 14, 14, 14, 14, 14, 15,
 15, 16, 15, 15, 15, 14, 15, 15,
 15, 15, 14, 14, 16, 14, 15, 14,
 14, 15, 15, 15, 15, 16, 15, 14,
 15, 15, 15, 16, 15, 15, 14, 14,
 14,  4,  7,  8,  8,  9,  9,  9,
  9, 10, 10, 11, 11, 11, 11, 11,
 11, 12, 11, 11, 11, 12, 12, 12,
 12, 12, 12, 12, 12, 12, 11, 12,
 12, 12, 13, 13, 13, 13, 13, 13,
 13, 12, 12, 13, 13, 13, 13, 14,
 14, 13, 14, 13, 13, 13, 14, 14,
 15, 15, 14, 13, 13, 13, 14, 14,
 15, 15, 15, 16, 14, 15, 17, 17,
 15, 15, 15, 15, 15, 14, 16, 14,
 16, 16, 16, 16, 16, 16, 15, 15,
 17, 15, 16, 15,  6,  8, 10, 10,
 10, 11, 11, 11, 12, 12, 13, 13,
 13, 13, 14, 13, 14, 13, 14, 14,
 14, 14, 14, 15, 15, 14, 14, 14,
 14, 14, 14, 15, 15, 15, 15, 16,
 15, 15, 16, 15, 15, 15, 14, 16,
 15, 15, 18, 17, 16, 17, 15, 14,
 15, 16, 16, 19, 17, 19, 16, 17,
 15,  7, 10, 11, 12, 12, 12, 12,
 13, 13, 13, 14, 15, 14, 15, 15,
 16, 15, 14, 14, 15, 16, 15, 16,
 16, 16, 16, 15, 15,  7, 11, 12,
 13, 13, 14, 14, 15, 15, 15,  8,
 11, 13, 14, 14, 15,  9, 12, 14,
 14, 15,  9, 13, 10, 13, 10, 14,
 10, 14, 11, 15, 11, 15, 11, 14,
 12, 15, 12, 13, 13, 13, 13, 13,
 13, 14, 13, 14, 14, 14, 14, 14,
 14, 15, 14, 15, 16, 15, 14, 15,
 16, 15, 15,
};

static const uint32_t coef2_huffcodes[1336] = {
 0x003e6, 0x000f6, 0x00000, 0x00002, 0x00006, 0x0000f, 0x0001b, 0x00028,
 0x00039, 0x0003f, 0x0006b, 0x00076, 0x000b7, 0x000e8, 0x000ef, 0x00169,
 0x001a7, 0x001d4, 0x001dc, 0x002c4, 0x00349, 0x00355, 0x00391, 0x003dc,
 0x00581, 0x005b2, 0x00698, 0x0070c, 0x00755, 0x0073a, 0x00774, 0x007cf,
 0x00b0a, 0x00b66, 0x00d2e, 0x00d5e, 0x00e1b, 0x00eac, 0x00e5a, 0x00f7e,
 0x00fa1, 0x0163e, 0x01a37, 0x01a52, 0x01c39, 0x01ab3, 0x01d5f, 0x01cb6,
 0x01f52, 0x01dd9, 0x02c04, 0x02c2e, 0x02c2d, 0x02c23, 0x03467, 0x034a3,
 0x0351b, 0x03501, 0x03a5d, 0x0351c, 0x03875, 0x03dea, 0x0397b, 0x039db,
 0x03df1, 0x039d8, 0x03bb4, 0x0580a, 0x0584d, 0x05842, 0x05b13, 0x058ea,
 0x0697d, 0x06a06, 0x068cc, 0x06ac7, 0x06a96, 0x072f4, 0x07543, 0x072b4,
 0x07d20, 0x0b003, 0x073b5, 0x07be6, 0x0d180, 0x07bd1, 0x07cb8, 0x07d06,
 0x07d25, 0x0d2f2, 0x0d19a, 0x0d334, 0x0e1dc, 0x0d529, 0x0d584, 0x0e1d2,
 0x0e5e3, 0x0eec4, 0x0e564, 0x0fa49, 0x16001, 0x0eedc, 0x0f7fa, 0x1a32c,
 0x16131, 0x16003, 0x0f9c8, 0x1ef80, 0x1d2a0, 0x1aa4b, 0x0f7ce, 0x1abfe,
 0x1aa50, 0x1a458, 0x1a816, 0x1cae4, 0x1d2fe, 0x1d52e, 0x1aa4c, 0x2c245,
 0x1d2a1, 0x1a35d, 0x1ca1b, 0x1d5d8, 0x1f531, 0x1ca1c, 0x1f389, 0x1f4af,
 0x3a5e7, 0x351fb, 0x2c24b, 0x34bce, 0x2c24d, 0x2c249, 0x2c24a, 0x72dfc,
 0x357ef, 0x35002, 0x3a5e6, 0x39431, 0x5843b, 0x34a77, 0x58431, 0x3a5f3,
 0x3a5dd, 0x3e5e5, 0x356bd, 0x3976e, 0x6a3d2, 0x3500d, 0x694c4, 0x580bd,
 0x3e5e8, 0x74b95, 0x34a6e, 0x3977c, 0x39432, 0x5b0d2, 0x6a3d8, 0x580b8,
 0x5b0cb, 0x5b0d7, 0x72dee, 0x72ded, 0x72dec, 0x74b9c, 0x3977f, 0x72dea,
 0x74b9e, 0x7be7d, 0x580bf, 0x5b0d5, 0x7cba8, 0x74b91, 0x3e5dd, 0xb6171,
 0xd46b3, 0xd46b9, 0x7cba1, 0x74b9f, 0x72de1, 0xe59f5, 0x3e5eb, 0x00004,
 0x00015, 0x00038, 0x00075, 0x000e8, 0x001d3, 0x00347, 0x0039c, 0x00690,
 0x0074a, 0x00b60, 0x00e93, 0x00f74, 0x0163d, 0x01a5a, 0x01d24, 0x01cbe,
 0x01f4b, 0x03468, 0x03562, 0x03947, 0x03e82, 0x05804, 0x05b12, 0x05803,
 0x0696d, 0x06a9e, 0x0697c, 0x06978, 0x06afb, 0x074b2, 0x072f5, 0x073c0,
 0x07541, 0x06944, 0x074b7, 0x070d3, 0x07ba9, 0x0b0b1, 0x0d1af, 0x0e1dd,
 0x0e5e2, 0x0e1a3, 0x0eec3, 0x1612f, 0x0e961, 0x0eeda, 0x0e78e, 0x0fa48,
 0x1612c, 0x0e511, 0x0e565, 0x0e953, 0x1aa4a, 0x0e59d, 0x1d52c, 0x1a811,
 0x1cae7, 0x1abfc, 0x1d52d, 0x1cacf, 0x1cf05, 0x2c254, 0x34a72, 0x1f4ac,
 0x3976b, 0x34a71, 0x2c6d9, 0x2d873, 0x34a6a, 0x357e7, 0x3464c, 0x3e5f5,
 0x58433, 0x1f53a, 0x3500a, 0x357ea, 0x34a73, 0x3942f, 0x357e5, 0x39775,
 0x694cd, 0x39772, 0x7cba5, 0x6a3ef, 0x35483, 0x74b98, 0x5b0c1, 0x39770,
 0x3a5d7, 0x39433, 0x39434, 0x694ce, 0x580be, 0x3e5ff, 0x6a3ec, 0xb616f,
 0xd46b1, 0x6a3d1, 0x72de5, 0x74b6e, 0x72de9, 0x3e700, 0xd46b6, 0x6a3e9,
 0x74b69, 0xe5675, 0xd46b8, 0x7cbaa, 0x3a5d1, 0x0000c, 0x0003c, 0x000eb,
 0x001f1, 0x003a4, 0x006a8, 0x007d5, 0x00d43, 0x00e77, 0x016c5, 0x01cb1,
 0x02c5d, 0x03a55, 0x03a56, 0x03e51, 0x03bb5, 0x05b0a, 0x06a9f, 0x074b8,
 0x07d28, 0x0d187, 0x0d40e, 0x0d52e, 0x0d425, 0x0eae3, 0x0e1d3, 0x1612e,
 0x0e59e, 0x0eec2, 0x0e578, 0x0e51a, 0x0e579, 0x0e515, 0x0e960, 0x0d183,
 0x0d220, 0x0d2cb, 0x0e512, 0x16c3e, 0x16002, 0x16c42, 0x1cae9, 0x3461a,
 0x1d2fa, 0x1a308, 0x1a849, 0x1cf07, 0x1f38f, 0x34b65, 0x2c253, 0x1ef9e,
 0x1cbc3, 0x1cbc1, 0x2c255, 0x1f384, 0x58435, 0x2c5cd, 0x3a5f7, 0x2c252,
 0x3959c, 0x2c6d8, 0x3a5d3, 0x6ad78, 0x6a3f2, 0x7cba9, 0xb6176, 0x72deb,
 0x39764, 0x3e5f6, 0x3a5d8, 0x74a8c, 0x6a3e6, 0x694d1, 0x6ad79, 0x1a4592,
 0xe59fb, 0x7cbb3, 0x5b0cd, 0x00017, 0x000b5, 0x002c3, 0x005b7, 0x00b1c,
 0x00e5c, 0x0163f, 0x01ab2, 0x01efa, 0x0348a, 0x0396e, 0x058da, 0x06963,
 0x06a30, 0x072cd, 0x073cf, 0x07ce7, 0x0d2ca, 0x0d2d8, 0x0e764, 0x0e794,
 0x16008, 0x16167, 0x1617e, 0x1aa49, 0x1a30b, 0x1a813, 0x2c6da, 0x1a580,
 0x1cbc2, 0x0f9ca, 0x1617f, 0x1d2fe, 0x0f7fc, 0x16c40, 0x0e513, 0x0eec5,
 0x0f7c3, 0x1d508, 0x1a81e, 0x1d2fd, 0x39430, 0x35486, 0x3e5fd, 0x2c24c,
 0x2c75a, 0x34a74, 0x3a5f4, 0x3464d, 0x694ca, 0x3a5f1, 0x1d509, 0x1d5c0,
 0x34648, 0x3464e, 0x6a3d5, 0x6a3e8, 0x6a3e7, 0x5b0c3, 0x2c248, 0x1f38a,
 0x3a5f2, 0x6a3e5, 0x00029, 0x00168, 0x0058c, 0x00b67, 0x00f9d, 0x01c3d,
 0x01cbf, 0x02c20, 0x0351d, 0x03df6, 0x06af9, 0x072b5, 0x0b1d7, 0x0b0b2,
 0x0d40a, 0x0d52b, 0x0e952, 0x0e797, 0x163c3, 0x1c3a0, 0x1f386, 0x1ca21,
 0x34655, 0x2c247, 0x1f53b, 0x2c250, 0x2c24f, 0x1f385, 0x1ef5d, 0x1cf15,
 0x1caea, 0x1ab0a, 0x1cf19, 0x1f53d, 0x1d5c2, 0x1d2fb, 0x1ef58, 0x34a78,
 0x357ec, 0x1f533, 0x3a5e1, 0x694d2, 0x58482, 0x3a5ee, 0x2c6dc, 0x357eb,
 0x5b0c4, 0x39778, 0x6a3e1, 0x7cbb4, 0x3a5e1, 0x74b68, 0x3a5ef, 0x3a5d2,
 0x39424, 0x72de2, 0xe59f6, 0xe59f7, 0x3e702, 0x3e5ec, 0x1f38b, 0x0003b,
 0x001f0, 0x00777, 0x00fa8, 0x01cb2, 0x02d84, 0x03a57, 0x03dd6, 0x06917,
 0x06a11, 0x07d07, 0x0eae2, 0x0e796, 0x0f9c9, 0x0f7fb, 0x16166, 0x16160,
 0x1ab1b, 0x1abfa, 0x2d87b, 0x1d2f7, 0x39768, 0x1f38c, 0x34653, 0x34651,
 0x6a3d9, 0x35001, 0x3abbd, 0x38742, 0x39426, 0x34a76, 0x3a5ec, 0x34a75,
 0x35000, 0x35488, 0x1cf10, 0x2c6db, 0x357ed, 0x357e8, 0x357e9, 0x3a5f0,
 0x694c2, 0xb6178, 0x72df5, 0x39425, 0x3942b, 0x74b6d, 0x74b6f, 0xb6177,
 0xb6179, 0x74b6a, 0xb6172, 0x58487, 0x3e5ee, 0x3e5ed, 0x72df2, 0x72df4,
 0x7cbae, 0x6a3ca, 0x70e86, 0x34bcf, 0x6a3c8, 0x00059, 0x00384, 0x00d5b,
 0x01c38, 0x03560, 0x0395b, 0x0584e, 0x06964, 0x073cd, 0x0b1e7, 0x0e798,
 0x0e78d, 0x0fa43, 0x1a848, 0x1a32f, 0x1aa4e, 0x3464a, 0x1f4ab, 0x1f38d,
 0x3a5eb, 0x3a5d4, 0x3548a, 0x6a3c7, 0x5b0d0, 0x6a3c5, 0x7cbb0, 0x694cb,
 0x3a5e5, 0x3e5e2, 0x3942c, 0x2d872, 0x1f4ae, 0x3a5d5, 0x694d3, 0x58481,
 0x35009, 0x39774, 0x58432, 0xb616c, 0x5b0db, 0x3548b, 0xb6174, 0x1d5d95,
 0xb004c, 0x7cbb2, 0x3a5e5, 0x74a8f, 0xe59f9, 0x72df6, 0xe59fd, 0x7cbad,
 0xd427d, 0x72cff, 0x3977a, 0x5b0d9, 0xb616d, 0xb616b, 0x1a4593, 0x7cbaf,
 0x5b0da, 0x00071, 0x003eb, 0x01603, 0x02c6c, 0x03961, 0x068c8, 0x06a31,
 0x072bd, 0x0d2c2, 0x0e51b, 0x0e5e6, 0x1abfb, 0x1d2ff, 0x1cae5, 0x1ef5c,
 0x1ef5e, 0x1cf13, 0x34a6d, 0x3976d, 0xb616a, 0x3e5f2, 0x6a3c4, 0xb6169,
 0x3e5dc, 0x580b9, 0x74b99, 0x75764, 0x58434, 0x3a5d9, 0x6945a, 0x69459,
 0x3548c, 0x3a5e9, 0x69457, 0x72df1, 0x6945e, 0x6a35e, 0x3e701, 0xb6168,
 0x5b0dd, 0x3a5de, 0x6a3c2, 0xd4278, 0x6a3cc, 0x72dfd, 0xb6165, 0x16009a,
 0x7cbb1, 0xd427c, 0xb6162, 0xe765e, 0x1cecbe, 0x7cbb6, 0x69454, 0xb6160,
 0xd427a, 0x1d5d96, 0xb1d6d, 0xe59f4, 0x72de8, 0x3a5db, 0x0007a, 0x006ae,
 0x01c3c, 0x03aba, 0x058e9, 0x072cc, 0x0d2dd, 0x0d22d, 0x0eec1, 0x0eedb,
 0x1d2a2, 0x1ef5b, 0x357e2, 0x3abbf, 0x1d2f9, 0x35004, 0x3a5dc, 0x351fc,
 0x3976c, 0x6a3c6, 0x6a3cb, 0x3e5ea, 0xe59f3, 0x6a3ce, 0x69452, 0xe59f0,
 0x74b90, 0xd4279, 0xd427b, 0x7cbb5, 0x5b0c5, 0x3a5e3, 0x3a5e2, 0x000d0,
 0x00775, 0x01efe, 0x03dd5, 0x0728c, 0x07cb9, 0x0e1a2, 0x0ea85, 0x0eed8,
 0x1a30a, 0x1aa4f, 0x3a5df, 0x35008, 0x3a5e0, 0x3e5f4, 0x3e5f7, 0xb1d6c,
 0x5843e, 0x34a70, 0x72df8, 0x74b6b, 0xd427f, 0x72df0, 0x5b0bf, 0x5b0c0,
 0xd46b0, 0x72def, 0xe59f8, 0x162e64, 0xb1d6f, 0x3a5e0, 0x39427, 0x69166,
 0x6a3e2, 0x6a3e3, 0x74a8d, 0xd427e, 0x1d5d97, 0xd46b4, 0x5b0d8, 0x6a3d3,
 0x000e0, 0x00b63, 0x034cc, 0x06a33, 0x073c9, 0x0e1a0, 0x0f7fd, 0x0f9cc,
 0x1617d, 0x1caeb, 0x1f4a9, 0x3abb3, 0x69450, 0x39420, 0x39777, 0x3e5e0,
 0x6a3d4, 0x6a3ed, 0xb6166, 0xe59f1, 0xb1d6e, 0xe5676, 0x6a3ea, 0xe5674,
 0xb6163, 0xd46b7, 0x7cba6, 0xd46ba, 0x1d5d94, 0xb6164, 0x6a3f1, 0x7cba2,
 0x69451, 0x72dfa, 0xd46bb, 0x72df7, 0x74b94, 0x1cecbf, 0xe59fa, 0x16009b,
 0x6a3e4, 0x000e6, 0x00e94, 0x03876, 0x070ef, 0x0d52a, 0x16015, 0x16014,
 0x1abf9, 0x1cf17, 0x34a79, 0x34650, 0x3e705, 0x6a3d0, 0x58430, 0x74b9d,
 0x7be7e, 0x5b0be, 0x39773, 0x6a3de, 0x000fb, 0x00f7b, 0x03dd7, 0x07bd0,
 0x0e59c, 0x0f9cd, 0x1cf18, 0x1d2ff, 0x34a7a, 0x39429, 0x3500c, 0x72de0,
 0x69456, 0x7be7c, 0xd46b5, 0xd46b2, 0x6a3dd, 0x001a2, 0x0163b, 0x06913,
 0x0b016, 0x0fa42, 0x1a32d, 0x1cf06, 0x34a7c, 0x34a7d, 0xb6161, 0x35481,
 0x3e5fa, 0x7cba0, 0x7be7f, 0x7cba3, 0x7cba7, 0x5b0d3, 0x72de6, 0x6a3dc,
 0x001a9, 0x01ab4, 0x06a34, 0x0d46a, 0x16130, 0x1ef5f, 0x1f532, 0x1f536,
 0x3942e, 0x58436, 0x6a3db, 0x6945b, 0x001c9, 0x01ca0, 0x0728b, 0x0eed9,
 0x1f539, 0x1ca1d, 0x39765, 0x39766, 0x58439, 0x6945d, 0x39767, 0x001d3,
 0x01f2c, 0x07bfc, 0x16161, 0x34652, 0x3a5ed, 0x3548d, 0x58438, 0x6a3da,
 0x002c1, 0x02c5e, 0x0d335, 0x1ab1a, 0x2d874, 0x35006, 0x35484, 0x5b0cc,
 0x74b9a, 0x72df3, 0x6a3d6, 0x002da, 0x034b3, 0x0d5ae, 0x1caee, 0x2d871,
 0x357e3, 0x74b97, 0x72df9, 0x580ba, 0x5b0d4, 0x0034d, 0x0354e, 0x0f750,
 0x1cbc0, 0x3a5e7, 0x3a5e4, 0x00385, 0x03a58, 0x16c41, 0x2c5cf, 0x3e5e1,
 0x74b6c, 0xe5677, 0x6a3df, 0x00390, 0x03e50, 0x163c2, 0x2d876, 0x35482,
 0x5b0d6, 0x5843a, 0x0039f, 0x0585e, 0x1a583, 0x3500f, 0x74b93, 0x39771,
 0x003e4, 0x06912, 0x16c43, 0x357e1, 0x0058a, 0x0696f, 0x1f538, 0x5b0c9,
 0x6a3cf, 0x005b6, 0x06af8, 0x1f534, 0x58483, 0x6a3e0, 0x00695, 0x07d02,
 0x1cae8, 0x58485, 0x006a2, 0x0754a, 0x357ee, 0x3977b, 0x00748, 0x074b2,
 0x34a7b, 0x00729, 0x0b1e0, 0x34649, 0x3e5e3, 0x0073d, 0x0d2c4, 0x3e5e6,
 0x007bb, 0x0b099, 0x39762, 0x5b0ce, 0x6945f, 0x007d1, 0x0d5ab, 0x39779,
 0x007d3, 0x0d52f, 0x39763, 0x6945c, 0x00b1a, 0x0d2c5, 0x35489, 0x00d23,
 0x0eaed, 0x3e5f8, 0x00d32, 0x16016, 0x3e5fb, 0x00d41, 0x0e768, 0x3a5ed,
 0x00e1f, 0x16017, 0x58027, 0x00ead, 0x0fa07, 0x69455, 0x00e54, 0x1612b,
 0x00e55, 0x1a581, 0x00f78, 0x1a32b, 0x580bc, 0x6a3ee, 0x00f79, 0x1abfd,
 0x00f95, 0x1ab18, 0x6a3f0, 0x01637, 0x1aa4d, 0x0162d, 0x1f53c, 0x6a3f3,
 0x01a31, 0x1a810, 0x39769, 0x01a50, 0x1caef, 0x01a36, 0x1a32e, 0x01a67,
 0x1f38e, 0x01a85, 0x1ef59, 0x01aa6, 0x1ef83, 0x01d51, 0x2c012, 0x01d53,
 0x2d879, 0x01d5e, 0x35005, 0x01cba, 0x1cf04, 0x69453, 0x01d2d, 0x351ff,
 0x01f2d, 0x2d86f, 0x01f29, 0x35007, 0x02c22, 0x351fa, 0x02c03, 0x3a5ec,
 0x02c5f, 0x3a5eb, 0x02c58, 0x34a6b, 0x03469, 0x356be, 0x02c59, 0x34a6c,
 0x0346a, 0x3a5ea, 0x034bd, 0x034bf, 0x356bf, 0x0386a, 0x03ab9, 0x5843f,
 0x0386b, 0x3a5f5, 0x03a4b, 0x39421, 0x03aa4, 0x3a5e9, 0x03a5a, 0x03960,
 0x3977e, 0x03de9, 0x03958, 0x03df7, 0x039e1, 0x3e5e4, 0x0395f, 0x69458,
 0x03e91, 0x03df2, 0x39428, 0x058f2, 0x03e80, 0x6a3c3, 0x03e93, 0x694c0,
 0x058b8, 0x5b0ca, 0x0584f, 0x694c1, 0x058f1, 0x068d6, 0x06a10, 0x06ac3,
 0x06a32, 0x070d2, 0x06911, 0x074b1, 0x07494, 0x06ad4, 0x06ad6, 0x072b8,
 0x06afa, 0x074b3, 0x07540, 0x073ce, 0x0b005, 0x074b3, 0x07495, 0x074b9,
 0x0d336, 0x07bff, 0x07763, 0x073c8, 0x07d29, 0x0b622, 0x0d221, 0x0d181,
 0x0b1d1, 0x074b8, 0x0b1d0, 0x0d19b, 0x0d2c3, 0x0b172, 0x0d2dc, 0x0b623,
 0x0d5aa, 0x0d426, 0x0d182, 0x0e795, 0x0e1d1, 0x0d337, 0x0e96c, 0x0e5e4,
 0x0e514, 0x0eaee, 0x16000, 0x0e767, 0x0e1a1, 0x0e78f, 0x16004, 0x0f7c2,
 0x0e799, 0x0e5e7, 0x0e566, 0x0e769, 0x0f751, 0x0eede, 0x0fa06, 0x16005,
 0x0fa9f, 0x1a5e6, 0x0e766, 0x1636f, 0x0eedd, 0x0eec0, 0x1a309, 0x1ceca,
 0x163cd, 0x0f9cb, 0x0eedf, 0x1a582, 0x1612d, 0x0e5e5, 0x1abf8, 0x1a30c,
 0x1ca1f, 0x163cc, 0x1a35c, 0x1ca1e, 0x1aa51, 0x163ac, 0x1a84e, 0x1a53f,
 0x1cf16, 0x1d2fc, 0x1a5b3, 0x1ab19, 0x1a81f, 0x1d5c3, 0x16c3f, 0x1d5c1,
 0x1d2fc, 0x1f4aa, 0x1a812, 0x1f535, 0x1cf12, 0x1a817, 0x1617c, 0x1ab0b,
 0x1d2f8, 0x1ef82, 0x2d87a, 0x1d52f, 0x1f530, 0x1aa48, 0x35487, 0x1d2fd,
 0x1f4ad, 0x1cf11, 0x3461b, 0x35485, 0x1ca20, 0x1caed, 0x1cae6, 0x1abff,
 0x3464f, 0x34a6f, 0x1ef81, 0x3464b, 0x39d96, 0x1f383, 0x1f537, 0x1cf14,
 0x2c5ce, 0x3500e, 0x2c251, 0x1caec, 0x1f387, 0x34654, 0x357e4, 0x2d878,
 0x3500b, 0x35480, 0x3a5e8, 0x3548e, 0x34b64, 0x1f4a8, 0x35003, 0x3e5df,
 0x2d870, 0x357e6, 0x3e5f0, 0x1ef5a, 0x3a5ea, 0x1f388, 0x3e703, 0x2c24e,
 0x3a5e2, 0x351fd, 0x2c6dd, 0x3e704, 0x351fe, 0x2d875, 0x5b0c7, 0x3976a,
 0x3a5e6, 0x39423, 0x58480, 0x2c246, 0x3a5e3, 0x2d877, 0x3e5f1, 0x3abbe,
 0x58489, 0x3e5f9, 0x357e0, 0x3abbc, 0x5b0c6, 0x69167, 0x69165, 0x3e5e9,
 0x39422, 0x3976f, 0x3977d, 0x3e5de, 0x6a3c9, 0x58b98, 0x3a5f6, 0x3a5d0,
 0x58486, 0x6a3c1, 0x3e5fc, 0x5b0dc, 0x3548f, 0x3942d, 0x694c9, 0x58484,
 0x3a5e8, 0x74b9b, 0x74b96, 0x694d0, 0x58488, 0x3a5e4, 0x3942a, 0x72ec2,
 0x39776, 0x5b0d1, 0x5b0cf, 0x3a5d6, 0xe59fc, 0x5b0c8, 0x3e5e7, 0x7cbb7,
 0x70e87, 0x7cbab, 0x5b0c2, 0x694c3, 0x74a8e, 0x3e5f3, 0x6a3cd, 0x72dfe,
 0x73b2e, 0x72ec0, 0x694c5, 0x58437, 0x694c8, 0x72dff, 0x39435, 0x5843d,
 0x6a3d7, 0x72ec1, 0xd22c8, 0x694cf, 0xb6173, 0x3e5fe, 0x580bb, 0xe59f2,
 0xb616e, 0xb6175, 0x3a5da, 0x5b0bd, 0x694cc, 0x5843c, 0x694c7, 0x74b92,
 0x72ec3, 0x694c6, 0xb6170, 0x7cbac, 0xb1733, 0x7cba4, 0xb6167, 0x72de7,
 0x72de4, 0x6a3c0, 0x3e5ef, 0x162e65, 0x72de3, 0x72dfb, 0x6a35f, 0x6a3eb,
};

static const uint8_t coef2_huffbits[1336] = {
 11,  9,  2,  3,  4,  4,  5,  6,
  6,  7,  7,  8,  8,  8,  9,  9,
  9,  9, 10, 10, 10, 10, 11, 11,
 11, 11, 11, 11, 11, 12, 12, 12,
 12, 12, 12, 12, 12, 12, 13, 13,
 13, 13, 13, 13, 13, 13, 13, 14,
 14, 14, 14, 14, 14, 14, 14, 14,
 14, 14, 14, 14, 14, 15, 15, 15,
 15, 15, 15, 15, 15, 15, 15, 15,
 15, 15, 15, 15, 15, 16, 15, 16,
 16, 16, 16, 16, 16, 16, 16, 16,
 16, 16, 16, 16, 16, 16, 16, 16,
 17, 17, 17, 17, 17, 17, 17, 17,
 17, 17, 17, 18, 17, 17, 17, 17,
 17, 17, 17, 18, 18, 17, 17, 18,
 17, 17, 18, 17, 18, 18, 18, 18,
 19, 18, 18, 18, 18, 18, 18, 20,
 18, 18, 18, 19, 19, 18, 19, 18,
 19, 19, 18, 19, 19, 18, 19, 19,
 19, 19, 18, 19, 19, 19, 19, 19,
 19, 19, 20, 20, 20, 19, 19, 20,
 19, 20, 19, 19, 20, 19, 19, 20,
 20, 20, 20, 19, 20, 21, 19,  3,
  5,  7,  8,  9,  9, 10, 11, 11,
 12, 12, 12, 13, 13, 13, 13, 14,
 14, 14, 14, 15, 15, 15, 15, 15,
 15, 15, 15, 15, 15, 15, 16, 16,
 15, 15, 15, 15, 16, 16, 16, 16,
 17, 16, 17, 17, 16, 17, 17, 17,
 17, 17, 17, 16, 17, 17, 17, 17,
 18, 17, 17, 18, 18, 18, 18, 18,
 19, 18, 18, 18, 18, 18, 18, 19,
 19, 18, 18, 18, 18, 19, 18, 19,
 19, 19, 20, 19, 18, 19, 19, 19,
 19, 19, 19, 19, 19, 19, 19, 20,
 20, 19, 20, 19, 20, 19, 20, 19,
 19, 21, 20, 20, 19,  4,  7,  8,
 10, 11, 11, 12, 12, 13, 13, 14,
 14, 14, 14, 15, 15, 15, 15, 15,
 16, 16, 16, 16, 16, 16, 16, 17,
 17, 17, 17, 17, 17, 17, 16, 16,
 16, 16, 17, 17, 17, 17, 18, 18,
 18, 17, 17, 18, 18, 18, 18, 18,
 18, 18, 18, 18, 19, 18, 18, 18,
 19, 18, 19, 19, 19, 20, 20, 20,
 19, 19, 19, 19, 19, 19, 19, 21,
 21, 20, 19,  5,  8, 10, 11, 12,
 13, 13, 13, 14, 14, 15, 15, 15,
 15, 16, 16, 16, 16, 16, 17, 17,
 17, 17, 17, 17, 17, 17, 18, 17,
 18, 17, 17, 17, 17, 17, 17, 17,
 17, 17, 17, 17, 19, 18, 19, 18,
 18, 18, 18, 18, 19, 18, 17, 17,
 18, 18, 19, 19, 19, 19, 18, 18,
 18, 19,  6,  9, 11, 12, 13, 13,
 14, 14, 14, 15, 15, 16, 16, 16,
 16, 16, 16, 17, 17, 17, 18, 18,
 18, 18, 18, 18, 18, 18, 18, 18,
 18, 17, 18, 18, 17, 18, 18, 18,
 18, 18, 18, 19, 19, 18, 18, 18,
 19, 19, 19, 20, 19, 19, 18, 19,
 19, 20, 21, 21, 19, 19, 18,  6,
 10, 12, 13, 14, 14, 14, 15, 15,
 15, 16, 16, 17, 17, 17, 17, 17,
 17, 17, 18, 18, 19, 18, 18, 18,
 19, 18, 18, 18, 19, 18, 18, 18,
 18, 18, 18, 18, 18, 18, 18, 18,
 19, 20, 20, 19, 19, 19, 19, 20,
 20, 19, 20, 19, 19, 19, 20, 20,
 20, 19, 19, 18, 19,  7, 10, 12,
 13, 14, 15, 15, 15, 16, 16, 17,
 17, 17, 17, 17, 17, 18, 18, 18,
 18, 19, 18, 19, 19, 19, 20, 19,
 18, 19, 19, 18, 18, 19, 19, 19,
 18, 19, 19, 20, 19, 18, 20, 21,
 20, 20, 19, 19, 21, 20, 21, 20,
 20, 20, 19, 19, 20, 20, 21, 20,
 19,  7, 11, 13, 14, 15, 15, 15,
 16, 16, 17, 17, 17, 17, 18, 18,
 18, 18, 18, 19, 20, 19, 19, 20,
 19, 19, 19, 19, 19, 19, 19, 19,
 18, 18, 19, 20, 19, 19, 19, 20,
 19, 19, 19, 20, 19, 20, 20, 21,
 20, 20, 20, 21, 22, 20, 19, 20,
 20, 21, 20, 21, 20, 19,  8, 11,
 13, 14, 15, 16, 16, 16, 17, 17,
 17, 18, 18, 18, 18, 18, 19, 18,
 19, 19, 19, 19, 21, 19, 19, 21,
 19, 20, 20, 20, 19, 18, 18,  8,
 12, 14, 15, 16, 16, 16, 16, 17,
 17, 17, 19, 18, 18, 19, 19, 20,
 19, 18, 20, 19, 20, 20, 19, 19,
 20, 20, 21, 21, 20, 19, 19, 19,
 19, 19, 19, 20, 21, 20, 19, 19,
  8, 12, 14, 15, 16, 16, 17, 17,
 17, 18, 18, 18, 19, 19, 19, 19,
 19, 19, 20, 21, 20, 21, 19, 21,
 20, 20, 20, 20, 21, 20, 19, 20,
 19, 20, 20, 20, 19, 22, 21, 21,
 19,  9, 12, 14, 15, 16, 17, 17,
 17, 18, 18, 18, 19, 19, 19, 19,
 20, 19, 19, 19,  9, 13, 15, 16,
 17, 17, 18, 18, 18, 19, 18, 20,
 19, 20, 20, 20, 19,  9, 13, 15,
 16, 17, 17, 18, 18, 18, 20, 18,
 19, 20, 20, 20, 20, 19, 20, 19,
  9, 13, 15, 16, 17, 18, 18, 18,
 19, 19, 19, 19, 10, 14, 16, 17,
 18, 18, 19, 19, 19, 19, 19, 10,
 14, 16, 17, 18, 18, 18, 19, 19,
 10, 14, 16, 17, 18, 18, 18, 19,
 19, 20, 19, 10, 14, 16, 18, 18,
 18, 19, 20, 19, 19, 10, 14, 17,
 18, 18, 18, 10, 15, 17, 18, 19,
 19, 21, 19, 11, 15, 17, 18, 18,
 19, 19, 11, 15, 17, 18, 19, 19,
 11, 15, 17, 18, 11, 15, 18, 19,
 19, 11, 15, 18, 19, 19, 11, 16,
 18, 19, 11, 15, 18, 19, 11, 16,
 18, 12, 16, 18, 19, 12, 16, 19,
 12, 16, 19, 19, 19, 12, 16, 19,
 12, 16, 19, 19, 12, 16, 18, 12,
 16, 19, 12, 17, 19, 12, 17, 19,
 12, 17, 19, 12, 17, 19, 13, 17,
 13, 17, 13, 17, 19, 19, 13, 17,
 13, 17, 19, 13, 17, 13, 18, 19,
 13, 17, 19, 13, 18, 13, 17, 13,
 18, 13, 18, 13, 18, 13, 18, 13,
 18, 13, 18, 14, 18, 19, 14, 18,
 14, 18, 14, 18, 14, 18, 14, 19,
 14, 19, 14, 18, 14, 18, 14, 18,
 14, 19, 14, 14, 18, 14, 14, 19,
 14, 18, 14, 19, 14, 19, 14, 15,
 19, 15, 15, 15, 15, 19, 15, 19,
 15, 15, 19, 15, 15, 19, 15, 19,
 15, 19, 15, 19, 15, 15, 15, 15,
 15, 15, 15, 15, 15, 15, 15, 16,
 15, 15, 15, 16, 16, 16, 15, 16,
 16, 16, 16, 16, 16, 16, 16, 16,
 16, 16, 16, 16, 16, 16, 16, 16,
 16, 16, 16, 17, 16, 16, 16, 17,
 17, 16, 17, 17, 16, 17, 17, 17,
 17, 17, 17, 17, 17, 17, 17, 17,
 17, 17, 17, 17, 17, 17, 17, 18,
 17, 17, 17, 17, 17, 17, 17, 17,
 18, 17, 17, 18, 17, 17, 17, 17,
 18, 18, 17, 17, 17, 17, 17, 17,
 17, 18, 17, 18, 18, 17, 17, 17,
 18, 18, 18, 17, 18, 17, 18, 18,
 18, 18, 18, 18, 18, 18, 18, 17,
 18, 18, 18, 18, 19, 18, 18, 18,
 18, 18, 18, 18, 18, 18, 18, 18,
 18, 18, 18, 18, 18, 18, 18, 19,
 18, 18, 19, 18, 18, 18, 19, 18,
 19, 18, 18, 19, 18, 18, 19, 19,
 19, 19, 19, 18, 19, 18, 19, 18,
 19, 19, 18, 18, 19, 19, 19, 19,
 19, 19, 19, 19, 19, 19, 18, 19,
 19, 19, 19, 19, 18, 19, 19, 19,
 19, 19, 19, 19, 19, 19, 19, 20,
 19, 19, 19, 19, 21, 19, 19, 20,
 19, 20, 19, 19, 19, 19, 19, 20,
 20, 20, 19, 19, 19, 20, 19, 19,
 19, 20, 20, 19, 20, 19, 19, 21,
 20, 20, 19, 19, 19, 19, 19, 19,
 20, 19, 20, 20, 20, 20, 20, 20,
 20, 19, 19, 21, 20, 20, 19, 19,
};

static const uint32_t coef3_huffcodes[1072] = {
 0x001b2, 0x00069, 0x00000, 0x00004, 0x00006, 0x0000e, 0x00014, 0x00019,
 0x00016, 0x0002b, 0x00030, 0x0003d, 0x0003c, 0x0005a, 0x0005f, 0x0006d,
 0x0007e, 0x0005f, 0x0007f, 0x000b6, 0x000bc, 0x000d8, 0x000f2, 0x000fe,
 0x000bc, 0x000fc, 0x00161, 0x0016e, 0x00174, 0x00176, 0x001a2, 0x001e3,
 0x001f3, 0x00174, 0x0017a, 0x001ea, 0x002a8, 0x002c4, 0x002e6, 0x00314,
 0x00346, 0x00367, 0x003e9, 0x002e5, 0x002ee, 0x003d6, 0x00555, 0x00554,
 0x00557, 0x005c3, 0x005d6, 0x006e0, 0x0062f, 0x006e2, 0x00799, 0x00789,
 0x007fa, 0x005ce, 0x007fe, 0x005ec, 0x007cc, 0x007af, 0x00aa7, 0x00b19,
 0x00b94, 0x00b85, 0x00b9f, 0x00c48, 0x00c45, 0x00dd8, 0x00c4c, 0x00c4b,
 0x00d99, 0x00d1f, 0x00dc2, 0x00f95, 0x00fa2, 0x00bb5, 0x00b9f, 0x00f5d,
 0x00bbf, 0x00f47, 0x0154a, 0x00fd5, 0x00f45, 0x00f7f, 0x0160d, 0x01889,
 0x01757, 0x01722, 0x018b3, 0x0172d, 0x01a39, 0x01a18, 0x01bb3, 0x01b30,
 0x01e63, 0x0173c, 0x01b35, 0x01723, 0x01e80, 0x01fee, 0x01761, 0x01ffc,
 0x01f7f, 0x02c7c, 0x01fa1, 0x0177b, 0x01755, 0x0175a, 0x01fa6, 0x02eab,
 0x0310a, 0x02c69, 0x03669, 0x03127, 0x03103, 0x02e43, 0x03662, 0x03165,
 0x03124, 0x0313b, 0x03111, 0x03668, 0x0343b, 0x03c52, 0x03efc, 0x02e6c,
 0x03fda, 0x03ef8, 0x02e7b, 0x03ee2, 0x03cc5, 0x03d72, 0x058c0, 0x03df8,
 0x02ea9, 0x03e7e, 0x0556d, 0x05c82, 0x03d71, 0x03e7b, 0x03c42, 0x058d7,
 0x03f4e, 0x06200, 0x03d70, 0x05cb2, 0x05c96, 0x05cb0, 0x03f45, 0x05cb1,
 0x02e6d, 0x03110, 0x02f68, 0x05c90, 0x07ca6, 0x07c88, 0x06204, 0x062c8,
 0x078a6, 0x07986, 0x079d5, 0x0b1ad, 0x07989, 0x0b079, 0x05cdd, 0x0aad4,
 0x05de8, 0x07dcd, 0x07987, 0x05d67, 0x05d99, 0x0b91d, 0x07cf1, 0x05d9b,
 0x079d7, 0x0b07b, 0x05c85, 0x05d9a, 0x07dcc, 0x07ebf, 0x07dce, 0x07dfb,
 0x07ec0, 0x07d1a, 0x07a07, 0x05c84, 0x0c471, 0x07cf2, 0x0baef, 0x0b9d2,
 0x05deb, 0x07bd6, 0x0b845, 0x05d98, 0x0b91a, 0x0bae8, 0x0c4e0, 0x0dc31,
 0x0f93d, 0x0bbce, 0x0d1d2, 0x0f7a9, 0x0d9b9, 0x0bbcb, 0x0b900, 0x0aad7,
 0x0babd, 0x0c4e1, 0x0f46f, 0x0c588, 0x0c58b, 0x160e6, 0x0bbcf, 0x0bac3,
 0x0f945, 0x0f7a3, 0x0d1c1, 0x0fb8e, 0x0f7a4, 0x0fb8c, 0x0f40c, 0x0c473,
 0x0fd72, 0x0bbcd, 0x0fffa, 0x0f940, 0x0bbc9, 0x0f7a8, 0x1a1ed, 0x0bbc5,
 0x1f26f, 0x163fd, 0x160c7, 0x1a1f5, 0x0f947, 0x163fc, 0x154b3, 0x0fff6,
 0x163f6, 0x160e9, 0x1a1f0, 0x0bab9, 0x0baba, 0x17086, 0x0b903, 0x0fd75,
 0x0f308, 0x176f3, 0x163ff, 0x0fd7d, 0x1bb78, 0x163fb, 0x188db, 0x1a1f7,
 0x154b2, 0x172fd, 0x163f4, 0x1bb73, 0x172ff, 0x0babc, 0x0f97d, 0x1a1f3,
 0x1bb6d, 0x1ffd5, 0x1a1f4, 0x1f272, 0x17380, 0x17382, 0x1ffe7, 0x0bac8,
 0x0bbc4, 0x188d3, 0x160e0, 0x0fd7b, 0x1725f, 0x172f5, 0x1bb79, 0x1fad9,
 0x1f269, 0x188d0, 0x0bac4, 0x0bac5, 0x31185, 0x188d2, 0x188cc, 0x31187,
 0x3e7fe, 0x188d1, 0x1bb6c, 0x1f268, 0x1fad2, 0x1ffd9, 0x1a1ea, 0x1bb68,
 0x1facb, 0x3fdb2, 0x1e81a, 0x188ce, 0x172fb, 0x1a1ef, 0x1face, 0x1bb70,
 0x0bac1, 0x1bb6b, 0x172f8, 0x1bb66, 0x1ffdf, 0x1bb6a, 0x1ffd7, 0x1f266,
 0x176f8, 0x37653, 0x1fa7e, 0x31182, 0x1fac8, 0x2c7e3, 0x370ee, 0x176ec,
 0x176e9, 0x2e4bc, 0x160c5, 0x3765a, 0x3ce9c, 0x17373, 0x176e8, 0x188d4,
 0x176f1, 0x176ef, 0x37659, 0x1bb7c, 0x1ffde, 0x176f2, 0x3118b, 0x2c7d4,
 0x37651, 0x5ce9f, 0x37650, 0x31191, 0x3f4f6, 0x3f4f5, 0x7a06c, 0x1fac1,
 0x5c97b, 0x2c7e0, 0x79d3a, 0x3e7fd, 0x2c7df, 0x3f4f0, 0x7a06d, 0x376c1,
 0x79d3b, 0x00004, 0x00014, 0x00059, 0x000ab, 0x000b8, 0x00177, 0x001f5,
 0x001f2, 0x00315, 0x003fc, 0x005bd, 0x0062d, 0x006e8, 0x007dd, 0x00b04,
 0x007cd, 0x00b1e, 0x00d1e, 0x00f15, 0x00f3b, 0x00f41, 0x01548, 0x018b0,
 0x0173b, 0x01884, 0x01a1c, 0x01bb4, 0x01f25, 0x017b5, 0x0176d, 0x01ef8,
 0x02e73, 0x03107, 0x03125, 0x03105, 0x02e49, 0x03ce8, 0x03ef9, 0x03e5e,
 0x02e72, 0x03471, 0x03fd9, 0x0623f, 0x078a0, 0x06867, 0x05cb3, 0x06272,
 0x068ec, 0x06e9a, 0x079d4, 0x06e98, 0x0b1aa, 0x06e1a, 0x07985, 0x068ee,
 0x06e9b, 0x05c88, 0x0b1ac, 0x07dfa, 0x05d65, 0x07cf0, 0x07cbf, 0x0c475,
 0x160eb, 0x1bb7e, 0x0f7a6, 0x1fedd, 0x160e3, 0x0fffb, 0x0fb8d, 0x0fff9,
 0x0d1c0, 0x0c58c, 0x1a1e9, 0x0bab8, 0x0f5cf, 0x0fff5, 0x376c5, 0x1a1ec,
 0x160ed, 0x1fede, 0x1fac9, 0x1a1eb, 0x1f224, 0x176ee, 0x0fd79, 0x17080,
 0x17387, 0x1bb7a, 0x1ffe9, 0x176f7, 0x17385, 0x17781, 0x2c7d5, 0x17785,
 0x1ffe3, 0x163f5, 0x1fac2, 0x3e7f9, 0x3118d, 0x3fdb1, 0x1ffe2, 0x1f226,
 0x3118a, 0x2c7d9, 0x31190, 0x3118c, 0x3f4f3, 0x1bb7f, 0x1bb72, 0x31184,
 0xb92f4, 0x3e7fb, 0x6e1d9, 0x1faca, 0x62300, 0x3fdb8, 0x3d037, 0x3e7fc,
 0x62301, 0x3f4f2, 0x1f26a, 0x0000e, 0x00063, 0x000f8, 0x001ee, 0x00377,
 0x003f7, 0x006e3, 0x005cc, 0x00b05, 0x00dd2, 0x00fd4, 0x0172e, 0x0172a,
 0x01e23, 0x01f2d, 0x01763, 0x01769, 0x0176c, 0x02e75, 0x03104, 0x02ec1,
 0x03e58, 0x0583f, 0x03f62, 0x03f44, 0x058c5, 0x0623c, 0x05cf4, 0x07bd7,
 0x05d9d, 0x0aad2, 0x05d66, 0x0b1a9, 0x0b078, 0x07cfe, 0x0b918, 0x0c46f,
 0x0b919, 0x0b847, 0x06e1b, 0x0b84b, 0x0aad8, 0x0fd74, 0x172f4, 0x17081,
 0x0f97c, 0x1f273, 0x0f7a0, 0x0fd7c, 0x172f7, 0x0fd7a, 0x1bb77, 0x172fe,
 0x1f270, 0x0fd73, 0x1bb7b, 0x1a1bc, 0x1bb7d, 0x0bbc3, 0x172f6, 0x0baeb,
 0x0fb8f, 0x3f4f4, 0x3fdb4, 0x376c8, 0x3e7fa, 0x1ffd0, 0x62303, 0xb92f5,
 0x1f261, 0x31189, 0x3fdb5, 0x2c7db, 0x376c9, 0x1fad6, 0x1fad1, 0x00015,
 0x000f0, 0x002e0, 0x0058e, 0x005d7, 0x00c4d, 0x00fa1, 0x00bdb, 0x01756,
 0x01f70, 0x02c19, 0x0313c, 0x0370f, 0x03cc0, 0x02ea8, 0x058c6, 0x058c7,
 0x02eb7, 0x058d0, 0x07d18, 0x0aa58, 0x0b848, 0x05d9e, 0x05d6c, 0x0b84c,
 0x0c589, 0x0b901, 0x163f8, 0x0bac9, 0x0b9c5, 0x0f93c, 0x188d8, 0x0bbc7,
 0x160ec, 0x0fd6f, 0x188d9, 0x160ea, 0x0f7a7, 0x0f944, 0x0baab, 0x0dc3a,
 0x188cf, 0x176fb, 0x2c7d8, 0x2c7d7, 0x1bb75, 0x5ce9e, 0x62302, 0x370ed,
 0x176f4, 0x1ffd1, 0x370ef, 0x3f4f8, 0x376c7, 0x1ffe1, 0x376c6, 0x176ff,
 0x6e1d8, 0x176f6, 0x17087, 0x0f5cd, 0x00035, 0x001a0, 0x0058b, 0x00aac,
 0x00b9a, 0x0175f, 0x01e22, 0x01e8c, 0x01fb2, 0x0310b, 0x058d1, 0x0552e,
 0x05c27, 0x0686e, 0x07ca7, 0x0c474, 0x0dc33, 0x07bf2, 0x05de9, 0x07a35,
 0x0baaa, 0x0b9eb, 0x0fb95, 0x0b9b8, 0x17381, 0x1f262, 0x188cd, 0x17088,
 0x172fa, 0x0f7a2, 0x1fad3, 0x0bac0, 0x3765c, 0x1fedf, 0x1f225, 0x1fad4,
 0x2c7da, 0x5ce9d, 0x3e7f8, 0x1e203, 0x188d7, 0x00054, 0x002c0, 0x007a1,
 0x00f78, 0x01b36, 0x01fa3, 0x0313a, 0x03436, 0x0343a, 0x07d1d, 0x07bd8,
 0x05cdf, 0x0b846, 0x0b189, 0x0d9b8, 0x0fff8, 0x0d9be, 0x0c58a, 0x05dea,
 0x0d1d3, 0x160e4, 0x1f26b, 0x188da, 0x1e202, 0x2c7d2, 0x163fe, 0x31193,
 0x17782, 0x376c2, 0x2c7d1, 0x3fdb0, 0x3765d, 0x2c7d0, 0x1fad0, 0x1e201,
 0x188dd, 0x2c7e2, 0x37657, 0x37655, 0x376c4, 0x376c0, 0x176ea, 0x0006f,
 0x003cf, 0x00dd5, 0x01f23, 0x02c61, 0x02ed0, 0x05d54, 0x0552d, 0x07883,
 0x0b1a8, 0x0b91c, 0x0babf, 0x0b902, 0x0f7aa, 0x0f7a5, 0x1a1e8, 0x1ffd6,
 0x0babe, 0x1a1bf, 0x163f3, 0x1ffd8, 0x1fad7, 0x1f275, 0x1ffdc, 0x0007d,
 0x005bc, 0x01549, 0x02a99, 0x03def, 0x06273, 0x079d6, 0x07d1b, 0x0aad3,
 0x0d0fc, 0x2c7dd, 0x188d6, 0x0bac2, 0x2c7e1, 0x1bb76, 0x1a1bd, 0x31186,
 0x0fd78, 0x1a1be, 0x31183, 0x3fdb6, 0x3f4f1, 0x37652, 0x1fad5, 0x3f4f9,
 0x3e7ff, 0x5ce9c, 0x3765b, 0x31188, 0x17372, 0x000bd, 0x0078b, 0x01f21,
 0x03c43, 0x03ded, 0x0aad6, 0x07ec1, 0x0f942, 0x05c86, 0x17089, 0x0babb,
 0x1ffe8, 0x2c7de, 0x1f26e, 0x1fac4, 0x3f4f7, 0x37656, 0x1fa7d, 0x376c3,
 0x3fdb3, 0x3118f, 0x1fac6, 0x000f8, 0x007ed, 0x01efd, 0x03e7a, 0x05c91,
 0x0aad9, 0x0baec, 0x0dc32, 0x0f46e, 0x1e200, 0x176fa, 0x3765e, 0x3fdb7,
 0x2c7d6, 0x3fdb9, 0x37654, 0x37658, 0x3118e, 0x1ffdb, 0x000f6, 0x00c43,
 0x03106, 0x068ef, 0x0b84d, 0x0b188, 0x0bbcc, 0x1f264, 0x1bb69, 0x17386,
 0x1fac0, 0x00171, 0x00f39, 0x03e41, 0x068ed, 0x0d9bc, 0x0f7a1, 0x1bb67,
 0x1ffdd, 0x176f9, 0x001b9, 0x00f7d, 0x03f63, 0x0d0fd, 0x0b9ea, 0x188dc,
 0x1fac3, 0x1a1f2, 0x31192, 0x1ffe4, 0x001f6, 0x01754, 0x06865, 0x0f309,
 0x160e5, 0x176f5, 0x3765f, 0x1facc, 0x001e9, 0x01a1a, 0x06201, 0x0f105,
 0x176f0, 0x002df, 0x01756, 0x05d6d, 0x163fa, 0x176ed, 0x00342, 0x02e40,
 0x0d0ff, 0x17082, 0x003cd, 0x02a98, 0x0fffc, 0x2c7dc, 0x1fa7f, 0x003fe,
 0x03764, 0x0fffd, 0x176fc, 0x1fac5, 0x002f7, 0x02ed1, 0x0fb97, 0x0058a,
 0x02edc, 0x0bbc8, 0x005d4, 0x0623d, 0x160e8, 0x0062e, 0x05830, 0x163f9,
 0x006eb, 0x06205, 0x1f274, 0x007de, 0x062c9, 0x1f265, 0x005c9, 0x05cde,
 0x1ffd3, 0x005d4, 0x07988, 0x007ce, 0x0b849, 0x00b1b, 0x05c89, 0x1fac7,
 0x00b93, 0x05c83, 0x00b9e, 0x0f14f, 0x00c4a, 0x0b9c7, 0x00dd4, 0x0c470,
 0x1f271, 0x00f38, 0x0fb96, 0x176eb, 0x00fa0, 0x163f7, 0x00bb2, 0x0b91b,
 0x00bbe, 0x0f102, 0x00f44, 0x0f946, 0x1facd, 0x00f79, 0x0d9bd, 0x0154d,
 0x0bbc6, 0x00fd2, 0x160e7, 0x0172b, 0x188cb, 0x0175e, 0x0fd76, 0x0175c,
 0x1bb71, 0x0189f, 0x1a1ee, 0x01f24, 0x1a1f6, 0x01ba7, 0x0bbca, 0x01f7d,
 0x0ffff, 0x01f2e, 0x1bb65, 0x01bb5, 0x172f9, 0x01fef, 0x1f26c, 0x01f3e,
 0x0fd77, 0x01762, 0x1bb6e, 0x01ef9, 0x172fc, 0x01fa0, 0x02ab7, 0x02e4a,
 0x1f267, 0x01fb3, 0x1ffda, 0x02e42, 0x03101, 0x17780, 0x0313d, 0x03475,
 0x17784, 0x03126, 0x1facf, 0x03c51, 0x17783, 0x03e40, 0x1ffe5, 0x03663,
 0x1ffe0, 0x03e8f, 0x1f26d, 0x0343c, 0x03cc1, 0x176fd, 0x03e45, 0x02ec0,
 0x03f61, 0x03dee, 0x03fd8, 0x0583e, 0x02e45, 0x03e59, 0x03d02, 0x05ce8,
 0x05568, 0x176fe, 0x02f69, 0x1fad8, 0x058c1, 0x05c83, 0x1ffe6, 0x06271,
 0x06e1c, 0x062c7, 0x068e1, 0x0552f, 0x06864, 0x06866, 0x06e99, 0x05cbc,
 0x07ca5, 0x078a1, 0x05c82, 0x07dcf, 0x0623b, 0x0623e, 0x068e8, 0x07a36,
 0x05d9c, 0x0b077, 0x07cf3, 0x07a34, 0x07ca4, 0x07d19, 0x079d2, 0x07d1c,
 0x07bd9, 0x0b84a, 0x0fb94, 0x0aad5, 0x0dc30, 0x07bf3, 0x0baee, 0x0b07a,
 0x0c472, 0x0b91e, 0x0d9ba, 0x05d9f, 0x0d0fe, 0x0b9c6, 0x05c87, 0x0f14e,
 0x0baed, 0x0b92e, 0x0f103, 0x0b9c4, 0x0fb91, 0x0d9bb, 0x0b1ab, 0x0c58d,
 0x0fffe, 0x0f93b, 0x0f941, 0x0baea, 0x0b91f, 0x0f5cc, 0x0d9bf, 0x0f943,
 0x0f104, 0x1f260, 0x0fb92, 0x0f93f, 0x0f3a6, 0x0bac7, 0x0f7ab, 0x0bac6,
 0x17383, 0x0fd6d, 0x0bae9, 0x0fd6e, 0x1e74f, 0x188ca, 0x1f227, 0x0fb93,
 0x0fb90, 0x0fff7, 0x17085, 0x17083, 0x160e1, 0x17084, 0x0f93e, 0x160e2,
 0x160c6, 0x1a1f1, 0x1bb6f, 0x17384, 0x0fd70, 0x1f263, 0x188d5, 0x173a6,
 0x0f5ce, 0x163f2, 0x0fd71, 0x1ffd2, 0x160c4, 0x1ffd4, 0x2c7d3, 0x1bb74,
};

static const uint8_t coef3_huffbits[1072] = {
  9,  7,  2,  3,  4,  4,  5,  5,
  6,  6,  6,  6,  7,  7,  7,  7,
  7,  8,  8,  8,  8,  8,  8,  8,
  9,  9,  9,  9,  9,  9,  9,  9,
  9, 10, 10, 10, 10, 10, 10, 10,
 10, 10, 10, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11,
 11, 12, 11, 12, 12, 12, 12, 12,
 12, 12, 12, 12, 12, 12, 12, 12,
 12, 12, 12, 12, 12, 13, 13, 13,
 13, 13, 13, 13, 13, 13, 13, 13,
 13, 13, 13, 13, 13, 13, 13, 13,
 13, 14, 13, 14, 14, 13, 14, 13,
 13, 14, 14, 14, 14, 14, 14, 14,
 14, 14, 14, 14, 14, 14, 14, 14,
 14, 14, 14, 14, 14, 14, 14, 15,
 14, 14, 15, 14, 14, 15, 15, 15,
 15, 15, 15, 15, 15, 15, 14, 15,
 15, 15, 15, 15, 15, 15, 15, 15,
 15, 14, 15, 15, 15, 15, 15, 15,
 15, 15, 15, 16, 15, 16, 16, 16,
 16, 15, 15, 16, 16, 16, 16, 16,
 15, 16, 16, 16, 15, 16, 15, 15,
 16, 15, 16, 16, 16, 16, 16, 16,
 16, 16, 16, 16, 16, 16, 16, 16,
 16, 17, 16, 17, 16, 17, 17, 16,
 17, 16, 17, 16, 16, 17, 17, 17,
 16, 17, 16, 16, 17, 16, 17, 16,
 17, 17, 16, 16, 17, 17, 17, 17,
 17, 17, 17, 17, 16, 17, 17, 16,
 17, 17, 17, 17, 17, 17, 17, 17,
 16, 18, 17, 17, 17, 17, 17, 17,
 17, 17, 17, 17, 17, 17, 16, 17,
 17, 17, 17, 17, 17, 17, 17, 17,
 17, 17, 17, 17, 17, 17, 17, 18,
 17, 17, 17, 17, 18, 17, 17, 18,
 19, 17, 17, 17, 18, 17, 17, 17,
 18, 18, 18, 17, 17, 17, 18, 17,
 17, 17, 17, 17, 17, 17, 17, 17,
 18, 18, 18, 18, 18, 18, 18, 18,
 18, 18, 17, 18, 18, 18, 18, 17,
 18, 18, 18, 17, 17, 18, 18, 18,
 18, 19, 18, 18, 19, 19, 20, 18,
 19, 18, 19, 19, 18, 19, 20, 18,
 19,  4,  6,  7,  8,  9,  9,  9,
 10, 10, 10, 11, 11, 11, 11, 12,
 12, 12, 12, 12, 12, 13, 13, 13,
 13, 13, 13, 13, 13, 14, 14, 14,
 14, 14, 14, 14, 14, 14, 14, 14,
 14, 14, 14, 15, 15, 15, 15, 15,
 15, 15, 15, 15, 16, 15, 15, 15,
 15, 16, 16, 15, 16, 16, 15, 16,
 17, 17, 17, 17, 17, 16, 16, 16,
 16, 16, 17, 17, 17, 16, 18, 17,
 17, 17, 18, 17, 17, 18, 17, 17,
 17, 17, 17, 18, 17, 18, 18, 18,
 17, 17, 18, 19, 18, 18, 17, 17,
 18, 18, 18, 18, 19, 17, 17, 18,
 20, 19, 19, 18, 19, 18, 19, 19,
 19, 19, 17,  5,  7,  9, 10, 10,
 11, 11, 12, 12, 12, 13, 13, 13,
 13, 13, 14, 14, 14, 14, 14, 15,
 14, 15, 15, 15, 15, 15, 16, 16,
 16, 16, 16, 16, 16, 16, 16, 16,
 16, 16, 15, 16, 16, 17, 17, 17,
 16, 17, 17, 17, 17, 17, 17, 17,
 17, 17, 17, 17, 17, 17, 17, 16,
 16, 19, 18, 18, 19, 17, 19, 20,
 17, 18, 18, 18, 18, 18, 18,  6,
  8, 10, 11, 12, 12, 12, 13, 13,
 13, 14, 14, 14, 14, 15, 15, 15,
 15, 15, 15, 16, 16, 16, 16, 16,
 16, 17, 17, 17, 16, 16, 17, 17,
 17, 17, 17, 17, 17, 16, 16, 16,
 17, 18, 18, 18, 17, 19, 19, 18,
 18, 17, 18, 19, 18, 17, 18, 18,
 19, 18, 17, 17,  6,  9, 11, 12,
 13, 13, 13, 14, 14, 14, 15, 15,
 15, 15, 15, 16, 16, 16, 16, 16,
 16, 17, 16, 17, 17, 17, 17, 17,
 17, 17, 18, 17, 18, 17, 17, 18,
 18, 19, 19, 17, 17,  7, 10, 12,
 13, 13, 14, 14, 14, 14, 15, 16,
 16, 16, 16, 16, 16, 16, 16, 16,
 16, 17, 17, 17, 17, 18, 17, 18,
 18, 18, 18, 18, 18, 18, 18, 17,
 17, 18, 18, 18, 18, 18, 18,  7,
 10, 12, 13, 14, 15, 15, 15, 15,
 16, 16, 17, 17, 17, 17, 17, 17,
 17, 17, 17, 17, 18, 17, 17,  8,
 11, 13, 14, 15, 15, 15, 15, 16,
 16, 18, 17, 17, 18, 17, 17, 18,
 17, 17, 18, 18, 19, 18, 18, 19,
 19, 19, 18, 18, 18,  8, 11, 13,
 14, 15, 16, 16, 16, 16, 17, 17,
 17, 18, 17, 18, 19, 18, 18, 18,
 18, 18, 18,  8, 12, 14, 15, 15,
 16, 16, 16, 17, 17, 18, 18, 18,
 18, 18, 18, 18, 18, 17,  9, 12,
 14, 15, 16, 16, 17, 17, 17, 17,
 18,  9, 12, 14, 15, 16, 17, 17,
 17, 18,  9, 13, 15, 16, 17, 17,
 18, 17, 18, 17,  9, 13, 15, 16,
 17, 18, 18, 18, 10, 13, 15, 16,
 18, 10, 14, 16, 17, 18, 10, 14,
 16, 17, 10, 14, 16, 18, 18, 10,
 14, 16, 18, 18, 11, 15, 16, 11,
 15, 17, 11, 15, 17, 11, 15, 17,
 11, 15, 17, 11, 15, 17, 12, 16,
 17, 12, 15, 12, 16, 12, 16, 18,
 12, 16, 12, 16, 12, 16, 12, 16,
 17, 12, 16, 18, 12, 17, 13, 16,
 13, 16, 13, 16, 18, 13, 16, 13,
 17, 13, 17, 13, 17, 13, 17, 13,
 17, 13, 17, 13, 17, 13, 17, 13,
 16, 13, 17, 13, 17, 13, 17, 14,
 17, 14, 17, 14, 17, 14, 14, 14,
 17, 14, 17, 14, 14, 18, 14, 14,
 18, 14, 18, 14, 18, 14, 17, 14,
 17, 14, 17, 14, 14, 18, 14, 15,
 15, 15, 14, 15, 15, 14, 15, 15,
 15, 18, 15, 18, 15, 15, 17, 15,
 15, 15, 15, 15, 15, 15, 15, 15,
 15, 15, 16, 15, 15, 15, 15, 16,
 16, 16, 16, 16, 15, 15, 15, 15,
 16, 16, 16, 16, 16, 16, 16, 16,
 16, 16, 16, 16, 16, 16, 16, 16,
 16, 16, 16, 16, 16, 16, 16, 16,
 16, 16, 16, 16, 16, 17, 16, 16,
 16, 17, 16, 16, 16, 17, 17, 17,
 17, 17, 16, 17, 17, 17, 17, 16,
 16, 16, 17, 17, 17, 17, 16, 17,
 17, 17, 17, 17, 17, 17, 17, 17,
 17, 17, 17, 17, 17, 17, 18, 17,
};

static const uint32_t coef4_huffcodes[476] = {
 0x00f01, 0x0001e, 0x00000, 0x00004, 0x00006, 0x0000d, 0x0000a, 0x00017,
 0x0001d, 0x00017, 0x0002c, 0x00031, 0x00039, 0x0003e, 0x00039, 0x0005a,
 0x00066, 0x00070, 0x0007b, 0x00070, 0x00077, 0x000af, 0x000c9, 0x000f2,
 0x000f4, 0x000b2, 0x000e3, 0x0015b, 0x0015d, 0x00181, 0x0019d, 0x001e3,
 0x001c5, 0x002b5, 0x002db, 0x00338, 0x003c3, 0x003cc, 0x003f0, 0x002cd,
 0x003fa, 0x003a1, 0x005b4, 0x00657, 0x007ab, 0x0074d, 0x0074c, 0x00ac1,
 0x00ac5, 0x0076b, 0x00ca8, 0x00f04, 0x00f00, 0x00fe3, 0x00f3c, 0x00f10,
 0x00f39, 0x00fe6, 0x00e26, 0x00e90, 0x016c5, 0x01827, 0x01954, 0x015c5,
 0x01958, 0x01f8a, 0x01c4a, 0x02b0f, 0x02b41, 0x02b0e, 0x033c6, 0x03050,
 0x01c4f, 0x02d88, 0x0305c, 0x03c18, 0x02b4f, 0x02cc2, 0x03a47, 0x05680,
 0x0569d, 0x06442, 0x06443, 0x06446, 0x0656e, 0x06444, 0x07120, 0x0748a,
 0x0c1ba, 0x07e22, 0x07aa6, 0x07f25, 0x07aa7, 0x07e20, 0x0c11b, 0x0c118,
 0x07aa5, 0x0ad0a, 0x0f389, 0x19ebb, 0x0caad, 0x0fe42, 0x0fe40, 0x16c34,
 0x2b4e5, 0x33d65, 0x16c30, 0x1e7ae, 0x1e25c, 0x18370, 0x1e703, 0x19eba,
 0x16c37, 0x0e234, 0x16c6e, 0x00004, 0x0002a, 0x00061, 0x00075, 0x000cb,
 0x000ff, 0x00190, 0x001eb, 0x001d1, 0x002b9, 0x00307, 0x00339, 0x0033f,
 0x003fb, 0x003b4, 0x0060c, 0x00679, 0x00645, 0x0067d, 0x0078a, 0x007e3,
 0x00749, 0x00ac4, 0x00ad2, 0x00ae3, 0x00c10, 0x00c16, 0x00ad1, 0x00cf4,
 0x00fe2, 0x01586, 0x00e9d, 0x019f1, 0x01664, 0x01e26, 0x01d38, 0x02b4d,
 0x033c5, 0x01fc2, 0x01fc3, 0x01d28, 0x03c1d, 0x0598e, 0x0f094, 0x07aa4,
 0x0ad38, 0x0ac0c, 0x0c11a, 0x079ea, 0x0c881, 0x0fe44, 0x0b635, 0x0ac0d,
 0x0b61e, 0x05987, 0x07121, 0x0f382, 0x0f387, 0x0e237, 0x0fe47, 0x0f383,
 0x0f091, 0x0f385, 0x0e233, 0x182ee, 0x19eb8, 0x1663e, 0x0f093, 0x00014,
 0x00058, 0x00159, 0x00167, 0x00300, 0x003d4, 0x005b5, 0x0079d, 0x0076a,
 0x00b67, 0x00b60, 0x00f05, 0x00cf0, 0x00f17, 0x00e95, 0x01822, 0x01913,
 0x016c2, 0x0182f, 0x01959, 0x01fcb, 0x01e27, 0x01c40, 0x033c7, 0x01e7b,
 0x01c49, 0x02d89, 0x01e23, 0x01660, 0x03f12, 0x02cc6, 0x033e1, 0x05b34,
 0x0609a, 0x06569, 0x07488, 0x07e21, 0x0cf5f, 0x0712c, 0x0389d, 0x067cf,
 0x07f28, 0x1663f, 0x33d67, 0x1663d, 0x1e25d, 0x3c1ab, 0x15c44, 0x16c36,
 0x0001f, 0x000ec, 0x00323, 0x005b2, 0x0079f, 0x00ac2, 0x00f16, 0x00e9e,
 0x01956, 0x01e0f, 0x019ea, 0x01666, 0x02b89, 0x02b02, 0x02d8c, 0x03c1b,
 0x03c19, 0x032b5, 0x03f9c, 0x02ccf, 0x03897, 0x05b35, 0x0ad02, 0x07f29,
 0x06441, 0x03884, 0x07888, 0x0784e, 0x06568, 0x0c1bb, 0x05986, 0x067cc,
 0x0fe49, 0x0fe48, 0x0c1bc, 0x0fe41, 0x18371, 0x1663c, 0x0e231, 0x0711e,
 0x0ad09, 0x0f092, 0x0002d, 0x001db, 0x00781, 0x00c1a, 0x00f55, 0x01580,
 0x01ea8, 0x02d9b, 0x032af, 0x03f16, 0x03c1c, 0x07834, 0x03c45, 0x0389c,
 0x067ce, 0x06445, 0x0c1b9, 0x07889, 0x07f3a, 0x0784f, 0x07f2b, 0x0ad0b,
 0x0f090, 0x0c11d, 0x0e94e, 0x0711f, 0x0e9f1, 0x0f38e, 0x079e9, 0x0ad03,
 0x0f09b, 0x0caae, 0x0fe46, 0x2b4e6, 0x0e9f0, 0x19eb6, 0x67ac1, 0x67ac0,
 0x33d66, 0x0f388, 0x00071, 0x003a0, 0x00ca9, 0x01829, 0x01d39, 0x02b43,
 0x02cc4, 0x06554, 0x0f09a, 0x0b61f, 0x067cd, 0x0711c, 0x0b636, 0x07f2a,
 0x0b634, 0x0c11f, 0x0cf5e, 0x0b61d, 0x0f06b, 0x0caab, 0x0c1be, 0x0e94c,
 0x0f099, 0x182ed, 0x0e94f, 0x0c119, 0x0e232, 0x2b4e4, 0x0f38a, 0x19eb4,
 0x1e25f, 0x0e94d, 0x000b7, 0x00785, 0x016cc, 0x03051, 0x033c4, 0x0656f,
 0x03891, 0x0711d, 0x0caaf, 0x0f097, 0x07489, 0x0f098, 0x0c880, 0x0caaa,
 0x0f386, 0x19eb7, 0x16c6f, 0x0f384, 0x182e8, 0x182e9, 0x0e230, 0x1e700,
 0x33d62, 0x33d63, 0x33d64, 0x16c33, 0x0e216, 0x000fd, 0x00c15, 0x01665,
 0x03c4a, 0x07f3b, 0x07896, 0x0c11c, 0x0e215, 0x16c32, 0x0f38b, 0x0f38d,
 0x182ea, 0x1e701, 0x712df, 0x15c46, 0x00194, 0x00fe0, 0x03f13, 0x0748b,
 0x0f096, 0x0cf80, 0x1e25e, 0xe25bd, 0x33d61, 0x16c31, 0x001f9, 0x01912,
 0x05710, 0x0f3d0, 0x0c1bf, 0x00301, 0x01e24, 0x0ad08, 0x003cd, 0x01c41,
 0x0c1bd, 0x00563, 0x03a52, 0x0f3d1, 0x00570, 0x02cce, 0x0e217, 0x0067b,
 0x0655d, 0x0074b, 0x06447, 0x00c12, 0x074fb, 0x00f08, 0x0b61c, 0x00e22,
 0x0fe43, 0x016c7, 0x01836, 0x019f2, 0x01c43, 0x01d3f, 0x01fcf, 0x02b4c,
 0x0304c, 0x032b6, 0x03a46, 0x05607, 0x03f17, 0x02cc5, 0x0609b, 0x0655c,
 0x07e23, 0x067c1, 0x07f26, 0x07f27, 0x0f095, 0x0e9f3, 0x0cf81, 0x0c11e,
 0x0caac, 0x0f38f, 0x0e9f2, 0x074fa, 0x0e236, 0x0fe45, 0x1c428, 0x0e235,
 0x182ef, 0x19eb5, 0x0f3d6, 0x182ec, 0x16c35, 0x0f38c, 0x2b4e7, 0x15c47,
 0xe25bc, 0x1e702, 0x1c4b6, 0x0e25a, 0x3c1aa, 0x15c45, 0x1c429, 0x19eb9,
 0x1e7af, 0x182eb, 0x1e0d4, 0x3896e,
};

static const uint8_t coef4_huffbits[476] = {
 12,  6,  2,  3,  4,  4,  5,  5,
  5,  6,  6,  6,  6,  6,  7,  7,
  7,  7,  7,  8,  8,  8,  8,  8,
  8,  9,  9,  9,  9,  9,  9,  9,
 10, 10, 10, 10, 10, 10, 10, 11,
 10, 11, 11, 11, 11, 12, 12, 12,
 12, 12, 12, 12, 12, 12, 12, 12,
 12, 12, 13, 13, 13, 13, 13, 13,
 13, 13, 14, 14, 14, 14, 14, 14,
 14, 14, 14, 14, 14, 15, 15, 15,
 15, 15, 15, 15, 15, 15, 16, 16,
 16, 15, 15, 15, 15, 15, 16, 16,
 15, 16, 16, 17, 16, 16, 16, 17,
 18, 18, 17, 17, 17, 17, 17, 17,
 17, 17, 17,  4,  6,  7,  8,  8,
  8,  9,  9, 10, 10, 10, 10, 10,
 10, 11, 11, 11, 11, 11, 11, 11,
 12, 12, 12, 12, 12, 12, 12, 12,
 12, 13, 13, 13, 14, 13, 14, 14,
 14, 13, 13, 14, 14, 16, 16, 15,
 16, 16, 16, 15, 16, 16, 16, 16,
 16, 16, 16, 16, 16, 17, 16, 16,
 16, 16, 17, 17, 17, 18, 16,  5,
  8,  9, 10, 10, 10, 11, 11, 12,
 12, 12, 12, 12, 12, 13, 13, 13,
 13, 13, 13, 13, 13, 14, 14, 13,
 14, 14, 13, 14, 14, 15, 14, 15,
 15, 15, 16, 15, 16, 16, 15, 15,
 15, 18, 18, 18, 17, 18, 17, 17,
  6,  9, 10, 11, 11, 12, 12, 13,
 13, 13, 13, 14, 14, 14, 14, 14,
 14, 14, 14, 15, 15, 15, 16, 15,
 15, 15, 15, 15, 15, 16, 16, 15,
 16, 16, 16, 16, 17, 18, 17, 16,
 16, 16,  7, 10, 11, 12, 12, 13,
 13, 14, 14, 14, 14, 15, 14, 15,
 15, 15, 16, 15, 15, 15, 15, 16,
 16, 16, 17, 16, 17, 16, 15, 16,
 16, 16, 16, 18, 17, 17, 19, 19,
 18, 16,  7, 11, 12, 13, 14, 14,
 15, 15, 16, 16, 15, 16, 16, 15,
 16, 16, 16, 16, 16, 16, 16, 17,
 16, 17, 17, 16, 17, 18, 16, 17,
 17, 17,  8, 11, 13, 14, 14, 15,
 15, 16, 16, 16, 16, 16, 16, 16,
 16, 17, 17, 16, 17, 17, 17, 17,
 18, 18, 18, 17, 17,  8, 12, 14,
 14, 15, 15, 16, 17, 17, 16, 16,
 17, 17, 20, 17,  9, 12, 14, 16,
 16, 16, 17, 21, 18, 17,  9, 13,
 15, 16, 16, 10, 13, 16, 10, 14,
 16, 11, 15, 16, 11, 15, 17, 11,
 15, 12, 15, 12, 16, 12, 16, 13,
 16, 13, 13, 13, 14, 14, 13, 14,
 14, 14, 15, 15, 14, 15, 15, 15,
 15, 15, 15, 15, 16, 17, 16, 16,
 16, 16, 17, 16, 17, 16, 18, 17,
 17, 17, 16, 17, 17, 16, 18, 17,
 21, 17, 18, 17, 18, 17, 18, 17,
 17, 17, 17, 19,
};

static const uint32_t coef5_huffcodes[435] = {
 0x00347, 0x0000b, 0x00001, 0x00001, 0x0000c, 0x00004, 0x00010, 0x00015,
 0x0001f, 0x0000b, 0x00023, 0x00026, 0x00029, 0x00035, 0x00037, 0x00001,
 0x00015, 0x0001a, 0x0001d, 0x0001c, 0x0001e, 0x0004e, 0x00049, 0x00051,
 0x00078, 0x00004, 0x00000, 0x00008, 0x0000d, 0x0007b, 0x00005, 0x00032,
 0x00095, 0x00091, 0x00096, 0x000a1, 0x000d9, 0x00003, 0x00019, 0x00061,
 0x00066, 0x00060, 0x00017, 0x0000e, 0x00063, 0x001a0, 0x001b7, 0x001e6,
 0x001e7, 0x001b6, 0x00018, 0x001e8, 0x00038, 0x00031, 0x00005, 0x0003d,
 0x00027, 0x001ea, 0x0001a, 0x000c5, 0x000f9, 0x000ff, 0x000db, 0x00250,
 0x000fc, 0x0025c, 0x00008, 0x00075, 0x003d7, 0x003d3, 0x001b0, 0x0007c,
 0x003ca, 0x00036, 0x00189, 0x004a6, 0x004a2, 0x004fb, 0x000c0, 0x0007f,
 0x0009a, 0x00311, 0x0006e, 0x0009b, 0x0068c, 0x006c0, 0x00484, 0x00012,
 0x000c3, 0x0094f, 0x00979, 0x009f9, 0x00d09, 0x00da6, 0x00da8, 0x00901,
 0x000c1, 0x00373, 0x00d08, 0x009fa, 0x00d8b, 0x00d85, 0x00d86, 0x000df,
 0x006e2, 0x000ce, 0x00f24, 0x009fe, 0x001f7, 0x007c1, 0x000cf, 0x009fc,
 0x009ff, 0x00d89, 0x00da9, 0x009fd, 0x001f8, 0x01a36, 0x0128c, 0x0129d,
 0x01a37, 0x00196, 0x003ea, 0x00f8b, 0x00d93, 0x01e45, 0x01e58, 0x01e4b,
 0x01e59, 0x013f1, 0x00309, 0x00265, 0x00308, 0x0243a, 0x027e1, 0x00f89,
 0x00324, 0x03cbc, 0x03c86, 0x03695, 0x0243c, 0x0243b, 0x0243e, 0x01e4a,
 0x003a5, 0x03468, 0x03428, 0x03c84, 0x027e0, 0x025e2, 0x01880, 0x00197,
 0x00325, 0x03cb7, 0x0791e, 0x007ec, 0x06c75, 0x004c8, 0x04bc7, 0x004c6,
 0x00983, 0x0481e, 0x01b53, 0x0251b, 0x01b58, 0x00984, 0x04fa8, 0x03cbb,
 0x00f8a, 0x00322, 0x0346a, 0x0243d, 0x00326, 0x03469, 0x0481f, 0x0481d,
 0x00746, 0x09032, 0x01b50, 0x01d13, 0x0d8e4, 0x0481b, 0x06c74, 0x0796b,
 0x07969, 0x00985, 0x0d8e3, 0x00986, 0x00fa2, 0x01301, 0x06c7c, 0x00987,
 0x03cb8, 0x0f4af, 0x00e88, 0x1b1c0, 0x00fce, 0x033eb, 0x03f6a, 0x03f69,
 0x00fcf, 0x0791f, 0x004c9, 0x04871, 0x00fcd, 0x00982, 0x00fcc, 0x00fa3,
 0x01d12, 0x0796c, 0x01b47, 0x00321, 0x0796a, 0x0d8e2, 0x04872, 0x04873,
 0x0000e, 0x00014, 0x0000a, 0x000a0, 0x00012, 0x0007d, 0x001a2, 0x0003b,
 0x0025f, 0x000dd, 0x0027c, 0x00343, 0x00368, 0x0036b, 0x0003e, 0x001fa,
 0x00485, 0x001b3, 0x0007f, 0x001b1, 0x0019e, 0x004ba, 0x007ad, 0x00339,
 0x00066, 0x007a4, 0x00793, 0x006c6, 0x0007e, 0x000f1, 0x00372, 0x009fb,
 0x00d83, 0x00d8a, 0x00947, 0x009f4, 0x001d0, 0x01b09, 0x01b4b, 0x007ec,
 0x003e1, 0x000ca, 0x003ec, 0x02539, 0x04fa9, 0x01b57, 0x03429, 0x03d2a,
 0x00d97, 0x003a7, 0x00dc0, 0x00d96, 0x00dc1, 0x007eb, 0x03cba, 0x00c43,
 0x00c41, 0x01b52, 0x007ef, 0x00323, 0x03cb9, 0x03c83, 0x007d0, 0x007ed,
 0x06c7f, 0x09033, 0x03f6c, 0x36383, 0x1e95d, 0x06c78, 0x00747, 0x01b51,
 0x00022, 0x00016, 0x00039, 0x00252, 0x00079, 0x00486, 0x00338, 0x00369,
 0x00d88, 0x00026, 0x00d87, 0x00f4b, 0x00d82, 0x00027, 0x001e1, 0x01a15,
 0x007c7, 0x012f0, 0x001e0, 0x006d0, 0x01a16, 0x01e44, 0x01e5f, 0x03690,
 0x00d90, 0x00c42, 0x00daf, 0x00d92, 0x00f80, 0x00cfb, 0x0342f, 0x0487f,
 0x01b46, 0x07968, 0x00d95, 0x00d91, 0x01b55, 0x03f68, 0x04bc6, 0x03cbd,
 0x00f81, 0x00320, 0x00069, 0x000fe, 0x006d5, 0x0033f, 0x000de, 0x007c6,
 0x01e40, 0x00d94, 0x00f88, 0x03c8e, 0x03694, 0x00dae, 0x00dad, 0x00267,
 0x003a6, 0x00327, 0x0487e, 0x007ee, 0x00749, 0x004c7, 0x03692, 0x01b56,
 0x00fd1, 0x07a56, 0x06c77, 0x09031, 0x00748, 0x06c7a, 0x0796d, 0x033ea,
 0x06c76, 0x00fd0, 0x36382, 0x1e417, 0x00745, 0x04faf, 0x0d8e1, 0x03f6b,
 0x1e95c, 0x04fad, 0x0009e, 0x004bd, 0x0067c, 0x01b08, 0x003eb, 0x01b45,
 0x03691, 0x0d8e5, 0x07904, 0x00981, 0x007ea, 0x019f4, 0x06c7d, 0x04fab,
 0x04fac, 0x06c7e, 0x01300, 0x06c7b, 0x0006f, 0x003f7, 0x03c85, 0x004c4,
 0x0001e, 0x006e1, 0x03693, 0x01b44, 0x00241, 0x01e46, 0x0019d, 0x00266,
 0x004bb, 0x02538, 0x007ac, 0x01b54, 0x00902, 0x04870, 0x00da7, 0x00900,
 0x00185, 0x06c79, 0x006e3, 0x003e9, 0x01e94, 0x003ed, 0x003f2, 0x0342e,
 0x0346b, 0x0251a, 0x004c5, 0x01881, 0x0481c, 0x01b59, 0x03c87, 0x04fae,
 0x007e9, 0x03f6d, 0x0f20a, 0x09030, 0x04faa, 0x0d8e6, 0x03f6f, 0x0481a,
 0x03f6e, 0x1e416, 0x0d8e7,
};

static const uint8_t coef5_huffbits[435] = {
 10,  4,  2,  4,  4,  5,  5,  5,
  5,  6,  6,  6,  6,  6,  6,  7,
  7,  7,  7,  7,  7,  7,  7,  7,
  7,  8,  8,  8,  8,  7,  8,  8,
  8,  8,  8,  8,  8,  9,  9,  9,
  9,  9,  9,  9,  9,  9,  9,  9,
  9,  9, 10,  9, 10, 10, 10, 10,
 10,  9, 10, 10, 10, 10, 10, 10,
 10, 10, 11, 11, 10, 10, 11, 11,
 10, 11, 11, 11, 11, 11, 12, 12,
 12, 12, 12, 12, 11, 11, 11, 12,
 12, 12, 12, 12, 12, 12, 12, 12,
 12, 12, 12, 12, 12, 12, 12, 13,
 13, 13, 12, 12, 13, 13, 13, 12,
 12, 12, 12, 12, 13, 13, 13, 13,
 13, 14, 14, 14, 14, 13, 13, 13,
 13, 13, 14, 14, 14, 14, 14, 14,
 15, 14, 14, 14, 14, 14, 14, 13,
 14, 14, 14, 14, 14, 14, 15, 14,
 15, 14, 15, 15, 15, 15, 15, 15,
 16, 15, 15, 14, 15, 16, 15, 14,
 14, 15, 14, 14, 15, 14, 15, 15,
 15, 16, 15, 17, 16, 15, 15, 15,
 15, 16, 16, 16, 16, 17, 15, 16,
 14, 16, 16, 17, 16, 16, 16, 16,
 16, 15, 15, 15, 16, 16, 16, 16,
 17, 15, 15, 15, 15, 16, 15, 15,
  4,  7,  8,  8,  9,  9,  9, 10,
 10, 10, 10, 10, 10, 10, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 12,
 12, 11, 11, 11, 12, 12, 12, 12,
 12, 12, 12, 12, 13, 13, 13, 13,
 12, 13, 14, 14, 15, 15, 14, 14,
 14, 14, 14, 14, 14, 15, 14, 14,
 14, 15, 15, 15, 14, 14, 15, 15,
 15, 16, 16, 18, 17, 15, 15, 15,
  6,  9, 10, 10, 11, 11, 12, 12,
 12, 13, 12, 12, 12, 13, 13, 13,
 13, 13, 13, 13, 13, 13, 13, 14,
 14, 14, 14, 14, 14, 14, 14, 15,
 15, 15, 14, 14, 15, 16, 15, 14,
 14, 15,  7, 10, 11, 12, 13, 13,
 13, 14, 14, 14, 14, 14, 14, 14,
 14, 15, 15, 15, 15, 15, 14, 15,
 16, 15, 15, 16, 15, 15, 15, 16,
 15, 16, 18, 17, 15, 15, 16, 16,
 17, 15,  8, 11, 13, 13, 14, 15,
 14, 16, 15, 16, 15, 15, 15, 15,
 15, 15, 17, 15,  9, 12, 14, 15,
 10, 13, 14, 15, 10, 13, 11, 14,
 11, 14, 11, 15, 12, 15, 12, 12,
 13, 15, 13, 14, 13, 14, 14, 14,
 14, 14, 15, 15, 15, 15, 14, 15,
 15, 16, 16, 16, 15, 16, 16, 15,
 16, 17, 16,
};

static const uint16_t levels0[60] = {
317, 92, 62, 60, 19, 17, 10,  7,
  6,  5,  5,  3,  3,  3,  2,  2,
  2,  2,  2,  2,  2,  1,  2,  2,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,
};

static const uint16_t levels1[40] = {
311, 91, 61, 28, 10,  6,  5,  2,
  2,  2,  2,  2,  2,  2,  2,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
};

static const uint16_t levels2[340] = {
181,110, 78, 63, 61, 62, 60, 61,
 33, 41, 41, 19, 17, 19, 12, 11,
  9, 11, 10,  6,  8,  7,  6,  4,
  5,  5,  4,  4,  3,  4,  3,  5,
  3,  4,  3,  3,  3,  3,  3,  3,
  2,  2,  4,  2,  3,  2,  3,  3,
  2,  2,  2,  2,  2,  2,  2,  2,
  3,  2,  2,  2,  2,  2,  2,  2,
  2,  2,  2,  1,  2,  1,  2,  2,
  2,  2,  1,  2,  1,  1,  1,  2,
  2,  1,  2,  1,  2,  2,  2,  2,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,
};

static const uint16_t levels3[180] = {
351,122, 76, 61, 41, 42, 24, 30,
 22, 19, 11,  9, 10,  8,  5,  5,
  4,  5,  5,  3,  3,  3,  3,  3,
  3,  3,  2,  2,  3,  2,  2,  2,
  3,  3,  2,  2,  2,  3,  2,  2,
  2,  2,  2,  2,  2,  2,  2,  2,
  2,  2,  2,  2,  2,  2,  1,  1,
  2,  2,  1,  2,  1,  2,  2,  2,
  2,  2,  2,  1,  2,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  2,
  2,  1,  2,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,
};

static const uint16_t levels4[70] = {
113, 68, 49, 42, 40, 32, 27, 15,
 10,  5,  3,  3,  3,  3,  2,  2,
  2,  2,  2,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,
};

static const uint16_t levels5[40] = {
214, 72, 42, 40, 18,  4,  4,  2,
  2,  2,  2,  2,  1,  1,  2,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,
};
    
static const CoefVLCTable coef_vlcs[6] = {
    { 
        sizeof(coef0_huffbits), coef0_huffcodes, coef0_huffbits, levels0,
    },
    { 
        sizeof(coef1_huffbits), coef1_huffcodes, coef1_huffbits, levels1,
    },
    { 
        sizeof(coef2_huffbits), coef2_huffcodes, coef2_huffbits, levels2,
    },
    { 
        sizeof(coef3_huffbits), coef3_huffcodes, coef3_huffbits, levels3,
    },
    { 
        sizeof(coef4_huffbits), coef4_huffcodes, coef4_huffbits, levels4,
    },
    { 
        sizeof(coef5_huffbits), coef5_huffcodes, coef5_huffbits, levels5,
    },
};


typedef struct AVInputStream
{
	int index;
	AVStream *st;
	int discard; 			/* true if stream data should be discarded */
	int64_t sample_index;  	/* current sample */
	int64_t   	start;		/* time when read started */
	unsigned long frame; 	/* current frame */
	AVFrac		next_pts;   /* synthetic pts for cases where pkt.pts == 0 */
	int64_t   	pts;		/* current pts */
} AVInputStream;

typedef struct AVInputFile
{
	int eof_reached;		/* true if eof reached */
	int buffer_size;		/* current total buffer size */
	int buffer_size_max;	/* buffer size at which we consider we can stop buffering */
} AVInputFile;

typedef struct AVW
{
	AVFormatContext *is;
	AVCodecContext *icodec;
	AVInputStream *ist, **ist_table;
	AVInputFile *file_table;
	AVFormatContext *stream_no_data;
} AVW;

char g_cstr[512];


/* don't seek on wma
int l_wmaseek = -1;
void wmaseek(int pos, int siz)
{
	if(l_wmaseek >= 0 || pos < 0 || siz <= 0)
		return;

	l_wmaseek = ((float)pos / (float)siz) * (float)w.is->file_size;	

} // wmaseek
*/

void maPlayWma()
{
	if(!g_state.inpath)
	{
		maLog("maPlayWma error null filename");
		return;
	}
	
	// format of this decoder output
#ifdef WORDS_BIGENDIAN
	g_state.inwordsbigendian = 1;
#else
	g_state.inwordsbigendian = 0;
#endif
	g_state.ininterleaved = 1;
	
	wmamain(&g_state);

	FlushFifo();
	
} // maPlayWma


signed long ThreadPlayWma(void *parameter)
{
#ifdef MAUNUSED
 (void)parameter;
#endif

	g_waitaudiodone = 1;

	maPlayWma();
		
	g_state.stop |= STOP_PLAYDECODE;
	
	// wait for output owned by this thread to finish
	while(!g_stopB && g_waitaudiodone)
		maSleep(100);

	return 0;

} // ThreadPlayWma


// ffmpeg:

// ffmpeg.c : stripped down for WMA play

static AVFormatContext *l_input_file;
static AVInputFormat *l_file_iformat;

#define MAX_AUDIO_PACKET_SIZE (128 * 1024)
#define CODECIDCAST (enum CodecID)
#define CODECTYPECAST (enum CodecType)
 
static int av_encode_input_open(AVFormatContext *input_file, MASTATE *pstate, AVW *pw)
{
	pw->file_table = (AVInputFile*) av_mallocz(sizeof(AVInputFile));
	if(!pw->file_table)
		return -11; //goto fail;

	/* input stream init */
		pw->is = input_file;

	pw->ist_table = (AVInputStream**)av_mallocz(sizeof(AVInputStream *));
	if(!pw->ist_table)
		return -11; //goto fail;
	
	pw->ist = (AVInputStream*)av_mallocz(sizeof(AVInputStream));
	if(!pw->ist)
		return -11; //goto fail;
	pw->ist_table[0] = pw->ist;

	pw->is = input_file;
	pw->ist = pw->ist_table[0];
	pw->ist->st = pw->is->stream;
	pw->ist->index = 0;
	pw->ist->discard = 1; /* the stream pw->is discarded by default (changed later) */
	
	if(pw->is->duration != (int64_t)AV_NOPTS_VALUE)
		pstate->duration = pw->is->duration / 1000;
	else
		pstate->duration = 0; // unknown
		
	//pstate->size = pw->is->file_size;

	return 0;
} // av_encode_input_open

static int av_encode_output_open(AVFormatContext *input_file, AVW *pw)
{
	int n;
	AVCodec *codec;

	n = 0;
	
	/* open each decoder */
		pw->ist = pw->ist_table[0];
			
			codec = avcodec_find_decoder(pw->ist->st->codec.codec_id);
			if(!codec)
			{
				sprintf(g_cstr, "maPlayWma av_encode_output_open unsupported codec (id=%d) for input stream #%d.%d", 
						pw->ist->st->codec.codec_id, 0, pw->ist->index);
				maLog(g_cstr);
				return 1;
			}
			if(avcodec_open(&pw->ist->st->codec, codec) < 0)
			{
				sprintf(g_cstr, "maPlayWma av_encode_output_open error while opening codec for input stream #%d.%d", 
						0, pw->ist->index);
				maLog(g_cstr);
				return 1;
			}

	/* init pts */
		pw->ist = pw->ist_table[0];
		pw->is = input_file;
		pw->ist->pts = 0;
		av_frac_init(&pw->ist->next_pts, 0, 0, (uint64_t)pw->is->pts_num * pw->ist->st->codec.sample_rate);
	
	/* compute buffer size max (should use a complete heuristic) */
	pw->file_table[0].buffer_size_max = 2048;

	pw->stream_no_data = 0;
	
	return 0;
	
} // av_encode_output_open

// The following code is the main loop of the file converter
static int av_encode(AVFormatContext *input_file, MASTATE *pstate, AVW *pw)
{
	int ret;
	
	pw->is = input_file;

	for(; !pstate->stop;)
	{				
		AVPacket pkt;
		uint8_t *ptr;
		int len;
		uint8_t *data_buf;
		int data_size;
		short samples[AVCODEC_MAX_AUDIO_FRAME_SIZE / 2];

		ThreadYield();
		
		/* don't seek on wma
		if(l_wmaseek >= 0)
		{
			if(l_wmaseek <= pw->is->file_size)
			{
				url_fseek(&pw->is->pb, l_wmaseek, SEEK_SET, pstate->nomem);
				// pstate->position = ((double)pw->is->pb.pos / (double)pw->is->file_size) * pstate->duration;
			}
			l_wmaseek = -1;
		}
		*/

	redo:

		/* read a packet from the stream and output it */

		if(av_read_packet(pw->is, &pkt, pstate->nomem) < 0)
		{
			pw->file_table[0].eof_reached = 1;
			break;
		}
		if(!pkt.size)
		{
			pw->stream_no_data = pw->is;
		}
		else
		{
			pw->stream_no_data = 0;
		}

		len = pkt.size;
		ptr = pkt.data;
		
		while(len > 0)
		{
			pstate->position = LROUND(((double)pw->is->pb.pos / (double)pw->is->file_size) * pstate->duration);
			
			/* decode the packet if needed */
			data_buf = NULL; /* fail safe */
			data_size = 0;

			/* XXX: could avoid copy if PCM 16 bits with same endianness as CPU */
			ret = avcodec_decode_audio(&pw->ist->st->codec, samples, &data_size, ptr, len);
			if(ret < 0)
			{
					//sprintf(g_cstr, "Error while decoding stream");
					//maLog(g_cstr);
					av_free_packet(&pkt);
					goto redo;
			}

			data_buf = (uint8_t *)samples;
			av_frac_add(&pw->ist->next_pts,
					(uint64_t)pw->is->pts_den * data_size / (2 * pw->ist->st->codec.channels));

			ptr += ret;
			len -= ret;

			maOutputWrite((char *)data_buf, data_size, pstate);
		}

		av_free_packet(&pkt);
				
	} // for

	/* close each decoder */
	pw->ist = pw->ist_table[0];
	avcodec_close(&pw->ist->st->codec);

	/* finished ! */
	
	ret = 0;

	av_free(pw->file_table);

	if(pw->ist_table)
	{
		pw->ist = pw->ist_table[0];
		av_free(pw->ist);
		av_free(pw->ist_table);
	}

	return ret;
	
} // av_encode


int opt_input_file(const char *filename, int nomem,
				AVFormatContext **pinput_file, AVInputFormat **pfile_iformat,
				int *paudio_channels, int *paudio_sample_rate)
{
	AVFormatContext *ic;
	AVFormatParameters params, *ap = &params;
	int err, ret;
	AVCodecContext *enc;

	/* get default parameters */
	memset(ap, 0, sizeof(*ap));
	ap->sample_rate = *paudio_sample_rate;
	ap->channels = *paudio_channels;

	/* open the input file with generic libav function */
	err = av_open_input_file(&ic, filename, l_file_iformat, 0, ap, nomem);
	if(err < 0)
	{
		if(err == AVERROR_NOFMT)
			sprintf(g_cstr, "maPlayWma av_open_input_file bad wma format '%s'", filename);
		else
			sprintf(g_cstr, "maPlayWma av_open_input_file err=%d '%s'", err, filename);
		maLog(g_cstr);
		return 1;
	}
	
	/* If not enough info to get the stream parameters, we decode the
		first frames to get it. (used in mpeg case for example) */
	ret = av_find_stream_info(ic, nomem);
	if(ret < 0)
	{
		sprintf(g_cstr, "maPlayWma '%s' could not find codec parameters", filename);
		maLog(g_cstr);
		return 1;
	}

	/* update the current parameters so that they match the one of the input stream */
	enc = &ic->stream->codec;
	*paudio_channels = enc->channels;
	*paudio_sample_rate = enc->sample_rate;
	
	*pinput_file = ic;

	*pfile_iformat = NULL;
	
	return 0;
	
} // opt_input_file


// wmamain moved below

// eof

// libavcodec:

//___av_freep
void __av_freep(void **ptr)
{
	av_free(*ptr);
	*ptr = NULL;
}

// _av_free
void av_free(void *ptr)
{
	/* XXX: this test should not be needed on most libcs */
	if(ptr)
		MAFREE(ptr);
}

// _av_mallocz
void *av_mallocz(unsigned int size)
{
	void *ptr;
	
	ptr = av_malloc(size);
	if(!ptr)
		return NULL;
	memset(ptr, 0, size);
	return ptr;
}

//_avcodec_close
int avcodec_close(AVCodecContext *avctx)
{
	if(avctx)
	{
		if(avctx->codec->close)
			avctx->codec->close(avctx);
		av_freep(&avctx->priv_data);
		avctx->codec = NULL;
	}
	else
		maLog("maPlayWma avcodec_close null avctx");
	return 0;
}

//_avcodec_decode_audio
/* decode an audio frame. return -1 if error, otherwise return the
   *number of bytes used. If no frame could be decompressed,
   *frame_size_ptr is zero. Otherwise, it is the decompressed frame
   *size in BYTES. */
int avcodec_decode_audio(AVCodecContext *avctx, int16_t *samples, 
 						int *frame_size_ptr,
 						uint8_t *buf, int buf_size)
{
	int ret;

	ret = avctx->codec->decode(avctx, samples, frame_size_ptr, buf, buf_size);
	avctx->frame_number++;
	return ret;
}

//_avcodec_find_decoder
AVCodec *avcodec_find_decoder(enum CodecID id)
{
	AVCodec *p;
	p = first_avcodec;
	while(p)
	{
		if(p->decode != NULL && p->id == (int)id)
			return p;
		p = p->next;
	}
	return NULL;
}


void avcodec_get_context_defaults(AVCodecContext *s);
void avcodec_get_context_defaults(AVCodecContext *s){
	s->bit_rate= 800*1000;
	s->frame_rate = 25;
}

//_avcodec_open
int avcodec_open(AVCodecContext *avctx, AVCodec *codec)
{
	int ret;

	if(avctx->codec)
		return -1;

	avctx->codec = codec;
	avctx->codec_id = CODECIDCAST codec->id;
	avctx->frame_number = 0;
	if(codec->priv_data_size > 0)
	{
		avctx->priv_data = av_mallocz(codec->priv_data_size);
		if(!avctx->priv_data) 
			return -ENOMEM;
	}
	else
		avctx->priv_data = NULL;
	ret = avctx->codec->init(avctx); // wma_decode_init
	if(ret < 0)
	{
		av_freep(&avctx->priv_data);
		return ret;
	}
	return 0;
}

//_first_avcodec
AVCodec *first_avcodec;

//_register_avcodec
void register_avcodec(AVCodec *format)
{
	AVCodec **p;
	p = &first_avcodec;
	while(*p != NULL) p = &(*p)->next;
	*p = format;
	format->next = NULL;
}

//_av_malloc mem.c
void *av_malloc(unsigned int size)
{
	void *ptr;
	
#if defined(HAVE_MEMALIGN)
	ptr = memalign(16,size);
	/* Why 64? 
   	Indeed, we should align it:
 		on 4 for 386
 		on 16 for 486
	 on 32 for 586, PPro - k6-III
	 on 64 for K7 (maybe for P3 too).
   	Because L1 and L2 caches are aligned on those values.
   	But I don't want to code such logic here!
 	*/
 	/* Why 16?
		because some cpus need alignment, for example SSE2 on P4, & most RISC cpus
		it will just trigger an exception and the unaligned load will be done in the
		exception handler or it will just segfault (SSE2 on P4)
		Why not larger? because i didnt see a difference in benchmarks ...
 	*/
 	/* benchmarks with p3
		memalign(64)+1		3071,3051,3032
		memalign(64)+2		3051,3032,3041
		memalign(64)+4		2911,2896,2915
		memalign(64)+8		2545,2554,2550
		memalign(64)+16		2543,2572,2563
		memalign(64)+32		2546,2545,2571
		memalign(64)+64		2570,2533,2558
		
		btw, MAMALLOC seems to do 8 byte alignment by default here
 	*/
#else
	ptr = MAMALLOC(size);
#endif
	return ptr;
}

/*
//_av_rescale
int64_t av_rescale(int64_t a, int b, int c){
	uint64_t h, l;
	
	if(a<0) return -av_rescale(-a, b, c);
	
	h= a>>32;
	if(h==0) return a*b/c;
	
	l= a&0xFFFFFFFF;
	l *= b;
	h *= b;

	l += (h%c)<<32;

	return ((h/c)<<32) + l/c;
}
*/

//_av_realloc
void *av_realloc(void *ptr, unsigned int size)
{
	return realloc(ptr, size);
}


//_BF
#define BF(pre, pim, qre, qim, pre1, pim1, qre1, qim1) \
{\
  int ax, ay, bx, by;\
  bx=pre1;\
  by=pim1;\
  ax=qre1;\
  ay=qim1;\
  pre = (bx + ax) >> 1;\
  pim = (by + ay) >> 1;\
  qre = (bx - ax) >> 1;\
  qim = (by - ay) >> 1;\
}

// common.c
/**
 * @file common.c
 * common internal api.
 */

const uint8_t ff_sqrt_tab[128]={
		0, 1, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5,
		5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
		8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9,
		9, 9, 9, 9,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,11,11,11,11,11,11,11
};

const uint8_t ff_log2_tab[256]={
		0,0,1,1,2,2,2,2,3,3,3,3,3,3,3,3,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,
		5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
		6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
		6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
		7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
		7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
		7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
		7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7
};

/**
 * init GetBitContext.
 * @param buffer bitstream buffer, must be FF_INPUT_BUFFER_PADDING_SIZE bytes larger then the actual read bits
 * because some optimized bitstream readers read 32 or 64 bit at once and could read over the end
 * @param bit_size the size of the buffer in bits
 */
void init_get_bits(GetBitContext *s,
				   const uint8_t *buffer, int bit_size)
{
	const int buffer_size= (bit_size+7)>>3;

	s->buffer= buffer;
	s->size_in_bits= bit_size;
	s->buffer_end= buffer + buffer_size;
#ifdef ALT_BITSTREAM_READER
	s->index=0;
#endif
	{
		OPEN_READER(re, s)
		UPDATE_CACHE(re, s)
		UPDATE_CACHE(re, s)
		CLOSE_READER(re, s)
	}
}

void align_get_bits(GetBitContext *s)
{
	int n= (-get_bits_count(s)) & 7;
	if(n) skip_bits(s, n);
}

/* VLC decoding */

#define GET_DATA(v, table, i, wrap, size) \
{\
	const uint8_t *ptr = (const uint8_t *)table + i * wrap;\
	switch(size) {\
	case 1:\
		v = *(const uint8_t *)ptr;\
		break;\
	case 2:\
		v = *(const uint16_t *)ptr;\
		break;\
	default:\
		v = *(const uint32_t *)ptr;\
		break;\
	}\
}


static int alloc_table(VLC *vlc, int size)
{
	int index;
	index = vlc->table_size;
	vlc->table_size += size;
	if(vlc->table_size > vlc->table_allocated)
	{
		vlc->table_allocated += (1 << vlc->bits);
		vlc->table = (short (*) [2])av_realloc(vlc->table,
								sizeof(VLC_TYPE) * 2 * vlc->table_allocated);
		if(!vlc->table)
			return -1;
	}
	return index;
}

static int build_table(VLC *vlc, int table_nb_bits,
					   int nb_codes,
					   const void *bits, int bits_wrap, int bits_size,
					   const void *codes, int codes_wrap, int codes_size,
					   uint32_t code_prefix, int n_prefix)
{
	int i, j, k, n, table_size, table_index, nb, n1, index;
	uint32_t code;
	VLC_TYPE (*table)[2];

	table_size = 1 << table_nb_bits;
	table_index = alloc_table(vlc, table_size);
	if(table_index < 0)
		return -1;
	table = &vlc->table[table_index];

	for(i=0;i<table_size;i++)
	{
		table[i][1] = 0; //bits
		table[i][0] = -1; //codes
	}

	/* first pass: map codes and compute auxillary table sizes */
	for(i=0;i<nb_codes;i++)
	{
		GET_DATA(n, bits, i, bits_wrap, bits_size);
		GET_DATA(code, codes, i, codes_wrap, codes_size);
		/* we accept tables with holes */
		if(n <= 0)
			continue;
		/* if code matches the prefix, it is in the table */
		n -= n_prefix;
		if(n > 0 && (code >> n) == code_prefix)
		{
			if(n <= table_nb_bits)
			{
				/* no need to add another table */
				j = (code << (table_nb_bits - n)) & (table_size - 1);
				nb = 1 << (table_nb_bits - n);
				for(k=0;k<nb;k++)
				{
					if(table[j][1] /*bits*/ != 0)
					{
						sprintf(g_cstr, "maPlayWma build_table incorrect codes");
						maLog(g_cstr);
						return -1; // need to free stuff from alloc_table //x av_abort();
					}
					table[j][1] = n; //bits
					table[j][0] = i; //code
					j++;
				}
			}
			else
			{
				n -= table_nb_bits;
				j = (code >> n) & ((1 << table_nb_bits) - 1);
				/* compute table size */
				n1 = -table[j][1]; //bits
				if(n > n1)
					n1 = n;
				table[j][1] = -n1; //bits
			}
		}
	}

	/* second pass : fill auxillary tables recursively */
	for(i=0;i<table_size;i++)
	{
		n = table[i][1]; //bits
		if(n < 0)
		{
			n = -n;
			if(n > table_nb_bits)
			{
				n = table_nb_bits;
				table[i][1] = -n; //bits
			}
			index = build_table(vlc, n, nb_codes,
								bits, bits_wrap, bits_size,
								codes, codes_wrap, codes_size,
								(code_prefix << table_nb_bits) | i,
								n_prefix + table_nb_bits);
			if(index < 0)
				return -1;
			/* note: realloc has been done, so reload tables */
			table = &vlc->table[table_index];
			table[i][0] = index; //code
		}
	}
	return table_index;
}


/* Build VLC decoding tables suitable for use with get_vlc().

   'nb_bits' set thee decoding table size (2^nb_bits) entries. The
   bigger it is, the faster is the decoding. But it should not be too
   big to save memory and L1 cache. '9' is a good compromise.
   
   'nb_codes' : number of vlcs codes

   'bits' : table which gives the size (in bits) of each vlc code.

   'codes' : table which gives the bit pattern of of each vlc code.

   'xxx_wrap' : give the number of bytes between each entry of the
   'bits' or 'codes' tables.

   'xxx_size' : gives the number of bytes of each entry of the 'bits'
   or 'codes' tables.

   'wrap' and 'size' allows to use any memory configuration and types
   (byte/word/long) to store the 'bits' and 'codes' tables.  
*/
int init_vlc(VLC *vlc, int nb_bits, int nb_codes,
			 const void *bits, int bits_wrap, int bits_size,
			 const void *codes, int codes_wrap, int codes_size)
{
	vlc->bits = nb_bits;
	vlc->table = NULL;
	vlc->table_allocated = 0;
	vlc->table_size = 0;

	if(build_table(vlc, nb_bits, nb_codes, bits, bits_wrap, bits_size,
							codes, codes_wrap, codes_size, 0, 0) < 0)
	{
		av_free(vlc->table);
		return -1;
	}
	return 0;
}


void free_vlc(VLC *vlc)
{
	av_free(vlc->table);
}

// wmadec.c
/*
 * WMA compatible decoder
 * Copyright (c) 2002 The FFmpeg Project.
 */

/**
 * @file wmadec.c
 * WMA compatible decoder.
 */

/* size of blocks */
#define BLOCK_MIN_BITS 7
#define BLOCK_MAX_BITS 11
#define BLOCK_MAX_SIZE (1 << BLOCK_MAX_BITS)

#define BLOCK_NB_SIZES (BLOCK_MAX_BITS - BLOCK_MIN_BITS + 1)

/* XXX: find exact max size */
#define HIGH_BAND_MAX_SIZE 16

/* XXX: is it a suitable value ? */ // 4096 is too small
#define MAX_CODED_SUPERFRAME_SIZE 16384

#define MAX_CHANNELS 2

#define NOISE_TAB_SIZE 8192

#define LSP_POW_BITS 7

typedef struct WMADecodeContext
{
	GetBitContext gb;
	int sample_rate;
	int nb_channels;
	int bit_rate;
	int version; /* 1 = 0x160 (WMAV1), 2 = 0x161 (WMAV2) */
	int block_align;
	int use_bit_reservoir;
	int use_variable_block_len;
	int use_exp_vlc;  /* exponent coding: 0 = lsp, 1 = vlc + delta */
	int use_noise_coding; /* true if perceptual noise is added */
	int byte_offset_bits;
	VLC exp_vlc;
	int exponent_sizes[BLOCK_NB_SIZES];
	uint16_t exponent_bands[BLOCK_NB_SIZES][25];
	int high_band_start[BLOCK_NB_SIZES]; /* index of first coef in high band */
	int coefs_start;			   /* first coded coef */
	int coefs_end[BLOCK_NB_SIZES]; /* max number of coded coefficients */
	int exponent_high_sizes[BLOCK_NB_SIZES];
	int exponent_high_bands[BLOCK_NB_SIZES][HIGH_BAND_MAX_SIZE]; 
	VLC hgain_vlc;
	
	/* coded values in high bands */
	int high_band_coded[MAX_CHANNELS][HIGH_BAND_MAX_SIZE];
	int high_band_values[MAX_CHANNELS][HIGH_BAND_MAX_SIZE];

	/* there are two possible tables for spectral coefficients */
	VLC coef_vlc[2];
	uint16_t *run_table[2];
	uint16_t *level_table[2];
	/* frame info */
	int frame_len;	   /* frame length in samples */
	int frame_len_bits;  /* frame_len = 1 << frame_len_bits */
	int nb_block_sizes;  /* number of block sizes */
	/* block info */
	int reset_block_lengths;
	int block_len_bits; /* log2 of current block length */
	int next_block_len_bits; /* log2 of next block length */
	int prev_block_len_bits; /* log2 of prev block length */
	int block_len; /* block length in samples */
	int block_num; /* block number in current frame */
	int block_pos; /* current position in frame */
	uint8_t ms_stereo; /* true if mid/side stereo mode */
	uint8_t channel_coded[MAX_CHANNELS]; /* true if channel is coded */
	float exponents[MAX_CHANNELS][BLOCK_MAX_SIZE] /*n __attribute__((aligned(16))) */;
	float max_exponent[MAX_CHANNELS];
	int16_t coefs1[MAX_CHANNELS][BLOCK_MAX_SIZE];
	float coefs[MAX_CHANNELS][BLOCK_MAX_SIZE] /*n __attribute__((aligned(16))) */;
	MDCTContext mdct_ctx[BLOCK_NB_SIZES];
	float *windows[BLOCK_NB_SIZES];
	FFTSample mdct_tmp[BLOCK_MAX_SIZE] /*n __attribute__((aligned(16))) */; /* temporary storage for imdct */
	/* output buffer for one frame and the last for IMDCT windowing */
	float frame_out[MAX_CHANNELS][BLOCK_MAX_SIZE * 2] /*n __attribute__((aligned(16))) */;
	/* last frame info */
	uint8_t last_superframe[MAX_CODED_SUPERFRAME_SIZE + 4]; /* padding added */
	int last_bitoffset;
	int last_superframe_len;
	float noise_table[NOISE_TAB_SIZE];
	int noise_index;
	float noise_mult; /* XXX: suppress that and integrate it in the noise array */
	/* lsp_to_curve tables */
	float lsp_cos_table[BLOCK_MAX_SIZE];
	float lsp_pow_e_table[256];
	float lsp_pow_m_table1[(1 << LSP_POW_BITS)];
	float lsp_pow_m_table2[(1 << LSP_POW_BITS)];

} WMADecodeContext;


static void wma_lsp_to_curve_init(WMADecodeContext *s, int frame_len);

/* XXX: use same run/length optimization as mpeg decoders */
static void init_coef_vlc(VLC *vlc, 
						  uint16_t **prun_table, uint16_t **plevel_table,
						  const CoefVLCTable *vlc_table)
{
	int n = vlc_table->n;
	const uint8_t *table_bits = vlc_table->huffbits;
	const uint32_t *table_codes = vlc_table->huffcodes;
	const uint16_t *levels_table = vlc_table->levels;
	uint16_t *run_table, *level_table;
	const uint16_t *p;
	int i, l, j, level;

	init_vlc(vlc, 9, n, table_bits, 1, 1, table_codes, 4, 4);

	run_table = (unsigned short *)av_malloc(n * sizeof(uint16_t));
	level_table = (unsigned short *)av_malloc(n * sizeof(uint16_t));
	p = levels_table;
	i = 2;
	level = 1;
	while(i < n)
	{
		l = *p++;
		for(j=0;j<l;j++)
		{
			run_table[i] = j;
			level_table[i] = level;
			i++;
		}
		level++;
	}
	*prun_table = run_table;
	*plevel_table = level_table;
}

static int wma_decode_init(AVCodecContext * avctx)
{
	WMADecodeContext *s = (WMADecodeContext *)avctx->priv_data;
	int i, flags1, flags2;
	float *window;
	uint8_t *extradata;
	float bps1, high_freq, bps;
	int sample_rate1;
	int coef_vlc_table;
	
	s->sample_rate = avctx->sample_rate;
	s->nb_channels = avctx->channels;
	s->bit_rate = avctx->bit_rate;
	s->block_align = avctx->block_align;

	if(avctx->codec->id == CODEC_ID_WMAV1)
		s->version = 1;
	else
		s->version = 2;
	
	/* extract flag infos */
	flags1 = 0;
	flags2 = 0;
	extradata = (unsigned char *)avctx->extradata;
	if(s->version == 1 && avctx->extradata_size >= 4)
	{
		flags1 = extradata[0] | (extradata[1] << 8);
		flags2 = extradata[2] | (extradata[3] << 8);
	}
	else if(s->version == 2 && avctx->extradata_size >= 6)
	{
		flags1 = extradata[0] | (extradata[1] << 8) | 
			(extradata[2] << 16) | (extradata[3] << 24);
		flags2 = extradata[4] | (extradata[5] << 8);
	}
	s->use_exp_vlc = flags2 & 0x0001;
	s->use_bit_reservoir = flags2 & 0x0002;
	s->use_variable_block_len = flags2 & 0x0004;

	/* compute MDCT block size */
	if(s->sample_rate <= 16000)
	{
		s->frame_len_bits = 9;
	} else if(s->sample_rate <= 22050 || 
			   (s->sample_rate <= 32000 && s->version == 1))
	{
		s->frame_len_bits = 10;
	} else
	{
		s->frame_len_bits = 11;
	}
	s->frame_len = 1 << s->frame_len_bits;
	if(s->use_variable_block_len)
	{
		int nb_max, nb;
		nb = ((flags2 >> 3) & 3) + 1;
		if((s->bit_rate / s->nb_channels) >= 32000)
			nb += 2;
		nb_max = s->frame_len_bits - BLOCK_MIN_BITS;
		if(nb > nb_max)
			nb = nb_max;
		s->nb_block_sizes = nb + 1;
	}
	else
	{
		s->nb_block_sizes = 1;
	}

	/* init rate dependant parameters */
	s->use_noise_coding = 1;
	high_freq = s->sample_rate * 0.5;

	/* if version 2, then the rates are normalized */
	sample_rate1 = s->sample_rate;
	if(s->version == 2)
	{
		if(sample_rate1 >= 44100) 
			sample_rate1 = 44100;
		else if(sample_rate1 >= 22050) 
			sample_rate1 = 22050;
		else if(sample_rate1 >= 16000) 
			sample_rate1 = 16000;
		else if(sample_rate1 >= 11025) 
			sample_rate1 = 11025;
		else if(sample_rate1 >= 8000) 
			sample_rate1 = 8000;
	}

	bps = (float)s->bit_rate / (float)(s->nb_channels * s->sample_rate);
	s->byte_offset_bits = av_log2((int)(bps * s->frame_len / 8.0)) + 2;

	/* compute high frequency value and choose if noise coding should
	   be activated */
	bps1 = bps;
	if(s->nb_channels == 2)
		bps1 = bps * 1.6;
	if(sample_rate1 == 44100)
	{
		if(bps1 >= 0.61)
			s->use_noise_coding = 0;
		else
			high_freq = high_freq * 0.4;
	} else if(sample_rate1 == 22050)
	{
		if(bps1 >= 1.16)
			s->use_noise_coding = 0;
		else if(bps1 >= 0.72) 
			high_freq = high_freq * 0.7;
		else
			high_freq = high_freq * 0.6;
	} else if(sample_rate1 == 16000)
	{
		if(bps > 0.5)
			high_freq = high_freq * 0.5;
		else
			high_freq = high_freq * 0.3;
	} else if(sample_rate1 == 11025)
	{
		high_freq = high_freq * 0.7;
	} else if(sample_rate1 == 8000)
	{
		if(bps <= 0.625)
		{
			high_freq = high_freq * 0.5;
		} else if(bps > 0.75)
		{
			s->use_noise_coding = 0;
		}
		else
		{
			high_freq = high_freq * 0.65;
		}
	}
	else
	{
		if(bps >= 0.8)
		{
			high_freq = high_freq * 0.75;
		}
		else if(bps >= 0.6)
		{
			high_freq = high_freq * 0.6;
		} else
		{
			high_freq = high_freq * 0.5;
		}
	}

	/* compute the scale factor band sizes for each MDCT block size */
	{
		int a, b, pos, lpos, k, block_len, i, j, n;
		const uint8_t *table;
		
		if(s->version == 1)
		{
			s->coefs_start = 3;
		}
		else
		{
			s->coefs_start = 0;
		}
		for(k = 0; k < s->nb_block_sizes; k++)
		{
			block_len = s->frame_len >> k;

			if(s->version == 1)
			{
				lpos = 0;
				for(i=0;i<25;i++)
				{
					a = wma_critical_freqs[i];
					b = s->sample_rate;
					pos = ((block_len * 2 * a)  + (b >> 1)) / b;
					if(pos > block_len) 
						pos = block_len;
					s->exponent_bands[0][i] = pos - lpos;
					if(pos >= block_len)
					{
						i++;
						break;
					}
					lpos = pos;
				}
				s->exponent_sizes[0] = i;
			}
			else
			{
				/* hardcoded tables */
				table = NULL;
				a = s->frame_len_bits - BLOCK_MIN_BITS - k;
				if(a < 3)
				{
					if(s->sample_rate >= 44100)
						table = exponent_band_44100[a];
					else if(s->sample_rate >= 32000)
						table = exponent_band_32000[a];
					else if(s->sample_rate >= 22050)
						table = exponent_band_22050[a];
				}
				if(table)
				{
					n = *table++;
					for(i=0;i<n;i++)
						s->exponent_bands[k][i] = table[i];
					s->exponent_sizes[k] = n;
				}
				else
				{
					j = 0;
					lpos = 0;
					for(i=0;i<25;i++)
					{
						a = wma_critical_freqs[i];
						b = s->sample_rate;
						pos = ((block_len * 2 * a)  + (b << 1)) / (4 * b);
						pos <<= 2;
						if(pos > block_len) 
							pos = block_len;
						if(pos > lpos)
							s->exponent_bands[k][j++] = pos - lpos;
						if(pos >= block_len)
							break;
						lpos = pos;
					}
					s->exponent_sizes[k] = j;
				}
			}

			/* max number of coefs */
			s->coefs_end[k] = (s->frame_len - ((s->frame_len * 9) / 100)) >> k;
			/* high freq computation */
			s->high_band_start[k] = (int)((block_len * 2 * high_freq) / 
										  s->sample_rate + 0.5);
			n = s->exponent_sizes[k];
			j = 0;
			pos = 0;
			for(i=0;i<n;i++)
			{
				int start, end;
				start = pos;
				pos += s->exponent_bands[k][i];
				end = pos;
				if(start < s->high_band_start[k])
					start = s->high_band_start[k];
				if(end > s->coefs_end[k])
					end = s->coefs_end[k];
				if(end > start)
					s->exponent_high_bands[k][j++] = end - start;
			}
			s->exponent_high_sizes[k] = j;
		}
	}

	/* init MDCT */
	for(i = 0; i < s->nb_block_sizes; i++)
		ff_mdct_init(&s->mdct_ctx[i], s->frame_len_bits - i + 1, 1);
	
	/* init MDCT windows : simple sinus window */
	for(i = 0; i < s->nb_block_sizes; i++)
	{
		int n, j;
		float alpha;
		n = 1 << (s->frame_len_bits - i);
		window = (float *)av_malloc(sizeof(float) * n);
		alpha = M_PI / (2.0 * n);
		for(j=0;j<n;j++)
		{
			window[n - j - 1] = sin((j + 0.5) * alpha);
		}
		s->windows[i] = window;
	}

	s->reset_block_lengths = 1;
	
	if(s->use_noise_coding)
	{

		/* init the noise generator */
		if(s->use_exp_vlc)
			s->noise_mult = 0.02;
		else
			s->noise_mult = 0.04;
			   
		{
			unsigned int seed;
			float norm;
			seed = 1;
#ifdef __MAWIN__
			norm = (1.0 / (float)(1L << 31)) * sqrt(3) * s->noise_mult;
#else
			norm = (1.0 / (float)(1LL << 31)) * sqrt(3) * s->noise_mult;
#endif
			for(i=0;i<NOISE_TAB_SIZE;i++)
			{
				seed = seed * 314159 + 1;
				s->noise_table[i] = (float)((int)seed) * norm;
			}
		}

		init_vlc(&s->hgain_vlc, 9, sizeof(hgain_huffbits), 
				 hgain_huffbits, 1, 1,
				 hgain_huffcodes, 2, 2);
	}

	if(s->use_exp_vlc)
	{
		init_vlc(&s->exp_vlc, 9, sizeof(scale_huffbits), 
				 scale_huffbits, 1, 1,
				 scale_huffcodes, 4, 4);
	}
	else
	{
		wma_lsp_to_curve_init(s, s->frame_len);
	}

	/* choose the VLC tables for the coefficients */
	coef_vlc_table = 2;
	if(s->sample_rate >= 32000)
	{
		if(bps1 < 0.72)
			coef_vlc_table = 0;
		else if(bps1 < 1.16)
			coef_vlc_table = 1;
	}

	init_coef_vlc(&s->coef_vlc[0], &s->run_table[0], &s->level_table[0],
				  &coef_vlcs[coef_vlc_table * 2]);
	init_coef_vlc(&s->coef_vlc[1], &s->run_table[1], &s->level_table[1],
				  &coef_vlcs[coef_vlc_table * 2 + 1]);
	return 0;
}

/* interpolate values for a bigger or smaller block. The block must
   have multiple sizes */
static void interpolate_array(float *scale, int old_size, int new_size)
{
	int i, j, jincr, k;
	float v;

	if(new_size > old_size)
	{
		jincr = new_size / old_size;
		j = new_size;
		for(i = old_size - 1; i >=0; i--)
		{
			v = scale[i];
			k = jincr;
			do
			{
				scale[--j] = v;
			} while(--k);
		}
	}
	else if(new_size < old_size)
	{
		j = 0;
		jincr = old_size / new_size;
		for(i = 0; i < new_size; i++)
		{
			scale[i] = scale[j];
			j += jincr;
		}
	}
}

/* compute x^-0.25 with an exponent and mantissa table. We use linear
   interpolation to reduce the mantissa table size at a small speed
   expense (linear interpolation approximately doubles the number of
   bits of precision). */
static inline float pow_m1_4(WMADecodeContext *s, float x)
{
	union
	{
		float f;
		unsigned int v;
	} u, t;
	unsigned int e, m;
	float a, b;

	u.f = x;
	e = u.v >> 23;
	m = (u.v >> (23 - LSP_POW_BITS)) & ((1 << LSP_POW_BITS) - 1);
	/* build interpolation scale: 1 <= t < 2. */
	t.v = ((u.v << LSP_POW_BITS) & ((1 << 23) - 1)) | (127 << 23);
	a = s->lsp_pow_m_table1[m];
	b = s->lsp_pow_m_table2[m];
	return s->lsp_pow_e_table[e] * (a + b * t.f);
}

static void wma_lsp_to_curve_init(WMADecodeContext *s, int frame_len)
{  
	float wdel, a, b;
	int i, e, m;

	wdel = M_PI / frame_len;
	for(i=0;i<frame_len;i++)
		s->lsp_cos_table[i] = 2.0f * cos(wdel * i);

	/* tables for x^-0.25 computation */
	for(i=0;i<256;i++)
	{
		e = i - 126;
		s->lsp_pow_e_table[i] = pow(2.0, e * -0.25);
	}

	/* NOTE: these two tables are needed to avoid two operations in pow_m1_4 */
	b = 1.0;
	for(i=(1 << LSP_POW_BITS) - 1;i>=0;i--)
	{
		m = (1 << LSP_POW_BITS) + i;
		a = (float)m * (0.5 / (1 << LSP_POW_BITS));
		a = pow(a, -0.25);
		s->lsp_pow_m_table1[i] = 2 * a - b;
		s->lsp_pow_m_table2[i] = b - a;
		b = a;
	}
}

/* NOTE: We use the same code as Vorbis here */
/* XXX: optimize it further with SSE/3Dnow */
static void wma_lsp_to_curve(WMADecodeContext *s, 
							 float *out, float *val_max_ptr, 
							 int n, float *lsp)
{
	int i, j;
	float p, q, w, v, val_max;

	val_max = 0;
	for(i=0;i<n;i++)
	{
		p = 0.5f;
		q = 0.5f;
		w = s->lsp_cos_table[i];
		for(j=1;j<NB_LSP_COEFS;j+=2){
			q *= w - lsp[j - 1];
			p *= w - lsp[j];
		}
		p *= p * (2.0f - w);
		q *= q * (2.0f + w);
		v = p + q;
		v = pow_m1_4(s, v);
		if(v > val_max)
			val_max = v;
		out[i] = v;
	}
	*val_max_ptr = val_max;
}

/* decode exponents coded with LSP coefficients (same idea as Vorbis) */
static void decode_exp_lsp(WMADecodeContext *s, int ch)
{
	float lsp_coefs[NB_LSP_COEFS];
	int val, i;

	for(i = 0; i < NB_LSP_COEFS; i++)
	{
		if(i == 0 || i >= 8)
			val = get_bits(&s->gb, 3);
		else
			val = get_bits(&s->gb, 4);
		lsp_coefs[i] = lsp_codebook[i][val];
	}

	wma_lsp_to_curve(s, s->exponents[ch], &s->max_exponent[ch],
					 s->block_len, lsp_coefs);
}

/* decode exponents coded with VLC codes */
static int decode_exp_vlc(WMADecodeContext *s, int ch)
{
	int last_exp, n, code;
	const uint16_t *ptr, *band_ptr;
	float v, *q, max_scale, *q_end;
	
	band_ptr = s->exponent_bands[s->frame_len_bits - s->block_len_bits];
	ptr = band_ptr;
	q = s->exponents[ch];
	q_end = q + s->block_len;
	max_scale = 0;
	if(s->version == 1)
	{
		last_exp = get_bits(&s->gb, 5) + 10;
		/* XXX: use a table */
		v = pow(10, last_exp * (1.0 / 16.0));
		max_scale = v;
		n = *ptr++;
		do
		{
			*q++ = v;
		} while(--n);
	}
	last_exp = 36;
	while(q < q_end)
	{
		code = get_vlc(&s->gb, &s->exp_vlc);
		if(code < 0)
			return -1;
		/* NOTE: this offset is the same as MPEG4 AAC ! */
		last_exp += code - 60;
		/* XXX: use a table */
		v = pow(10, last_exp * (1.0 / 16.0));
		if(v > max_scale)
			max_scale = v;
		n = *ptr++;
		do
		{
			*q++ = v;
		} while(--n);
	}
	s->max_exponent[ch] = max_scale;
	return 0;
}

/* return 0 if OK. return 1 if last block of frame. return -1 if
   unrecorrable error. */
static int wma_decode_block(WMADecodeContext *s)
{
	int n, v, a, ch, code, bsize;
	int coef_nb_bits, total_gain, parse_exponents;
	float window[BLOCK_MAX_SIZE * 2];
	int nb_coefs[MAX_CHANNELS];
	float mdct_norm;

	/* compute current block length */
	if(s->use_variable_block_len)
	{
		n = av_log2(s->nb_block_sizes - 1) + 1;
	
		if(s->reset_block_lengths)
		{
			s->reset_block_lengths = 0;
			v = get_bits(&s->gb, n);
			if(v >= s->nb_block_sizes)
				return -1;
			s->prev_block_len_bits = s->frame_len_bits - v;
			v = get_bits(&s->gb, n);
			if(v >= s->nb_block_sizes)
				return -1;
			s->block_len_bits = s->frame_len_bits - v;
		}
		else
		{
			/* update block lengths */
			s->prev_block_len_bits = s->block_len_bits;
			s->block_len_bits = s->next_block_len_bits;
		}
		v = get_bits(&s->gb, n);
		if(v >= s->nb_block_sizes)
			return -1;
		s->next_block_len_bits = s->frame_len_bits - v;
	}
	else
	{
		/* fixed block len */
		s->next_block_len_bits = s->frame_len_bits;
		s->prev_block_len_bits = s->frame_len_bits;
		s->block_len_bits = s->frame_len_bits;
	}

	/* now check if the block length is coherent with the frame length */
	s->block_len = 1 << s->block_len_bits;
	if((s->block_pos + s->block_len) > s->frame_len)
		return -1;

	if(s->nb_channels == 2)
	{
		s->ms_stereo = get_bits(&s->gb, 1);
	}
	v = 0;
	for(ch = 0; ch < s->nb_channels; ch++)
	{
		a = get_bits(&s->gb, 1);
		s->channel_coded[ch] = a;
		v |= a;
	}
	/* if no channel coded, no need to go further */
	/* XXX: fix potential framing problems */
	if(!v)
		goto next;

	bsize = s->frame_len_bits - s->block_len_bits;

	/* read total gain and extract corresponding number of bits for
	   coef escape coding */
	total_gain = 1;
	for(;;)
	{
		a = get_bits(&s->gb, 7);
		total_gain += a;
		if(a != 127)
			break;
	}
	
	if(total_gain < 15)
		coef_nb_bits = 13;
	else if(total_gain < 32)
		coef_nb_bits = 12;
	else if(total_gain < 40)
		coef_nb_bits = 11;
	else if(total_gain < 45)
		coef_nb_bits = 10;
	else
		coef_nb_bits = 9;

	/* compute number of coefficients */
	n = s->coefs_end[bsize] - s->coefs_start;
	for(ch = 0; ch < s->nb_channels; ch++)
		nb_coefs[ch] = n;

	/* complex coding */
	if(s->use_noise_coding)
	{
		for(ch = 0; ch < s->nb_channels; ch++)
		{
			if(s->channel_coded[ch])
			{
				int i, n, a;
				n = s->exponent_high_sizes[bsize];
				for(i=0;i<n;i++)
				{
					a = get_bits(&s->gb, 1);
					s->high_band_coded[ch][i] = a;
					/* if noise coding, the coefficients are not transmitted */
					if(a)
						nb_coefs[ch] -= s->exponent_high_bands[bsize][i];
				}
			}
		}
		for(ch = 0; ch < s->nb_channels; ch++)
		{
			if(s->channel_coded[ch])
			{
				int i, n, val, code;

				n = s->exponent_high_sizes[bsize];
				val = (int)0x80000000;
				for(i=0;i<n;i++)
				{
					if(s->high_band_coded[ch][i])
					{
						if(val == (int)0x80000000)
						{
							val = get_bits(&s->gb, 7) - 19;
						}
						else
						{
							code = get_vlc(&s->gb, &s->hgain_vlc);
							if(code < 0)
								return -1;
							val += code - 18;
						}
						s->high_band_values[ch][i] = val;
					}
				}
			}
		}
	}
		   
	/* exposant can be interpolated in short blocks. */
	parse_exponents = 1;
	if(s->block_len_bits != s->frame_len_bits)
	{
		parse_exponents = get_bits(&s->gb, 1);
	}
	
	if(parse_exponents)
	{
		for(ch = 0; ch < s->nb_channels; ch++)
		{
			if(s->channel_coded[ch])
			{
				if(s->use_exp_vlc)
				{
					if(decode_exp_vlc(s, ch) < 0)
						return -1;
				}
				else
				{
					decode_exp_lsp(s, ch);
				}
			}
		}
	}
	else
	{
		for(ch = 0; ch < s->nb_channels; ch++)
		{
			if(s->channel_coded[ch])
			{
				interpolate_array(s->exponents[ch], 1 << s->prev_block_len_bits, 
								  s->block_len);
			}
		}
	}

	/* parse spectral coefficients : just RLE encoding */
	for(ch = 0; ch < s->nb_channels; ch++)
	{
		if(s->channel_coded[ch])
		{
			VLC *coef_vlc;
			int level, run, sign, tindex;
			int16_t *ptr, *eptr;
			const int16_t *level_table, *run_table;

			/* special VLC tables are used for ms stereo because
			   there is potentially less energy there */
			tindex = (ch == 1 && s->ms_stereo);
			coef_vlc = &s->coef_vlc[tindex];
			run_table = (const short *)s->run_table[tindex];
			level_table = (const short *)s->level_table[tindex];
			/* XXX: optimize */
			ptr = &s->coefs1[ch][0];
			eptr = ptr + nb_coefs[ch];
			memset(ptr, 0, s->block_len * sizeof(int16_t));
			for(;;)
			{
				code = get_vlc(&s->gb, coef_vlc);
				if(code < 0)
					return -1;
				if(code == 1)
				{
					/* EOB */
					break;
				}
				else if(code == 0)
				{
					/* escape */
					level = get_bits(&s->gb, coef_nb_bits);
					/* NOTE: this is rather suboptimal. reading
					   block_len_bits would be better */
					run = get_bits(&s->gb, s->frame_len_bits);
				}
				else
				{
					/* normal code */
					run = run_table[code];
					level = level_table[code];
				}
				sign = get_bits(&s->gb, 1);
				if(!sign)
					level = -level;
				ptr += run;
				if(ptr >= eptr)
					return -1;
				*ptr++ = level;
				/* NOTE: EOB can be omitted */
				if(ptr >= eptr)
					break;
			}
		}
		if(s->version == 1 && s->nb_channels >= 2)
		{
			align_get_bits(&s->gb);
		}
	}
	 
	/* normalize */
	{
		int n4 = s->block_len / 2;
		mdct_norm = 1.0 / (float)n4;
		if(s->version == 1)
		{
			mdct_norm *= sqrt(n4);
		}
	}

	/* finally compute the MDCT coefficients */
	for(ch = 0; ch < s->nb_channels; ch++)
	{
		if(s->channel_coded[ch])
		{
			int16_t *coefs1;
			float *coefs, *exponents, mult, mult1, noise, *exp_ptr;
			int i, j, n, n1, last_high_band;
			float exp_power[HIGH_BAND_MAX_SIZE];

			coefs1 = s->coefs1[ch];
			exponents = s->exponents[ch];
			mult = pow(10, total_gain * 0.05) / s->max_exponent[ch];
			mult *= mdct_norm;
			coefs = s->coefs[ch];
			if(s->use_noise_coding)
			{
				mult1 = mult;
				/* very low freqs : noise */
				for(i = 0;i < s->coefs_start; i++)
				{
					*coefs++ = s->noise_table[s->noise_index] * (*exponents++) * mult1;
					s->noise_index = (s->noise_index + 1) & (NOISE_TAB_SIZE - 1);
				}
				
				n1 = s->exponent_high_sizes[bsize];

				/* compute power of high bands */
				exp_ptr = exponents + 
					s->high_band_start[bsize] - 
					s->coefs_start;
				last_high_band = 0; /* avoid warning */
				for(j=0;j<n1;j++)
				{
					n = s->exponent_high_bands[s->frame_len_bits - 
											  s->block_len_bits][j];
					if(s->high_band_coded[ch][j])
					{
						float e2, v;
						e2 = 0;
						for(i = 0;i < n; i++)
						{
							v = exp_ptr[i];
							e2 += v * v;
						}
						exp_power[j] = e2 / n;
						last_high_band = j;
					}
					exp_ptr += n;
				}

				/* main freqs and high freqs */
				for(j=-1;j<n1;j++)
				{
					if(j < 0)
					{
						n = s->high_band_start[bsize] - 
							s->coefs_start;
					}
					else
					{
						n = s->exponent_high_bands[s->frame_len_bits - 
												  s->block_len_bits][j];
					}
					if(j >= 0 && s->high_band_coded[ch][j])
					{
						/* use noise with specified power */
						mult1 = sqrt(exp_power[j] / exp_power[last_high_band]);
						/* XXX: use a table */
						mult1 = mult1 * pow(10, s->high_band_values[ch][j] * 0.05);
						mult1 = mult1 / (s->max_exponent[ch] * s->noise_mult);
						mult1 *= mdct_norm;
						for(i = 0;i < n; i++)
						{
							noise = s->noise_table[s->noise_index];
							s->noise_index = (s->noise_index + 1) & (NOISE_TAB_SIZE - 1);
							*coefs++ = (*exponents++) * noise * mult1;
						}
					}
					else
					{
						/* coded values + small noise */
						for(i = 0;i < n; i++)
						{
							noise = s->noise_table[s->noise_index];
							s->noise_index = (s->noise_index + 1) & (NOISE_TAB_SIZE - 1);
							*coefs++ = ((*coefs1++) + noise) * (*exponents++) * mult;
						}
					}
				}

				/* very high freqs : noise */
				n = s->block_len - s->coefs_end[bsize];
				mult1 = mult * exponents[-1];
				for(i = 0; i < n; i++)
				{
					*coefs++ = s->noise_table[s->noise_index] * mult1;
					s->noise_index = (s->noise_index + 1) & (NOISE_TAB_SIZE - 1);
				}
			}
			else
			{
				/* XXX: optimize more */
				for(i = 0;i < s->coefs_start; i++)
					*coefs++ = 0.0;
				n = nb_coefs[ch];
				for(i = 0;i < n; i++)
				{
					*coefs++ = coefs1[i] * exponents[i] * mult;
				}
				n = s->block_len - s->coefs_end[bsize];
				for(i = 0;i < n; i++)
					*coefs++ = 0.0;
			}
		}
	}

	if(s->ms_stereo && s->channel_coded[1])
	{
		float a, b;
		int i;

		/* nominal case for ms stereo: we do it before mdct */
		/* no need to optimize this case because it should almost
		   never happen */
		if(!s->channel_coded[0])
		{
			memset(s->coefs[0], 0, sizeof(float) * s->block_len);
			s->channel_coded[0] = 1;
		}
		
		for(i = 0; i < s->block_len; i++)
		{
			a = s->coefs[0][i];
			b = s->coefs[1][i];
			s->coefs[0][i] = a + b;
			s->coefs[1][i] = a - b;
		}
	}

	/* build the window : we ensure that when the windows overlap
	   their squared sum is always 1 (MDCT reconstruction rule) */
	/* XXX: merge with output */
	{
		int i, next_block_len, block_len, prev_block_len, n;
		float *wptr;

		block_len = s->block_len;
		prev_block_len = 1 << s->prev_block_len_bits;
		next_block_len = 1 << s->next_block_len_bits;

		/* right part */
		wptr = window + block_len;
		if(block_len <= next_block_len)
		{
			for(i=0;i<block_len;i++)
				*wptr++ = s->windows[bsize][i];
		}
		else
		{
			/* overlap */
			n = (block_len / 2) - (next_block_len / 2);
			for(i=0;i<n;i++)
				*wptr++ = 1.0;
			for(i=0;i<next_block_len;i++)
				*wptr++ = s->windows[s->frame_len_bits - s->next_block_len_bits][i];
			for(i=0;i<n;i++)
				*wptr++ = 0.0;
		}

		/* left part */
		wptr = window + block_len;
		if(block_len <= prev_block_len)
		{
			for(i=0;i<block_len;i++)
				*--wptr = s->windows[bsize][i];
		}
		else
		{
			/* overlap */
			n = (block_len / 2) - (prev_block_len / 2);
			for(i=0;i<n;i++)
				*--wptr = 1.0;
			for(i=0;i<prev_block_len;i++)
				*--wptr = s->windows[s->frame_len_bits - s->prev_block_len_bits][i];
			for(i=0;i<n;i++)
				*--wptr = 0.0;
		}
	}

	
	for(ch = 0; ch < s->nb_channels; ch++)
	{
		if(s->channel_coded[ch])
		{
			FFTSample output[BLOCK_MAX_SIZE * 2] /*n __attribute__((aligned(16))) */;
			float *ptr;
			int i, n4, index, n;

			n = s->block_len;
			n4 = s->block_len / 2;
			ff_imdct_calc(&s->mdct_ctx[bsize], 
						  output, s->coefs[ch], s->mdct_tmp);

			/* XXX: optimize all that by build the window and
			   multipying/adding at the same time */
			/* multiply by the window */
			for(i=0;i<n * 2;i++)
			{
				output[i] *= window[i];
			}

			/* add in the frame */
			index = (s->frame_len / 2) + s->block_pos - n4;
			ptr = &s->frame_out[ch][index];
			for(i=0;i<n * 2;i++)
			{
				*ptr += output[i];
				ptr++;
			}

			/* specific fast case for ms-stereo : add to second
			   channel if it is not coded */
			if(s->ms_stereo && !s->channel_coded[1])
			{
				ptr = &s->frame_out[1][index];
				for(i=0;i<n * 2;i++)
				{
					*ptr += output[i];
					ptr++;
				}
			}
		}
	}
 next:
	/* update block number */
	s->block_num++;
	s->block_pos += s->block_len;
	if(s->block_pos >= s->frame_len)
		return 1;
	else
		return 0;
}

/* decode a frame of frame_len samples */
static int wma_decode_frame(WMADecodeContext *s, int16_t *samples)
{
	int ret, i, n, a, ch, incr;
	int16_t *ptr;
	float *iptr;

	/* read each block */
	s->block_num = 0;
	s->block_pos = 0;
	for(;;)
	{
		ret = wma_decode_block(s);
		if(ret < 0) 
			return -1;
		if(ret)
			break;
	}

	/* convert frame to integer */
	n = s->frame_len;
	incr = s->nb_channels;
	for(ch = 0; ch < s->nb_channels; ch++)
	{
		ptr = samples + ch;
		iptr = s->frame_out[ch];

		for(i=0;i<n;i++)
		{
			a = lrintf(*iptr++);
			if(a > 32767)
				a = 32767;
			else if(a < -32768)
				a = -32768;
			*ptr = a;
			ptr += incr;
		}
		/* prepare for next block */
		memmove(&s->frame_out[ch][0], &s->frame_out[ch][s->frame_len],
				s->frame_len * sizeof(float));
		/* XXX: suppress this */
		memset(&s->frame_out[ch][s->frame_len], 0, 
			   s->frame_len * sizeof(float));
	}

	return 0;
}

static int wma_decode_superframe(AVCodecContext *avctx, 
								 void *data, int *data_size,
								 uint8_t *buf, int buf_size)
{
	WMADecodeContext *s = (WMADecodeContext *)avctx->priv_data;
	int nb_frames, bit_offset, i, pos, len;
	uint8_t *q;
	int16_t *samples;
	
	samples = (short *)data;

	init_get_bits(&s->gb, buf, buf_size*8);
	
	if(s->use_bit_reservoir)
	{
		/* read super frame header */
		get_bits(&s->gb, 4); /* super frame index */
		nb_frames = get_bits(&s->gb, 4) - 1;

		bit_offset = get_bits(&s->gb, s->byte_offset_bits + 3);

		if(s->last_superframe_len > 0)
		{
			// sprintf(g_cstr, "skip=%d", s->last_bitoffset);
			/* add bit_offset bits to last frame */
			if((s->last_superframe_len + ((bit_offset + 7) >> 3)) > 
				MAX_CODED_SUPERFRAME_SIZE)
				goto fail;
			q = s->last_superframe + s->last_superframe_len;
			len = bit_offset;
			while(len > 0)
			{
				*q++ = (get_bits)(&s->gb, 8);
				len -= 8;
			}
			if(len > 0)
			{
				*q++ = (get_bits)(&s->gb, len) << (8 - len);
			}
			
			/* XXX: bit_offset bits into last frame */
			init_get_bits(&s->gb, s->last_superframe, MAX_CODED_SUPERFRAME_SIZE*8);
			/* skip unused bits */
			if(s->last_bitoffset > 0)
				skip_bits(&s->gb, s->last_bitoffset);
			/* this frame is stored in the last superframe and in the
			   current one */
			if(wma_decode_frame(s, samples) < 0)
				goto fail;
			samples += s->nb_channels * s->frame_len;
		}

		/* read each frame starting from bit_offset */
		pos = bit_offset + 4 + 4 + s->byte_offset_bits + 3;
		init_get_bits(&s->gb, buf + (pos >> 3), (MAX_CODED_SUPERFRAME_SIZE - (pos >> 3))*8);
		len = pos & 7;
		if(len > 0)
			skip_bits(&s->gb, len);
	
		s->reset_block_lengths = 1;
		for(i=0;i<nb_frames;i++)
		{
			if(wma_decode_frame(s, samples) < 0)
				goto fail;
			samples += s->nb_channels * s->frame_len;
		}

		/* we copy the end of the frame in the last frame buffer */
		pos = get_bits_count(&s->gb) + ((bit_offset + 4 + 4 + s->byte_offset_bits + 3) & ~7);
		s->last_bitoffset = pos & 7;
		pos >>= 3;
		len = buf_size - pos;
		if(len > MAX_CODED_SUPERFRAME_SIZE || len < 0)
		{
			goto fail;
		}
		s->last_superframe_len = len;
		memcpy(s->last_superframe, buf + pos, len);
	}
	else
	{
		/* single frame decode */
		if(wma_decode_frame(s, samples) < 0)
			goto fail;
		samples += s->nb_channels * s->frame_len;
	}
	*data_size = (int8_t *)samples - (int8_t *)data;
	return s->block_align;
 fail:
	/* when error, we reset the bit reservoir */
	s->last_superframe_len = 0;
	return -1;
}

static int wma_decode_end(AVCodecContext *avctx)
{
	WMADecodeContext *s = (WMADecodeContext *)avctx->priv_data;
	int i;

	for(i = 0; i < s->nb_block_sizes; i++)
		ff_mdct_end(&s->mdct_ctx[i]);
	for(i = 0; i < s->nb_block_sizes; i++)
		av_free(s->windows[i]);

	if(s->use_exp_vlc)
	{
		free_vlc(&s->exp_vlc);
	}
	if(s->use_noise_coding)
	{
		free_vlc(&s->hgain_vlc);
	}
	for(i = 0;i < 2; i++)
	{
		free_vlc(&s->coef_vlc[i]);
		av_free(s->run_table[i]);
		av_free(s->level_table[i]);
	}
	
	return 0;
}

AVCodec g_wmav1_decoder =
{
	"wmav1",
	CODEC_TYPE_AUDIO,
	CODEC_ID_WMAV1,
	sizeof(WMADecodeContext),
	wma_decode_init,
	NULL,
	wma_decode_end,
	wma_decode_superframe,
};

AVCodec g_wmav2_decoder =
{
	"wmav2",
	CODEC_TYPE_AUDIO,
	CODEC_ID_WMAV2,
	sizeof(WMADecodeContext),
	wma_decode_init,
	NULL,
	wma_decode_end,
	wma_decode_superframe,
};

AVCodec g_wmav1_decoderB =
{
	"wmav1B",
	CODEC_TYPE_AUDIO,
	CODEC_ID_WMAV1,
	sizeof(WMADecodeContext),
	wma_decode_init,
	NULL,
	wma_decode_end,
	wma_decode_superframe,
};

AVCodec g_wmav2_decoderB =
{
	"wmav2B",
	CODEC_TYPE_AUDIO,
	CODEC_ID_WMAV2,
	sizeof(WMADecodeContext),
	wma_decode_init,
	NULL,
	wma_decode_end,
	wma_decode_superframe,
};


// mdct.c
/**
 * @file mdct.c
 * MDCT/IMDCT transforms.
 */

/**
 * init MDCT or IMDCT computation.
 */
int ff_mdct_init(MDCTContext *s, int nbits, int inverse)
{

	int n, n4, i;
	float alpha;

	memset(s, 0, sizeof(*s));
	n = 1 << nbits;
	s->nbits = nbits;
	s->n = n;
	n4 = n >> 2;
	s->tcos = (float *)av_malloc(n4 * sizeof(FFTSample));
	if(!s->tcos)
		goto fail;
	s->tsin = (float *)av_malloc(n4 * sizeof(FFTSample));
	if(!s->tsin)
		goto fail;

	for(i=0;i<n4;i++)
	{
		alpha = 2 * M_PI * (i + 1.0 / 8.0) / n;
		s->tcos[i] = -cos(alpha);
		s->tsin[i] = -sin(alpha);
	}

	if(wmafft_init(&s->fft, s->nbits - 2, inverse) < 0)
		goto fail;
	return 0;
	
 fail:
	av_freep(&s->tcos);
	av_freep(&s->tsin);
	return -1;
}

/* complex multiplication: p = a * b */
#define CMUL(pre, pim, are, aim, bre, bim) \
{\
	float _are = (are);\
	float _aim = (aim);\
	float _bre = (bre);\
	float _bim = (bim);\
	(pre) = _are * _bre - _aim * _bim;\
	(pim) = _are * _bim + _aim * _bre;\
}

/**
 * Compute inverse MDCT of size N = 2^nbits
 * @param output N samples
 * @param input N/2 samples
 * @param tmp N/2 samples
 */
void ff_imdct_calc(MDCTContext *s, FFTSample *output, 
				   const FFTSample *input, FFTSample *tmp)
{
	int k, n8, n4, n2, n, j;
	const uint16_t *revtab = s->fft.revtab;
	const FFTSample *tcos = s->tcos;
	const FFTSample *tsin = s->tsin;
	const FFTSample *in1, *in2;
	FFTComplex *z = (FFTComplex *)tmp;

	n = 1 << s->nbits;
	n2 = n >> 1;
	n4 = n >> 2;
	n8 = n >> 3;

	/* pre rotation */
	in1 = input;
	in2 = input + n2 - 1;
	for(k = 0; k < n4; k++)
	{
		j=revtab[k];
		CMUL(z[j].re, z[j].im, *in2, *in1, tcos[k], tsin[k]);
		in1 += 2;
		in2 -= 2;
	}
	fft_calc(&s->fft, z);

	/* post rotation + reordering */
	/* XXX: optimize */
	for(k = 0; k < n4; k++)
	{
		CMUL(z[k].re, z[k].im, z[k].re, z[k].im, tcos[k], tsin[k]);
	}
	for(k = 0; k < n8; k++)
	{
		output[2*k] = -z[n8 + k].im;
		output[n2-1-2*k] = z[n8 + k].im;

		output[2*k+1] = z[n8-1-k].re;
		output[n2-1-2*k-1] = -z[n8-1-k].re;

		output[n2 + 2*k]=-z[k+n8].re;
		output[n-1- 2*k]=-z[k+n8].re;

		output[n2 + 2*k+1]=z[n8-k-1].im;
		output[n-2 - 2 * k] = z[n8-k-1].im;
	}
}

/**
 * Compute MDCT of size N = 2^nbits
 * @param input N samples
 * @param out N/2 samples
 * @param tmp temporary storage of N/2 samples
 */
void ff_mdct_calc(MDCTContext *s, FFTSample *out, 
				  const FFTSample *input, FFTSample *tmp)
{
	int i, j, n, n8, n4, n2, n3;
	FFTSample re, im, re1, im1;
	const uint16_t *revtab = s->fft.revtab;
	const FFTSample *tcos = s->tcos;
	const FFTSample *tsin = s->tsin;
	FFTComplex *x = (FFTComplex *)tmp;

	n = 1 << s->nbits;
	n2 = n >> 1;
	n4 = n >> 2;
	n8 = n >> 3;
	n3 = 3 * n4;

	/* pre rotation */
	for(i=0;i<n8;i++)
	{
		re = -input[2*i+3*n4] - input[n3-1-2*i];
		im = -input[n4+2*i] + input[n4-1-2*i];
		j = revtab[i];
		CMUL(x[j].re, x[j].im, re, im, -tcos[i], tsin[i]);

		re = input[2*i] - input[n2-1-2*i];
		im = -(input[n2+2*i] + input[n-1-2*i]);
		j = revtab[n8 + i];
		CMUL(x[j].re, x[j].im, re, im, -tcos[n8 + i], tsin[n8 + i]);
	}

	fft_calc(&s->fft, x);
  
	/* post rotation */
	for(i=0;i<n4;i++)
	{
		re = x[i].re;
		im = x[i].im;
		CMUL(re1, im1, re, im, -tsin[i], -tcos[i]);
		out[2*i] = im1;
		out[n2-1-2*i] = re1;
	}
}

void ff_mdct_end(MDCTContext *s)
{
	av_freep(&s->tcos);
	av_freep(&s->tsin);
	fft_end(&s->fft);
}

// fft.c
/**
 * @file fft.c
 * FFT/IFFT transforms.
 */

/**
 * The size of the FFT is 2^nbits. If inverse is TRUE, inverse FFT is done 
 */
int wmafft_init(FFTContext *s, int nbits, int inverse)
{
	int i, j, m, n;
	float alpha, c1, s1, s2;
	
	s->nbits = nbits;
	n = 1 << nbits;

	s->exptab = (FFTComplex *)av_malloc((n / 2) * sizeof(FFTComplex));
	if(!s->exptab)
		goto fail;
	s->revtab = (unsigned short *)av_malloc(n * sizeof(uint16_t));
	if(!s->revtab)
		goto fail;
	s->inverse = inverse;

	s2 = inverse ? 1.0 : -1.0;
		
	for(i=0;i<(n/2);i++)
	{
		alpha = 2 * M_PI * (float)i / (float)n;
		c1 = cos(alpha);
		s1 = sin(alpha) * s2;
		s->exptab[i].re = c1;
		s->exptab[i].im = s1;
	}
	s->fft_calc = fft_calc_c;
	s->exptab1 = NULL;

	/* compute bit reverse table */

	for(i=0;i<n;i++)
	{
		m=0;
		for(j=0;j<nbits;j++)
		{
			m |= ((i >> j) & 1) << (nbits-j-1);
		}
		s->revtab[i]=m;
	}
	return 0;

fail:
	av_freep(&s->revtab);
	av_freep(&s->exptab);
	av_freep(&s->exptab1);
	return -1;
}


/* butterfly op */
#undef BF
#define BF(pre, pim, qre, qim, pre1, pim1, qre1, qim1) \
{\
  FFTSample ax, ay, bx, by;\
  bx=pre1;\
  by=pim1;\
  ax=qre1;\
  ay=qim1;\
  pre = (bx + ax);\
  pim = (by + ay);\
  qre = (bx - ax);\
  qim = (by - ay);\
}

#define MUL16(a,b) ((a) * (b))

#undef CMUL
#define CMUL(pre, pim, are, aim, bre, bim) \
{\
   pre = (MUL16(are, bre) - MUL16(aim, bim));\
   pim = (MUL16(are, bim) + MUL16(bre, aim));\
}

/**
 * Do a complex FFT with the parameters defined in wmafft_init(). The
 * input data must be permuted before with s->revtab table. No
 * 1.0/sqrt(n) normalization is done.  
 */
void fft_calc_c(FFTContext *s, FFTComplex *z)
{
	int ln = s->nbits;
	int	j, np, np2;
	int	nblocks, nloops;
	register FFTComplex *p, *q;
	FFTComplex *exptab = s->exptab;
	int l;
	FFTSample tmp_re, tmp_im;

	np = 1 << ln;

	/* pass 0 */

	p=&z[0];
	j=(np >> 1);
	do
	{
		BF(p[0].re, p[0].im, p[1].re, p[1].im, 
		   p[0].re, p[0].im, p[1].re, p[1].im);
		p+=2;
	} while(--j != 0);

	/* pass 1 */

	
	p=&z[0];
	j=np >> 2;
	if(s->inverse)
	{
		do
		{
			BF(p[0].re, p[0].im, p[2].re, p[2].im, 
			   p[0].re, p[0].im, p[2].re, p[2].im);
			BF(p[1].re, p[1].im, p[3].re, p[3].im, 
			   p[1].re, p[1].im, -p[3].im, p[3].re);
			p+=4;
		} while(--j != 0);
	}
	else
	{
		do
		{
			BF(p[0].re, p[0].im, p[2].re, p[2].im, 
			   p[0].re, p[0].im, p[2].re, p[2].im);
			BF(p[1].re, p[1].im, p[3].re, p[3].im, 
			   p[1].re, p[1].im, p[3].im, -p[3].re);
			p+=4;
		} while(--j != 0);
	}
	/* pass 2 .. ln-1 */

	nblocks = np >> 3;
	nloops = 1 << 2;
	np2 = np >> 1;
	do
	{
		p = z;
		q = z + nloops;
		for(j = 0; j < nblocks; ++j)
		{
			BF(p->re, p->im, q->re, q->im,
			   p->re, p->im, q->re, q->im);
			
			p++;
			q++;
			for(l = nblocks; l < np2; l += nblocks)
			{
				CMUL(tmp_re, tmp_im, exptab[l].re, exptab[l].im, q->re, q->im);
				BF(p->re, p->im, q->re, q->im,
				   p->re, p->im, tmp_re, tmp_im);
				p++;
				q++;
			}

			p += nloops;
			q += nloops;
		}
		nblocks = nblocks >> 1;
		nloops = nloops << 1;
	} while(nblocks != 0);
}

void fft_end(FFTContext *s)
{
	av_freep(&s->revtab);
	av_freep(&s->exptab);
	av_freep(&s->exptab1);
}

// fft_altivec.c
/*
 * FFT/IFFT transforms
 * AltiVec-enabled
 * Copyright (c) 2003 Romain Dolbeau <romain@dolbeau.org>
 * Based on code Copyright (c) 2002 Fabrice Bellard.
 */


/*
  those three macros are from libavcodec/fft.c and are required for the reference C code
*/
/* butter fly op */
#undef BF
#define BF(pre, pim, qre, qim, pre1, pim1, qre1, qim1) \
{\
  FFTSample ax, ay, bx, by;\
  bx=pre1;\
  by=pim1;\
  ax=qre1;\
  ay=qim1;\
  pre = (bx + ax);\
  pim = (by + ay);\
  qre = (bx - ax);\
  qim = (by - ay);\
}
#define MUL16(a,b) ((a) * (b))
#undef CMUL
#define CMUL(pre, pim, are, aim, bre, bim) \
{\
   pre = (MUL16(are, bre) - MUL16(aim, bim));\
   pim = (MUL16(are, bim) + MUL16(bre, aim));\
}


// libvformat:
// utils.c

AVInputFormat *first_iformat;

void av_register_input_format(AVInputFormat *format)
{
	AVInputFormat **p;
	p = &first_iformat;
	while(*p != NULL) p = &(*p)->next;
	*p = format;
	format->next = NULL;
}


int match_ext(const char *filename, const char *extensions);
int match_ext(const char *filename, const char *extensions)
{
	const char *ext, *p;
	char ext1[32], *q;

	ext = strrchr(filename, '.');
	if(ext)
	{
		ext++;
		p = extensions;
		for(;;)
		{
			q = ext1;
			while(*p != '\0' && *p != ',') 
				*q++ = *p++;
			*q = '\0';			
			if(!mastrcasecmp(ext1, (char *)ext)) 
				return 1;
			if(*p == '\0') 
				break;
			p++;
		}
	}
	return 0;
}

/* memory handling */

/**
 * Default packet destructor 
 */
static void av_destruct_packet(AVPacket *pkt)
{
	av_free(pkt->data);
	pkt->data = NULL; pkt->size = 0;
}

/**
 * Allocate the payload of a packet and intialized its fields to default values.
 *
 * @param pkt packet
 * @param size wanted payload size
 * @return 0 if OK. AVERROR_xxx otherwise.
 */
int av_new_packet(AVPacket *pkt, int size)
{
	void *data = av_malloc(size + FF_INPUT_BUFFER_PADDING_SIZE);
	if(!data)
		return AVERROR_NOMEM;
	memset((char *)data + size, 0, FF_INPUT_BUFFER_PADDING_SIZE);

	av_init_packet(pkt);
	pkt->data = (unsigned char *)data; 
	pkt->size = size;
	pkt->destruct = av_destruct_packet;
	return 0;
}

/* guess file format */
AVInputFormat *av_probe_input_format(AVProbeData *pd, int is_opened)
{
	AVInputFormat *fmt1, *fmt;
	int score, score_max;

	fmt = NULL;
	score_max = 0;
	for(fmt1 = first_iformat; fmt1 != NULL; fmt1 = fmt1->next)
	{
		if(!is_opened && !(fmt1->flags & AVFMT_NOFILE))
			continue;
		score = 0;
		if(fmt1->read_probe)
		{
			score = fmt1->read_probe(pd);
		}
		else if(fmt1->extensions)
		{
			if(match_ext(pd->filename, fmt1->extensions))
			{
				score = 50;
			}
		} 

		if(score > score_max)
		{
			score_max = score;
			fmt = fmt1;
		}
	}
	return fmt;
}

/************************************************************/
/* input media file */

#define PROBE_BUF_SIZE 2048

/**
 * Open a media file as input. The codec are not opened. Only the file
 * header (if present) is read.
 *
 * @param ic_ptr the opened media file handle is put here
 * @param filename filename to open.
 * @param fmt if non NULL, force the file format to use
 * @param buf_size optional buffer size (zero if default is OK)
 * @param ap additionnal parameters needed when opening the file (NULL if default)
 * @return 0 if OK. AVERROR_xxx otherwise.
 */
int av_open_input_file(AVFormatContext **ic_ptr, const char *filename, 
					   AVInputFormat *fmt,
					   int buf_size,
					   AVFormatParameters *ap, int nomem)
{
	AVFormatContext *ic = NULL;
	int err, must_open_file;
	char buf[PROBE_BUF_SIZE];
	AVProbeData probe_data, *pd = &probe_data;

	ic = (AVFormatContext *)av_mallocz(sizeof(AVFormatContext));
	if(!ic)
	{
		err = AVERROR_NOMEM;
		goto fail;
	}
	ic->duration = AV_NOPTS_VALUE;
	ic->start_time = AV_NOPTS_VALUE;
	pstrcpy(ic->filename, sizeof(ic->filename), filename);
	pd->filename = ic->filename;
	pd->buf = (unsigned char *)buf;
	pd->buf_size = 0;

	if(!fmt)
	{
		/* guess format if no file can be opened  */
		fmt = av_probe_input_format(pd, 0);
	}

	/* do not open file if the format does not need it. XXX: specific
	   hack needed to handle RTSP/TCP */
	must_open_file = 1;
	if((fmt && (fmt->flags & AVFMT_NOFILE)))
	{
		must_open_file = 0;
	}

	if(!fmt || must_open_file)
	{
		/* if no file needed do not try to open one */
		if(url_fopen(&ic->pb, filename, URL_RDONLY, nomem) < 0)
		{
			err = AVERROR_IO;
			goto fail;
		}
		if(buf_size > 0)
		{
			url_setbufsize(&ic->pb, buf_size);
		}
		if(!fmt)
		{
			/* read probe data */
			pd->buf_size = get_buffer(&ic->pb, (unsigned char *)buf, PROBE_BUF_SIZE, nomem);
			url_fseek(&ic->pb, 0, SEEK_SET, nomem);
		}
	}
	
	/* guess file format */
	if(!fmt)
	{
		fmt = av_probe_input_format(pd, 1);
	}

	/* if still no format found, error */
	if(!fmt)
	{
		err = AVERROR_NOFMT;
		goto fail1;
	}
		
	ic->iformat = fmt;

	/* allocate private data */
	if(fmt->priv_data_size > 0)
	{
		ic->priv_data = av_mallocz(fmt->priv_data_size);
		if(!ic->priv_data)
		{
			err = AVERROR_NOMEM;
		goto fail1;
		}
	} else
		ic->priv_data = NULL;

	/* default pts settings is MPEG like */
	av_set_pts_info(ic, 33, 1, 90000);

	err = ic->iformat->read_header(ic, ap, nomem);
	if(err < 0)
		goto fail1;
	*ic_ptr = ic;
	return 0;
 fail1:
	if(!fmt || must_open_file)
	{
		url_fclose(&ic->pb, nomem);
	}
 fail:
	if(ic)
	{
		av_freep(&ic->priv_data);
	}
	av_free(ic);
	*ic_ptr = NULL;
	return err;
}

/**
 * Read a packet from a media file
 * @param s media file handle
 * @param pkt is filled 
 * @return 0 if OK. AVERROR_xxx if error.
 */
int av_read_packet(AVFormatContext *s, AVPacket *pkt, int nomem)
{
	AVPacketList *pktl;

	pktl = s->packet_buffer;
	if(pktl)
	{
		/* read packet from packet buffer, if there is data */
		*pkt = pktl->pkt;
		s->packet_buffer = pktl->next;
		av_free(pktl);
		return 0;
	}
	else
	{
		return s->iformat->read_packet(s, pkt, nomem);
	}
}


/* return TRUE if the stream has accurate timings for at least one component */
static int av_has_timings(AVFormatContext *ic)
{
	AVStream *st;

		st = ic->stream;
		if(st->start_time != (int64_t)AV_NOPTS_VALUE && st->duration != (int64_t)AV_NOPTS_VALUE)
			return 1;

	return 0;
}

/* estimate the stream timings from the one of each components. Also
   compute the global bitrate if possible */
static void av_update_stream_timings(AVFormatContext *ic)
{
	int64_t start_time, end_time, end_time1;
	AVStream *st;

	start_time = MAXINT64;
	end_time = MININT64;
		st = ic->stream;
		if(st->start_time != (int64_t)AV_NOPTS_VALUE)
		{
			if(st->start_time < start_time)
				start_time = st->start_time;
			if(st->duration != (int64_t)AV_NOPTS_VALUE)
			{
				end_time1 = st->start_time + st->duration;
				if(end_time1 > end_time)
					end_time = end_time1;
			}
		}
	
	if(start_time != MAXINT64)
	{
		ic->start_time = start_time;
		if(end_time != MAXINT64)
		{
			ic->duration = end_time - start_time;
			if(ic->file_size > 0)
			{
				/* compute the bit rate */
				ic->bit_rate = (int64_t)((double)ic->file_size * 8.0 * AV_TIME_BASE / ic->duration);
			}
		}
	}

}

static void fill_all_stream_timings(AVFormatContext *ic)
{
	AVStream *st;

	av_update_stream_timings(ic);
		st = ic->stream;
		if(st->start_time == (int64_t)AV_NOPTS_VALUE)
		{
			st->start_time = ic->start_time;
			st->duration = ic->duration;
		}
}

static void av_estimate_timings_from_bit_rate(AVFormatContext *ic)
{
	int64_t filesize, duration;
	int bit_rate;
	AVStream *st;

	/* if bit_rate is already set, we believe it */
	if(ic->bit_rate == 0)
	{
		bit_rate = 0;
			st = ic->stream;
			bit_rate += st->codec.bit_rate;

		ic->bit_rate = bit_rate;
	}

	/* if duration is already set, we believe it */
	if(ic->duration == (int64_t)AV_NOPTS_VALUE && ic->bit_rate != 0 && ic->file_size != 0)
	{
		filesize = ic->file_size;
		if(filesize > 0)
		{
			duration = (int64_t)((8 * AV_TIME_BASE * (double)filesize) / (double)ic->bit_rate);

				st = ic->stream;
				if(st->start_time == (int64_t)AV_NOPTS_VALUE || st->duration == (int64_t)AV_NOPTS_VALUE)
				{
					st->start_time = 0;
					st->duration = duration;
				}
			}
		}
}

static void av_estimate_timings(AVFormatContext *ic, int nomem)
{
	URLContext *h;
	int64_t file_size;

	/* get the file size, if possible */
	if(ic->iformat->flags & AVFMT_NOFILE)
	{
		file_size = 0;
	}
	else
	{
		h = url_fileno(&ic->pb);
		file_size = url_filesize(h, nomem);
		if(file_size < 0)
			file_size = 0;
	}
	ic->file_size = file_size;

	if(av_has_timings(ic))
	{
		/* at least one components has timings - we use them for all the components */
		fill_all_stream_timings(ic);
	}
	else
	{
		/* less precise: use bit rate info */
		av_estimate_timings_from_bit_rate(ic);
	}
	av_update_stream_timings(ic);
}

/* state for codec information */
#define CSTATE_NOTFOUND	0
#define CSTATE_DECODING	1
#define CSTATE_FOUND	   2

static int has_codec_parameters(AVCodecContext *enc)
{
	int val;
	switch(enc->codec_type)
	{
		case CODEC_TYPE_AUDIO:
			val = enc->sample_rate;
		break;
		default:
			val = 1;
		break;
	}
	return (val != 0);
}

/**
 * Read the beginning of a media file to get stream information. This
 * is useful for file formats with no headers such as MPEG.
 *
 * @param ic media file handle
 * @return >=0 if OK. AVERROR_xxx if error.  
 */
int av_find_stream_info(AVFormatContext *ic, int nomem)
{
	int count, ret, size, read_size;
	int got_picture = 0;
	AVCodec *codec;
	AVStream *st;
	AVPacket *pkt;
	AVPacketList *pktl=NULL, **ppktl;
	short samples[AVCODEC_MAX_AUDIO_FRAME_SIZE / 2];
	uint8_t *ptr;
	int min_read_size, max_read_size;

	min_read_size = 250000;

	/* max read size is 2 seconds of video max */
	max_read_size = min_read_size * 10;

	/* set initial codec state */
	st = ic->stream;
	if(has_codec_parameters(&st->codec))
		st->codec_info_state = CSTATE_FOUND;
	else
		st->codec_info_state = CSTATE_NOTFOUND;

	count = 0;
	read_size = 0;
	ppktl = &ic->packet_buffer;
	for(;;)
	{
		/* NOTE: if the format has no header, then we need to read
		some packets to get most of the streams, so we cannot
		stop here */
		if(!(ic->iformat->flags & AVFMT_NOHEADER) || read_size >= min_read_size)
		{
			/* if we found the info for all the codecs, we can stop */
			ret = count;
			break;
		}
		pktl = (AVPacketList *)av_mallocz(sizeof(AVPacketList));
		if(!pktl)
		{
			ret = AVERROR_NOMEM;
			break;
		}

		/* add the packet in the buffered packet list */
		*ppktl = pktl;
		ppktl = &pktl->next;

		/* NOTE: a new stream can be added there if no header in file (AVFMT_NOHEADER) */
		pkt = &pktl->pkt;
		if(ic->iformat->read_packet(ic, pkt, nomem) < 0)
		{
			/* EOF or error */
			ret = -1; /* we could not have all the codec parameters before EOF */
			if((ic->iformat->flags & AVFMT_NOHEADER))
				ret = 0;
			break;
		}
		read_size += pkt->size;

		/* open new codecs */
			st = ic->stream;
			if(st->codec_info_state == CSTATE_NOTFOUND)
			{
				/* set to found in case of error */
				st->codec_info_state = CSTATE_FOUND; 
				codec = avcodec_find_decoder(st->codec.codec_id);
				if(codec)
				{
					/*
					if(codec->capabilities & CODEC_CAP_TRUNCATED)
						st->codec.flags |= CODEC_FLAG_TRUNCATED;
					*/
					ret = avcodec_open(&st->codec, codec);
					if(ret >= 0)
						st->codec_info_state = CSTATE_DECODING;
				}
			}

		st = ic->stream;
		if(st->codec_info_state == CSTATE_DECODING)
		{
			/* decode the data and update codec parameters */
			ptr = pkt->data;
			size = pkt->size;
			while(size > 0)
			{
				switch(st->codec.codec_type)
				{
					case CODEC_TYPE_AUDIO:
						ret = avcodec_decode_audio(&st->codec, samples, &got_picture, ptr, size);
					break;
					default:
						ret = -1;
					break;
				}
				if(ret < 0)
				{
					/* if error, simply ignore because another packet may be OK */
					break;
				}
				if(got_picture)
				{
					{
						//n close_codec:
						st->codec_info_state = CSTATE_FOUND;
						avcodec_close(&st->codec);
						break;
					}
				}
				ptr += ret;
				size -= ret;
			}
		}
		count++;
	}

	/* close each codec if there are opened */
	st = ic->stream;
	if(st->codec_info_state == CSTATE_DECODING)
		avcodec_close(&st->codec);

	av_estimate_timings(ic, nomem);
	return ret;
}

/**
 * Close a media file (but not its codecs)
 *
 * @param s media file handle
 */
void av_close_input_file(AVFormatContext *s, int nomem)
{
	int must_open_file;

	if(s->iformat->read_close)
		s->iformat->read_close(s);
		av_free(s->stream);
	
	if(s->packet_buffer)
	{
		AVPacketList *p, *p1;
		p = s->packet_buffer;
		while(p != NULL)
		{
			p1 = p->next;
			av_free_packet(&p->pkt);
			av_free(p);
			p = p1;
		}
		s->packet_buffer = NULL;
	}
	must_open_file = 1;
	if((s->iformat->flags & AVFMT_NOFILE))
	{
		must_open_file = 0;
	}
	if(must_open_file)
	{
		url_fclose(&s->pb, nomem);
	}
	av_freep(&s->priv_data);
	av_free(s);
}

/**
 * Add a new stream to a media file. Can only be called in the
 * read_header function. If the flag AVFMT_NOHEADER is in the format
 * description, then new streams can be added in read_packet too.
 *
 *
 * @param s media file handle
 * @param id file format dependent stream id
 */
AVStream *av_new_stream(AVFormatContext *s, int id) // clobbers any previous, s.b. only one anyway...
{
	AVStream *st;

	st = (AVStream *)av_mallocz(sizeof(AVStream));
	if(!st)
		return NULL;
	avcodec_get_context_defaults(&st->codec);
	if(s->iformat)
	{
		/* no default bitrate if decoding */
		st->codec.bit_rate = 0;
	}
	st->index = 0;
	st->id = id;
	st->start_time = AV_NOPTS_VALUE;
	st->duration = AV_NOPTS_VALUE;
	s->stream = st;
	return st;
}

/************************************************************/
/* output media file */

/**
 * Set the pts for a given stream
 * @param s stream 
 * @param pts_wrap_bits number of bits effectively used by the pts
 *		(used for wrap control, 33 is the value for MPEG) 
 * @param pts_num numerator to convert to seconds (MPEG: 1) 
 * @param pts_den denominator to convert to seconds (MPEG: 90000)
 */
void av_set_pts_info(AVFormatContext *s, int pts_wrap_bits,
					 int pts_num, int pts_den)
{
	s->pts_wrap_bits = pts_wrap_bits;
	s->pts_num = pts_num;
	s->pts_den = pts_den;
}

/* fraction handling */

/**
 * f = val + (num / den) + 0.5. 'num' is normalized so that it is such
 * as 0 <= num < den.
 *
 * @param f fractional number
 * @param val integer value
 * @param num must be >= 0
 * @param den must be >= 1 
 */
void av_frac_init(AVFrac *f, int64_t val, int64_t num, int64_t den)
{
	num += (den >> 1);
	if(num >= den)
	{
		val += num / den;
		num = num % den;
	}
	f->val = val;
	f->num = num;
	f->den = den;
}

/**
 * Fractionnal addition to f: f = f + (incr / f->den)
 *
 * @param f fractional number
 * @param incr increment, can be positive or negative
 */
void av_frac_add(AVFrac *f, int64_t incr)
{
	int64_t num, den;

	num = f->num + incr;
	den = f->den;
	if(num < 0)
	{
		f->val += num / den;
		num = num % den;
		if(num < 0)
		{
			num += den;
			f->val--;
		}
	} else if(num >= den)
	{
		f->val += num / den;
		num = num % den;
	}
	f->num = num;
}

// asf.c
#define PACKET_SIZE 3200
#define FRAME_HEADER_SIZE 17

typedef struct
{
	int num;
	int seq;
	/* use for reading */
	AVPacket pkt;
	int frag_offset;
	int timestamp;
	int64_t duration;

	int ds_span;		/* descrambling  */
	int ds_packet_size;
	int ds_chunk_size;
	int ds_data_size;
	int ds_silence_data;

} ASFStream;

typedef struct
{
	uint32_t v1;
	uint16_t v2;
	uint16_t v3;
	uint8_t v4[8];
} WMAGUID;

typedef struct
{
	WMAGUID guid;			// generated by client computer
	uint64_t file_size;		// in bytes
								// invalid if broadcasting
	uint64_t create_time;	// time of creation, in 100-nanosecond units since 1.1.1601
								// invalid if broadcasting
	uint64_t packets_count;	// how many packets are there in the file
								// invalid if broadcasting
	uint64_t play_time;		// play time, in 100-nanosecond units
								// invalid if broadcasting
	uint64_t send_time;		// time to send file, in 100-nanosecond units
								// invalid if broadcasting (could be ignored)
	uint32_t preroll;		// timestamp of the first packet, in milliseconds
					// if nonzero - substract from time
	uint32_t ignore;			// preroll is 64bit - but let's just ignore it
	uint32_t flags;		// 0x01 - broadcast
					// 0x02 - seekable
								// rest is reserved should be 0
	uint32_t min_pktsize;	// size of a data packet
								// invalid if broadcasting
	uint32_t max_pktsize;	// shall be the same as for min_pktsize
								// invalid if broadcasting
	uint32_t max_bitrate;	// bandwith of stream in bps
					// should be the sum of bitrates of the
								// individual media streams
} ASFMainHeader;


typedef struct
{
	int packet_size;
	int asfid2avid[128];		/* conversion table from asf ID 2 AVStream ID */
	ASFStream streams[1]; //n 128];	/* it's max number and it's not that big */
	/* non streamed additonnal info */
	int64_t nb_packets;
	/* packet filling */
	int packet_size_left;
	uint8_t packet_buf[PACKET_SIZE];
	ByteIOContext pb;
	/* only for reading */
	uint64_t data_offset; /* begining of the first data packet */

	ASFMainHeader hdr;

	int packet_flags;
	int packet_property;
	int packet_timestamp;
	int packet_segsizetype;
	int packet_segments;
	int packet_seq;
	int packet_replic_size;
	int packet_key_frame;
	int packet_padsize;
	int packet_frag_offset;
	int packet_frag_size;
	int packet_frag_timestamp;
	int packet_multi_size;
	int packet_obj_size;
	int packet_time_delta;
	int packet_time_start;

	int stream_index;
	ASFStream* asf_st; /* currently decoded stream */
} ASFContext;

static const WMAGUID asf_header = {
	0x75B22630, 0x668E, 0x11CF, { 0xA6, 0xD9, 0x00, 0xAA, 0x00, 0x62, 0xCE, 0x6C },
};

static const WMAGUID file_header = {
	0x8CABDCA1, 0xA947, 0x11CF, { 0x8E, 0xE4, 0x00, 0xC0, 0x0C, 0x20, 0x53, 0x65 },
};

static const WMAGUID stream_header = {
	0xB7DC0791, 0xA9B7, 0x11CF, { 0x8E, 0xE6, 0x00, 0xC0, 0x0C, 0x20, 0x53, 0x65 },
};

static const WMAGUID audio_stream = {
	0xF8699E40, 0x5B4D, 0x11CF, { 0xA8, 0xFD, 0x00, 0x80, 0x5F, 0x5C, 0x44, 0x2B },
};

static const WMAGUID comment_header = {
	0x75b22633, 0x668e, 0x11cf, { 0xa6, 0xd9, 0x00, 0xaa, 0x00, 0x62, 0xce, 0x6c },
};

static const WMAGUID data_header = {
	0x75b22636, 0x668e, 0x11cf, { 0xa6, 0xd9, 0x00, 0xaa, 0x00, 0x62, 0xce, 0x6c },
};

/**********************************/
/* decoding */

static void get_guid(ByteIOContext *s, WMAGUID *g, int nomem)
{
	int i;

	g->v1 = get_le32(s, nomem);
	g->v2 = get_le16(s, nomem);
	g->v3 = get_le16(s, nomem);
	for(i=0;i<8;i++)
		g->v4[i] = get_byte(s, nomem);
}

static void get_str16_nolen(ByteIOContext *pb, int len, char *buf, int buf_size, int nomem)
{
	int c;
	char *q;

	q = buf;
	while(len > 0)
	{
		c = get_le16(pb, nomem);
		if((q - buf) < buf_size - 1)
			*q++ = c;
		len-=2;
	}
	*q = '\0';
}

static int asf_probe(AVProbeData *pd)
{
	WMAGUID g;
	const unsigned char *p;
	int i;

	/* check file header */
	if(pd->buf_size <= 32)
		return 0;
	p = pd->buf;
	g.v1 = p[0] | (p[1] << 8) | (p[2] << 16) | (p[3] << 24);
	p += 4;
	g.v2 = p[0] | (p[1] << 8);
	p += 2;
	g.v3 = p[0] | (p[1] << 8);
	p += 2;
	for(i=0;i<8;i++)
		g.v4[i] = *p++;

	if(!memcmp(&g, &asf_header, sizeof(WMAGUID)))
		return AVPROBE_SCORE_MAX;
	else
		return 0;
}

static int asf_read_header(AVFormatContext *s, AVFormatParameters *ap, int nomem)
{
#ifdef MAUNUSED
 (void)ap;
#endif
	
	ASFContext *asf = (ASFContext *)s->priv_data;
	WMAGUID g;
	ByteIOContext *pb = &s->pb;
	AVStream *st;
	ASFStream *asf_st;
	int64_t gsize;

	av_set_pts_info(s, 32, 1, 1000); /* 32 bit pts in ms */

	get_guid(pb, &g, nomem);
	if(memcmp(&g, &asf_header, sizeof(WMAGUID)))
		goto fail;
	get_le64(pb, nomem);
	get_le32(pb, nomem);
	get_byte(pb, nomem);
	get_byte(pb, nomem);
	memset(&asf->asfid2avid, -1, sizeof(asf->asfid2avid));
	for(;;)
	{
		get_guid(pb, &g, nomem);
		gsize = get_le64(pb, nomem);
		if(gsize < 24)
			goto fail;
		if(!memcmp(&g, &file_header, sizeof(WMAGUID)))
		{
			get_guid(pb, &asf->hdr.guid, nomem);
			asf->hdr.file_size		= get_le64(pb, nomem);
			asf->hdr.create_time	= get_le64(pb, nomem);
			asf->hdr.packets_count	= get_le64(pb, nomem);
			asf->hdr.play_time		= get_le64(pb, nomem);
			asf->hdr.send_time		= get_le64(pb, nomem);
			asf->hdr.preroll		= get_le32(pb, nomem);
			asf->hdr.ignore			= get_le32(pb, nomem);
			asf->hdr.flags			= get_le32(pb, nomem);
			asf->hdr.min_pktsize	= get_le32(pb, nomem);
			asf->hdr.max_pktsize	= get_le32(pb, nomem);
			asf->hdr.max_bitrate	= get_le32(pb, nomem);
			asf->packet_size = asf->hdr.max_pktsize;
			asf->nb_packets = asf->hdr.packets_count;
		}
		else if(!memcmp(&g, &stream_header, sizeof(WMAGUID)))
		{
			int type, total_size, type_specific_size;
			int64_t pos1, pos2;

			pos1 = url_ftell(pb, nomem);

			st = av_new_stream(s, 0);
			if(!st)
				goto fail;
			asf_st = (ASFStream *)av_mallocz(sizeof(ASFStream));
			if(!asf_st)
				goto fail;
			st->priv_data = asf_st;
			st->start_time = asf->hdr.preroll / (10000000 / AV_TIME_BASE);
			st->duration = (asf->hdr.send_time - asf->hdr.preroll) / (10000000 / AV_TIME_BASE);
			get_guid(pb, &g, nomem);
			if(!memcmp(&g, &audio_stream, sizeof(WMAGUID)))
			{
				type = CODEC_TYPE_AUDIO;
			}
			else
			{
				goto fail;
			}
			get_guid(pb, &g, nomem);
			total_size = get_le64(pb, nomem);
			type_specific_size = get_le32(pb, nomem);
			get_le32(pb, nomem);
			st->id = get_le16(pb, nomem) & 0x7f; /* stream id */
			// mapping of asf ID to AV stream ID;
			asf->asfid2avid[st->id] = 0;

			get_le32(pb, nomem);
			st->codec.codec_type = CODECTYPECAST type;
			st->codec.frame_rate = 15 * s->pts_den / s->pts_num; // 15 fps default
			if(type == CODEC_TYPE_AUDIO)
			{
				get_wav_header(pb, &st->codec, type_specific_size, nomem);
				/* We have to init the frame size at some point .... */
				pos2 = url_ftell(pb, nomem);
				if(gsize > (pos2 + 8 - pos1 + 24))
				{
					asf_st->ds_span = get_byte(pb, nomem);
					asf_st->ds_packet_size = get_le16(pb, nomem);
					asf_st->ds_chunk_size = get_le16(pb, nomem);
					asf_st->ds_data_size = get_le16(pb, nomem);
					asf_st->ds_silence_data = get_byte(pb, nomem);
				}
				//sprintf(g_cstr, "Descrambling: ps:%d cs:%d ds:%d s:%d  sd:%d",
				//	   asf_st->ds_packet_size, asf_st->ds_chunk_size,
				//	   asf_st->ds_data_size, asf_st->ds_span, asf_st->ds_silence_data);
				if(asf_st->ds_span > 1)
				{
					if(!asf_st->ds_chunk_size || (asf_st->ds_packet_size/asf_st->ds_chunk_size <= 1))
					asf_st->ds_span = 0; // disable descrambling
				}

				/*
				switch (st->codec.codec_id)
				{
					case CODEC_ID_PCM_S16LE:
					case CODEC_ID_PCM_S16BE:
					case CODEC_ID_PCM_U16LE:
					case CODEC_ID_PCM_U16BE:
					case CODEC_ID_PCM_S8:
					case CODEC_ID_PCM_U8:
					case CODEC_ID_PCM_ALAW:
					case CODEC_ID_PCM_MULAW:
						st->codec.frame_size = 1;
					break;
					default:
				*/
						/* This is probably wrong, but it prevents a crash later */
						st->codec.frame_size = 1;
				/*
					break;
				}
				*/
			}
			else
			{
			}
			pos2 = url_ftell(pb, nomem);
			url_fskip(pb, gsize - (pos2 - pos1 + 24), nomem);
		} else if(!memcmp(&g, &data_header, sizeof(WMAGUID)))
		{
			break;
		} else if(!memcmp(&g, &comment_header, sizeof(WMAGUID)))
		{
			int len1, len2, len3, len4, len5;

			len1 = get_le16(pb, nomem);
			len2 = get_le16(pb, nomem);
			len3 = get_le16(pb, nomem);
			len4 = get_le16(pb, nomem);
			len5 = get_le16(pb, nomem);
			get_str16_nolen(pb, len1, s->title, sizeof(s->title), nomem);
			get_str16_nolen(pb, len2, s->author, sizeof(s->author), nomem);
			get_str16_nolen(pb, len3, s->copyright, sizeof(s->copyright), nomem);
			get_str16_nolen(pb, len4, s->comment, sizeof(s->comment), nomem);
			url_fskip(pb, len5, nomem);
		}
		else if(url_feof(pb))
		{
			goto fail;
		}
		else
		{
			url_fseek(pb, gsize - 24, SEEK_CUR, nomem);
		}
	}
	get_guid(pb, &g, nomem);
	get_le64(pb, nomem);
	get_byte(pb, nomem);
	get_byte(pb, nomem);
	if(url_feof(pb))
		goto fail;
	asf->data_offset = url_ftell(pb, nomem);
	asf->packet_size_left = 0;

	return 0;

 fail:
	{
		AVStream *st = s->stream;
		
		if(st)
		{
			av_free(st->priv_data);
			av_free(st->codec.extradata);
		}
		av_free(st);
	}
	return -1;
}

#define DO_2BITS(bits, var, defval) \
	switch (bits & 3) \
	{ \
	case 3: var = get_le32(pb, nomem); rsize += 4; break; \
	case 2: var = get_le16(pb, nomem); rsize += 2; break; \
	case 1: var = get_byte(pb, nomem); rsize++; break; \
	default: var = defval; break; \
	}

static int asf_get_packet(AVFormatContext *s, int nomem)
{
	ASFContext *asf = (ASFContext *)s->priv_data;
	ByteIOContext *pb = &s->pb;
	uint32_t packet_length, padsize;
	int rsize = 11;
	int c = get_byte(pb, nomem);
	if(c != 0x82)
	{
		if(!url_feof(pb))
		{
			sprintf(g_cstr, "maPlayWma asf_get_packet ff asf bad header %x  at:%lld", c, url_ftell(pb, nomem));
			maLog(g_cstr);
		}
		return -EIO;
	}
	
	if((c & 0x0f) == 2)
	{ // always true for now
		if(get_le16(pb, nomem) != 0)
		{
			if(!url_feof(pb))
			{
				sprintf(g_cstr, "maPlayWma ff asf bad non zero");
				maLog(g_cstr);
			}
			return -EIO;
		}
	}

	asf->packet_flags = get_byte(pb, nomem);
	asf->packet_property = get_byte(pb, nomem);

	DO_2BITS(asf->packet_flags >> 5, packet_length, asf->packet_size);
	DO_2BITS(asf->packet_flags >> 1, padsize, 0); // sequence ignored
	DO_2BITS(asf->packet_flags >> 3, padsize, 0); // padding length

	asf->packet_timestamp = get_le32(pb, nomem);
	get_le16(pb, nomem); /* duration */
	// rsize has at least 11 bytes which have to be present

	if(asf->packet_flags & 0x01)
	{
		asf->packet_segsizetype = get_byte(pb, nomem); rsize++;
		asf->packet_segments = asf->packet_segsizetype & 0x3f;
	}
	else
	{
		asf->packet_segments = 1;
		asf->packet_segsizetype = 0x80;
	}
	asf->packet_size_left = packet_length - padsize - rsize;
	if(packet_length < asf->hdr.min_pktsize)
		padsize += asf->hdr.min_pktsize - packet_length;
	asf->packet_padsize = padsize;
	return 0;
}

static int asf_read_packet(AVFormatContext *s, AVPacket *pkt, int nomem)
{
	ASFContext *asf = (ASFContext *)s->priv_data;
	ASFStream *asf_st = 0;
	ByteIOContext *pb = &s->pb;
	//static int pc = 0;
	for(;;)
	{
		int rsize = 0;
		if(asf->packet_size_left < FRAME_HEADER_SIZE || asf->packet_segments < 1)
		{
			//asf->packet_size_left <= asf->packet_padsize) {
			int ret = asf->packet_size_left + asf->packet_padsize;
			//sprintf(g_cstr, "PacketLeftSize:%d  Pad:%d Pos:%Ld", asf->packet_size_left, asf->packet_padsize, url_ftell(pb, nomem));
			/* fail safe */
			url_fskip(pb, ret, nomem);
			ret = asf_get_packet(s, nomem);
			//sprintf(g_cstr, "READ ASF PACKET  %d   r:%d   c:%d", ret, asf->packet_size_left, pc++);
			if(ret < 0 || url_feof(pb))
				return -EIO;
			asf->packet_time_start = 0;
			continue;
		}
		if(asf->packet_time_start == 0)
		{
		/* read frame header */
			int num = get_byte(pb, nomem);
		asf->packet_segments--;
		rsize++;
		asf->packet_key_frame = (num & 0x80) >> 7;
		asf->stream_index = asf->asfid2avid[num & 0x7f];
		// sequence should be ignored!
		DO_2BITS(asf->packet_property >> 4, asf->packet_seq, 0);
		DO_2BITS(asf->packet_property >> 2, asf->packet_frag_offset, 0);
		DO_2BITS(asf->packet_property, asf->packet_replic_size, 0);

		if(asf->packet_replic_size > 1)
		{
				// it should be always at least 8 bytes - FIXME validate
			asf->packet_obj_size = get_le32(pb, nomem);
			asf->packet_frag_timestamp = get_le32(pb, nomem); // timestamp
			if(asf->packet_replic_size > 8)
				url_fskip(pb, asf->packet_replic_size - 8, nomem);
			rsize += asf->packet_replic_size; // FIXME - check validity
		}
		else
		{
			// multipacket - frag_offset is begining timestamp
			asf->packet_time_start = asf->packet_frag_offset;
			asf->packet_frag_offset = 0;
			asf->packet_frag_timestamp = asf->packet_timestamp;

			if(asf->packet_replic_size == 1)
			{
				asf->packet_time_delta = get_byte(pb, nomem);
				rsize++;
			}
		}
		if(asf->packet_flags & 0x01)
		{
			DO_2BITS(asf->packet_segsizetype >> 6, asf->packet_frag_size, 0); // 0 is illegal
#undef DO_2BITS
			//sprintf(g_cstr, "Fragsize %d", asf->packet_frag_size);
		}
		else
		{
			asf->packet_frag_size = asf->packet_size_left - rsize;
			//sprintf(g_cstr, "Using rest  %d %d %d", asf->packet_frag_size, asf->packet_size_left, rsize);
		}
		if(asf->packet_replic_size == 1)
		{
			asf->packet_multi_size = asf->packet_frag_size;
			if(asf->packet_multi_size > asf->packet_size_left)
			{
				asf->packet_segments = 0;
				continue;
			}
		}
		asf->packet_size_left -= rsize;
		//sprintf(g_cstr, "___objsize____  %d   %d	rs:%d", asf->packet_obj_size, asf->packet_frag_offset, rsize);

		if(asf->stream_index < 0)
		{
			asf->packet_time_start = 0;
			/* unhandled packet (should not happen) */
			url_fskip(pb, asf->packet_frag_size, nomem);
			asf->packet_size_left -= asf->packet_frag_size;
			//sprintf(g_cstr, "ff asf skip %d  %d", asf->packet_frag_size, num & 0x7f);
			continue;
		}
		asf->asf_st = (ASFStream *)s->stream->priv_data;
	}
	asf_st = asf->asf_st;

	if((asf->packet_frag_offset != asf_st->frag_offset
		 || (asf->packet_frag_offset
		 && asf->packet_seq != asf_st->seq)) // seq should be ignored
	   )
	{
		/* cannot continue current packet: free it */
		// FIXME better check if packet was already allocated
		sprintf(g_cstr, "maPlayWma ff asf parser skips: %d - %d	 o:%d - %d	%d %d   fl:%d",
		   asf_st->pkt.size, asf->packet_obj_size, asf->packet_frag_offset, asf_st->frag_offset,
		   asf->packet_seq, asf_st->seq, asf->packet_frag_size);
		maLog(g_cstr);
		if(asf_st->pkt.size)
			av_free_packet(&asf_st->pkt);
		asf_st->frag_offset = 0;
		if(asf->packet_frag_offset != 0)
		{
			url_fskip(pb, asf->packet_frag_size, nomem);
			sprintf(g_cstr, "maPlayWma ff asf parser skiping %db", asf->packet_frag_size);
			maLog(g_cstr);
			asf->packet_size_left -= asf->packet_frag_size;
			continue;
		}
	}
	if(asf->packet_replic_size == 1)
	{
		// frag_offset is here used as the begining timestamp
		asf->packet_frag_timestamp = asf->packet_time_start;
		asf->packet_time_start += asf->packet_time_delta;
		asf->packet_obj_size = asf->packet_frag_size = get_byte(pb, nomem);
		asf->packet_size_left--;
		asf->packet_multi_size--;
		if(asf->packet_multi_size < asf->packet_obj_size)
		{
			asf->packet_time_start = 0;
			url_fskip(pb, asf->packet_multi_size, nomem);
			asf->packet_size_left -= asf->packet_multi_size;
			continue;
		}
		asf->packet_multi_size -= asf->packet_obj_size;
		//sprintf(g_cstr, "COMPRESS size  %d  %d  %d   ms:%d", asf->packet_obj_size, asf->packet_frag_timestamp, asf->packet_size_left, asf->packet_multi_size);
	}
	if(asf_st->frag_offset == 0)
	{
		/* new packet */
		av_new_packet(&asf_st->pkt, asf->packet_obj_size);
		asf_st->seq = asf->packet_seq;
		asf_st->pkt.pts = asf->packet_frag_timestamp - asf->hdr.preroll;
		asf_st->pkt.stream_index = asf->stream_index;
		if(asf->packet_key_frame)
			asf_st->pkt.flags |= PKT_FLAG_KEY;
	}

	/* read data */
	//sprintf(g_cstr, "READ PACKET s:%d  os:%d  o:%d,%d  l:%d   DATA:%p",
	//	   asf->packet_size, asf_st->pkt.size, asf->packet_frag_offset,
	//	   asf_st->frag_offset, asf->packet_frag_size, asf_st->pkt.data);
	asf->packet_size_left -= asf->packet_frag_size;
	if(asf->packet_size_left < 0)
			continue;
	get_buffer(pb, asf_st->pkt.data + asf->packet_frag_offset,
		   asf->packet_frag_size, nomem);
	asf_st->frag_offset += asf->packet_frag_size;
	/* test if whole packet is read */
	if(asf_st->frag_offset == asf_st->pkt.size)
	{
		/* return packet */
		if(asf_st->ds_span > 1)
		{
			/* packet descrambling */
			char* newdata = (char *)av_malloc(asf_st->pkt.size);
			if(newdata)
			{
				int offset = 0;
				while(offset < asf_st->pkt.size)
				{
					int off = offset / asf_st->ds_chunk_size;
					int row = off / asf_st->ds_span;
					int col = off % asf_st->ds_span;
					int idx = row + col * asf_st->ds_packet_size / asf_st->ds_chunk_size;
					//sprintf(g_cstr, "off:%d  row:%d  col:%d  idx:%d", off, row, col, idx);
					memcpy(newdata + offset,
						asf_st->pkt.data + idx * asf_st->ds_chunk_size,
						asf_st->ds_chunk_size);
					offset += asf_st->ds_chunk_size;
					}
					av_free(asf_st->pkt.data);
					asf_st->pkt.data = (unsigned char *)newdata;
				}
			}
			asf_st->frag_offset = 0;
			memcpy(pkt, &asf_st->pkt, sizeof(AVPacket));
			//sprintf(g_cstr, "packet %d %d", asf_st->pkt.size, asf->packet_frag_size);
			asf_st->pkt.size = 0;
			asf_st->pkt.data = 0;
			break; // packet completed
		}
	}
	return 0;
}

static int asf_read_close(AVFormatContext *s)
{
	AVStream *st = s->stream;
	av_free(st->priv_data);
	av_free(st->codec.extradata);

	return 0;
}

static int asf_read_seek(AVFormatContext *s, int64_t pts)
{
#ifdef MAUNUSED
 (void)s;
#endif
	
	sprintf(g_cstr, "SEEK TO %lld", pts);
	maLog(g_cstr);

	return -1;
}

static AVInputFormat asf_iformat =
{
	"asf",
	"asf format",
	sizeof(ASFContext),
	asf_probe,
	asf_read_header,
	asf_read_packet,
	asf_read_close,
	asf_read_seek,
};

static AVInputFormat asf_iformatB =
{
	"asf",
	"asf format",
	sizeof(ASFContext),
	asf_probe,
	asf_read_header,
	asf_read_packet,
	asf_read_close,
	asf_read_seek,
};

int asf_init(AVInputFormat *p)
{
	av_register_input_format(p);
	return 0;
}

// ffmpeg.c
int wmamain(MASTATE *pstate)
{
	AVW wmaw;

	pstate->inchannels = 2; // default
	pstate->insamplerate = 44100; // default
	
	register_avcodec(&g_wmav1_decoder);
	register_avcodec(&g_wmav2_decoder);
	
	asf_init(&asf_iformat);	
			
	if(opt_input_file(pstate->inpath, pstate->nomem, &l_input_file, &l_file_iformat,
						&(pstate->inchannels), &(pstate->insamplerate)))
	{
		maLog("maPlayWma wmamain opt_input_file failed");
		return 1;
	}

	if(maOutputOpen() < 0)
	{
		maLog("maPlayWma wmamain maOutputOpen failed");
		return 1;
	}
	
	if(!av_encode_input_open(l_input_file, pstate, &wmaw))
		if(!av_encode_output_open(l_input_file, &wmaw))
		{
			if(!g_posendon)
				g_posend = pstate->duration;
			
			av_encode(l_input_file, pstate, &wmaw); // wma decode loop
		}
		
	av_close_input_file(l_input_file, pstate->nomem);

	pstate->position = 0;
	// don't seek on l_wmaseek = -1;

	return 0;
	
} // wmamain


//avienc.c
enum CodecID codec_get_id(const CodecTag *tags, unsigned int tag);
enum CodecID codec_get_id(const CodecTag *tags, unsigned int tag)
{
	while(tags->id != 0)
	{
		if(   toupper((tag >> 0)&0xFF) == toupper((tags->tag >> 0)&0xFF)
		   && toupper((tag >> 8)&0xFF) == toupper((tags->tag >> 8)&0xFF)
		   && toupper((tag >>16)&0xFF) == toupper((tags->tag >>16)&0xFF)
		   && toupper((tag >>24)&0xFF) == toupper((tags->tag >>24)&0xFF))
			return CODECIDCAST tags->id;
		tags++;
	}
	return CODEC_ID_NONE;
}

// wav.c
const CodecTag codec_wav_tags[] =
{
	/*
	{ CODEC_ID_MP2, 0x50 },
	{ CODEC_ID_MP3, 0x55 },
	{ CODEC_ID_AC3, 0x2000 },
	{ CODEC_ID_PCM_S16LE, 0x01 },
	{ CODEC_ID_PCM_U8, 0x01 }, /w* must come after s16le in this list *w/
	{ CODEC_ID_PCM_ALAW, 0x06 },
	{ CODEC_ID_PCM_MULAW, 0x07 },
	{ CODEC_ID_ADPCM_MS, 0x02 },
	{ CODEC_ID_ADPCM_IMA_WAV, 0x11 },
	{ CODEC_ID_ADPCM_IMA_DK4, 0x61 },  /w* rogue format number *w/
	{ CODEC_ID_ADPCM_IMA_DK3, 0x62 },  /w* rogue format number *w/
	*/
	{ CODEC_ID_WMAV1, 0x160 },
	{ CODEC_ID_WMAV2, 0x161 },
	{ 0, 0 },
};

/* We could be given one of the three possible structures here:
 * WAVEFORMAT, PCMWAVEFORMAT or WAVEFORMATEX. Each structure
 * is an expansion of the previous one with the fields added
 * at the bottom. PCMWAVEFORMAT adds 'WORD wBitsPerSample' and
 * WAVEFORMATEX adds 'WORD  cbSize' and basically makes itself
 * an openended structure.
 */
void get_wav_header(ByteIOContext *pb, AVCodecContext *codec, int size, int nomem) 
{
	int id;

	id = get_le16(pb, nomem);
	codec->codec_type = CODEC_TYPE_AUDIO;
	codec->codec_tag = id;
	codec->channels = get_le16(pb, nomem);
	codec->sample_rate = get_le32(pb, nomem);
	codec->bit_rate = get_le32(pb, nomem) * 8;
	codec->block_align = get_le16(pb, nomem);
	if(size == 14)
	{  /* We're dealing with plain vanilla WAVEFORMAT */
		codec->bits_per_sample = 8;
	}else
		codec->bits_per_sample = get_le16(pb, nomem);
	codec->codec_id = CODECIDCAST wav_codec_get_id(id, codec->bits_per_sample);

	if(size > 16)
	{  /* We're obviously dealing with WAVEFORMATEX */
		codec->extradata_size = get_le16(pb, nomem);
		if(codec->extradata_size > 0)
		{
			if(codec->extradata_size > size - 18)
				codec->extradata_size = size - 18;
			codec->extradata = av_mallocz(codec->extradata_size);
			get_buffer(pb, (unsigned char *)codec->extradata, codec->extradata_size, nomem);
		}
		else
			codec->extradata_size = 0;
	
		/* It is possible for the chunk to contain garbage at the end */
		if(size - codec->extradata_size - 18 > 0)
			url_fskip(pb, size - codec->extradata_size - 18, nomem);
	}
}

int wav_codec_get_id(unsigned int tag, int bps)
{
#ifdef MAUNUSED
 (void)bps;
#endif
	
	return codec_get_id(codec_wav_tags, tag);
}

// file.c

/* standard file protocol */

static int file_open(URLContext *h, const char *filename, int flags, int nomem)
{
#ifdef MAUNUSED
 (void)flags;
#endif
	
	int access;
	int fd;
	
	access = O_RDONLY;
	
#if defined(CONFIG_WIN32)
	access |= O_BINARY;
#endif

	if(nomem)
	{
#if defined(__MAOSX__)  
		char bsdname[MAX_PATH];
		strcpy(bsdname, filename);
		StrOSXtoStrBSD(bsdname);
		fd = open(bsdname, access, 0666);
#else
		fd = open(filename, access, 0666);
#endif
	}
	else
		fd = (int)memfopen((char *)filename, "rb");
	
	if(fd < 0)
		return -ENOENT;
		
	h->priv_data = (void *)fd;
	
	return 0;
}

static int file_read(URLContext *h, unsigned char *buf, int size, int nomem)
{
	int ret;
	int fd = (int)h->priv_data;

	if(nomem)
		ret = read(fd, buf, size);
	else
		ret = memread(fd, (char *)buf, size);
		
	return ret;
}


/* XXX: use llseek */
static offset_t file_seek(URLContext *h, offset_t pos, int whence, int nomem)
{
	int fd = (int)h->priv_data;
	
	if(!nomem)
		return memlseek(fd, (off_t)pos, whence);
	
#ifdef CONFIG_WIN32
	return _lseeki64(fd, pos, whence);
#else
	return lseek(fd, pos, whence);
#endif
}

static int file_close(URLContext *h, int nomem)
{
	int fd = (int)h->priv_data;

	if(nomem)
		return close(fd);
	else
		return memfclose((FILE *)fd);
}

//aviobuf.c
#define AV_IO_BUFFER_SIZE 8192 //n 32768

int init_put_byte(ByteIOContext *s,
				  unsigned char *buffer,
				  int buffer_size,
				  void *opaque,
				  int (*read_packet)(void *opaque, uint8_t *buf, int buf_size, int nomem),
				  int (*seek)(void *opaque, offset_t offset, int whence, int nomem))
{
	s->buffer = buffer;
	s->buffer_size = buffer_size;
	s->buf_ptr = buffer;
	s->buf_end = buffer;
	s->opaque = opaque;
	s->read_packet = read_packet;
	s->seek = seek;
	s->pos = 0;
	s->eof_reached = 0;
	s->max_packet_size = 0;
	return 0;
}
				  

offset_t url_fseek(ByteIOContext *s, offset_t offset, int whence, int nomem)
{
	offset_t offset1;

	if(whence != SEEK_CUR && whence != SEEK_SET)
		return -EINVAL;

		if(whence == SEEK_CUR)
		{
			offset1 = s->pos - (s->buf_end - s->buffer) + (s->buf_ptr - s->buffer);
			if(offset == 0)
				return offset1;
			offset += offset1;
		}
		offset1 = offset - (s->pos - (s->buf_end - s->buffer));
		if(offset1 >= 0 && offset1 <= (s->buf_end - s->buffer))
		{
			/* can do the seek inside the buffer */
			s->buf_ptr = s->buffer + offset1;
		} else
		{
			if(!s->seek)
				return -32; //-EPIPE;
			s->buf_ptr = s->buffer;
			s->buf_end = s->buffer;
			s->seek(s->opaque, offset, SEEK_SET, nomem);
			s->pos = offset;
		}
		s->eof_reached = 0;

	return offset;
}

void url_fskip(ByteIOContext *s, offset_t offset, int nomem)
{
	url_fseek(s, offset, SEEK_CUR, nomem);
}

offset_t url_ftell(ByteIOContext *s, int nomem)
{
	return url_fseek(s, 0, SEEK_CUR, nomem);
}

int url_feof(ByteIOContext *s)
{
	return s->eof_reached;
}

/* Input stream */

static void fill_buffer(ByteIOContext *s, int nomem)
{
	int len;

	/* no need to do anything if EOF already reached */
	if(s->eof_reached)
		return;
	len = s->read_packet(s->opaque, s->buffer, s->buffer_size, nomem);
	if(len <= 0)
	{
		/* do not modify buffer if EOF reached so that a seek back can
		   be done without rereading data */
		s->eof_reached = 1;
	}
	else
	{
		s->pos += len;
		s->buf_ptr = s->buffer;
		s->buf_end = s->buffer + len;
	}
}

/* NOTE: return 0 if EOF, so you cannot use it if EOF handling is necessary */
/* XXX: put an inline version */
int get_byte(ByteIOContext *s, int nomem)
{
	if(s->buf_ptr < s->buf_end)
	{
		return *s->buf_ptr++;
	} else
	{
		fill_buffer(s, nomem);
		if(s->buf_ptr < s->buf_end)
			return *s->buf_ptr++;
		else
			return 0;
	}
}

int get_buffer(ByteIOContext *s, unsigned char *buf, int size, int nomem)
{
	int len, size1;

	size1 = size;
	while(size > 0)
	{
		len = s->buf_end - s->buf_ptr;
		if(len > size)
			len = size;
		if(len == 0)
		{
			fill_buffer(s, nomem);
			len = s->buf_end - s->buf_ptr;
			if(len == 0)
				break;
		}
		else
		{
			memcpy(buf, s->buf_ptr, len);
			buf += len;
			s->buf_ptr += len;
			size -= len;
		}
	}
	return size1 - size;
}

unsigned int get_le16(ByteIOContext *s, int nomem)
{
	unsigned int val;
	val = get_byte(s, nomem);
	val |= get_byte(s, nomem) << 8;
	return val;
}

unsigned int get_le32(ByteIOContext *s, int nomem)
{
	unsigned int val;
	val = get_byte(s, nomem);
	val |= get_byte(s, nomem) << 8;
	val |= get_byte(s, nomem) << 16;
	val |= get_byte(s, nomem) << 24;
	return val;
}

uint64_t get_le64(ByteIOContext *s, int nomem)
{
	uint64_t val;
	val = (uint64_t)get_le32(s, nomem);
	val |= (uint64_t)get_le32(s, nomem) << 32;
	return val;
}

/* link with avio functions */

static int url_read_packet(void *opaque, uint8_t *buf, int buf_size, int nomem)
{
	URLContext *h = (URLContext *)opaque;
	return url_read(h, buf, buf_size, nomem);
}

static int url_seek_packet(void *opaque, int64_t offset, int whence, int nomem)
{
	URLContext *h = (URLContext *)opaque;
	url_seek(h, offset, whence, nomem);
	return 0;
}

int url_fdopen(ByteIOContext *s, URLContext *h)
{
	uint8_t *buffer;
	int buffer_size, max_packet_size;

	
	max_packet_size = url_get_max_packet_size(h);
	if(max_packet_size)
	{
		buffer_size = max_packet_size; /* no need to bufferize more than one packet */
	}
	else
		buffer_size = AV_IO_BUFFER_SIZE;
	buffer = (unsigned char *)av_malloc(buffer_size);
	if(!buffer)
		return -ENOMEM;

	if(init_put_byte(s, buffer, buffer_size,
					  h,
					  url_read_packet,
					  url_seek_packet) < 0)
	{
		av_free(buffer);
		return -EIO;
	}
	s->max_packet_size = max_packet_size;
	return 0;
}

/* XXX: must be called before any I/O */
int url_setbufsize(ByteIOContext *s, int buf_size)
{
	uint8_t *buffer;
	buffer = (unsigned char *)av_malloc(buf_size);
	if(!buffer)
		return -ENOMEM;

	av_free(s->buffer);
	s->buffer = buffer;
	s->buffer_size = buf_size;
	s->buf_ptr = buffer;
	s->buf_end = buffer;

	return 0;
}

/* NOTE: when opened as read/write, the buffers are only used for reading */
int url_fopen(ByteIOContext *s, const char *filename, int flags, int nomem)
{
	URLContext *h;
	int err;

	err = url_open(&h, filename, flags, nomem);
	if(err < 0)
		return err;
	err = url_fdopen(s, h);
	if(err < 0)
	{
		url_close(h, nomem);
		return err;
	}
	return 0;
}

int url_fclose(ByteIOContext *s, int nomem)
{
	URLContext *h = (URLContext *)s->opaque;
	
	av_free(s->buffer);
	memset(s, 0, sizeof(ByteIOContext));
	return url_close(h, nomem);
}

URLContext *url_fileno(ByteIOContext *s)
{
	return (URLContext *)s->opaque;
}


// cutils.c
/**
 * Copy the string str to buf. If str length is bigger than buf_size -
 * 1 then it is clamped to buf_size - 1.
 * NOTE: this function does what strncpy should have done to be
 * useful. NEVER use strncpy.
 * 
 * @param buf destination buffer
 * @param buf_size size of destination buffer
 * @param str source string
 */
void pstrcpy(char *buf, int buf_size, const char *str)
{
	int c;
	char *q = buf;

	if(buf_size <= 0)
		return;

	for(;;)
	{
		c = *str++;
		if(c == 0 || q >= buf + buf_size - 1)
			break;
		*q++ = c;
	}
	*q = '\0';
}

//avio.c

int url_open(URLContext **puc, const char *filename, int flags, int nomem)
{
	URLContext *uc;
	int err;

	uc = (URLContext *)av_malloc(sizeof(URLContext) + strlen(filename));
	if(!uc)
	{
		err = -ENOMEM;
		goto fail;
	}
	strcpy(uc->filename, filename);

	uc->flags = flags;

	uc->max_packet_size = 0; /* default: stream file */
	
	err = file_open(uc, filename, flags, nomem);
	if(err < 0)
	{
		av_free(uc);
		*puc = NULL;
		return err;
	}
	*puc = uc;
	return 0;
 fail:
	*puc = NULL;
	return err;
}

int url_read(URLContext *h, unsigned char *buf, int size, int nomem)
{
	int ret;
	if(h->flags & URL_WRONLY)
		return -EIO;
	ret = file_read(h, buf, size, nomem);
	return ret;
}

offset_t url_seek(URLContext *h, offset_t pos, int whence, int nomem)
{
	offset_t ret;

	ret = file_seek(h, pos, whence, nomem);
	return ret;
}

int url_close(URLContext *h, int nomem)
{
	int ret;

	ret = file_close(h, nomem);
	av_free(h);
	return ret;
}

offset_t url_filesize(URLContext *h, int nomem)
{
	offset_t pos, size;
	
	pos = url_seek(h, 0, SEEK_CUR, nomem);
	size = url_seek(h, 0, SEEK_END, nomem);
	url_seek(h, pos, SEEK_SET, nomem);
	return size;
}

/* 
 * Return the maximum packet size associated to packetized file
 * handle. If the file is not packetized (stream like http or file on
 * disk), then 0 is returned.
 * 
 * @param h file handle
 * @return maximum packet size in bytes
 */
int url_get_max_packet_size(URLContext *h)
{
	return h->max_packet_size;
}


int DumpInfoWma(MASTATE *pstate, FILE *dumpfp)
{
	AVW wmaw;
	char tstr[32];
	char cstrx[512];
	AVFormatContext *g_input_fileB;
	AVInputFormat *g_file_iformatB;
	
	pstate->inchannels = 2; // default
	pstate->insamplerate = 44100; // default

	register_avcodec(&g_wmav1_decoderB);
	register_avcodec(&g_wmav2_decoderB);
	
	asf_init(&asf_iformatB);	
			
	if(opt_input_file(pstate->inpath, pstate->nomem, &g_input_fileB, &g_file_iformatB,
								&(pstate->inchannels), &(pstate->insamplerate)))
		return 1;

	av_encode_input_open(g_input_fileB, pstate, &wmaw);

	sprintf(cstrx, "Duration : %s", SecondsToStrHMS(tstr, pstate->duration / MSPERSEC));
	DumpString(dumpfp, cstrx);
	
	sprintf(cstrx, "Samplerate : %d", pstate->insamplerate);
	DumpString(dumpfp, cstrx);

	sprintf(cstrx, "Bitrate : %d", wmaw.is->bit_rate);
	DumpString(dumpfp, cstrx);

	sprintf(cstrx, "Channels : %d", pstate->inchannels);
	DumpString(dumpfp, cstrx);

	if(wmaw.is->author[0])
	{
		sprintf(cstrx, "Author : %s", wmaw.is->author);
		DumpString(dumpfp, cstrx);
	}
	
	if(wmaw.is->copyright[0])
	{
		sprintf(cstrx, "Copyright : %s", wmaw.is->copyright);
		DumpString(dumpfp, cstrx);
	}

	if(wmaw.is->comment[0])
	{
		sprintf(cstrx, "Comment : %s", wmaw.is->comment);
		DumpString(dumpfp, cstrx);
	}

	if(wmaw.is->album[0])
	{
		sprintf(cstrx, "Album : %s", wmaw.is->album);
		DumpString(dumpfp, cstrx);
	}
	
	if(wmaw.is->year)
	{
		sprintf(cstrx, "Year : %d", wmaw.is->year);
		DumpString(dumpfp, cstrx);
	}
	
	if(wmaw.is->track)
	{
		sprintf(cstrx, "Track : %d", wmaw.is->track);
		DumpString(dumpfp, cstrx);
	}

	if(wmaw.is->genre[0])
	{
		sprintf(cstrx, "Genre : %s", wmaw.is->genre);
		DumpString(dumpfp, cstrx);
	}
	
	av_close_input_file(g_input_fileB, pstate->nomem);

	return 0;
	
} // DumpInfoWma

#if defined(__cplusplus)
}
#endif

#endif // NODECODEWMA

// eof
